/* $Id: aabc.c,v 1.5 2017/08/05 11:31:47 rlr Exp rlr $ */

/*
 * Copyright (c) 2016, 2017 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <ctype.h>
#include <limits.h>
#include <string.h>

#include "aabc.h"


#define HARD_BLANK  '$'
#define SP          ' '

#define WIDTH_01   1
#define WIDTH_02   2
#define WIDTH_03   3
#define WIDTH_04   4
#define WIDTH_05   5
#define WIDTH_06   6
#define WIDTH_07   7
#define WIDTH_08   8
#define WIDTH_09   9
#define WIDTH_10  10
#define WIDTH_11  11
#define WIDTH_12  12
#define WIDTH_13  13
#define WIDTH_14  14

#define CH_IGNORE  (size_t)UCHAR_MAX

/* Representation of a "big" character. */
struct	big_ch {
	const size_t	 w;  /* width [= strlen(a[0]), etc.] */
	const char const	*a[AABC_FONT_HEIGHT];

};

/* Function prototypes. */
static bool	 can_m_cols(char (*f)(char, char), char *[], size_t,
    const struct big_ch *, size_t);
static size_t	 ch_index(char);
static void	 determine_append_info(char *[], size_t, const struct big_ch *,
    size_t *, bool *);
static char	 kern_ch(char, char);
static size_t	 line_len_no_smush(const char *);
static void	 make_no_smush(const char *, char *, size_t);
static void	 make_smush(const char *, char *, size_t);
static void	 replace_hard_blanks(char *);
static char	 smush_ch(char, char);


/*
 * The font here is converted from (a subset of) the font "Big"
 * (file fonts/big.flf) of the (BSD licensed) figlet distribution.  That file
 * contains the following text:
 * ====
 * Big by Glenn Chappell 4/93 -- based on Standard
 * Includes ISO Latin-1
 * Greek characters by Bruce Jakeway <pbjakeway@neumann.uwaterloo.ca>
 * figlet release 2.2 -- November 1996
 * Permission is hereby given to modify this font, as long as the
 * modifier's name is placed on a comment line.
 *
 * Modified by Paul Burton <solution@earthlink.net> 12/96 to include new
 * parameter
 * supported by FIGlet and FIGWin.  May also be slightly modified for better use
 * of new full-width/kern/smush alternatives, but default output is NOT changed.
 * ====
 * Adapted for C use by Ross L Richardson, 2016.  The leading space of (each
 * row of each) character has been elided to facilitate the expected smushing.
 * In the non-smushing case, the code re-inserts it in the ouput.
 */

/*
 * '$' characters represent "hardblanks".  These are shown as spaces in the
 * output, but prevent "smushing"/kerning (in that cell).
 */

static const struct	big_ch	font[] = {
	{  /* ' ' */
		.w = WIDTH_01,
		.a = {
			"$",
			"$",
			"$",
			"$",
			"$",
			"$",
			"$",
			"$"
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			" _ ",
			"| |",
			"| |",
			"| |",
			"|_|",
			"(_)",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			" _ _ ",
			"( | )",
			" V V ",
			"  $  ",
			"  $  ",
			"  $  ",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_10,
		.a = {
			"   _  _   ",
			" _| || |_ ",
			"|_  __  _|",
			" _| || |_ ",
			"|_  __  _|",
			"  |_||_|  ",
			"          ",
			"          "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"  _  ",
			" | | ",
			"/ __)",
			"\\__ \\",
			"(   /",
			" |_| ",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" _   __",
			"(_) / /",
			"   / / ",
			"  / /  ",
			" / / _ ",
			"/_/ (_)",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"        ",
			"  ___   ",
			" ( _ )  ",
			" / _ \\/\\",
			"| (_>  <",
			" \\___/\\/",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			" _ ",
			"( )",
			"|/ ",
			" $ ",
			" $ ",
			" $ ",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_04,
		.a = {
			"  __",
			" / /",
			"| | ",
			"| | ",
			"| | ",
			"| | ",
			" \\_\\",
			"    "
		}
	},
	{
		.w = WIDTH_04,
		.a = {
			"__  ",
			"\\ \\ ",
			" | |",
			" | |",
			" | |",
			" | |",
			"/_/ ",
			"    "
		}
	},
	{
		.w = WIDTH_09,
		.a = {
			"    _    ",
			" /\\| |/\\ ",
			" \\ ` ' / ",
			"|_     _|",
			" / , . \\ ",
			" \\/|_|\\/ ",
			"         ",
			"         "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"   _   ",
			" _| |_ ",
			"|_   _|",
			"  |_|  ",
			"   $   ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			"   ",
			"   ",
			"   ",
			"   ",
			" _ ",
			"( )",
			"|/ ",
			"   "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"        ",
			"        ",
			" ______ ",
			"|______|",
			"    $   ",
			"    $   ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			"   ",
			"   ",
			"   ",
			"   ",
			" _ ",
			"(_)",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"     __",
			"    / /",
			"   / / ",
			"  / /  ",
			" / /   ",
			"/_/    ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"  ___  ",
			" / _ \\ ",
			"| | | |",
			"| | | |",
			"| |_| |",
			" \\___/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_04,
		.a = {
			" __ ",
			"/_ |",
			" | |",
			" | |",
			" | |",
			" |_|",
			"    ",
			"    "
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			" ___  ",
			"|__ \\ ",
			"  $) |",
			"  / / ",
			" / /_ ",
			"|____|",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" ____  ",
			"|___ \\ ",
			"  __) |",
			" |__ < ",
			" ___) |",
			"|____/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _  _   ",
			"| || |  ",
			"| || |_ ",
			"|__   _|",
			"   | |  ",
			"   |_|  ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" _____ ",
			"| ____|",
			"| |__  ",
			"|___ \\ ",
			" ___) |",
			"|____/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"   __  ",
			"  / /  ",
			" / /_  ",
			"| '_ \\ ",
			"| (_) |",
			" \\___/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" ______ ",
			"|____  |",
			"   $/ / ",
			"   / /  ",
			"  / /   ",
			" /_/    ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"  ___  ",
			" / _ \\ ",
			"| (_) |",
			" > _ < ",
			"| (_) |",
			" \\___/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"  ___  ",
			" / _ \\ ",
			"| (_) |",
			" \\__, |",
			"   / / ",
			"  /_/  ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			"   ",
			" _ ",
			"(_)",
			" $ ",
			" _ ",
			"(_)",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			"   ",
			" _ ",
			"(_)",
			" $ ",
			" _ ",
			"( )",
			"|/ ",
			"   "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"   __",
			"  / /",
			" / / ",
			"< <  ",
			" \\ \\ ",
			"  \\_\\",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"        ",
			" ______ ",
			"|______|",
			" ______ ",
			"|______|",
			"        ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"__   ",
			"\\ \\  ",
			" \\ \\ ",
			"  > >",
			" / / ",
			"/_/  ",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_06,
		.a = {

			" ___  ",
			"|__ \\ ",
			"   ) |",
			"  / / ",
			" |_|  ",
			" (_)  ",
			"      ",
			"      "
		}
	},
	{ /* '@' */
		.w = WIDTH_09,
		.a = {
			"         ",
			"   ____  ",
			"  / __ \\ ",
			" / / _` |",
			"| | (_| |",
			" \\ \\__,_|",
			"  \\____/ ",
			"         "
		}
	},
	{
		.w = WIDTH_10,
		.a = {
			"          ",
			"    /\\    ",
			"   /  \\   ",
			"  / /\\ \\  ",
			" / ____ \\ ",
			"/_/    \\_\\",
			"          ",
			"          "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" ____  ",
			"|  _ \\ ",
			"| |_) |",
			"|  _ < ",
			"| |_) |",
			"|____/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"  _____ ",
			" / ____|",
			"| | $   ",
			"| | $   ",
			"| |____ ",
			" \\_____|",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _____  ",
			"|  __ \\ ",
			"| |  | |",
			"| |  | |",
			"| |__| |",
			"|_____/ ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" ______ ",
			"|  ____|",
			"| |__   ",
			"|  __|  ",
			"| |____ ",
			"|______|",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" ______ ",
			"|  ____|",
			"| |__   ",
			"|  __|  ",
			"| |     ",
			"|_|     ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"  _____ ",
			" / ____|",
			"| |  __ ",
			"| | |_ |",
			"| |__| |",
			" \\_____|",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _    _ ",
			"| |  | |",
			"| |__| |",
			"|  __  |",
			"| |  | |",
			"|_|  |_|",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" _____ ",
			"|_   _|",
			"  | |  ",
			"  | |  ",
			" _| |_ ",
			"|_____|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"      _ ",
			"     | |",
			"     | |",
			" _   | |",
			"| |__| |",
			" \\____/ ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			" _  __",
			"| |/ /",
			"| ' / ",
			"|  <  ",
			"| . \\ ",
			"|_|\\_\\",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _      ",
			"| |     ",
			"| |     ",
			"| |     ",
			"| |____ ",
			"|______|",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" __  __ ",
			"|  \\/  |",
			"| \\  / |",
			"| |\\/| |",
			"| |  | |",
			"|_|  |_|",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" _   _ ",
			"| \\ | |",
			"|  \\| |",
			"| . ` |",
			"| |\\  |",
			"|_| \\_|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"  ____  ",
			" / __ \\ ",
			"| |  | |",
			"| |  | |",
			"| |__| |",
			" \\____/ ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _____  ",
			"|  __ \\ ",
			"| |__) |",
			"|  ___/ ",
			"| |     ",
			"|_|     ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"  ____  ",
			" / __ \\ ",
			"| |  | |",
			"| |  | |",
			"| |__| |",
			" \\___\\_\\",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _____  ",
			"|  __ \\ ",
			"| |__) |",
			"|  _  / ",
			"| | \\ \\ ",
			"|_|  \\_\\",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			"  _____ ",
			" / ____|",
			"| (___  ",
			" \\___ \\ ",
			" ____) |",
			"|_____/ ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_09,
		.a = {
			" _______ ",
			"|__   __|",
			"   | |   ",
			"   | |   ",
			"   | |   ",
			"   |_|   ",
			"         ",
			"         "
		}
	},
	{
		.w = WIDTH_08,
		.a = {
			" _    _ ",
			"| |  | |",
			"| |  | |",
			"| |  | |",
			"| |__| |",
			" \\____/ ",
			"        ",
			"        "
		}
	},
	{
		.w = WIDTH_10,
		.a = {
			"__      __",
			"\\ \\    / /",
			" \\ \\  / / ",
			"  \\ \\/ /  ",
			"   \\  /   ",
			"    \\/    ",
			"          ",
			"          "
		}
	},
	{
		.w = WIDTH_14,
		.a = {
			"__          __",
			"\\ \\        / /",
			" \\ \\  /\\  / / ",
			"  \\ \\/  \\/ /  ",
			"   \\  /\\  /   ",
			"    \\/  \\/    ",
			"              ",
			"              "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"__   __",
			"\\ \\ / /",
			" \\ V / ",
			"  > <  ",
			" / . \\ ",
			"/_/ \\_\\",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_09,
		.a = {
			"__     __",
			"\\ \\   / /",
			" \\ \\_/ / ",
			"  \\   /  ",
			"   | |   ",
			"   |_|   ",
			"         ",
			"         "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" ______",
			"|___  /",
			"  $/ / ",
			"  / /  ",
			" / /__ ",
			"/_____|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			" ___ ",
			"|  _|",
			"| |  ",
			"| |  ",
			"| |  ",
			"| |_ ",
			"|___|",
			"     "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"__     ",
			"\\ \\    ",
			" \\ \\   ",
			"  \\ \\  ",
			"   \\ \\ ",
			"    \\_\\",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			" ___ ",
			"|_  |",
			"  | |",
			"  | |",
			"  | |",
			" _| |",
			"|___|",
			"     "
		}
	},
	{
		.w = WIDTH_04,
		.a = {
			" /\\ ",
			"|/\\|",
			"  $ ",
			"  $ ",
			"  $ ",
			"  $ ",
			"    ",
			"    "
		}
	},
	{ /* '_' */
		.w = WIDTH_08,
		.a = {
			"        ",
			"        ",
			"        ",
			"        ",
			"        ",
			"    $   ",
			" ______ ",
			"|______|"
		}
	},
	{ /* '`' */
		.w = WIDTH_03,
		.a = {
			" _ ",
			"( )",
			" \\|",
			" $ ",
			" $ ",
			" $ ",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			"  __ _ ",
			" / _` |",
			"| (_| |",
			" \\__,_|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" _     ",
			"| |    ",
			"| |__  ",
			"| '_ \\ ",
			"| |_) |",
			"|_.__/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			"      ",
			"      ",
			"  ___ ",
			" / __|",
			"| (__ ",
			" \\___|",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"     _ ",
			"    | |",
			"  __| |",
			" / _` |",
			"| (_| |",
			" \\__,_|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			"      ",
			"      ",
			"  ___ ",
			" / _ \\",
			"|  __/",
			" \\___|",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"  __ ",
			" / _|",
			"| |_ ",
			"|  _|",
			"| |  ",
			"|_|  ",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			"  __ _ ",
			" / _` |",
			"| (_| |",
			" \\__, |",
			"  __/ |",
			" |___/ "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			" _     ",
			"| |    ",
			"| |__  ",
			"| '_ \\ ",
			"| | | |",
			"|_| |_|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			" _ ",
			"(_)",
			" _ ",
			"| |",
			"| |",
			"|_|",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"   _ ",
			"  (_)",
			"   _ ",
			"  | |",
			"  | |",
			"  | |",
			" _/ |",
			"|__/ "
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			" _    ",
			"| |   ",
			"| | __",
			"| |/ /",
			"|   < ",
			"|_|\\_\\",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			" _ ",
			"| |",
			"| |",
			"| |",
			"| |",
			"|_|",
			"   ",
			"   "
		}
	},
	{
		.w = WIDTH_11,
		.a = {
			"           ",
			"           ",
			" _ __ ___  ",
			"| '_ ` _ \\ ",
			"| | | | | |",
			"|_| |_| |_|",
			"           ",
			"           "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			" _ __  ",
			"| '_ \\ ",
			"| | | |",
			"|_| |_|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			"  ___  ",
			" / _ \\ ",
			"| (_) |",
			" \\___/ ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			" _ __  ",
			"| '_ \\ ",
			"| |_) |",
			"| .__/ ",
			"| |    ",
			"|_|    "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			"  __ _ ",
			" / _` |",
			"| (_| |",
			" \\__, |",
			"    | |",
			"    |_|"
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			"      ",
			"      ",
			" _ __ ",
			"| '__|",
			"| |   ",
			"|_|   ",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"     ",
			"     ",
			" ___ ",
			"/ __|",
			"\\__ \\",
			"|___/",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			" _   ",
			"| |  ",
			"| |_ ",
			"| __|",
			"| |_ ",
			" \\__|",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			" _   _ ",
			"| | | |",
			"| |_| |",
			" \\__,_|",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			"__   __",
			"\\ \\ / /",
			" \\ V / ",
			"  \\_/  ",
			"       ",
			"       "
		}
	},
	{
		.w = WIDTH_10,
		.a = {
			"          ",
			"          ",
			"__      __",
			"\\ \\ /\\ / /",
			" \\ V  V / ",
			"  \\_/\\_/  ",
			"          ",
			"          "
		}
	},
	{
		.w = WIDTH_06,
		.a = {
			"      ",
			"      ",
			"__  __",
			"\\ \\/ /",
			" >  < ",
			"/_/\\_\\",
			"      ",
			"      "
		}
	},
	{
		.w = WIDTH_07,
		.a = {
			"       ",
			"       ",
			" _   _ ",
			"| | | |",
			"| |_| |",
			" \\__, |",
			"  __/ |",
			" |___/ "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"     ",
			"     ",
			" ____",
			"|_  /",
			" / / ",
			"/___|",
			"     ",
			"     "
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"   __",
			"  / /",
			" | | ",
			"/ /  ",
			"\\ \\  ",
			" | | ",
			"  \\_\\",
			"     "
		}
	},
	{
		.w = WIDTH_03,
		.a = {
			" _ ",
			"| |",
			"| |",
			"| |",
			"| |",
			"| |",
			"| |",
			"|_|"
		}
	},
	{
		.w = WIDTH_05,
		.a = {
			"__   ",
			"\\ \\  ",
			" | | ",
			"  \\ \\",
			"  / /",
			" | | ",
			"/_/  ",
			"     "
		}
	},
	{ /* '~' */
		.w = WIDTH_05,
		.a = {
			" /\\/|",
			"|/\\/ ",
			"  $  ",
			"  $  ",
			"  $  ",
			"  $  ",
			"     ",
			"     "
		}
	}
};


/* Strings for "smushing" rules. */
static const char	*rule_2_chars = "|/\\[]{}()<>";
static const char	*rule_3_chars = "||/\\[]{}()<>";


/*
 * "Smushing" consists of two steps:
 *     - Kerning until (big) characters touch (i.e. adjacent cells are
 *       occupied by non-space sub-characters of different (big) characters).
 *     - If possible according to the "smushing" rules, the creation of one
 *       additional column of overlap by "smushing"/merging of sub-characters.
 *
 * Note that the number of cells overlapped in "smushing"/kerning is limited
 * to the width of the later (big) character.
 */

/*
 * We deal only with a single line of left-to-right text.  Conceptually,
 * things look a little like the following (where the '.' characters
 * represent *zero-width* cell boundaries.
 *
 *     .__          __.   .
 *     .\ \        / /.   .
 *     . \ \  /\  / / .   .
 *     .  \ \/  \/ /  .   .
 *     .   \  /\  /   . _ .
 *     .    \/  \/    .( ).
 *     .              .|/ .
 *     .              .   .
 *
 * The first (big) character is placed hard to the left.  Subsequent (big)
 * characters are moved as far to the left as the "smushing" rules permit.
 * In the above, the comma can be moved to the left.
 *
 * It may be possible to move a (big) character to the left by more than
 * a single (small) character, but we have to check _all_ overlapped cells
 * for "smushing"/kerning compatibility.
 */

/*
 * The following are the subset of rules which are applicable to horizontal
 * "smushing" of the font "Big".  They are taken (verbatim) from the
 * "figfont.txt" file of the figlet distribution.
 * ====
 * Rule 1: EQUAL CHARACTER SMUSHING (code value 1)
 *
 *     Two sub-characters are smushed into a single sub-character
 *     if they are the same.  This rule does not smush
 *     hardblanks.  (See "Hardblanks" below.)
 *
 * Rule 2: UNDERSCORE SMUSHING (code value 2)
 *
 *     An underscore ("_") will be replaced by any of: "|", "/",
 *     "\", "[", "]", "{", "}", "(", ")", "<" or ">".
 *
 * Rule 3: HIERARCHY SMUSHING (code value 4)
 *
 *     A hierarchy of six classes is used: "|", "/\", "[]", "{}",
 *     "()", and "<>".  When two smushing sub-characters are
 *     from different classes, the one from the latter class
 *     will be used.
 *
 * Rule 4: OPPOSITE PAIR SMUSHING (code value 8)
 *
 *     Smushes opposing brackets ("[]" or "]["), braces ("{}" or
 *     "}{") and parentheses ("()" or ")(") together, replacing
 *     any such pair with a vertical bar ("|").
 *
 */


/*
 * Returns the result of attempting to apply kerning to a and b:
 * either the resultant character or HARD_BLANK if kerning is not possible.
 */
static char
kern_ch(char a, char b)
{
	return a == SP ? b : (b == SP ? a : HARD_BLANK);
}


/*
 * Returns result of applying "smushing" rules to the given characters;
 * HARD_BLANK is returned if "smushing" is not possible.
 */
static char
smush_ch(char a, char b)
{
	char	*p, *q;
	char	 smushed;

	if (a == HARD_BLANK || b == HARD_BLANK) {
		smushed = HARD_BLANK;
	} else if (a == SP) {
		smushed = b;  /* kerning */
	} else if (b == SP) {
		smushed = a;  /* kerning */
	} else if (a == b) {
		smushed = a;  /* Rule 1 */
	} else if (a == '_' && strchr(rule_2_chars, b) != NULL) {
		smushed = b;  /* Rule 2 */
	} else if (b == '_' && strchr(rule_2_chars, a) != NULL) {
		smushed = a;  /* Rule 2 */
	} else if ((p = strchr(rule_3_chars, a)) != NULL
	    && (q = strchr(rule_3_chars, b)) != NULL
	    && (p - rule_3_chars) / 2 != (q - rule_3_chars) / 2) {
		smushed = p > q ? a : b;  /* Rule 3 */
	} else if ((a == '[' && b == ']') || (a == ']' && b == '[')
	    || (a == '{' && b == '}') || (a == '}' && b == '{')
	    || (a == '(' && b == ')') || (a == ')' && b == '(')) {
		smushed = '|';  /* Rule 4 */
	} else {
		smushed = HARD_BLANK;
	}

	return smushed;
}


/*
 * If a representation of c is found in the font array, returns the index.
 * Otherwise, returns the index of ' ' for all whitespace characters, or
 * CH_IGNORE indicating an out-of-range character.
 */
static size_t
ch_index(char c)
{
	unsigned char	 i;

	if (SP <= c && c <= '~') {
		i = c - SP;
	} else if (isspace((unsigned char)c)) {
		i = 0;
	} else {
		i = CH_IGNORE;
	}

	return i;
}


/*
 * Returns the length in (sub-) characters of each line (including added
 * trailing newline) of the big character representation of s, in the case
 * no "smushing"/kerning is performed.
 */
static size_t
line_len_no_smush(const char *s)
{
	size_t	 len;
	size_t	 c, i;

	len = 0;

	/* Get the line length (without any "smushing"). */
	i = 0;
	while (s[i] != '\0') {
		if ((c = ch_index(s[i])) != CH_IGNORE) {
			/*
			 * Allow one extra character for the leading space
			 * elided in converting the font to C.
			 */
			len += 1 + font[c].w;
		}
		i++;
	}

	/* Increment for the newline. */
	len += 1;

	return len;
}


size_t
aabc_req_buf_sz_no_smush(const char *s)
{
	size_t	 total;

	total = line_len_no_smush(s);

	/* Multiply by the number of lines. */
	total *= AABC_FONT_HEIGHT;

	/* Increment for the terminating NUL. */
	total += 1;

	return total;
}


/*
 * Replaces all hardblank characters in s with spaces.
 */
static void
replace_hard_blanks(char *s)
{
	size_t	 i;

	i = 0;
	while (s[i] != '\0') {
		if (s[i] == HARD_BLANK) {
			s[i] = SP;
		}
		i++;
	}
}


/*
 * Returns non-zero if and only if f returns HARD_BLANK for no case when called
 * for the n x AABC_FONT_HEIGHT cells at p[i] and b->a.  Used to test whether
 * corresponding cells can be "smushed"/kerned.
 */
static bool
can_m_cols(char (*f)(char, char), char *p[], size_t i, const struct big_ch *b,
    size_t n)
{
	size_t	 x;
	size_t	 y;
	bool	 ok;

	ok = true;


	y = 0;
	while (y < AABC_FONT_HEIGHT && ok) {
		x = 0;
		while (x < n && ok) {
			ok = f(p[y][i + x], b->a[y][x]) != HARD_BLANK;
			x++;
		}
		y++;
	}

	return ok;
}


/*
 * Determine the two pieces of information required for adding a new
 * (big) character:
 *     - the ("smushing"/kerning) offset with existing text (with zero
 *       indicating that no "smushing"/kerning is possible)
 *     - if the offset is non-zero, whether "smushing" (rather than just
 *       kerning) is possible
 */
static void
determine_append_info(char *p[], size_t x, const struct big_ch *b,
    size_t *offset, bool *smush)
{
	size_t	 o;

	o = 1;
	while (o < x && o <= b->w && can_m_cols(kern_ch, p, x - o, b, o)) {
		o++;
	}

	if (o >= x || o > b->w) {
		/* Limit of space available or char width has been reached. */
		*offset = o - 1;
		*smush  = false;
	} else if (can_m_cols(smush_ch, p, x - o, b, o <= b->w ? o : b->w)) {
		/* "Smushing" is possible. */
		*offset = o;
		*smush  = true;
	} else {
		/* "Smushing" is not possible. */
		*offset = o - 1;
		*smush  = false;
	}
}


/*
 * Copies to buf (of size buf_sz, which must be sufficient) the big character
 * representation of s with "smushing"/kerning.
 */
static void
make_smush(const char *s, char *buf, __unused size_t buf_sz)
{
	char	*p[AABC_FONT_HEIGHT];
	size_t	 len;
	size_t	 c, h, i, x, y;
	size_t	 offset;
	bool	 can_smush;

	/* Set up pointers for each of the rows of the (intermediate) matrix. */
	len = line_len_no_smush(s);
	for (y = 0; y < AABC_FONT_HEIGHT; y++) {
		p[y] = buf + y * len;
	} /* for */;

	x = 0;  /* No cells filled so far. */

	/* Process each character of the input string. */
	i = 0;
	while (s[i] != '\0') {
		if ((c = ch_index(s[i])) != CH_IGNORE) {
			determine_append_info(p, x, &font[c], &offset,
			    &can_smush);

			/* Add/merge/"smush" the new big character. */
			for (y = 0; y < AABC_FONT_HEIGHT; y++) {
				/* Process merged cells */
				for (h = 0; h < offset && h < font[c].w; h++) {
					p[y][x - offset + h]
					    = can_smush
					    ? smush_ch(p[y][x - offset + h],
					    font[c].a[y][h])
					    : kern_ch(p[y][x - offset + h],
					    font[c].a[y][h]);
				} /* for */;

				/* Append non-merged cells. */
				for (h = offset; h < font[c].w; h++) {
					p[y][x - offset + h] = font[c].a[y][h];
				}
			} /* for */;

			/*
			 * If new cells have been filled, increase the
			 * row length count accordingly.
			 */
			if (offset < font[c].w) {
				x += font[c].w - offset;
			}
		}
		i++;
	}

	/* Append newline to each line of the intermediate matrix. */
	for (y = 0; y < AABC_FONT_HEIGHT; y++) {
		p[y][x] = '\n';
	}
	x++;

	/*
	 * Note that we do not NUL-terminate each line as there may be
	 * insufficient room.
	 */

	/* Consolidate the buffer into one contiguous string. */
	for (y = 1; y < AABC_FONT_HEIGHT; y++) {
		memmove(buf + y * x, p[y], x);
	}

	/* NUL-terminate the buffer. */
	buf[AABC_FONT_HEIGHT * x] = '\0';

	replace_hard_blanks(buf);
}


/*
 * Copies to buf (of size buf_sz, which must be sufficient) the big character
 * representation of s without "smushing"/kerning.
 */
static void
make_no_smush(const char *s, char *buf, size_t buf_sz)
{
	char	*p;
	size_t	 c, i, y;

	p = buf;

	for (y = 0; y < AABC_FONT_HEIGHT; y++) {
		/* Process one row. */
		i = 0;
		while (s[i] != '\0') {
			if ((c = ch_index(s[i])) != CH_IGNORE) {
				/* Insert space elided in conversion to C. */
				*p++ = SP;
				strlcpy(p, font[c].a[y], buf_sz - (p - buf));
				p += font[c].w;
			}
			i++;
		}
		/* Add newline. */
		*p++ = '\n';
	} /* for */;

	/* NUL-terminate. */
	*p++ = '\0';

	replace_hard_blanks(buf);
}


void
aabc_make(const char *s, bool smush, char *buf, size_t buf_sz)
{
	if (buf_sz < aabc_req_buf_sz_no_smush(s)) {
		/* nbg */
		buf[0] = '\0';
	} else if (smush) {
		make_smush(s, buf, buf_sz);
	} else {
		make_no_smush(s, buf, buf_sz);
	}
}

