/* $Id: captcha.c,v 1.17 2017/08/05 11:45:39 rlr Exp $ */

/*
 * Copyright (c) 2006 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*
 * Copyright (c) 2016, 2017 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "base64.h"
#include "captcha.h"
#include "fdk_hmac.h"
#include "aabc.h"


#ifdef PRINT_ERR_MESSAGES
#define MSG(...)  fprintf(stderr, __VA_ARGS__)
#else
#define MSG(...)
#endif


#define FMT_MSG  "%s%s"
#define FMT_SUM  "%s:%s"

#define TINTERVAL_EXPIRY    (15 * 60)
#define TINTERVAL_TOO_SOON  5


/* Function prototypes. */
static bool	 is_all_digits(const char *, size_t);


int
generate_captcha(const char *kf, size_t len, char *buf, size_t buflen,
    char *vs, size_t vs_sz)
{
	const char	*charset = "abcdefghijkmnpqrstuvwxyz23456789";
	char	 m[256], h[FDK_HMAC_B64_SIZE], ts[32], w[CAPTCHA_LEN_MAX + 1];
	time_t	 now;
	size_t	 charset_len, i;
	int	 err;

	err = CAPTCHA_ERR_NONE;

	/* Create string of characters randomly chose from our charset. */
	charset_len = strlen(charset);
	for (i = 0; i < len; ++i) {
		w[i] = charset[arc4random_uniform(charset_len)];
	}
	w[i] = '\0';

	now = time(NULL);
	if (now == -1) {
		err = CAPTCHA_ERR_CLOCK;
	} else {
		/* Get string version of the clock. */
		snprintf(ts, sizeof(ts), "%lld", now);

		/* Create message from time and CAPTCHA strings. */
		snprintf(m, sizeof(m), FMT_MSG, ts, w);

		/* Get HMAC of message. */
		fdk_hmac_generate_b64(kf, m, strlen(m), h);

		if (h[0] == '\0') {
			MSG("%s() error: fdk_hmac_generate_b64() failed",
			    __func__);
			err = CAPTCHA_ERR_HMAC;
		} else {
			snprintf(vs, vs_sz, FMT_SUM, ts, h);
			aabc_make(w, AABC_SMUSH, buf, buflen);

			explicit_bzero(w, sizeof(w));
			explicit_bzero(m, sizeof(m));
		}
	}

	return err;
}


int
verify_captcha(const char *kf, const char *reply, const char *vs)
{
	char	 m[64], d[32];
	const char	 *es;
	char	*p;
	time_t	 now, t;

	if (!captcha_is_well_formed_vs(vs)) {
		MSG("%s() warning: vs not well-formed", __func__);
		return CAPTCHA_ERR_VS_NOT_WF;
	}

	/*
	 * As we know the verification string is well-formed, we can skip
	 * many checks.
	 */
	p = strchr(vs, ':');
	strlcpy(d, vs, p - vs + 1);

	p++;

	if ((t = strtonum(d, 0, LLONG_MAX, &es)) == 0 && es != NULL) {
		MSG("%s() warning: invalid time", __func__);
		return CAPTCHA_ERR_VS_NOT_WF;
	}

	snprintf(m, sizeof(m), FMT_MSG, d, reply);

	if (!fdk_hmac_validate_b64(kf, m, strlen(m), p)) {
		MSG("%s() warning: mismatch\n", __func__);
		return CAPTCHA_ERR_MISMATCH;
	}

	now = time(NULL);
	if (now == -1) {
		return CAPTCHA_ERR_CLOCK;
	}

	if (t < now - TINTERVAL_EXPIRY || t >= now - TINTERVAL_TOO_SOON) {
		MSG("%s() warning: time out of range - (given) %s versus (now)"
		    " %lld", __func__, vs, now);
		return CAPTCHA_ERR_TIME;
	}

	return CAPTCHA_ERR_NONE;
}


static bool
is_all_digits(const char *s, size_t len)
{
	size_t	 i;

	i = 0;
	while (i < len && isdigit((unsigned char)s[i])) {
		i++;
	}

	return i == len;
}


bool
captcha_is_well_formed_vs(const char *vs)
{
	const char	*h;
	size_t	 len;

	return vs != NULL && (h = strchr(vs, ':')) != NULL
	    && (len = h - vs) >= 10 && len <= 30 && is_all_digits(vs, len)
	    && strlen(h + 1) == FDK_HMAC_B64_LEN
	    && base64_is_well_formed(h + 1);
}


