/* $Id: fdk_hmac.c,v 1.6 2017/08/05 12:05:10 rlr Exp $ */

/*
 * Copyright (c) 2016, 2017 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <openssl/hmac.h>
#include <sha2.h>
#include <string.h>

#include "fdk_hmac.h"
#include "base64.h"


void
fdk_hmac_generate(const char *kfp, const void *m, size_t m_sz,
    unsigned char *hmac, size_t *hmac_sz)
{
	char	 key[SHA256_DIGEST_STRING_LENGTH];
	unsigned int	 len;

	if (SHA256File(kfp, key) == NULL) {
		*hmac_sz = 0;
	} else {
		if (HMAC(EVP_sha256(), key, strlen(key),
		    (const unsigned char *)m, m_sz, hmac, &len) == NULL) {
			/* Failed. */
			*hmac_sz = 0;
		} else {
			*hmac_sz = len;
		}

		/* Clean up. */
		explicit_bzero(key, sizeof(key));
	}
}


void
fdk_hmac_generate_b64(const char *kfp, const void *m, size_t m_sz, char *b64)
{
	unsigned char	 md[FDK_HMAC_BIN_SIZE];
	size_t	 sz;

	fdk_hmac_generate(kfp, m, m_sz, md, &sz);

	if (sz != FDK_HMAC_BIN_SIZE) {
		/* Failure occurred. */
		b64[0] = '\0';
	} else if (base64_encode(md, sz, b64, FDK_HMAC_B64_SIZE) == -1) {
		/* Failure in base64 conversion. */
		explicit_bzero(b64, FDK_HMAC_B64_SIZE);
	}

	explicit_bzero(md, sizeof(md));
}


bool
fdk_hmac_validate(const char *kfp, const void *m, size_t m_sz,
    const unsigned char *d)
{
	unsigned char	 md[FDK_HMAC_BIN_SIZE];
	size_t	 sz;
	bool	 valid;

	fdk_hmac_generate(kfp, m, m_sz, md, &sz);

	valid = sz == FDK_HMAC_BIN_SIZE
	    && timingsafe_bcmp(md, d, FDK_HMAC_BIN_SIZE) == 0;

	explicit_bzero(md, sizeof(md));

	return valid;
}


bool
fdk_hmac_validate_b64(const char *kfp, const void *m, size_t m_sz,
    const char *b)
{
	unsigned char	 md[FDK_HMAC_BIN_SIZE];
	bool	 valid;

	valid = b != NULL && strlen(b) == FDK_HMAC_B64_LEN
	    && base64_decode(b, md, sizeof(md)) == FDK_HMAC_BIN_SIZE
	    && fdk_hmac_validate(kfp, m, m_sz, md);

	explicit_bzero(md, sizeof(md));

	return valid;
}


