/* $Id: file_utils.c,v 1.5 2017/08/05 12:12:49 rlr Exp $ */

/*
 * Copyright (c) 2016, 2017 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <fts.h>
#include <unistd.h>

#include "file_utils.h"


int
touch_file(const char *path, time_t t)
{
	struct	stat	 info;
	struct	timespec	 up[2];
	int	 err;
	int	 fd;

	err = 0;

	/* Open the file, creating it if it does not exist. */
	fd = open(path, O_WRONLY | O_CREAT, DEFFILEMODE);
	if (fd == -1) {
		err = errno;
	} else if (fstat(fd, &info) == -1) {
		err = errno;
		(void)close(fd);
	} else if (info.st_mtim.tv_sec >= t) {
		/* Modification time no earlier than t, so just close. */
		if (close(fd) == -1) {
			err = errno;
		}
	} else {
		/* Need to update modification time. */
		up[0].tv_sec = t;
		up[0].tv_nsec = UTIME_OMIT;  /* Do not update access time. */
		up[1].tv_sec = t;
		up[1].tv_nsec = 0;  /* _Do_ update modification time. */

		if (futimens(fd, up) == -1) {
			err = errno;
		}

		(void)close(fd);
	}

	return err;
}


FILE *
fopen_temp(char *template, const char *mode)
{
	FILE	*f;
	int	 fd;

	if ((fd = mkstemp(template)) == -1 || (f = fdopen(fd, mode)) == NULL) {
		if (fd != -1) {
			(void)unlink(template);
			(void)close(fd);
		}
		f = NULL;
	}

	return f;
}


int
delete_dir(char *path)
{
	FTS     *hier;
	FTSENT  *ent;
	char    *paths[2];
	int	 e;

	paths[0] = path;
	paths[1] = NULL;

	e = 0;

	hier = fts_open(paths, (FTS_LOGICAL | FTS_NOSTAT), NULL);
	if (hier == NULL) {
		e = errno;
	} else {
		while (e == 0 && (ent = fts_read(hier)) != NULL) {
			if (ent->fts_info == FTS_DNR
			    || ent->fts_info == FTS_ERR
			    || ent->fts_info == FTS_NS) {
				e = ent->fts_errno;
			} else if (ent->fts_info == FTS_D) {
				/* This is a directory, seen in pre-order. */
			} else if (ent->fts_info == FTS_DP) {
				/* This is a directory, seen in post-order. */
				if (rmdir(ent->fts_path) != 0) {
					e = errno;
				}
			} else {
				/* It is some other file. */
				if (unlink(ent->fts_path) != 0) {
					e = errno;
				}
			}
		}

		if (fts_close(hier) != 0) {
			if (e == 0) {
				e = errno;
			}
		}
	}

	return e;
}

