/* $Id: html_utils.c,v 1.2 2017/08/05 12:17:42 rlr Exp $ */

/*
 * Copyright (c) 2017 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "html_utils.h"


/* Paragraph/line-break states. */
enum tth_state {
	TTH_START,
	TTH_IN_P,
	TTH_BR,
	TTH_EOP
};


void
text_to_html(const char *t, bool nlx, char *h, size_t *sz)
{
	size_t	 i;
	size_t	 j;
	enum tth_state	 st;

	if (t == NULL) {
		*sz = 0;	
		if (h != NULL) {
			*h = '\0';
		}
	} else {
		/*
		 * Output of '\n' to <br> and <p>/</p> conversion is performed
		 * lazily, so that the minimum required markup is emitted.
		 */

		st = TTH_START;

		i = 0;
		j = 0;

		while (t[i] != '\0') {
			if (nlx && t[i] != '\r') {
				switch (st) {

				case TTH_START :
					if (t[i] != '\n') {
						/* Lazy start of paragraph. */
						if (h != NULL) {
							strlcpy(&h[j], "<p>",
							    4);
						}
						j += 3;
						st = TTH_IN_P;
					}
				break;

				case TTH_IN_P :
					if (t[i] == '\n') {
						st = TTH_BR;
					}
				break;

				case TTH_BR :
					if (t[i] == '\n') {
						st = TTH_EOP;
					} else {
						if (h != NULL) {
							/* Lazy linebreak. */
							strlcpy(&h[j],
							    "<br>\n", 6);
						}
						j += 5;
						st = TTH_IN_P;
					}
				break;

				case TTH_EOP :
					if (t[i] != '\n') {
						if (h != NULL) {
							/*
							 * Lazy end/start para.
							 */
							strlcpy(&h[j],
							    "</p>\n<p>", 9);
						}
						j += 8;
						st = TTH_IN_P;
					}
				break;

				}
			}

			switch (t[i]) {

			case '"' :
				if (h != NULL) {
					strlcpy(&h[j], "&quot;", 7);
				}
				j += 6;
			break;

			case '&' :
				if (h != NULL) {
					strlcpy(&h[j], "&amp;", 6);
				}
				j += 5;
			break;

			case '<' :
				if (h != NULL) {
					strlcpy(&h[j], "&lt;", 5);
				}
				j += 4;
			break;

			case '>' :
				if (h != NULL) {
					strlcpy(&h[j], "&gt;", 5);
				}
				j += 4;
			break;

			default :
				if (!nlx || (t[i] != '\n' && t[i] != '\r')) {
					if (h != NULL) {
						h[j] = t[i];
					}
					j += 1;
				}
			break;

			}

			i++;
		}

		if (nlx) {
			/* Complete lazy markup. */
			if (st == TTH_START) {
				if (h != NULL) {
					strlcpy(&h[j], "<p>", 4);
				}
				j += 3;
			}
			if (h != NULL) {
				strlcpy(&h[j], "</p>\n", 6);
			}
			j += 5;
		}

		if (h != NULL) {
			h[j] = '\0';
		}
		j += 1;

		*sz = j;
	}
}


char *
alloc_text_to_html(const char *t, bool nlx)
{
	char	*s;
	size_t	 n;

	s = NULL;

	if (t != NULL) {
		/* Determine required buffer size. */
		text_to_html(t, nlx, NULL, &n);

		if (n > 0 && (s = malloc(n)) != NULL) {
			/* Perform conversion. */
			text_to_html(t, nlx, s, &n);
		}
	}

	return s;
}


