/*	$Id: mail.c,v 1.6 2017/07/19 00:54:43 rlr Exp $ */

/*
 * Copyright (c) 2004-2006 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * Reworked by Ross Richardson in 2016, 2017.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mail.h"


#ifdef PRINT_ERR_MESSAGES
#define MSG(...)  fprintf(stderr, __VA_ARGS__)
#else
#define MSG(...)
#endif


int
mail_send(int (*cb)(const int, void *), void *a)
{
	struct	timespec	 rqt;
	char	*argv[] = { MAIL_MTA, "-t", NULL };
	int	 fd[2], r, status;
	pid_t	 pid;
	time_t	 t;

	if (cb == NULL && a == NULL) {
		MSG("%s() error: cb and a both NULL", __func__);
		return MAIL_ERR_INVALID;
	}
	if (pipe(fd)) {
		MSG("%s() error: pipe: %s", __func__, strerror(errno));
		return MAIL_ERR_PIPE;
	}
	if ((pid = fork()) < 0) {
		MSG("%s() error: fork: %s", __func__, strerror(errno));
		close(fd[0]);
		close(fd[1]);
		return MAIL_ERR_FORK;
	}
	if (!pid) {
		close(fd[1]);
		dup2(fd[0], STDIN_FILENO);
		dup2(open("/dev/null", O_WRONLY, 0), STDOUT_FILENO);
		dup2(open("/dev/null", O_WRONLY, 0), STDERR_FILENO);
		execv(argv[0], argv);
		exit(EXIT_FAILURE);
	}
	close(fd[0]);

	if (cb != NULL) {
		if (cb(fd[1], a) != 0) {
			MSG("%s() error: callback reported failure", __func__);
			close(fd[1]);
			return MAIL_ERR_WRITE;
		}
	} else /* (if a != NULL) */ {
		if (write(fd[1], a, strlen((const char *)a)) == -1) {
			MSG("%s() error: write: %s", __func__,
			    strerror(errno));
			close(fd[1]);
			return MAIL_ERR_WRITE;
		}
	}
	close(fd[1]);

	rqt.tv_sec = 0;
	rqt.tv_nsec = 10000 * 1000;
	t = time(NULL) + 10;
	do {
		(void)nanosleep(&rqt, NULL);
		r = waitpid(pid, &status, WNOHANG);
		if (r < 0) {
			if (errno != EINTR && errno != EAGAIN) {
				MSG("%s() error: waitpid: %s", __func__,
				    strerror(errno));
			} else {
				r = 0;
			}
		}
	} while (r == 0 && time(NULL) < t);
	if (r == 0) {
		MSG("%s() error: timeout waiting for child %u", __func__,
		    (unsigned)pid);
		return MAIL_ERR_WAIT;
	}
	if (r != pid) {
		MSG("%s() error: waitpid returned %u for child %u", __func__,
		    (unsigned)r, (unsigned)pid);
		return MAIL_ERR_WAIT;
	}
	if (!WIFEXITED(status)) {
		MSG("%s() error: abnormal exit of child %u", __func__,
		    (unsigned)pid);
		return MAIL_ERR_WAIT;
	}
	if (WEXITSTATUS(status)) {
		MSG("%s() error: exit status %d from child %u", __func__,
		    (int)WEXITSTATUS(status), (unsigned)pid);
		return MAIL_ERR_WAIT;
	}

	return MAIL_ERR_NONE;
}

