/* $Id: rhf.c,v 1.4 2017/08/05 12:23:44 rlr Exp $ */

/*
 * Copyright (c) 2001, 2017 Ross L Richardson <ross@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <ctype.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "rhf.h"


/* Characters of note in html_print_cleansed() */
#define EXCLAM       '!'
#define QUOTE        '"'
#define DASH         '-'
#define SLASH        '/'
#define ANGLE_OPEN   '<'
#define EQUALS       '='
#define ANGLE_CLOSE  '>'

/* FSM states for html_print_cleansed() */
enum state {
	S_NBG,
	S_START,
	S_BEGIN_MARKUP,
	S_COMMENT,
	S_ELEMENT_NAME,
	S_PRE_ATTRIBUTE,
	S_ATTRIBUTE_NAME,
	S_POST_ATTRIBUTE_NAME,
	S_ATTRIBUTE_EQUALS,
	S_ATTRIBUTE_VALUE,
	S_END_ATTRIBUTE_VALUE,
	S_BODY
};

/* Representation of HTML elements. */
enum element {
	EL_A,
	EL_B,
	EL_BR,
	EL_EM,
	EL_I,
	EL_P,
	EL_PRE,
	EL_STRONG,
	EL__NBG
};

/* The HTML elements acceptable to the validator. */
static const char	*element_name[] = {
	"a",
	"b",
	"br",
	"em",
	"i",
	"p",
	"pre",
	"strong"
};

#define NUM_ELEMENTS  (sizeof(element_name) / sizeof(element_name[0]))

/* The HTML character references acceptable to the validator. */
static const char	*char_ref[] = {
	"&amp;",
	"&gt;",
	"&lt;",
	"&quot;",
	NULL
};

/* Function prototypes */
static size_t	 check_attribute(const char *);
static enum element	 find_element(const char *);
static int	 is_on_stack(enum element, const enum element st[], size_t)
    __attribute__ ((__bounded__ ( __buffer__, 2, 3 )));
static size_t	 match_char_ref(const char *);


/*
 * Attempts to match the initial substring of s with an acceptable HTML
 * character reference (in char_ref[]).  If matched successfully, the
 * length of the match is returned; otherwise, zero is returned.
 */
static size_t
match_char_ref(const char *s)
{
	size_t	 i;

	i = 0;
	while (char_ref[i] != NULL
	    && strncmp(s, char_ref[i], strlen(char_ref[i])) != 0) {
		i++;
	}

	return char_ref[i] == NULL ? 0 : strlen(char_ref[i]);
}


/*
 * Attempts to match NUL-terminated e with an acceptable HTML element (in
 * element_name[].  Returns the corresponding element on a match, or
 * EL_NBG if e does not match any of the elements.
 */
static enum element
find_element(const char *e)
{
	size_t	 n;

	n = 0;
	while (n < NUM_ELEMENTS && strcmp(e, element_name[n]) != 0) {
		n++;
	}

	return (enum element)n;
}


/* Returns non-zero if and only if one of the first n elements of st is e. */
static int
is_on_stack(enum element e, const enum element st[], size_t n)
{
	while (n > 0 && st[n - 1] != e) {
		n--;
	}

	return n > 0;
}


/*
 * Attempts to parse the initial substring of s for
 *   href sp '=' sp '"' url '"' sp '>'
 * where
 *   href is case insensitive "href"
 *   sp is zero or more isspace() characters
 *   url is string with char refs for '"', '&', '<', '>'
 *   things in single quotes are literals
 * Note that the string is expected to start with href.
 *
 * On success, returns the length of the matching string;
 * otherwise, returns zero.
 */
static size_t
check_attribute(const char *s)
{
	size_t	 n, r;

	n = 0;

	/* Deal with expected "href". */
	if (strncasecmp(&s[n], "href", 4) != 0) {
		return 0;
	}
	n += 4;

	/* Deal with any whitespace. */
	while (isspace((unsigned char)s[n])) {
		n++;
	}

	/* Deal with expected '=' */
	if (s[n] == '\0') {
		return 0;
	} else if (s[n] != '=') {
		return 0;
	}
	/* It must be '-'. */
	n++;

	/* Deal with any whitespace. */
	while (isspace((unsigned char)s[n])) {
		n++;
	}

	/* Deal with expected '"' */
	if (s[n] == '\0') {
		return 0;
	} else if (s[n] != '"') {
		return 0;
	}
	/* It must be '"'. */
	n++;

	/* Deal with value of href attribute. */
	while (s[n] != '\0' && s[n] != '"') {
		if (s[n] == '<' || s[n] == '>') {
			return 0;
		} else if (s[n] == '&') {
			r = match_char_ref(&s[n]);
			if (r == 0) {
				/* Failed to match. */
				return 0;
			}
			n += r - 1;
		}
		n++;
	}

	if (s[n] == '\0') {
		return 0;
	}

	/* Must be '"', so skip it. */
	n++;

	/* Deal with any whitespace. */
	while (isspace((unsigned char)s[n])) {
		n++;
	}

	if (s[n] == '\0') {
		return 0;
	} else if (s[n] != '>') {
		return 0;
	}

	/* Account for the '>'. */
	n++;

	return n;
}


void
rhf_validate(const char *m, int *status, size_t *pos)
{
	char	 ename[8];
	size_t	 j;
	size_t	 skip, stack_depth;
	/* NUM_ELMENTS is sufficent because not all nest. */
	enum element	 stack[NUM_ELEMENTS];
	enum element	 elmt;
	bool	 empty_a, is_closing, in_a, in_tag;

	stack_depth = 0;
	in_tag = false;
	in_a = false;
	empty_a = false;
	is_closing = false;
	*status = RHF_OK;
	*pos = 0;
	/*
	 * Alternate attempting to parse non-tag and tag until end-of-string
	 * or error.
	 */
	while (*status == RHF_OK && m[*pos] != '\0') {
		/* Deal with non-tag data. */
		while (m[*pos] != '\0' && m[*pos] != '<') {
			if (m[*pos] == '"' || m[*pos] == '>') {
				*status = RHF_CHAR_REF_REQ;
				return;
			} else {
				if (in_a) {
					empty_a = false;
				}
				if (m[*pos] == '&') {
					j = match_char_ref(&m[*pos]);
					if (j == 0) {
						/* Failed to match. */
						*status = RHF_UNKNOWN_CHAR_REF;
						return;
					}
					(*pos) += j - 1;
				}
			}
			(*pos)++;
		}

		if (m[*pos] == '\0') {
			break;
		}

		/* Deal with a tag. */
		(*pos)++;  /* Skip '<'. */
		in_tag = true;
		is_closing = false;

		/* Whitespace is forbidden here. */

		if (m[*pos] == '\0') {
			*status = RHF_UNCLOSED_TAG;
			return;
		} else if (m[*pos] == '/') {
			if (stack_depth == 0) {
				*status = RHF_UNBAL_CLOSE_ELEMENT;
				return;
			}
			is_closing = true;
			(*pos)++;
		}

		/* Whitespace is forbidden here. */

		/* Determine type of element. */
		j = 0;
		while (isalnum((unsigned char)m[*pos])
		    && j < sizeof(ename) - 1) {
			ename[j++] = tolower((unsigned char)m[(*pos)++]);
		}
		if (m[*pos] == '\0') {
			/* Reached EOS without '>'. */
			*status = RHF_UNCLOSED_TAG;
			return;
		} else if (j >= sizeof(ename) - 1) {
			/* Name is longer than that of any accepted element. */
			*status = RHF_UNKNOWN_ELEMENT;
			return;
		} else if (!isspace((unsigned char)m[*pos])
		    && m[*pos] != '>') {
			/* Invalid character found. */
			*status = RHF_SYNTAX_ERR;
			return;
		} else {
			ename[j] = '\0';
			elmt = find_element(ename);
			if (elmt == EL__NBG) {
				*status = RHF_UNKNOWN_ELEMENT;
				return;
			} else if (is_closing) {
				if (elmt == EL_BR) {
					*status = RHF_INVAL_CLOSE_ELEMENT;
					return;
				} else if (stack[stack_depth - 1] != elmt) {
					*status = RHF_NONMATCH_CLOSE_ELEMENT;
					return;
				} else if (elmt == EL_A && empty_a) {
					*status = RHF_EMPTY_ANCHOR;
					return;
				} else {
					if (elmt == EL_A) {
						in_a = false;
						empty_a = true;
					}
					/* Pop from stack. */
					stack_depth--;
				}
			} else if (elmt == EL_BR) {
				/* Don't push on stack. */
			} else if (elmt == EL_P) {
				if (stack_depth > 0) {
					if (stack[stack_depth - 1] == EL_P) {
						/* OK, but don't push. */
					} else {
						/* Nested <p> forbidden. */
						*status = RHF_BAD_HIER;
						return;
					}
				} else {
					/* Push on stack. */
					stack[stack_depth++] = elmt;
				}
			} else if (elmt == EL_PRE) {
				if (stack_depth == 0) {
					/* Push on stack. */
					stack[stack_depth++] = elmt;
				} else if (stack_depth == 1
				    && stack[0] == EL_P) {
					/* Pop then push (effectively). */
					stack[0] = EL_PRE;
				} else {
					/* Nested <pre> forbidden. */
					*status = RHF_BAD_HIER;
					return;
				}
			} else if (is_on_stack(elmt, stack, stack_depth)) {
				/* Element nested within same element. */
				*status = RHF_BAD_HIER;
				return;
			} else {
				/* Push on stack. */
				stack[stack_depth++] = elmt;
				if (elmt == EL_A) {
					in_a = true;
					empty_a = true;
				}
			}
		}
		while (isspace((unsigned char)m[*pos])) {
			(*pos)++;
		}
		if (m[*pos] == '\0') {
			*status = RHF_UNCLOSED_TAG;
			return;
		} else if (m[*pos] == '>') {
			/* No attribute(s) found, so fail if <a> */
			if (elmt == EL_A && !is_closing) {
				*status = RHF_MISSING_ATTR;
				return;
			} else {
				/* Skip the '>'. */
				(*pos)++;
				in_tag = false;
			}
		} else {
			/* Not EOS, '>', or isspace(). */
			if (elmt == EL_A && !is_closing) {
				skip = check_attribute(&m[*pos]);
				if (skip == 0) {
					*status = RHF_BAD_ATTR;
					return;
				} else {
					(*pos) += skip;
				}
			} else {
				*status = RHF_ATTR_NOT_PERM;
				return;
			}
		}
	}

	if (stack_depth != 0 && stack[0] != EL_P) {
		*status = RHF_UNBAL_OPEN_ELEMENT;
	}
}


void
rhf_print_cleansed(FILE *f, const char *s, const char *inj_a)
{
	size_t	 i;
	int	 is_a, is_closing;
	enum state	 st;

	/* Initialise to avoid compiler warnings. */
	is_a = 0;
	is_closing = 0;

	st = S_START;  /* Initialise FSM. */
	i = 0;
	while (s[i] != '\0' && st != S_NBG) {
		switch (st) {

		case S_START :
		case S_BODY :
			switch (s[i]) {

			case ANGLE_OPEN :
				putc(s[i], f);
				st = S_BEGIN_MARKUP;
				is_closing = 0;  /* To be determined. */
			break;

			case ANGLE_CLOSE :
				st = S_NBG;
			break;

			default :
				putc(s[i], f);
				st = S_BODY;
			break;

			}
		break;

		case S_BEGIN_MARKUP :
			switch (s[i]) {

			case ANGLE_OPEN :
			case ANGLE_CLOSE :
			case QUOTE :
				st = S_NBG;
			break;

			case SLASH :
				if (is_closing) {
					/* SLASH already seen. */
					st = S_NBG;
				} else {
					putc(s[i], f);
					/*
					 * Strictly, we should have a separate
					 * state.  However, as is_closing
					 * records that the SLASH has been
					 * seen, we can do without.
					 */
					/* st = S_BEGIN_MARKUP; */
					is_closing = 1;
				}
			break;

			case EXCLAM :
				putc(s[i], f);
				st = S_COMMENT;
			break;

			default :
				if (!isalpha((unsigned char)s[i])) {
					st = S_NBG;
				} else {
					putc(tolower((unsigned char)s[i]), f);
					st = S_ELEMENT_NAME;
					/*
					 * Determine (pedantically) whether
					 * this is an 'a' element.
					 */
					is_a = tolower((unsigned char)s[i])
					    == 'a' && (s[i + 1] == ANGLE_CLOSE
					    || isspace((unsigned char)
					    s[i + 1]));
				}
			break;

			}
		break;

		case S_COMMENT :
			switch (s[i]) {

			case ANGLE_OPEN :
				st = S_NBG;
			break;

			case ANGLE_CLOSE :
				putc(s[i], f);
				st = S_BODY;
			break;

			default :
				putc(s[i], f);
				/* st = S_COMMENT; */
			break;

			}
		break;

		case S_ELEMENT_NAME :
			switch (s[i]) {

			case ANGLE_OPEN :
			case EQUALS :
			case QUOTE :
				st = S_NBG;
			break;

			case ANGLE_CLOSE :
				putc(s[i], f);
				st = S_BODY;
			break;

			default :
				if (isalnum((unsigned char)s[i])
				    || s[i] == DASH) {
					putc(tolower(s[i]), f);
					/* st = S_ELEMENT_NAME; */
				} else if (isspace((unsigned char)s[i])) {
					st = S_PRE_ATTRIBUTE;
				} else {
					st = S_NBG;
				}
			break;

			}
		break;

		case S_PRE_ATTRIBUTE :
			switch (s[i]) {

			case ANGLE_OPEN :
			case QUOTE :
				st = S_NBG;
			break;

			case ANGLE_CLOSE :
				if (inj_a != NULL && is_a && !is_closing) {
					fprintf(f, " %s", inj_a);
				}
				putc(s[i], f);
				st = S_BODY;
			break;

			default :
				if (isspace((unsigned char)s[i])) {
					/* st = S_PRE_ATTRIBUTE; */
				} else {
					putc(' ', f);
					putc(tolower((unsigned char)s[i]), f);
					st = S_ATTRIBUTE_NAME;
				}
			break;

			}
		break;

		case S_ATTRIBUTE_NAME :
		case S_POST_ATTRIBUTE_NAME :
			switch (s[i]) {

			case ANGLE_OPEN :
			case ANGLE_CLOSE :
			case QUOTE :
				st = S_NBG;
			break;

			case EQUALS :
				putc(s[i], f);
				st = S_ATTRIBUTE_EQUALS;
			break;

			default :
				if (st == S_ATTRIBUTE_NAME
				    && (isalnum((unsigned char)s[i])
				    || s[i] == DASH)) {
					putc(tolower((unsigned char)s[i]), f);
					/* st = S_ATTRIBUTE_NAME; */
				} else if (isspace((unsigned char)s[i])) {
					/* Elide whitespace. */
					st = S_POST_ATTRIBUTE_NAME;
				} else {
					st = S_NBG;
				}
			break;

			}
		break;

		case S_ATTRIBUTE_EQUALS :
			switch (s[i]) {

			case QUOTE :
				putc(s[i], f);
				st = S_ATTRIBUTE_VALUE;
			break;

			default :
				if (!isspace((unsigned char)s[i])) {
					st = S_NBG;
				} else {
					/*
					 * Elide the whitespace and just keep
					 * the same state.
					 */
					/* st = S_ATTRIBUTE_EQUALS; */
				}
			break;

			}
		break;

		case S_ATTRIBUTE_VALUE :
			switch (s[i]) {

			case ANGLE_OPEN :
			case ANGLE_CLOSE :
				st = S_NBG;
			break;

			case QUOTE :
				putc(s[i], f);
				st = S_END_ATTRIBUTE_VALUE;
			break;

			default :
				putc(s[i], f);
				/* st = S_ATTRIBUTE_VALUE; */
			break;

			}
		break;

		case S_END_ATTRIBUTE_VALUE :
			switch (s[i]) {

			case ANGLE_OPEN :
				st = S_NBG;
			break;

			case ANGLE_CLOSE :
				if (inj_a != NULL && is_a && !is_closing) {
					fprintf(f, " %s", inj_a);
				}
				putc(s[i], f);
				st = S_BODY;
			break;

			default :
				if (isspace((unsigned char)s[i])) {
					/* Elide whitespace. */
					st = S_PRE_ATTRIBUTE;
				} else {
					st = S_NBG;
				}
			break;

			}
		break;

		default :
			/* Should not happen. */
			st = S_NBG;
		break;

		}

		i++;
	}
	if (s[i] != '\0') {
		/* State must be S_NBG. */
		fprintf(f, "[INTERNAL ERROR: \"%s\"]", &s[i]);
	}
}

