/* $Id: tmpl.c,v 1.3 2017/01/19 03:20:39 rlr Exp rlr $ */

/*
 * Copyright (c) 2017 Ross Richardson <ross@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/mman.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "tmpl.h"


/* Arg for the C string wrappers. */
struct repl_cs_info {
	int (*cb)(const char *, void *);  /* the callback function */
	void	*arg;                     /* the "real" arg */
};

static int	 cs_wrapper(const void *, size_t, void *);


/*
 * Finite State Machine for tmpl_buf()
 *
 * ========
 * State
 *     input	action	==> new state
 * md : the macro delimiting octet
 * ch is any other octet
 * zz is the end of buffer
 * ========
 *
 * TS_START
 *     zz	-	==> TS_END
 *     md	-	==> TS_PLAIN_MD
 *     ch	-	==> TS_PLAIN
 *
 * TS_PLAIN
 *     zz	emit	==> TS_END
 *     md	-	==> TS_PLAIN_MD
 *     ch	-	-
 *
 * TS_PLAIN_MD
 *     zz	emit	==> TS_END
 *     md	emit	==> TS_MACRO
 *     ch	-	==> TS_PLAIN
 *
 * TS_MACRO
 *     zz	-	==> TS_ERR
 *     md	-	==> TS_MACRO_MD
 *     ch	-	==> -
 *
 * TS_MACRO_MD
 *     zz	-	==> TS_ERR
 *     md	repl	==> TS_PLAIN
 *     ch	-	==> TS_MACRO
 *
 * TS_END
 *
 * TS_ERR
 *
 */

enum fsm_state {
	TS_START,
	TS_PLAIN,       /* outside macro name, normal data */
	TS_PLAIN_MD,    /* outside macro name, have one md */
	TS_MACRO,       /* inside macro name */
	TS_MACRO_MD,    /* inside macro name, have one md */
	TS_END,
	TS_ERR
};


int
tmpl_buf(const void *buf, size_t len, const unsigned char md,
    int (*emit)(const void *, size_t, void *), void *ae,
    int (*repl)(const void *, size_t, void *), void *ar)
{
	const unsigned char	*d;
	size_t	 i, p0, m0;
	int	 result;
	enum fsm_state	 st;

	if (len == 0) {
		result = TMPL_NO_ERR;
	} else if (emit == NULL) {
		result = TMPL_INVALID;
	} else if (repl == NULL) {
		/* NULL macro replacement function, so just emit buf. */
		if (emit(buf, len, ae) != 0) {
			result = TMPL_EMIT_ERR;
		} else {
			result = TMPL_NO_ERR;
		}
	} else {
		d = buf;
		m0 = 0; p0 = 0;  /* Avoid compiler warnings. */

		/* Implement the FSM. */
		result = TMPL_NO_ERR;
		i = 0;
		st = TS_START;
		while (result == TMPL_NO_ERR && st != TS_END && st != TS_ERR) {
			switch (st) {

			case TS_START :
				if (i == len) {
					st = TS_END;
				} else if (d[i] == md) {
					st = TS_PLAIN_MD;
				} else {
					p0 = i;
					st = TS_PLAIN;
				}
			break;

			case TS_PLAIN :
				if (i == len) {
					if (i > p0 && emit(&d[p0], i - p0, ae)
					    != 0) {
						result = TMPL_EMIT_ERR;
					}
					st = TS_END;
				} else if (d[i] == md) {
					st = TS_PLAIN_MD;
				} else {
					/* Just keep going. */
				}
			break;

			case TS_PLAIN_MD :
				if (i == len) {
					if (i > p0 && emit(&d[p0], i - p0, ae)
					    != 0) {
						result = TMPL_EMIT_ERR;
					}
					st = TS_END;
				} else if (d[i] == md) {
					if (i > p0 + 1 && emit(&d[p0],
					    i - 1 - p0, ae) != 0) {
						result = TMPL_EMIT_ERR;
					}
					m0 = i + 1;
					st = TS_MACRO;
				} else {
					st = TS_PLAIN;
				}
			break;

			case TS_MACRO :
				if (i == len) {
					st = TS_ERR;
				} else if (d[i] == md) {
					st = TS_MACRO_MD;
				} else {
					/* Just keep going. */
				}
			break;

			case TS_MACRO_MD :
				if (i == len) {
					st = TS_ERR;
				} else if (d[i] == md) {
					if (repl(&d[m0], i - 1 - m0, ar) != 0)
					    {
						result = TMPL_REPL_ERR;
					}
					p0 = i + 1;
					st = TS_PLAIN;
				} else {
					st = TS_MACRO;
				}
			break;

			default :
				/* Internal error occurred. */
				exit(1);
			break;

			}

			i++;
		}

		if (st == TS_ERR) {
			result = TMPL_INPUT_NOT_WF;
		}
	}

	return result;
}


int
tmpl_file(const char *filepath, const unsigned char md,
    int (*emit)(const void *, size_t, void *), void *ae,
    int (*repl)(const void *, size_t, void *), void *ar)
{
	struct	stat	 st;
	unsigned char	*d;
	int	 fd;
	int	 result;

	if ((fd = openat(AT_FDCWD, filepath, O_RDONLY, 0)) == -1) {
		result = TMPL_OPEN_ERR;
	} else if (fstat(fd, &st) == -1) {
		result = TMPL_STAT_ERR;
	} else if (st.st_size > SSIZE_MAX) {
		result = TMPL_TOO_BIG_ERR;
	} else if (st.st_size <= 0) {
		/* File is empty so there is nothign to do. */
		result = TMPL_NO_ERR;
	} else if ((d = mmap(NULL, (size_t)st.st_size, PROT_READ, MAP_SHARED,
	    fd, 0)) == MAP_FAILED) {
		result = TMPL_MMAP_ERR;
	} else {
		result = tmpl_buf(d, (size_t)st.st_size, md,
		    emit, ae, repl, ar);
		munmap(d, (size_t)st.st_size);
	}

	if (fd != -1) {
		close(fd);
	}

	return result;
}


/* Conversion function for C string wrappers. */
static int
cs_wrapper(const void *mv, size_t len, void *arg)
{
	struct	repl_cs_info	*wrap_info;
	char	*s;
	int	 ret;

	if (mv == NULL) {
		ret = 0;
	} else if (arg == NULL) {
		ret = 1;
	} else if (memchr(mv, '\0', len) != NULL) {
		/* Embedded NUL found, so cannot make a valid C string. */
		ret = 1;
	} else if ((s = strndup((const char *)mv, len)) == NULL) {
		/* Failed to make C string copy of data. */
		ret = 1;
	} else {
		wrap_info = (struct repl_cs_info *)arg;
		if (wrap_info->cb == NULL) {
			/* Caller should prevent this. */
			ret = 1;
		} else {
			ret = wrap_info->cb(s, wrap_info->arg);
		}
		free(s);
	}

	return ret;
}


int
tmpl_buf_cstr(const void *buf, size_t len, const unsigned char md,
    int (*emit)(const void *, size_t, void *), void *ae,
    int (*repl)(const char*, void *), void *ar)
{
	struct	repl_cs_info	 wrap_info;

	if (repl == NULL) {
		return tmpl_buf(buf, len, md, emit, ae, NULL, NULL);
	} else {
		wrap_info.cb = repl;
		wrap_info.arg = ar;

		return tmpl_buf(buf, len, md, emit, ae, cs_wrapper,
		    &wrap_info);
	}
}


int
tmpl_file_cstr(const char *filepath, const unsigned char md,
    int (*emit)(const void *, size_t, void *), void *ae,
    int (*repl)(const char*, void *), void *ar)
{
	struct	repl_cs_info	 wrap_info;

	if (repl == NULL) {
		return tmpl_file(filepath, md, emit, ae, NULL, NULL);
	} else {
		wrap_info.cb = repl;
		wrap_info.arg = ar;

		return tmpl_file(filepath, md, emit, ae, cs_wrapper,
		    &wrap_info);
	}
}


