/*	$Id: user.c,v 1.34 2017/09/09 11:49:35 rlr Exp $ */

/*
 * Copyright (c) 2004-2006 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
*    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * Copyright (c) 2016, 2017 Ross L Richardson <ross@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sha1.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "user.h"


#define PW_UNMATCHABLE  '*'

/*
 * Passwords are currently stored as the SHA-1 digest of the so-called "secret"
 * and the password string.  Accounts are (reversibly) locked by prepending an
 * asterisk.  SHA1_DIGEST_STRING_LENGTH includes space for a terminating NUL,
 * so we just add one (for the possible asterisk) to have room in all cases.
 */
#define PW_STORED_SIZE   (SHA1_DIGEST_STRING_LENGTH + 1)


/* Prototypes */
static int	 user_find_x(const char *, const char *, struct user *,
    bool *);
static void      user_hash(const char *, char *);
static int	 user_update_x(const char *, const struct user *,
    const char *);


/* Secret for hash creation. */
static const char	*secret
    = "lw1N0mN0t8a24UR4NY12AT55p/23FxJ9H6+NReXed8nTZJraO0g";

/* Prefix used to lock password/account. */
static const char	 pw_lock_prfx[] = "*";
/* Value used to indicate that password/account should be unlocked. */
static const char	 pw_unlock[] = "";

/* User database file path.  Set by user_init(). */
static char	 user_db_path[PATH_MAX] = "";


bool
user_is_valid_id(const char *id)
{
	size_t	 n;
	int	 has_alpha;
	bool	 valid;

	if (id == NULL) {
		valid = false;
	} else {
		has_alpha = 0;
		n = 0;
		while (n < USER_ID_MAX_LEN && (isalnum((unsigned char)id[n])
		    || id[n] == '-' || id[n] == '.' || id[n] == '_')) {
			if (isalpha((unsigned char)id[n])) {
				has_alpha = 1;
			}
			n++;
		}
		valid = n >= USER_ID_MIN_LEN && n <= USER_ID_MAX_LEN
		    && id[n] == '\0' && has_alpha;
	}

	return valid;
}


bool
user_is_valid_password(const char *pw)
{
	return pw != NULL && strlen(pw) >= USER_PW_MIN_LEN;
}


bool
user_is_valid_group(const char *g)
{
	size_t	 n;
	bool	 valid;

	if (g == NULL) {
		valid = false;
	} else {
		n = 0;
		while (n < USER_GROUP_MAX_LEN && (isalnum((unsigned char)g[n])
		    || g[n] == '_')) {
			n++;
		}
		valid = n >= USER_GROUP_MIN_LEN && n <= USER_GROUP_MAX_LEN
		    && g[n] == '\0';
	}

	return valid;
}


bool
user_is_valid_name(const char *name)
{
	size_t	 n;
	bool	 valid;

	if (name == NULL) {
		valid = false;
	} else {
		n = 0;
		while (n < USER_NAME_SIZE && name[n] != '\0'
		    && isprint((unsigned char)name[n]) && name[n] != USER_SEP
		    && (name[n] == ' ' || !isspace((unsigned char)name[n]))) {
			n++;
		}
		valid = n >= USER_NAME_MIN_LEN && n < USER_NAME_SIZE
		    && name[n] == '\0' && name[0] != ' ' && name[n - 1] != ' ';
	}

	return valid;
}


bool
user_is_valid_email(const char *em)
{
	return em != NULL && em[0] != '\0' && strchr(em, USER_SEP) == NULL
	    && strlen(em) < USER_EMAIL_SIZE;;
}


bool
user_is_valid_href(const char *href)
{
	size_t	 n;
	bool	 valid;

	if (href == NULL) {
		valid = false;
	} else {
		n = 0;
		while (n < USER_HREF_SIZE && href[n] != '\0'
		    && isprint((unsigned char)href[n]) && href[n] != USER_SEP
		    && href[n] != '<' && href[n] != '>' && href[n] != '"'
		    && (href[n] == ' ' || !isspace((unsigned char)href[n]))) {
			n++;
		}
		valid = n < USER_HREF_SIZE && href[n] == '\0' && href[0] != ' '
		    && (n == 0 || href[n - 1] != ' ');
	}

	return valid;
}


void
user_init(const char *path)
{
	strlcpy(user_db_path, path, sizeof(user_db_path));
}


static FILE *
user_file_open(const char *filename, int writable, int *uerr)
{
	struct	timespec	 rqt;
	FILE	*f;
	int	 fd, pass;

	rqt.tv_sec = 0;
	rqt.tv_nsec = 500000 * 1000;
	f = NULL;
	pass = 0;
	while (pass < 20 && (fd = open(filename,
	    (writable ? O_RDWR : O_RDONLY)
	    | (O_RDONLY | O_EXLOCK | O_NONBLOCK | O_SYNC))) < 0
	    && errno == EWOULDBLOCK) {
		pass++;
		(void)nanosleep(&rqt, NULL);
	}

	if (pass >= 20) {
		*uerr = USER_ERR_FILE_LOCKED;
	} else if (fd >= 0) {
		if ((f = fdopen(fd, writable ? "rw" : "r")) == NULL) {
			*uerr = USER_ERR_FILE_OTHER;
			close(fd);
		}
		*uerr = USER_ERR_NONE;
	} else {
		if (errno == ENOENT) {
			*uerr = USER_ERR_NONEXISTENT;
		} else {
			*uerr = USER_ERR_FILE_OTHER;
		}
	}

	return f;
}


static void
user_token(char **c, char *buf, size_t len)
{
	size_t	 i;

	i = 0;
	while (**c && **c != USER_SEP && i + 1 < len) {
		buf[i++] = *(*c)++;
	}
	if (**c == USER_SEP) {
		(*c)++;
	}
	buf[i] = '\0';
}


static void
user_hash(const char *password, char *hash)
{
	SHA1_CTX	 ctx;

	SHA1Init(&ctx);
	SHA1Update(&ctx, (const u_int8_t *)secret, strlen(secret));
	SHA1Update(&ctx, (const u_int8_t *)password, strlen(password));
	SHA1End(&ctx, hash);

	explicit_bzero(&ctx, sizeof(ctx));
}


int
user_find(const char *id, struct user *u, bool *acct_locked)
{
	return user_find_x(id, NULL, u, acct_locked);
}


int
user_check_pw(const char *id, const char *pw, struct user *u)
{
	struct	user	 ignored;
	bool	 locked;

	if (id == NULL || id[0] == '\0' || pw == NULL || pw[0] == '\0') {
		return USER_ERR_INVALID;
	} else {
		return user_find_x(id, pw, u != NULL ? u : &ignored, &locked);
	}
}


static int
user_find_x(const char *id, const char *pw, struct user *u, bool *acct_locked)
{
	char	 s[8192];
	char	 hash_db[PW_STORED_SIZE];
	char	 hash_pw[PW_STORED_SIZE];
	char	*p;
	FILE	*f;
	size_t	 i;
	int	 found;
	int	 uerr;

	if (id == NULL || id[0] == '\0' || u == NULL
	    || user_db_path[0] == '\0') {
		uerr = USER_ERR_INVALID;
	} else if ((f = user_file_open(user_db_path, 0, &uerr)) != NULL) {
		found = 0;
		while (!found && fgets(s, sizeof(s), f) != NULL) {
			i = strlen(s);
			if (i > 0 && s[i - 1] == '\n') {
				s[i - 1] = '\0';
			}
			p = s;
			user_token(&p, u->id, sizeof(u->id));
			found = strcasecmp(u->id, id) == 0;
			if (found) {
				user_token(&p, u->group, sizeof(u->group));
				user_token(&p, hash_db, sizeof(hash_db));
				user_token(&p, u->name, sizeof(u->name));
				user_token(&p, u->email, sizeof(u->email));
				user_token(&p, u->href, sizeof(u->href));
			}
		}
		fclose(f);

		if (!found) {
			uerr = USER_ERR_NO_SUCH_USER;
		} else if (pw == NULL) {
			*acct_locked = hash_db[0] == PW_UNMATCHABLE;
			uerr = USER_ERR_NONE;
		} else {
			*acct_locked = hash_db[0] == PW_UNMATCHABLE;
			user_hash(pw, hash_pw);
			uerr = timingsafe_bcmp(hash_pw, hash_db,
			    SHA1_DIGEST_STRING_LENGTH - 1) == 0
			    ? USER_ERR_NONE : USER_ERR_PW_INCORRECT;
			explicit_bzero(hash_pw, sizeof(hash_pw));
		}

		explicit_bzero(hash_db, sizeof(hash_db));
		explicit_bzero(s, sizeof(s));
	}

	return uerr;
}


int
user_add(const struct user *u, const char *pw)
{
	char	 s[8192], fn[PATH_MAX], hash[PW_STORED_SIZE], id[USER_ID_SIZE];
	char	*p;
	FILE	*f, *g;
	int	 fd;
	int	 uerr;

	if (u == NULL || pw == NULL || user_db_path[0] == '\0') {
		return USER_ERR_INVALID;
	}
	if ((f = user_file_open(user_db_path, 1, &uerr)) == NULL) {
		return uerr;
	}
	snprintf(fn, sizeof(fn), "%s.tmp.XXXXXXXXXX", user_db_path);
	if ((fd = mkstemp(fn)) == -1) {
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	if (fchmod(fd, 0660) != 0) {
		close(fd);
		unlink(fn);
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	if ((g = fdopen(fd, "w+")) == NULL) {
		close(fd);
		unlink(fn);
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	while (fgets(s, sizeof(s), f) != NULL) {
		p = s;
		user_token(&p, id, sizeof(id));
		if (strcasecmp(id, u->id) == 0) {
			fclose(f);
			unlink(fn);
			fclose(g);
			return USER_ERR_USER_EXISTS;
		}
		fputs(s, g);
	}
	user_hash(pw, hash);
	fprintf(g, "%s;%s;%s;%s;%s;%s\n", u->id, u->group,
	    hash, u->name, u->email, u->href);
	explicit_bzero(hash, sizeof(hash));
	fflush(g);
	fclose(g);
	if (rename(fn, user_db_path) != 0) {
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	fclose(f);
	return USER_ERR_NONE;
}


/*
 * Function for various forms of updating the account id.  If u is non-NULL,
 * account details are updated to match those specified.  If pw is NULL, the
 * password is not changed; if the string referenced by pw is equal to
 * pw_unlock, the account is unlocked; if the string referenced by pw is equal
 * to pw_lock_prfx, the account is locked; any other value is treated as a
 * normal change of password, with the stored value being updated accordingly.
 */
int
user_update_x(const char *id, const struct user *u,
    const char *pw)
{
	struct	user	 u_db;
	char	 s[8192], fn[PATH_MAX];
	char	 hash_db[PW_STORED_SIZE], hash_pw[PW_STORED_SIZE];
	char	*p;
	FILE	*f, *g;
	size_t	 i;
	int	 fd, found;
	int	 uerr;

	found = 0;

	if (id == NULL || (u == NULL && pw == NULL)
	    || user_db_path[0] == '\0') {
		return USER_ERR_INVALID;
	}
	if ((f = user_file_open(user_db_path, 1, &uerr)) == NULL) {
		return uerr;
	}
	snprintf(fn, sizeof(fn), "%s.tmp.XXXXXXXXXX", user_db_path);
	if ((fd = mkstemp(fn)) == -1) {
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	if (fchmod(fd, 0660) != 0) {
		close(fd);
		unlink(fn);
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	if (!(g = fdopen(fd, "w+"))) {
		close(fd);
		unlink(fn);
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	while (fgets(s, sizeof(s), f) != NULL) {
		p = s;
		user_token(&p, u_db.id, sizeof(u_db.id));
		if (strcasecmp(u_db.id, id) != 0) {
			/* Not target user, so just copy over to new file. */
			fputs(s, g);
		} else if (!found) {
			/* Found the user to update. */

			/* Strip trailing newline (if any). */
			i = strlen(s);
			if (i > 0 && s[i - 1] == '\n') {
				s[i - 1] = '\0';
			}

			/* Parse rest of entry. */
			user_token(&p, u_db.group, sizeof(u_db.group));
			user_token(&p, hash_db, sizeof(hash_db));
			user_token(&p, u_db.name, sizeof(u_db.name));
			user_token(&p, u_db.email, sizeof(u_db.email));
			user_token(&p, u_db.href, sizeof(u_db.href));

			if (pw != NULL) {
				if (strcmp(pw, pw_lock_prfx) == 0) {
					/* Ensure account is locked. */
					if (strncmp(hash_db, pw_lock_prfx,
					    strlen(pw_lock_prfx)) == 0) {
						/* Account alreadly locked. */
						(void)strlcpy(hash_pw, hash_db,
						    sizeof(hash_pw));
					} else {
						/* Create unmatchable value. */
						(void)strlcpy(hash_pw,
						    pw_lock_prfx,
						    sizeof(hash_pw));
						(void)strlcat(hash_pw, hash_db,
						    sizeof(hash_pw));
					}
				} else if (strcmp(pw, pw_unlock) == 0) {
					/* Ensure account is unlocked. */
					if (strncmp(hash_db, pw_lock_prfx,
					    strlen(pw_lock_prfx)) == 0) {
						/* Account is locked. */
						(void)strlcpy(hash_pw, hash_db
						    + strlen(pw_lock_prfx),
						    sizeof(hash_pw));
					} else {
						/* Account is not locked. */
						strlcpy(hash_pw, hash_db,
						    sizeof(hash_pw));
					}
				} else {
					user_hash(pw, hash_pw);
				}
			}

			if (u != NULL) {
				fprintf(g, "%s;%s;%s;%s;%s;%s\n", id, u->group,
				    pw != NULL ? hash_pw : hash_db,
				    u->name, u->email, u->href);
			} else {
				fprintf(g, "%s;%s;%s;%s;%s;%s\n", id,
				    u_db.group,
				    pw != NULL ? hash_pw : hash_db,
				    u_db.name, u_db.email, u_db.href);
			}

			/* Wipe new password hash. */
			explicit_bzero(hash_pw, sizeof(hash_pw));
			/* Wipe old password hash. */
			explicit_bzero(hash_db, sizeof(hash_db));

			found = 1;
		}
	}

	/* Wipe line buffer. */
	explicit_bzero(s, sizeof(s));

	if (!found) {
		fclose(f);
		unlink(fn);
		fclose(g);
		return USER_ERR_NO_SUCH_USER;
	}
	fflush(g);
	fclose(g);
	if (rename(fn, user_db_path) != 0) {
		fclose(f);
		return USER_ERR_FILE_OTHER;
	}
	fclose(f);

	return USER_ERR_NONE;
}


int
user_update(const struct user *u)
{
	return user_update_x(u->id, u, NULL);
}


int
user_change_pw(const char *id, const char *pw)
{
	return user_is_valid_password(pw) ? user_update_x(id, NULL, pw)
	    : USER_ERR_INVALID;
}


int
user_account_lock(const char *id)
{
	return user_update_x(id, NULL, pw_lock_prfx);
}


int
user_account_unlock(const char *id)
{
	return user_update_x(id, NULL, pw_unlock);
}


int
user_delete(const char *id)
{
	char	 s[8192], fn[PATH_MAX], db_id[USER_ID_SIZE];
	char	*p;
	FILE	*f, *g;
	int	 fd;
	int	 found;
	int	 uerr;

	snprintf(fn, sizeof(fn), "%s.tmp.XXXXXXXXXX", user_db_path);

	if (id == NULL || user_db_path[0] == '\0') {
		uerr = USER_ERR_INVALID;
	} else if ((f = user_file_open(user_db_path, 1, &uerr)) == NULL) {
		/* nothing */
	} else if ((fd = mkstemp(fn)) == -1) {
		fclose(f);
		uerr = USER_ERR_FILE_OTHER;
	} else if (fchmod(fd, 0660) != 0) {
		fclose(f);
		uerr = USER_ERR_FILE_OTHER;
	} else if ((g = fdopen(fd, "w+")) == NULL) {
		fclose(f);
		uerr = USER_ERR_FILE_OTHER;
	} else {
		found = 0;
		while (fgets(s, sizeof(s), f) != NULL) {
			p = s;
			user_token(&p, db_id, sizeof(db_id));
			if (strcasecmp(db_id, id) == 0) {
				found = 1;
				/* Do not copy over. */
			} else {
				/* Just copy over. */
				fputs(s, g);
			}
		}
		fflush(g);
		fclose(g);
		if (!found) {
			uerr = USER_ERR_NO_SUCH_USER;
		} else if (rename(fn, user_db_path) != 0) {
			uerr = USER_ERR_FILE_OTHER;
		} else {
			uerr = USER_ERR_NONE;
		}
		fclose(f);
	}

	return uerr;
}


int
user_each_do(void (*dtu)(const struct user *, void *), void *ref)
{
	struct	user	 u;
	char	 s[8192];
	char	 hash_db[PW_STORED_SIZE];
	char	*p;
	FILE	*f;
	size_t	 i;
	int	 uerr;

	if (dtu == NULL || user_db_path[0] == '\0') {
		uerr = USER_ERR_INVALID;
	} else if ((f = user_file_open(user_db_path, 0, &uerr)) != NULL) {
		while (fgets(s, sizeof(s), f) != NULL) {
			i = strlen(s);
			if (i > 0 && s[i - 1] == '\n') {
				s[i - 1] = '\0';
			}
			p = s;
			user_token(&p, u.id, sizeof(u.id));
			user_token(&p, u.group, sizeof(u.group));
			user_token(&p, hash_db, sizeof(hash_db));
			user_token(&p, u.name, sizeof(u.name));
			user_token(&p, u.email, sizeof(u.email));
			user_token(&p, u.href, sizeof(u.href));

			(*dtu)(&u, ref);
		}
		fclose(f);

		/* Wipe hash(es). */
		explicit_bzero(s, sizeof(s));
		explicit_bzero(hash_db, sizeof(hash_db));
	}

	return uerr;
}

