/*	$Id: user.h,v 1.18 2017/08/05 11:17:48 rlr Exp $ */

/*
 * Copyright (c) 2004-2006 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * Copyright (c) 2016, 2017 Ross L Richardson <ross@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _USER_H_
#define _USER_H_

#include <stdbool.h>


/* Field separator in user database (text) file. */
#define USER_SEP  ';'

/* Error values. */
#define USER_ERR_NONE          0
#define USER_ERR_INVALID       1  /* invalid argument */
#define USER_ERR_NONEXISTENT   2  /* file/dir nonexistent */
#define USER_ERR_FILE_LOCKED   3
#define USER_ERR_FILE_OTHER    4  /* other file/IO error */
#define USER_ERR_NO_SUCH_USER  5
#define USER_ERR_USER_EXISTS   6
#define USER_ERR_PW_INCORRECT  7

/* Sizes and lengths. */
#define USER_ID_SIZE      32
#define USER_GROUP_SIZE   32
#define USER_NAME_SIZE   128
#define USER_EMAIL_SIZE  128
#define USER_HREF_SIZE   128
#define USER_ID_MIN_LEN      2
#define USER_ID_MAX_LEN     14
#define USER_PW_MIN_LEN      8
#define USER_GROUP_MIN_LEN   2
#define USER_GROUP_MAX_LEN  14
#define USER_NAME_MIN_LEN    2


struct user {
	char	 id[USER_ID_SIZE];
	char	 group[USER_GROUP_SIZE];
	char	 name[USER_NAME_SIZE];
	char	 email[USER_EMAIL_SIZE];
	char	 href[USER_HREF_SIZE];
};


/*
 * bool	 user_is_valid_id(const char *id);
 *
 * Returns true if and only if id is (non-NULL and) of length between
 * USER_ID_MIN_LEN and USER_ID_MAX_LEN, contains only '-', '.', '_', and
 * alphanumeric characters, and contains at least one letter.
 */
bool	 user_is_valid_id(const char *);


/*
 * bool	 user_is_valid_password(const char *pw);
 *
 * Returns true if and only if pw is (non-NULL and) of length greater than or
 * equal to USER_PW_MIN_LEN.
 */
bool	 user_is_valid_password(const char *);


/*
 * bool	 user_is_valid_group(const char *g);
 *
 * Returns true if and only if g is (non-NULL and) of length between
 * USER_GROUP_MIN_LEN and USER_GROUP_MAX_LEN, and contains only '_' and
 * alphanumeric characters.
 */
bool	 user_is_valid_group(const char *);


/*
 * bool	 user_is_valid_name(const char *name);
 *
 * Returns true if and only if name is (non-NULL and) of length greater than or
 * equal to USER_NAME_MIN_LEN and less than USER_NAME_SIZE, and contains only
 * printable characters, no USER_SEP characters, no whitespace other than ' ',
 * and no leading or trailing ' '.
 */
bool	 user_is_valid_name(const char *);


/*
 * bool	 user_is_valid_email(const char *em);
 *
 * Returns true if and only if em is (non-NULL and) of length greater than zero
 * and less than USER_EMAIL_SIZE, and contains no USER_SEP characters.
 */
bool	 user_is_valid_email(const char *);


/*
 * bool	 user_is_valid_href(const char *href);
 *
 * Returns true if and only if href is (non-NULL and) of length less than
 * USER_HREF_SIZE, contains only printable characters, contains no USER_SEP,
 * '<', '>', and '"' characters, contains no whitespace characters other than
 * ' ', and contains lo leading or trailing ' '.
 */
bool	 user_is_valid_href(const char *);


/*
 * void	 user_init(const char *path);
 *
 * Saves path as the location of the user database file.  Must be called before
 * any calls are made to functions which use the user database file.
 */
void	 user_init(const char *);


/*
 * int	 user_find(const char *id, struct user *u, bool *acct_locked);
 *
 * Attempts to find in the user database the user with username id.  If found
 * (and no error occurs), account details are assigned to *u, *acct_locked
 * indicates whether the account is locked, and USER_ERR_NONE is returned;
 * otherwise, an appropriate error code is returned;
 */
int	 user_find(const char *, struct user *, bool *);


/*
 * int	 user_check_pw(const char *id, const char *pw, struct user *u);
 *
 * Checks the validity of password pw for user id.  In the absence of error,
 * returns USER_ERR_NONE for correct password or USER_ERR_PW_INCORRECT for
 * incorrect password, and, if u is non-NULL, details for the account are
 * assigned to *u; other values are returned in event of error.
 */
int	 user_check_pw(const char *, const char *, struct user *);


/*
 * int	 user_add(const struct user *u, const char *pw);
 *
 * Adds a new user account with the details of *u and password pw, returning an
 * indication of success or the error which prevent it.
 */
int	 user_add(const struct user *, const char *);


/*
 * int	 user_update(const struct user *u);
 *
 * Updates details for user u->id, returning an indication of success or the
 * error which prevented it.
 */
int	 user_update(const struct user *);


/*
 * int	 user_change_pw(const char *id, const char *pw);
 *
 * Changes the password for user id to pw, returning an indication of success
 * or the error which prevented it.
 */
int	 user_change_pw(const char *, const char *);


/*
 * int	 user_account_lock(const char *id);
 *
 * Locks account of user id, returning an indication of success or the error
 * which prevented it.
 */
int	 user_account_lock(const char *);


/*
 * int	 user_account_unlock(const char *id);
 *
 * Unlocks account of user id, returning an indication of success or the error
 * which prevented it.
 */
int	 user_account_unlock(const char *);


/*
 * int	 user_delete(const char *id);
 *
 * Deletes account of user id, returning an indication of success or the error
 * which prevented it.
 */
int	 user_delete(const char *);


/*
 * int	 user_each_do(void (*dtu)(const struct user *, void *), void *ref);
 *
 * Invokes dtu() with account details and ref for each user in the database,
 * returning an indication of success or the error which prevented it.
 */
int	 user_each_do(void (*)(const struct user *, void *), void *);


#endif /* !_USER_H_ */

