/* $Id: fwobac.c,v 1.249 2017/09/06 00:38:45 rlr Exp $ */

/*
 * Copyright (c) 2004-2006 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * Copyright (c) 2016, 2017 Ross Richardson <ross@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <sys/file.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fts.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <vis.h>

#include <kcgi.h>

#include "base64.h"
#include "captcha.h"
#include "conf_parse.h"
#include "fdk_hmac.h"
#include "file_utils.h"
#include "fwobac.h"
#include "html_utils.h"
#include "mail.h"
#include "rhf.h"
#ifdef FEATURE_THRES
#include "thres.h"
#endif /* FEATURE_THRES */
#include "tmpl.h"
#include "user.h"
#include "version.h"


/* Constants */

/* Bits per hexadecimal character */
#define BITS_PER_XDIG  4

/* Flags for strnvis(3) */
#define VFLAGS  (VIS_WHITE | VIS_CSTYLE)

/* Flags for struct entry */
#define FLAG_POLL	0x0000001  /* Is a poll */
#define FLAG_SUBMISSION	0x0000002  /* Submission rather than article */
#define FLAG_REPORTED	0x0000004  /* Comment reported to editors */
#define FLAG_PLAIN	0x0000008  /* Plain text rather than HTML */
#define FLAG_EDIT	0x0000010  /* Comment reported to editors */

/* Flags for url() */
#define URL_F_ABS  0x0001  /* Absolute URL */
#define URL_F_ESC  0x0001  /* HTML-escape */

/* pledge() promises */
#define PROMISES_MIN  "stdio"
#define PROMISES_R    "stdio rpath flock"
#define PROMISES_RW   "stdio rpath wpath cpath fattr flock"
#define PROMISES_RWX  "stdio rpath wpath cpath fattr flock proc exec"
#define PROMISES_MAX  "stdio rpath wpath cpath tmppath fattr flock proc exec"

/* Request schemes. */
#define SCHEME_HTTPS  "https"
#define SCHEME_HTTP   "http"

/* Standard ports. */
#define STD_PORT_N_HTTP    80
#define STD_PORT_N_HTTPS  443
#define STD_PORT_S_HTTP   "80"
#define STD_PORT_S_HTTPS  "443"

/* Tests for known (logged-in) or unkown/anonymous user. */
#define KNOWN_USER()  (SNB(A_DEAUTH) & ent_act)
#define ANON_USER()  (!KNOWN_USER())

/* String matching for pointer plus length against C string */
#define STRNEQ(s, z, c)  (strncmp(s, c, z) == 0 && strlen(c) == z)

/* SID to month (1 - 12) */
#define SID_MONTH(s)  (10 * (s[4] - '0') + (s[5] - '0'))

#ifdef DEBUG
static bool	 rt_debug = true;
#define DMSG(r, ...)  if (rt_debug) { msg(r, MSG_DEBUG, __VA_ARGS__); }
#else /* DEBUG */
#define DMSG(...)
#endif /* DEBUG */

#define SP       " "
#define EQ       "="

#ifdef FEATURE_A_PREV_NEXT_LINKS
#define NAV_ITEM_BEG  ""
#define NAV_ITEM_END  ""
#define NAV_ITEM_SEP  ""
#endif /* FEATURE_A_PREV_NEXT_LINKS */

#ifdef FIELDVAL_SEP_AMP
#define SEP      "&"
#define ESCD_SEP "&amp;"
#else /* FIELDVAL_SEP_AMP */
#define SEP      ";"
#define ESCD_SEP ";"
#endif /* FIELDVAL_SEP_AMP */

/* Size for client address string in cookie information. */
#define AUTH_COOKIE_REMOTE_SIZE  128

#ifdef COMPAT_LEGACY_ID
#define LSAFE_ID  vis_id
#else /* COMPAT_LEGACY_ID */
#define LSAFE_ID  user.id
#endif /* COMPAT_LEGACY_ID */

#if defined(COMPAT_LEGACY_R) || defined(COMPAT_LEGACY_W)
#define CTIME_FMT  "%a %b %d %H:%M:%S %Y"
#endif /* defined(COMPAT_LEGACY_R) || defined(COMPAT_LEGACY_W) */

/* Primary template macro delimiting character. */
#define TMDCP  (get_conf_val(CI_TEMPLATE_MACRO_CHARS)[0])

/* Size of subject field in struct entry. */
#define SUBJECT_SZ  256


/* Types */

/* Levels for msg() */
enum msg_level {
	MSG_DEBUG,
	MSG_INFO,
	MSG_NOTE,
	MSG_WARNING,
	MSG_ERROR,
	MSG_CRITICAL
};


/* Representation of article/comment */
struct	entry {
	struct	tm	 time_authored;
	struct	tm	 time_edited;
	char	 fn[PATH_MAX];
	char	 subject[SUBJECT_SZ];
	char	 href[128];
	char	 addr_author[64];
	char	 addr_editor[64];
	char	 dept[64];
	char	 email[64];
#ifdef COMPAT_LEGACY_R
	char	 host[64];
#endif /* COMPAT_LEGACY_R */
	char	 name[64];
	char	 id_author[USER_ID_SIZE];
	char	 id_editor[USER_ID_SIZE];
	char	 topic[32];
	char	 sid[TIMESTAMP_LEN + 1];
	char	 pid[TIMESTAMP_LEN + 1];  /* not a timestamp */
	char	 pubdate[TIMESTAMP_LEN + 1];
	unsigned int	 level;
	unsigned int	 flags;
#ifdef FEATURE_MODERATE
	int	 modcount;
	int	 modsum;
	int	 modable;
#endif /* FEATURE_MODERATE */
	char	*body;
	char	*more;
};


/* Details of a CAPTCHA */
struct	captcha_details {
	char	 text[1024];
	char	 check[128];
};


/* Parsed cookie format used by valid_auth_cookie() on success */
struct	auth_cookie {
	char	 id[USER_ID_SIZE];
	time_t	 t;
	char	 remote[AUTH_COOKIE_REMOTE_SIZE];
	unsigned char	 hmac[FDK_HMAC_BIN_SIZE];
};


/* struct for passing template macro replacement callback information */
struct	macro_cb_info {
	struct	kreq	*r;
	void	(*f)(struct kreq *, char, const char *, size_t, const void *);
	const void	*a;
	char	 c;
};


/*
 * struct for callback for registration email macro substitution.
 * Logically, the file descriptor (fd) doesn't belong in this structure.
 * However, as this type is used for just a single purpose, we save the
 * complexity of adding another type by including it.
 */
struct	reg_mail_info {
	const char	*id;
	const char	*hmac;
	const char	*rcpt;
	const char	*ref;
	const char	*remote;
	const char	*snom;
	const char	*srvr;
	const char	*ua;
	int	 fd;
};


/* For archive index generation */
struct	ai_years {
	int	 first;
	int	 last;
	int	 target;
};


/* For archive index template rendering. */
struct	ai_tmpl_info {
	struct	tm	 yt;
	struct	dirent	**de;  /* articles for year */
	const struct	kreq	*r;
	int	 mas[12];  /* article count by month */
	int	 fd;
	int	 i;
	char	 c;
};

/* Article rendering information */
struct a_rend_info {
	struct	entry	*e;
	time_t	 last_change;
	time_t	 last_read;
	unsigned int	 cmts_count;
	bool	 cmts_open;
};

/* Information on articles (in article list) for one day. */
struct	al_day_info {
	size_t	fst;
	size_t	cnt;
};

/* For rendering article lists. */
struct	al_info {
	struct	a_rend_info	*ari;
	struct	al_day_info	*di;
	size_t	 len;
};

/* Article rendering list. */
struct	al_article_list {
	struct	a_rend_info	*dai;
	size_t	 na;
};

/* For storing configuration entries. */
struct	lstr {
	size_t	 l;
	const char	*s;
};


/* Request field bitset of errors */
struct	field_errs {
	uint64_t	 invalid;
	uint64_t	 missing;
};


/* Request entry information */
struct	req_ent_info {
	struct	entry	 e;
	struct	field_errs	 b;
};


/* Comment rendering information */
struct c_rend_info {
	struct	entry	*e;
	time_t	 last_read;
	bool	 cmts_open;
};


/* User-editable fields in the article edit form */
#define EDIT_A_FORM_USER_K_MASK \
	( SNB(K_CONTENT0) \
	| SNB(K_CONTENT1) \
	| SNB(K_DEPT) \
	| SNB(K_EMAIL) \
	| SNB(K_HREF) \
	| SNB(K_NAME) \
	| SNB(K_POLL) \
	| SNB(K_SUBJECT) \
	| SNB(K_TOPIC) \
	)

/* User-editable fields in the submission edit form */
#define EDIT_S_FORM_USER_K_MASK \
	( SNB(K_CONTENT0) \
	| SNB(K_CONTENT1) \
	| SNB(K_DEPT) \
	| SNB(K_EMAIL) \
	| SNB(K_HREF) \
	| SNB(K_NAME) \
	| SNB(K_POLL) \
	| SNB(K_PUBDATE) \
	| SNB(K_SUBJECT) \
	| SNB(K_TOPIC) \
	)

/* User-editable fields in the registration form */
#define REG_FORM_USER_K_MASK \
    ( SNB(K_CAPTCHA) \
    | SNB(K_EMAIL) \
    | SNB(K_HREF) \
    | SNB(K_ID) \
    | SNB(K_NAME) \
    | SNB(K_PW) \
    | SNB(K_PWCONFIRM) \
    )

/* User-editable fields in the submit form */
#define SUBMIT_FORM_USER_K_MASK \
    ( SNB(K_CAPTCHA) \
    | SNB(K_CONTENT0) \
    | SNB(K_EMAIL) \
    | SNB(K_HREF) \
    | SNB(K_NAME) \
    | SNB(K_SUBJECT) \
    )

/* User-editable fields in the reply form */
#ifdef FEATURE_COMMENT_SUBJECT
#define REPLY_FORM_USER_K_MASK  SUBMIT_FORM_USER_K_MASK
#else /* FEATURE_COMMENT_SUBJECT */
#define REPLY_FORM_USER_K_MASK \
    ( SNB(K_CAPTCHA) \
    | SNB(K_CONTENT0) \
    | SNB(K_EMAIL) \
    | SNB(K_HREF) \
    | SNB(K_NAME) \
    )
#endif /* FEATURE_COMMENT_SUBJECT */

/* Request requirement flags for struct action_details */
#define RR_NONE          0x00
#define RR_SCHEME_HTTPS  0x01
#define RR_METHOD_POST   0x02

/* Structure describing an action */
struct	action_details {
	const char	* const	 name;
	const char	* const	 promises;  /* pledge(2) promises */
	void	(*handler)(struct kreq *);
	const uint8_t	 rreq;  /* request requirements */
};


/* Function prototypes */
static bool	 are_comments_open(const char *);
static bool	 article_exists(const char *);
static void	 build_entry_from_request(struct kreq *,
    struct req_ent_info *);
static void	 check_auth_cookie(struct kreq *, struct user *, uint64_t *);
static void	 check_config(uint16_t *, uint16_t *);
static bool	 comment_exists(const char *, const char *);
static int	 compare_action_name(const void *, const void *);
static int	 compare_ci_name(const void *, const void *);
static int	 compare_fts(const FTSENT **, const FTSENT **);
#ifdef FEATURE_A_PREV_NEXT_LINKS
static int	 compare_name_asc_fts(const FTSENT **, const FTSENT **);
#endif /* FEATURE_A_PREV_NEXT_LINKS */
static int	 compare_name_des_fts(const FTSENT **, const FTSENT **);
static unsigned int	 count_comments(const char *);
static int	 create_comment(struct entry *);
static void	 create_flag_file(const struct kreq *, const char *);
static void	 create_new_index_template(const struct kreq *, int);
static int	 create_submission(struct entry *);
static int	 current_year(void);
static int	 delete_article(const char *);
static int	 delete_comment(const char *, const char *);
static int	 delete_submission(const char *);
#ifdef FEATURE_QUOTE_PARENT_COMMENT
static void	 de_html(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
#endif /* FEATURE_QUOTE_PARENT_COMMENT */
static void	 do_action(struct kreq *, enum action);
static void	 do_action_about(struct kreq *);
static void	 do_action_archive(struct kreq *);
static void	 do_action_article(struct kreq *);
static void	 do_action_auth(struct kreq *);
static void	 do_action_changepw(struct kreq *);
static void	 do_action_confirm(struct kreq *);
static void	 do_action_deauth(struct kreq *);
static void	 do_action_delete_comment(struct kreq *);
#ifdef UNDEADLY_TRANSITION
static void	 do_action_errata(struct kreq *);
#endif /* UNDEADLY_TRANSITION */
static void	 do_action_front(struct kreq *);
static void	 do_action_login(struct kreq *);
#ifdef COMPAT_BRAINDEAD_FB
static void	 do_action_mangled_article_req(struct kreq *);
#endif /* COMPAT_BRAINDEAD_FB */
#ifdef FEATURE_MODERATE
static void	 do_action_moderate(struct kreq *);
#endif /* FEATURE_MODERATE */
static void	 do_action_prefs(struct kreq *);
static void	 do_action_prefspost(struct kreq *);
static void	 do_action_publish(struct kreq *);
static void	 do_action_register(struct kreq *);
static void	 do_action_registerpost(struct kreq *);
static void	 do_action_reply(struct kreq *);
static void	 do_action_rss(struct kreq *);
#ifdef FEATURE_SEARCH
static void	 do_action_search(struct kreq *);
#endif /* FEATURE_SEARCH */
static void	 do_action_show_doc(struct kreq *);
static void	 do_action_submission(struct kreq *);
static void	 do_action_submissions(struct kreq *);
static void	 do_action_submit(struct kreq *);
#ifdef FEATURE_VOTE
static void	 do_action_vote(struct kreq *);
#endif /* FEATURE_VOTE */
static void	 do_common_blockunblock_addr(struct kreq *);
static void	 do_common_delete_sa(struct kreq *);
static void	 do_common_lockunlock_account(struct kreq *);
static void	 do_common_openclose_comments(struct kreq *);
static void	 do_common_contribpost(struct kreq *);
static void	 do_common_edit(struct kreq *);
static void	 do_common_editpost(struct kreq *);
static void	 do_common_openclose_report(struct kreq *);
static void	 emit_http_end_headers(struct kreq *);
static void	 emit_http_locn_header(struct kreq *, const char *);
static void	 emit_http_start_headers(struct kreq *, enum khttp,
    enum kmime);
static void	 expand_cond_macro(struct kreq *, bool, char, const char *,
    size_t, void (*)(struct kreq *, char, const char *, size_t, const void *),
    const void *);
#ifdef FEATURE_A_PREV_NEXT_LINKS
static void	 find_article_adjacent(const struct kreq *, const char *,
    struct entry *, struct entry *);
#endif /* FEATURE_A_PREV_NEXT_LINKS */
static void	 find_articles(const struct kreq *, char *, struct entry *,
    size_t, bool, struct entry *);
#ifdef FEATURE_A_PREV_NEXT_LINKS
static void	 find_comment_adjacent(const struct kreq *, const char *,
    const char *, struct entry *, struct entry *);
#endif /* FEATURE_A_PREV_NEXT_LINKS */
#ifdef FEATURE_SHOW_THREAD
static void	 find_comments(const struct kreq *, const char *, const char *,
    struct entry *, size_t);
#else /* FEATURE_SHOW_THREAD */
static void	 find_comments(const struct kreq *, const char *,
    struct entry *, size_t);
#endif /* FEATURE_SHOW_THREAD */
static long long	 find_max_pid(const char *);
static time_t	 find_newest_change(const char *, bool);
#ifdef FEATURE_A_PREV_NEXT_LINKS
static void	 find_parent(const struct kreq *, const char *, const char *,
    struct entry *);
#endif /* FEATURE_A_PREV_NEXT_LINKS */
static int	 find_pid_path(const char *, const char *, char *, size_t);
static inline enum action	 get_action(const struct kreq *);
static uint64_t	 get_action_entitlements(const struct kreq *, const char *);
#ifdef DEBUG
static bool	 get_boolean_key_val(const struct kreq *, enum key, bool,
    bool);
#endif /* DEBUG */
static inline const char	*get_conf_val(enum conf_item);
static const char	*get_header_stringne(struct kreq *, enum krequ);
static time_t	 get_if_modified_since(struct kreq *);
static int	 get_lastread(const char *, time_t *, int);
static void	 get_subject(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
#ifdef FEATURE_THEMES
static void	 get_theme(struct kreq *, const char **, size_t *);
#endif /* FEATURE_THEMES */
static void	 include_template(struct kreq *, char, const char *,
    const char *, const char *, size_t, void (*cb)(struct kreq *, char,
    const char *, size_t, const void *), const void *);
static int	 index_filter(const struct dirent *);
static bool	 invoked_as_cgi(void);
static bool	 is_all_digits(const char *);
#ifdef COMPAT_BRAINDEAD_FB
static bool	 is_all_digits_n(const char *, size_t);
#endif /* COMPAT_BRAINDEAD_FB */
#ifdef UNUSED
static bool	 is_comment_reported(const char *, const char *);
#endif /* UNUSED */
static bool	 is_inet_addr(const char *);
static bool	 is_future_timestamp(const char *);
#ifdef COMPAT_BRAINDEAD_FB
static bool	 is_mangled_article_req(const struct kpair *);
#endif /* COMPAT_BRAINDEAD_FB */
static bool	 is_nonempty_file(const char *);
static bool	 is_safe_rel_path(const char *, size_t);
static inline void	 make_articles_dir_path(char *, size_t)
	__attribute__ ((__bounded__ (__string__, 1, 2 )));
static inline void	 make_article_dir_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_article_file_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_article_more_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_comments_open_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_article_poll_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_comment_report_path(const char *, const char *,
    char *, size_t)
	__attribute__ ((__bounded__ (__string__, 3, 4 )));
static inline void	 make_current_timestamp(char *, size_t)
	__attribute__ ((__bounded__ (__string__, 1, 2 )));
static inline void	 make_duration_str(time_t, char*, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_last_article_change_path(int, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_last_read_path(const char *sid, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_rss_log_dir_path(char *, size_t)
	__attribute__ ((__bounded__ (__string__, 1, 2 )));
static inline void	 make_rss_log_file_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_submissions_dir_path(char *, size_t)
	__attribute__ ((__bounded__ (__string__, 1, 2 )));
static inline void	 make_submission_dir_path(const char *, char *, size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_submission_file_path(const char *, char *,
    size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static inline void	 make_submission_more_path(const char *, char *,
    size_t)
	__attribute__ ((__bounded__ (__string__, 2, 3 )));
static const char	*md_match(const char *, const char *, bool);
static bool	 might_be_url(const char *);
#ifdef FEATURE_MODERATE
static int	 moderate_comment(const char *, const char *, const char *,
    int);
#endif /* FEATURE_MODERATE */
static void	 msg(const struct kreq *, enum msg_level, const char *, ...)
    __attribute__ ((format (printf, 3, 4)));
static bool	 need_new_index_template(int);
static int	 non_cgi_main(int, char *[]);
static void	 non_cgi_usage(void);
static void	 notify_article_change(const struct kreq *, const char *);
static void	 output_as_html(struct kreq *, const char *);
static void	 output_current_time(struct kreq *, const char *);
static int	 output_macro(const void *, size_t, void *);
static int	 output_plain(const void *, size_t, void *);
static void	 output_time(struct kreq *, const struct tm *, const char *);
static void	 prepare_and_render_al(struct kreq *, char,
    const struct entry *, struct entry [], size_t, const char *)
	__attribute__ ((__bounded__ ( __buffer__, 4, 5 )));
static void	 print_bad_keys(struct kreq *, uint64_t);
static void	 print_group_rights_info(FILE *);
static void	 print_group_rights_info_item(FILE *, const char *, uint64_t);
#ifdef FEATURE_A_PREV_NEXT_LINKS
static void	 print_nav_item(struct kreq *, const char *, const char *);
#endif /* FEATURE_A_PREV_NEXT_LINKS */
static void	 print_sample_config(FILE *);
static int	 read_file(const struct kreq *, struct entry *);
static void	 read_poll(const char *, int *, int, int *);
static void	 render_ai_year(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_al(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_al_day_item(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_archive_index(struct kreq *, char, const char *,
    size_t, const void *);
static void	 render_article(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_article_comment(struct kreq *, char, const char *,
    size_t, const void *);
static void	 render_captcha(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_common(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_common_contrib_or_edit(struct kreq *, char,
    const char *, size_t, const void *);
static void	 render_common_entry(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_common_form(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_edit(struct kreq *, char, const char *, size_t,
    const void *a);
static void	 render_front(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_login(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_prefs(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_register(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_reply(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_rss(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_submit(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_extra_html(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_extra_simple(struct kreq *, char, const char *, size_t,
    const void *);
static void	 render_extra_template(struct kreq *, char, const char *,
    size_t, const void *);
static void	 resp_bad_request(struct kreq *);
static void	 resp_client_redirect(struct kreq *, const char *,
    const char *);
#ifdef UNDEADLY_TRANSITION
static void	 resp_client_redirect_perm(struct kreq *, const char *,
    const char *);
#endif /* UNDEADLY_TRANSITION */
static void	 resp_edit(struct kreq *, const struct req_ent_info *);
static void	 resp_forbidden(struct kreq *, const char *);
static void	 resp_insufficient_rights(struct kreq *);
static void	 resp_internal_error(struct kreq *);
static void	 resp_login(struct kreq *);
static void	 resp_method_not_allowed(struct kreq *);
static void	 resp_not_modified(struct kreq *, time_t);
static void	 resp_not_found(struct kreq *);
static void	 resp_ok(struct kreq *, const char *, const char *);
static void	 resp_prefs(struct kreq *, const struct field_errs *);
static void	 resp_redirect_to_article(struct kreq *, const char *);
static void	 resp_redirect_to_front(struct kreq *, const char *);
static void	 resp_register(struct kreq *, const struct field_errs *);
static void	 resp_reply(struct kreq *,
    const struct req_ent_info *);
static void	 resp_service_unavailable(struct kreq *);
static void	 resp_submit(struct kreq *, const struct req_ent_info *);
static bool	 rss_rate_limit(const struct kreq *, const char *);
static inline const char	*scheme_str(const struct kreq *);
static int	 send_conf_req_email(struct kreq *, const char *, const char *,
    const char *);
static void	 set_comments_open(const struct kreq *, const char *, int);
static inline void	 set_conf_val(enum conf_item, const char *);
static void	 set_report_comment(const struct kreq *, const char *,
    const char *, int);
static int	 store_config_entry(const char *, size_t, const char *, size_t,
    void *);
static void	 sub_render_latest_list(struct kreq *, char);
static bool	 submission_exists(const char *);
static char	*url(const struct kreq *, unsigned int, char *, size_t,
    enum action, const char *, const char *, const char *, ...)
	__attribute__ ((__bounded__ (__string__, 3, 4 )))
	__attribute__ ((format (printf, 8, 9)));
static int	 use_template_file(struct kreq *, char, const char *,
    void (*)(struct kreq *, char, const char *, size_t, const void *),
    const void *);
static int	 use_template_buf(struct kreq *, char, const char *, size_t,
    void (*)(struct kreq *, char, const char *, size_t, const void *),
    const void *);
static int	 valid_action(struct kpair *);
static int	 valid_addr(struct kpair *);
static int	 valid_auth_cookie(struct kpair *);
static int	 valid_boolean(struct kpair *);
static int	 valid_hmac(struct kpair *);
static int	 valid_id(struct kpair *);
static int	 valid_name(struct kpair *);
static int	 valid_pid(struct kpair *);
static int	 valid_sid(struct kpair *);
static int	 valid_subject(struct kpair *);
#ifdef FEATURE_THEMES
static int	 valid_theme(struct kpair *);
#endif /* FEATURE_THEMES */
static int	 valid_type(struct kpair *);
#ifdef FEATURE_VOTE
static int	 vote_poll(const char *, const char *, int);
#endif /* FEATURE_VOTE */
static void	 write_file(const struct kreq *, FILE *, const struct entry *,
    bool);
static void	 write_ai_template(const struct kreq *, int, int);
static int	 write_ai_meta_macro(const void *, size_t, void *);
static int	 write_data(const void *, size_t, void *);
static int	 write_reg_mail_macro(const void *, size_t, void *);
static int	 write_registration_mail(const int, void *);
static int	 write_string(int, const char *);
static void	 xpledge(const char *);
static time_t	 xtime(void);
static void	 xutc_tm_now(struct tm *);
static int	 year_of_earliest_article(const struct kreq *);


/* Globals */

/* See malloc.conf(5). */
extern char	*malloc_options;

/* msg() level strings */
static const char	*msg_level_s[] = {
	"debug",
	"info",
	"note",
	"warning",
	"error",
	"critical"
};

/* Configuration entries */
static const char	*conf_name[] = {
	CI_N_ARTICLES_DIR,
	CI_N_CACHE_DIR,
	CI_N_CAPTCHA_KEY,
	CI_N_COOKIE_KEY,
	CI_N_DEFAULT_TOPIC,
	CI_N_DOMAIN,
	CI_N_GROUP_RIGHTS,
	CI_N_HTTPS_PORT,
	CI_N_REGISTRATION_KEY,
	CI_N_SUBMISSIONS_DIR,
	CI_N_TEMPLATE_DIR,
	CI_N_TEMPLATE_MACRO_CHARS,
#ifdef FEATURE_THEMES
	CI_N_THEMES,
#endif /* FEATURE_THEMES */
	CI_N_USER_DB_FILE
};

static const char	*conf_val[sizeof(conf_name) / sizeof(conf_name[0])];
static char	 utc_name[] = UTC_NAME;

static struct	entry	 newest[64];
static struct	entry	 poll;
static struct	user	 user;
#ifdef COMPAT_LEGACY_ID
static char	 vis_id[USER_ID_MAX_LEN * 4 + 1] = "";
#endif /* COMPAT_LEGACY_ID */
static char	 index_now[TIMESTAMP_LEN + 1];  /* for index_filter() (Y10K) */
static char	 index_year[5];  /* index year for index_filter (Y10K) */
static uint64_t	 ent_act = 0;


static const char	* const key_desc[] = {
	KD_ACTION,
	KD_ADDR_AUTHOR,
	KD_ADDR_EDITOR,
	KD_ASK,
	KD_AUTH_COOKIE,
	KD_CAPTCHA,
	KD_CAPTCHA_CHK,
	KD_CONTENT0,
	KD_CONTENT1,
#ifdef DEBUG
	KD_DEBUG,
#endif /* DEBUG */
	KD_DEPT,
	KD_EMAIL,
	KD_H,
	KD_HREF,
	KD_ID,
	KD_ID_AUTHOR,
	KD_ID_EDITOR,
	KD_ITEMS,
	KD_NAME,
	KD_PID,
	KD_POLL,
	KD_PREVIEW,
	KD_PUBDATE,
	KD_PW,
	KD_PWCONFIRM,
	KD_Q,
	KD_SID,
	KD_SUBJECT,
#ifdef FEATURE_THEMES
	KD_THEME,
#endif /* FEATURE_THEMES */
#ifdef FEATURE_THRES
	KD_THRES,
#endif /* FEATURE_THRES */
	KD_TOPIC,
	KD_TS_AUTHORED,
	KD_TYPE,
#if defined(FEATURE_MODERATE) || defined(FEATURE_VOTE)
	KD_VAL,
#endif /* defined(FEATURE_MODERATE) || defined(FEATURE_VOTE) */
	KD_YEAR
};


/* For khttp_parse(). */
static const struct	kvalid	keys[] = {
	{ valid_action,      KN_ACTION },       /* K_ACTION  */
	{ valid_addr,        KN_ADDR_AUTHOR },  /* K_ADDR_AUTHOR */
	{ valid_addr,        KN_ADDR_EDITOR },  /* K_ADDR_EDITOR */
	{ valid_boolean,     KN_ASK },          /* K_ASK */
	{ valid_auth_cookie, KN_AUTH_COOKIE },  /* K_AUTH_COOKIE */
	{ kvalid_string,     KN_CAPTCHA },      /* K_CAPTCHA */
	{ kvalid_string,     KN_CAPTCHA_CHK },  /* and so on... */
	{ kvalid_stringne,   KN_CONTENT0 },
	{ kvalid_string,     KN_CONTENT1 },
#ifdef DEBUG
	{ valid_boolean,     KN_DEBUG },
#endif /* DEBUG */
	{ kvalid_stringne,   KN_DEPT },
	{ kvalid_email,      KN_EMAIL },
	{ valid_hmac,        KN_H },
	{ kvalid_string,     KN_HREF },
	{ valid_id,          KN_ID },
	{ valid_id,          KN_ID_AUTHOR },
	{ valid_id,          KN_ID_EDITOR },
	{ kvalid_int,        KN_ITEMS },
	{ valid_name,        KN_NAME },
	{ valid_pid,         KN_PID },
	{ kvalid_stringne,   KN_POLL },
	{ kvalid_stringne,   KN_PREVIEW },
	{ valid_sid,         KN_PUBDATE },
	{ kvalid_stringne,   KN_PW },
	{ kvalid_stringne,   KN_PWCONFIRM },
	{ kvalid_stringne,   KN_Q },
	{ valid_sid,         KN_SID },
	{ valid_subject,     KN_SUBJECT },
#ifdef FEATURE_THEMES
	{ valid_theme,       KN_THEME },
#endif /* FEATURE_THEMES */
#ifdef FEATURE_THRES
	{ kvalid_stringne,   KN_THRES },
#endif /* FEATURE_THRES */
	{ kvalid_stringne,   KN_TOPIC },
	{ valid_sid,         KN_TS_AUTHORED },
	{ valid_type,        KN_TYPE },
#if defined(FEATURE_MODERATE) || defined(FEATURE_VOTE)
	{ kvalid_int,        KN_VAL },
#endif /* defined(FEATURE_MODERATE) || defined(FEATURE_VOTE) */
	{ kvalid_int,        KN_YEAR }
};


static const struct	action_details	 action_table[] = {
	{
		.name     = AN_ABOUT,
		.handler  = do_action_about,
		.promises = PROMISES_R,
		.rreq     = RR_NONE,
	},
	{
		.name     = AN_ARCHIVE,
		.handler  = do_action_archive,
		.promises = PROMISES_RW,
		.rreq     = RR_NONE,
	},
	{
		.name     = AN_ARTICLE,
		.handler  = do_action_article,
		.promises = PROMISES_RW,  /* W - get_lastread() with update */
		.rreq     = RR_NONE,
	},
	{
		.name     = AN_AUTH,
		.handler  = do_action_auth,
		.promises = PROMISES_R,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_BLOCK_ADDR,
		.handler  = do_common_blockunblock_addr,
		.promises = PROMISES_R, /*XXX will need more */
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_CHANGEPW,
		.handler  = do_action_changepw,
		.promises = PROMISES_RW,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_CLOSE_COMMENTS,
		.handler  = do_common_openclose_comments,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_CLOSE_REPORT,
		.handler  = do_common_openclose_report,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_CONFIRM,
		.handler  = do_action_confirm,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_DEAUTH,
		.handler  = do_action_deauth,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_DELETE_A,
		.handler  = do_common_delete_sa,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_DELETE_C,
		.handler  = do_action_delete_comment,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_DELETE_S,
		.handler  = do_common_delete_sa,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_EDIT_A,
		.handler  = do_common_edit,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_EDIT_S,
		.handler  = do_common_edit,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_EDITPOST_A,
		.handler  = do_common_editpost,
		.promises = PROMISES_RW,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_EDITPOST_S,
		.handler  = do_common_editpost,
		.promises = PROMISES_RW,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
#ifdef UNDEADLY_TRANSITION
	{
		.name     = AN_ERRATA,
		.handler  = do_action_errata,
		.promises = PROMISES_R,
		.rreq     = RR_NONE,
	},
#endif /* UNDEADLY_TRANSITION */
	{
		.name     = AN_FRONT,
		.handler  = do_action_front,
		.promises = PROMISES_R,
		.rreq     = RR_NONE,
	},
	{
		.name     = AN_LOCK_ACCT,
		.handler  = do_common_lockunlock_account,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_LOGIN,
		.handler  = do_action_login,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
#ifdef FEATURE_MODERATE
	{
		.name     = AN_MODERATE,
		.handler  = do_action_moderate,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
#endif /* FEATURE_MODERATE */
	{
		.name     = AN_OPEN_COMMENTS,
		.handler  = do_common_openclose_comments,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_PREFS,
		.handler  = do_action_prefs,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_PREFSPOST,
		.handler  = do_action_prefspost,
		.promises = PROMISES_RW,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_PUBLISH,
		.handler  = do_action_publish,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_REGISTER,
		.handler  = do_action_register,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_REGISTERPOST,
		.handler  = do_action_registerpost,
		.promises = PROMISES_RWX,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_REPLY,
		.handler  = do_action_reply,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_REPLYPOST,
		.handler  = do_common_contribpost,
		.promises = PROMISES_RW,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_REPORT,
		.handler  = do_common_openclose_report,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_RSS,
		.handler  = do_action_rss,
		.promises = PROMISES_RW,
		.rreq     = RR_NONE,
	},
#ifdef FEATURE_SEARCH
	{
		.name     = AN_SEARCH,
		.handler  = do_action_search,
		.promises = PROMISES_R,
		.rreq     = RR_NONE,
	},
#endif /* FEATURE_SEARCH */
	{
		.name     = AN_SHOW_DOC,
		.handler  = do_action_show_doc,
		.promises = PROMISES_R,
		.rreq     = RR_NONE,
	},
	{
		.name     = AN_SUBMISSION,
		.handler  = do_action_submission,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_SUBMISSIONS,
		.handler  = do_action_submissions,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_SUBMIT,
		.handler  = do_action_submit,
		.promises = PROMISES_R,
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_SUBMITPOST,
		.handler  = do_common_contribpost,
		.promises = PROMISES_RW,
		.rreq     = (RR_SCHEME_HTTPS | RR_METHOD_POST),
	},
	{
		.name     = AN_UNBLOCK_ADDR,
		.handler  = do_common_blockunblock_addr,
		.promises = PROMISES_R, /*XXX will need more */
		.rreq     = RR_SCHEME_HTTPS,
	},
	{
		.name     = AN_UNLOCK_ACCT,
		.handler  = do_common_lockunlock_account,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
#ifdef FEATURE_VOTE
	},
	{
		.name     = AN_VOTE,
		.handler  = do_action_vote,
		.promises = PROMISES_RW,
		.rreq     = RR_SCHEME_HTTPS,
#endif /* FEATURE_VOTE */
	}
};


/*
 * Returns the bitset of actions to which membership of the specified group
 * entitles a user (based on defaults and/or the specification in configuration
 * item CI_GROUP_RIGHTS.
 */
static uint64_t
get_action_entitlements(const struct kreq *r, const char *group)
{
	const char	*cnf;
	const char	*t;
	char	*e;
	uint64_t	 ae;
	size_t	 len;
	bool	 matched;

	if (strcmp(group, GROUP_PART_REG) == 0) {
		/* This should not be possible. */
		fprintf(stderr, "%s: fatal: %s() called when group is "
		    GROUP_PART_REG "\n", getprogname(), __func__);
		exit(EXIT_FAILURE);
	} else {
		/* Check for configuration file setting. */
		len = strlen(group);
		cnf = get_conf_val(CI_GROUP_RIGHTS);

		matched = false;
		t = cnf;

		while (!matched && (t = strstr(t, group)) != NULL) {
			matched = (t == cnf || *(t - 1)
			    == GROUP_RIGHTS_LIST_SEP) && t[len]
			    == GROUP_RIGHTS_SPEC_SEP;
		}

		if (matched) {
			/* Skip group name and separator. */
			t += len + 1;

			/* Check that initial syntax is correct. */
			if (t[0] != '0' || (tolower((unsigned char)t[1])
			    != 'x') || !isxdigit((unsigned char)t[2])) {
				msg(r, MSG_CRITICAL, "%s() invalid"
				    " specification for group \"%s\"",
				    __func__, group);
				matched = false;
			} else {
				/* Skip the part we don't need. */
				t += 2;

				errno = 0;
				ae = strtoull(t, &e, 16);
				if (t[0] == '\0'
				    || (*e != '\0' && *e
				    != GROUP_RIGHTS_LIST_SEP) || (errno
				    == ERANGE && ae == ULLONG_MAX)) {
					msg(r, MSG_CRITICAL, "%s() poorly "
					"formed rights specification for group"
					" \"%s\"", __func__, group);
					matched = false;
				}
			}
		}

		if (!matched) {
			/* Use default values. */
			if (strcmp(group, PSEUDO_GROUP_ANON) == 0) {
				ae = DEFAULT_RIGHTS_ANON;
			} else {
				ae = DEFAULT_RIGHTS_USER;
			}
		}
	}

	return ae;
}


/*
 * Returns true if and only if u is non-NULL and has a form suggesting that
 * it might be a URL with one of the following schemes: https, http, ftp,
 * ssh, telnet, or mailto.
 */
static bool
might_be_url(const char *u)
{
	return u != NULL && u[0] != '\0' && strchr(u, ' ') == NULL
	    && ((strncasecmp(u, "https://", 8) == 0 && u[8] != '\0')
	    || (strncasecmp(u, "http://", 7) == 0 && u[7] != '\0')
	    || (strncasecmp(u, "ftp://", 6) == 0 && u[6] != '\0')
	    || (strncasecmp(u, "ssh://", 6) == 0 && u[6] != '\0')
	    || (strncasecmp(u, "telnet://", 9) == 0 && u[9] != '\0')
	    || (strncasecmp(u, "mailto:", 7) == 0 && u[7] != '\0'));
}


/* pledge(), writing an error and terminating abormally on failure. */
static void
xpledge(const char *promises)
{
	if (pledge(promises, NULL) == -1) {
		fprintf(stderr, "%s: fatal: pledge: %s\n", getprogname(),
		    strerror(errno));
		exit(EXIT_FAILURE);
	}
}


/* Returns result of time(NULL) call, or exits if that fails. */
static time_t
xtime(void)
{
	time_t	 ct;

	if ((ct = time(NULL)) == (time_t)-1) {
		fprintf(stderr, "%s: fatal: time: %s\n", getprogname(),
		    strerror(errno));
		exit(EXIT_FAILURE);
	}

	return ct;
}


/*
 * Sets now to the current UTC time (with now->tm_zone set to utc_name) or
 * aborts on error.
 */
static void
xutc_tm_now(struct tm *now)
{
	struct	tm	*ts;
	time_t	 t;

	t = xtime();
	ts = gmtime_r(&t, now);

	if (ts == NULL) {
		fprintf(stderr, "%s: fatal: gmtime_r failed\n", getprogname());
		exit(EXIT_FAILURE);
	} else {
		now->tm_isdst = 0;
		now->tm_gmtoff = 0;
		now->tm_zone = utc_name;
	}
}


/*
 * Returns true if and only if p is non-empty, does not start or end with '/',
 * and not contain "../".
 */
static bool
is_safe_rel_path(const char *p, size_t len)
{
	return len != 0 && p[0] != '\0' && p[0] != '/' && p[len - 1] != '/'
	    && memmem(p, len, "../", 3) == NULL;
}


/*
 * Returns the value of configuration item i.
 */
static inline const char *
get_conf_val(enum conf_item i)
{
	return conf_val[i];
}


/*
 * Sets the value of configuration item i to s.
 */
static inline void
set_conf_val(enum conf_item i, const char *s)
{
	conf_val[i] = s;
}


/*
 * Prints to stderr a message consisting of the tab-separated fields:
 * - ISO 8601 timestamp (or '-' in the event of formatting failure)
 * - concatentation of basename of program, '[', process ID, and ']'
 * - client IP address (or '-' if unknown)
 * - username (or '-' if unknown)
 * - the string specified in the arguments
 */
static void
msg(const struct kreq *r, enum msg_level lvl, const char *fmt, ...)
{
	struct	tm	 ts;
	char	 buf[32];
	va_list	 ap;

	xutc_tm_now(&ts);
	if (strftime(buf, sizeof(buf), ISO_8601_FMT, &ts) == 0) {
		strlcpy(buf, "-", sizeof(buf));
	}

	fprintf(stderr, "%s\t%s[%u]\t%s\t%s\t%s: ", buf, getprogname(),
	    (unsigned)getpid(), r != NULL ? r->remote : "-",
	    LSAFE_ID[0] == '\0' ? "-" : LSAFE_ID, msg_level_s[lvl]);

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	fprintf(stderr, "\n");
}


/*
 * Attempts to create a new file of the specified path, and write to it
 * the tab-separated fields:
 * - ISO 8601 timestamp (or '-' in the event of formatting failure)
 * - username (or '-' if unknown)
 */
static void
create_flag_file(const struct kreq *r, const char *path)
{
	char	 buf[256];
	struct	tm	 ts;
	FILE	 *f;
	int	 fd;

	DMSG(r, "%s() path = \"%s\"", __func__, path);

	xutc_tm_now(&ts);
	if (strftime(buf, sizeof(buf), ISO_8601_FMT, &ts) == 0) {
		DMSG(r, "%s() strftime failed", __func__);
		strlcpy(buf, "-", sizeof(buf));
	}

	if ((fd = open(path, O_WRONLY | O_CREAT | O_EXCL, 0644)) == -1) {
		msg(r, MSG_ERROR, "%s() open \"%s\" failed \"%s\"", __func__,
		    path, strerror(errno));
	} else if ((f = fdopen(fd, "w")) == NULL) {
		msg(r, MSG_ERROR, "%s() fdopen \"%s\" failed \"%s\"", __func__,
		    path, strerror(errno));
		close(fd);
	} else {
		DMSG(r, "%s() path = \"%s\", id = \"%s\"", __func__, path,
		    LSAFE_ID);
		fprintf(f, "%s\t%s\n", buf, LSAFE_ID);
		fclose(f);
	}
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * file used to indicate the time of the latest change to any article of
 * year yr.  (Time of change is indicated by modification time of the file.)
 * The file is in the directory specified by configuration item CI_CACHE_DIR.
 */
static void
make_last_article_change_path(int yr, char *path, size_t sz)
{
	snprintf(path, sz, FMT_LAST_ART_CH, get_conf_val(CI_CACHE_DIR), yr);
}


/*
 * In the absence of error, updates the modification time of the last-change
 * indicator file for the year of timestamp sid, to the later of the time
 * of sid and the current time.
 */
static void
notify_article_change(const struct kreq *r, const char *sid)
{
	struct	tm	 ts;
	char	 path[PATH_MAX];
	time_t	 n, t;
	int	 err;

	DMSG(r, "%s() sid = \"%s\"", __func__, sid);

	memset(&ts, 0, sizeof(ts));
	if (strptime(sid, TIMESTAMP_FMT, &ts) == NULL) {
		msg(r, MSG_CRITICAL, "%s() strptime failed", __func__);
	} else if ((t = mktime(&ts)) == -1) {
		msg(r, MSG_CRITICAL, "%s() mktime failed", __func__);
	} else {
		make_last_article_change_path(ts.tm_year + 1900, path,
		    sizeof(path));
		/*
		 * Get the later of:
		 * - the time represented by timestamp sid
		 * - the current time
		 */
		n = xtime();
		if (t > n) {
			n = t;
		}

		err = touch_file(path, n);

		if (err != 0) {
			msg(r, MSG_CRITICAL, "%s() touch_file \"%s\" returned "
			    " %d", __func__, path, err);
		} else {
			DMSG(r, "%s() touched \"%s\"", __func__, path);
		}
	}
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * directory containing (published) articles, as specified by the
 * configuration item CI_ARTICLES_DIR.
 */
static inline void
make_articles_dir_path(char *path, size_t sz)
{
	strlcpy(path, get_conf_val(CI_ARTICLES_DIR), sz);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * directory for (published) article sid - that is, the sid subdirectory
 * of the directory specified by configuration item CI_ARTICLES_DIR.
 */
static inline void
make_article_dir_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s", get_conf_val(CI_ARTICLES_DIR), sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * main content file for (published) article sid - that is, ARTICLE_FILE
 * within the directory which would be given by make_article_dir_path().
 */
static inline void
make_article_file_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" ARTICLE_FILE,
	    get_conf_val(CI_ARTICLES_DIR), sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * additional content file for (published) article sid - that is, the path
 * produced by make_article_file_path() with MORE_SUFFIX appended.
 */
static inline void
make_article_more_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" ARTICLE_FILE MORE_SUFFIX,
	    get_conf_val(CI_ARTICLES_DIR), sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * poll file for (published) article/poll sid - that is, POLL_FILE within
 * the directory which would be given by make_article_dir_path().
 */
static inline void
make_article_poll_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" POLL_FILE, get_conf_val(CI_ARTICLES_DIR),
	    sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * directory for files used for rate-limiting RSS requests - the RSS_LOG
 * subdirectory of that specified by configuration item CI_CACHE_DIR.
 */
static inline void
make_rss_log_dir_path(char *path, size_t sz)
{
	snprintf(path, sz, "%s/" RSS_LOG, get_conf_val(CI_CACHE_DIR));
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * file used for rate-limiting RSS requests from IP address addr.
 */
static inline void
make_rss_log_file_path(const char *addr, char *path, size_t sz)
{
	snprintf(path, sz, "%s/" RSS_LOG "/%s", get_conf_val(CI_CACHE_DIR),
	    addr);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * directory containing unpublished articles ("submissions"), as specified
 * by the configuration item CI_SUBMISSIONS_DIR.
 */
static inline void
make_submissions_dir_path(char *path, size_t sz)
{
	strlcpy(path, get_conf_val(CI_SUBMISSIONS_DIR), sz);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * directory for submission sid - that is, the sid subdirectory of the
 * directory specified by configuration item CI_SUBMISSIONS_DIR.
 */
static inline void
make_submission_dir_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s", get_conf_val(CI_SUBMISSIONS_DIR), sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * main content file for submission sid - that is, ARTICLE_FILE
 * within the directory which would be given by make_submisison_dir_path().
 */
static inline void
make_submission_file_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" ARTICLE_FILE,
	    get_conf_val(CI_SUBMISSIONS_DIR), sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * additional content file for submission sid - that is, the path produced
 * by make_submission_file_path() with MORE_SUFFIX appended.
 */
static inline void
make_submission_more_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" ARTICLE_FILE MORE_SUFFIX,
	    get_conf_val(CI_SUBMISSIONS_DIR), sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the
 * file used to indicate that comments are open for article sid.
 */
static inline void
make_comments_open_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" COMMENTS_OPEN,
	    get_conf_val(CI_ARTICLES_DIR), sid);
}


/*
 * Returns true if and only if comments are open for the specified article.
 */
static bool
are_comments_open(const char *sid)
{
	char	 path[PATH_MAX];

	make_comments_open_path(sid, path, sizeof(path));
	return access(path, F_OK) == 0;
}


/*
 * Attempts to mark comments for article sid as closed (for o equal to zero)
 * or open (for non-zero o).
 */
static void
set_comments_open(const struct kreq *r, const char *sid, int o)
{
	char	 path[PATH_MAX];

	DMSG(r, "%s() sid = \"%s\", o = %d", __func__, sid, o);

	make_comments_open_path(sid, path, sizeof(path));
	if (o) {
		create_flag_file(r, path);
	} else {
		(void)unlink(path);
	}
	msg(r, MSG_NOTE, "%s() comments for \"%s\" %s by \"%s\"", __func__,
	    sid, o ? "opened" : "closed", LSAFE_ID);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the file
 * used to time of last reading information (for users) for article sid.
 */
static inline void
make_last_read_path(const char *sid, char *path, size_t sz)
{
	snprintf(path, sz, "%s/%s/" LAST_READ, get_conf_val(CI_ARTICLES_DIR),
	    sid);
}


/*
 * Copies to path (which must be of size no less than sz) the path of the file
 * used to record that comment sid/pid has been reported as inappropriate.
 */
static inline void
make_comment_report_path(const char *sid, const char *pid, char *path,
    size_t sz)
{
	if (find_pid_path(sid, pid, path, sz) != 0) {
		path[0] = '\0';
	} else {
		if (strlcat(path, "/" COMMENT_REPORT, sz) >= sz) {
			path[0] = '\0';
		}
	}
}


#ifdef UNUSED
/*
 * Returns true if and only if a comment report file exists for the
 * comment identified by sid/pid.
 */
static bool
is_comment_reported(const char *sid, const char *pid)
{
	char	 path[PATH_MAX];

	make_comment_report_path(sid, pid, path, sizeof(path));

	return path[0] != '\0' && access(path, F_OK) == 0;
}
#endif /* UNUSED */


/*
 * For the comment identified by sid/pid, attempts to create (for non-zero rep)
 * or remove (for rep equal to zero) the comment report file.
 */
static void
set_report_comment(const struct kreq *r, const char *sid, const char *pid,
    int rep)
{
	char	 path[PATH_MAX];

	DMSG(r, "%s() sid = \"%s\", pid = \"%s\", rep = %d", __func__, sid,
	    pid, rep);

	make_comment_report_path(sid, pid, path, sizeof(path));
	if (path[0] == '\0') {
		DMSG(r, "%s() make_comment_report_path() failed", __func__);
	} else {
		if (rep) {
			create_flag_file(r, path);
		} else {
			if (unlink(path) != 0) {
				msg(r, MSG_ERROR, "%s() unlink \"%s\": %s",
				    __func__, path, strerror(errno));
			}
		}
	}
}


/*
 * Returns true if and only if a main article file exists for submission sid.
 */
static bool
submission_exists(const char *sid)
{
	char	 path[PATH_MAX];

	make_submission_file_path(sid, path, sizeof(path));
	return access(path, F_OK) == 0;
}


/*
 * Returns true if and only if a main article file exists for article sid.
 */
static bool
article_exists(const char *sid)
{
	char	 path[PATH_MAX];

	make_article_file_path(sid, path, sizeof(path));
	return access(path, F_OK) == 0;
}


/*
 * Returns true if and only if the a comment sid/pid exists.
 */
static bool
comment_exists(const char *sid, const char *pid)
{
	char	 path[PATH_MAX];

	return find_pid_path(sid, pid, path, sizeof(path)) == 0;
}


/*
 * Returns the HTTP request header string for h if it exists and is non-emtpy;
 * otherwise, returns NULL.
 */
static const char*
get_header_stringne(struct kreq *r, enum krequ h)
{
	const char	*v;

	if (r->reqmap[h] != NULL && r->reqmap[h]->val != NULL
	    && r->reqmap[h]->val[0] != '\0') {
		v = r->reqmap[h]->val;
	} else {
		v = NULL;
	}

	return v;
}


/* Compares string k with name field of action_table entry v. */
static int
compare_action_name(const void *k, const void *v)
{
	return strcmp((const char*)k, ((const struct action_details*)v)->name);
}


/*
 * khttp_parse validator function for field K_ACTION.  Performs binary search
 * of action_table.
 */
static int
valid_action(struct kpair *kp)
{
	int	 valid;
	const struct	action_details	*a;

	valid = 0;  /* Validation failed, unless proven otherwise. */

	if (kvalid_stringne(kp)) {
		a = bsearch(kp->val, action_table,
		    sizeof(action_table) / sizeof(struct action_details),
		    sizeof(struct action_details), compare_action_name);
		if (a != NULL) {
			/*
			 * Valid action: for simplicity, piggy back on integer
			 * parsed value.
			 */
			kp->type = KPAIR_INTEGER;
			kp->parsed.i = (a - action_table);

			valid = 1;
		}
	}
	return valid;
}


/*
 * khttp_parse validator function for field IPv4/IPv6 addresses.
 */
static int
valid_addr(struct kpair *kp)
{
	return kvalid_stringne(kp) && is_inet_addr(kp->val);
}


/*
 * Returns true if and only if the s is non-empty and contains only digits.
 * s must be non-NULL.
 */
static bool
is_all_digits(const char *s)
{
	size_t	 i;

	i = 0;
	while (s[i] != '\0' && isdigit((unsigned char)s[i])) {
		i++;
	}
	return i > 0 && s[i] == '\0';
}


/*
 * khttp_parse validator function for the authorisation cookie.
 */
static int
valid_auth_cookie(struct kpair *kp)
{
	struct	auth_cookie	*cookie;
	char	*i, *t, *r, *h, *b, *e;
	const char	*err;
	long long	 n;
	size_t	 len;
	int	 valid;

	valid = 0;  /* Failed, until proven otherwise. */

	if (kvalid_stringne(kp)) {
		/*
		 * Passed kvalid_stringne(), so we don't have to worry
		 * about embedded NUL characters.
		 */
		i = t = r = h = NULL;
		b = kp->val;
		if ((e = strchr(b, '/')) != NULL) {
			i = b; *e = '\0'; b = e + 1;
			if ((e = strchr(b, '/')) != NULL) {
				t = b; *e = '\0'; b = e + 1;
				if ((e = strchr(b, '/')) != NULL) {
					r = b; *e = '\0'; h = e + 1;
				}
			}
		}

		if (i != NULL && t != NULL && r != NULL && h != NULL
		    && user_is_valid_id(i) && is_all_digits(t)
		    && (len = strlen(r)) >= 7
		    && len < AUTH_COOKIE_REMOTE_SIZE
		    && strlen(h) == FDK_HMAC_B64_LEN) {
			/* Extract the expiry. */
			n = strtonum(t, 0, LLONG_MAX, &err);

			if ((n != 0 || err == NULL)
			    && base64_is_well_formed(h)) {
				/*
				 * Number and base64-encoded HMAC are
				 * reasonable.
				 */
				cookie = kcalloc(1,
				    sizeof(struct auth_cookie));

				if (base64_decode(h, cookie->hmac,
				    sizeof(cookie->hmac))
				    != FDK_HMAC_BIN_SIZE) {
					/* Failed. */
					free(cookie);
				} else {
					strlcpy(cookie->id, i, USER_ID_SIZE);
					cookie->t = n;
					strlcpy(cookie->remote, r,
					    AUTH_COOKIE_REMOTE_SIZE);

					kp->type = KPAIR__MAX;
					kp->val = (char*)cookie;
					kp->valsz = sizeof(*cookie);
					valid = 1;
				}
			}
		}
	}

	return valid;
}


/*
 * khttp_parse validator function for boolean values.
 */
static int
valid_boolean(struct kpair *kp)
{
	int	 valid;

	valid = 0;  /* Invalid until found otherwise. */

	if (kvalid_int(kp)) {
		if (kp->parsed.i != 0) {
			kp->parsed.i = 1;
		}
		valid = 1;
	} else if (kvalid_stringne(kp)) {
		if (strcasecmp(kp->parsed.s, "off") == 0
		    || strcasecmp(kp->parsed.s, "no") == 0
		    || strcasecmp(kp->parsed.s, "n") == 0
		    || strcasecmp(kp->parsed.s, "false") == 0
		    || strcasecmp(kp->parsed.s, "f") == 0) {
			kp->type = KPAIR_INTEGER;
			kp->parsed.i = 0;
			valid = 1;
		} else if (strcasecmp(kp->parsed.s, "on") == 0
		    || strcasecmp(kp->parsed.s, "yes") == 0
		    || strcasecmp(kp->parsed.s, "y") == 0
		    || strcasecmp(kp->parsed.s, "true") == 0
		    || strcasecmp(kp->parsed.s, "t") == 0) {
			kp->type = KPAIR_INTEGER;
			kp->parsed.i = 1;
			valid = 1;
		}
	}

	return valid;
}


/*
 * khttp_parse validator function for user IDs.
 */
static int
valid_id(struct kpair *kp)
{
#ifdef COMPAT_LEGACY_ID
	return kvalid_stringne(kp)
	    && kp->valsz >= USER_ID_MIN_LEN && kp->valsz <= USER_ID_MAX_LEN
	    && strcspn(kp->val, " ;<>&\n\r\t") == kp->valsz;
#else /* COMPAT_LEGACY_ID */
	return kvalid_stringne(kp) && user_is_valid_id(kp->val);
#endif /* COMPAT_LEGACY_ID */
}


/*
 * khttp_parse validator function for base64 HMAC strings (as used by the
 * fdk_hmac module).  The validator only checks that the value is well-formed
 * base64 data of the correct length.
 */
static int
valid_hmac(struct kpair *kp)
{
	return kvalid_stringne(kp) && kp->valsz == FDK_HMAC_B64_LEN
	    && base64_is_well_formed(kp->val);
}


/*
 * khttp_parse validator for user names (not IDs).
 */
static int
valid_name(struct kpair *kp)
{
	return kvalid_stringne(kp) && user_is_valid_name(kp->val);
}


/*
 * khttp_parse validator function for K_PID values (non-negative integers).
 */
static int
valid_pid(struct kpair *kp)
{
	int	 valid;

	valid = 0;  /* Assume failed unless proven otherwise. */

	/*
	 * The length check seems arbitrary but was used in the legacy code
	 * (perhaps to match checking sid against TIMESTAMP_LEN).  It is highly
	 * unlikely to present a problem in practice.
	 */
	if (kvalid_int(kp) && kp->parsed.i > 0 && kp->valsz <= TIMESTAMP_LEN) {
		kp->type = KPAIR_STRING;
		kp->parsed.s = kp->val;
		valid = 1;
	}

	return valid;
}


/*
 * khttp_parse validator function for K_SID values (timestamps).
 */
static int
valid_sid(struct kpair *kp)
{
	struct	tm	 ts;

	return kvalid_stringne(kp) && kp->valsz == TIMESTAMP_LEN
	    && strptime(kp->val, TIMESTAMP_FMT, &ts) != NULL;
}


/*
 * khttp_parse validator for subject.
 */
static int
valid_subject(struct kpair *kp)
{
	size_t	 n;
	int	 v;

	v = kvalid_stringne(kp);

	if (v) {
		n = 0;
		while (n < SUBJECT_SZ && kp->val[n] != '\0'
		    && isprint((unsigned char)kp->val[n]) && (kp->val[n] == ' '
		    || !isspace((unsigned char)kp->val[n]))) {
			n++;
		}
		v = n > 0 && n < SUBJECT_SZ && kp->val[n] == '\0'
		    && kp->val[0] != ' ' && kp->val[n - 1] != ' ';
	}

	return v;
}


#ifdef FEATURE_THEMES
/*
 * If the request contained a valid theme specification, gets that theme name,
 * otherwise gets the name of the default theme.
 * Sets *t to the a character sequence starting with the name of the theme,
 * and *n to the length of the name.
 */
static void
get_theme(struct kreq *r, const char **t, size_t *n)
{
	const char	*s;

	if (r->fieldmap[K_THEME] != NULL) {
		*t = r->fieldmap[K_THEME]->parsed.s;
		*n = r->fieldmap[K_THEME]->valsz;
	} else {
		DMSG(r, "%s() using default theme", __func__);
		*t = get_conf_val(CI_THEMES);
		s = strchr(*t, THEME_LIST_SEP);
		if (s != NULL) {
			*n = s - *t;
		} else {
			*n = strlen(*t);
		}
	}
}


/*
 * khttp_parse validator function for K_THEME values.
 */
static int
valid_theme(struct kpair *kp)
{
	const char	*cnf;
	const char	*t;
	int	 valid;
	bool	 matched;

	valid = 0;  /* Invalid until proven otherwise. */

	if (kvalid_stringne(kp)) {
		cnf = get_conf_val(CI_THEMES);

		matched = false;
		t = cnf;

		while (!matched && (t = strstr(t, kp->val)) != NULL) {
			matched = (t == cnf || *(t - 1) == THEME_LIST_SEP)
			    && (t[kp->valsz] == THEME_LIST_SEP
			    || t[kp->valsz] == '\0');
		}

		if (matched) {
			valid = 1;
		}
	}

	return valid;
}
#endif /* FEATURE_THEMES */


/*
 * khttp_parse validator function for K_TYPE values (HTML or text).
 */
static int
valid_type(struct kpair *kp)
{
	int	 valid;

	valid = 0;  /* Invalid until proven otherwise. */

	if (kvalid_stringne(kp) && (strcmp(kp->val, KV_TYPE_PLAIN) == 0
	    || strcmp(kp->val, KV_TYPE_HTML) == 0)) {
		kp->type = KPAIR_INTEGER;
		kp->parsed.i = strcmp(kp->val, KV_TYPE_PLAIN) == 0;
		valid = 1;
	}

	return valid;
}


/*
 * Prerequisite: r is not NULL and contains a valid action specification.
 * Return the action specified in the request.
 */
static inline enum action
get_action(const struct kreq *r)
{
	return(enum action)r->fieldmap[K_ACTION]->parsed.i;
}


/* Return "http" or "https", depenending on which is in use. */
static inline const char *
scheme_str(const struct kreq *r)
{
	return (r->scheme == KSCHEME_HTTPS ? SCHEME_HTTPS : SCHEME_HTTP);
}


/* Emits CGI/HTTP headers for status st and content-type ct */
static void
emit_http_start_headers(struct kreq *r, enum khttp st, enum kmime ct)
{
	DMSG(r, "%s() status=\"%s\", content-type=\"%s\"", __func__,
	    khttps[st], kmimetypes[ct]);

	khttp_head(r, kresps[KRESP_STATUS], "%s", khttps[st]);
	khttp_head(r, kresps[KRESP_CONTENT_TYPE], "%s", kmimetypes[ct]);
}


/* Emits HTTP Location header for l */
static void
emit_http_locn_header(struct kreq *r, const char *l)
{
	/* DMSG(r, "%s()", __func__); */

	khttp_head(r, kresps[KRESP_LOCATION], "%s", l);
}


/* Emits end of HTTP headers. */
static void
emit_http_end_headers(struct kreq *r)
{
	/* DMSG(r, "%s()", __func__); */

	khttp_body(r);
}


/*
 * Sends an HTTP response consisting of headers (with status 200 and type HTML)
 * and a document based on template t, with the rendering function being
 * render_common() if extra is NULL, or render_extra_html if extra is non-NULL.
 */
static void
resp_ok(struct kreq *r, const char *t, const char *extra)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	if (extra != NULL) {
		use_template_file(r, TMDCP, t, render_extra_html, extra);
	} else {
		use_template_file(r, TMDCP, t, render_common, NULL);
	}

	xpledge(PROMISES_MIN);
}


/*
 * Sends an HTTP response consisting of headers (with status 302 and type HTML)
 * with an authorisation cookie based on as (if non-NULL), and a document
 * based on template TMPL_S_302 using rendering function render_extra_html()
 * and targ.
 *
 * targ MUST be an _absolute_ URL - client redirection is performed.
 */
static void
resp_client_redirect(struct kreq *r, const char *targ, const char *as)
{
	DMSG(r, "%s() target = \"%s\"", __func__, targ);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_302, KMIME_TEXT_HTML);
	emit_http_locn_header(r, targ);

	if (as != NULL) {
		/* Set auth (or deauth) cookie. */
		DMSG(r, "%s() cookie = \"" AUTH_COOKIE_FMT "\"", __func__, as,
		    get_conf_val(CI_DOMAIN));

		khttp_head(r, kresps[KRESP_SET_COOKIE], AUTH_COOKIE_FMT,
		    as, get_conf_val(CI_DOMAIN));
	}
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_302, render_extra_html, targ);

	xpledge(PROMISES_MIN);
}


#ifdef UNDEADLY_TRANSITION
/*
 * Sends an HTTP response consisting of headers (with status 301 and type HTML)
 * with an authorisation cookie based on as (if non-NULL), and a document
 * based on template TMPL_S_301 using rendering function render_extra_html()
 * and targ.
 *
 * targ MUST be an _absolute_ URL - client redirection is performed.
 */
static void
resp_client_redirect_perm(struct kreq *r, const char *targ, const char *as)
{
	DMSG(r, "%s() target = \"%s\"", __func__, targ);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_301, KMIME_TEXT_HTML);
	emit_http_locn_header(r, targ);

	if (as != NULL) {
		/* Set auth (or deauth) cookie. */
		DMSG(r, "%s() cookie = \"" AUTH_COOKIE_FMT "\"", __func__, as,
		    get_conf_val(CI_DOMAIN));

		khttp_head(r, kresps[KRESP_SET_COOKIE], AUTH_COOKIE_FMT,
		    as, get_conf_val(CI_DOMAIN));
	}
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_301, render_extra_html, targ);

	xpledge(PROMISES_MIN);
}
#endif /* UNDEADLY_TRANSITION */


/*
 * Sends an HTTP response (with HTML) redirecting to article sid.
 */
static void
resp_redirect_to_article(struct kreq *r, const char *sid)
{
	char	 u[1024];

	resp_client_redirect(r, url(r, URL_F_ABS, u, sizeof(u), A_ARTICLE,
	    sid, NULL, NULL), NULL);
}


/*
 * Sends an HTTP response (with HTML) redirecting to the "front" page.
 */
static void
resp_redirect_to_front(struct kreq *r, const char *as)
{
	char	 u[1024];

	DMSG(r, "%s()", __func__);

	resp_client_redirect(r, url(r, URL_F_ABS, u, sizeof(u), A_FRONT, NULL,
	    NULL, NULL), as);
}


/*
 * Sends an HTTP status 400 (with HTML) response indicating that the request
 * was invalid.
 */
static void
resp_bad_request(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_400, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_400, render_common, NULL);

	xpledge(PROMISES_MIN);
}


/*
 * Sends an HTTP status 403 (with HTML) response indicating that the request
 * was forbidden. t specifies an extra template used with
 * render_extra_template() and TMPL_S_403.
 */
static void
resp_forbidden(struct kreq *r, const char *t)
{
	/* DMSG(r, "%s()", __func__); */

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_403, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_403, render_extra_template, t);

	xpledge(PROMISES_MIN);
}


/*
 * Sends an HTTP status 403 (with HTML) response indicating that the request
 * was forbidden because the user has insufficient rights to carry out the
 * requested action.
 */
static void
resp_insufficient_rights(struct kreq *r)
{
	resp_forbidden(r, TMPL_S_403_INSUF_RIGHTS);
}


/*
 * Sends an HTTP status 404 (with HTML) response indicating that the requested
 * resource was not found.
 */
static void
resp_not_found(struct kreq *r)
{
	/* DMSG(r, "%s()", __func__); */

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_404, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_404, render_common, NULL);

	xpledge(PROMISES_MIN);
}


/*
 * Emits CGI/HTTP headers and an HTML document indicating that the HTTP
 * request method was not supported.
 */
static void
resp_method_not_allowed(struct kreq *r)
{
	/* DMSG(r, "%s()", __func__); */

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_405, KMIME_TEXT_HTML);
	khttp_head(r, kresps[KRESP_ALLOW], "GET, HEAD, POST");
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_405, render_common, NULL);

	xpledge(PROMISES_MIN);
}


/*
 * Sends an HTTP status 500 (with HTML) response indicating that an internal
 * error ocurred.
 */
static void
resp_internal_error(struct kreq *r)
{
	/* DMSG(r, "%s()", __func__); */

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_500, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_500, render_common, NULL);

	xpledge(PROMISES_MIN);
}


/*
 * Sends an HTTP status 304 response.  If last is greater than zero and
 * is successfully converted to a time string, this is sent as a
 * "Last-Modified" header.
 */
static void
resp_not_modified(struct kreq *r, time_t last)
{
	struct	tm	 mod;
	char	 buf[256];

	/* DMSG(r, "%s()", __func__); */

	xpledge(PROMISES_R);

	/* No message body permitted with 304. */
	khttp_head(r, kresps[KRESP_STATUS], "%s", khttps[KHTTP_304]);

	/* Send a Last-Modified header if we can. */
	if (last > 0 && gmtime_r(&last, &mod) != NULL
	    && strftime(buf, sizeof(buf), RFC_822_FMT, &mod) != 0) {
		khttp_head(r, kresps[KRESP_LAST_MODIFIED], "%s", buf);
	}

	khttp_body(r);

	xpledge(PROMISES_MIN);
}


/*
 * Sends an HTTP status 503 (with HTML) response indicating that the service
 * is currently unavailable.
 */
static void
resp_service_unavailable(struct kreq *r)
{
	/* DMSG(r, "%s()", __func__); */

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_503, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_S_503, render_common, NULL);

	xpledge(PROMISES_MIN);
}


/*
 * Comparison callback function for binary searching names of configuration
 * items in store_config_entry().
 */
static int
compare_ci_name(const void *k, const void *i)
{
	int	 r;

	r = strncmp(((const struct lstr*)k)->s, *(const char * const *)i,
	    ((const struct lstr*)k)->l);

	/*
	 * Using strncmp() alone is insuffient: if it returns 0, we must check
	 * whether the possible match is of greater length.
	 */
	if (r == 0 && strlen(*(const char * const *)i)
	    > ((const struct lstr*)k)->l) {
		r = -1;
	}

	return r;
}


/*
 * conf_parse_file() callback function for storing configuration values in
 * global configuration.  Strings n (of length nlen) and v (of length vlen)
 * specify the name and value.  If a name matching n is found in the global
 * configuration and no value is already stored for that name (and no error
 * occurs), the v string is duplicated (to a NUL-terminated string) and stored,
 * and zero is returned; otherwise, non-zero is returned.  If a is non-NULL,
 * it is interpreted as being of type int *, and a detailed status code is
 * assigned to the dereferenced value.
 */
static int
store_config_entry(const char *n, size_t nlen, const char *v, size_t vlen,
    void *a)
{
	struct	lstr	 key;
	const char	**cname;
	const char	*s;
	int	 r;
	enum conf_item	 ci;

	key.l = nlen;
	key.s = n;

	r = CPCB_INTERNAL_ERR;  /* Until proven otherwise, assume failure. */

	cname = bsearch(&key, conf_name,
	    sizeof(conf_name) / sizeof(conf_name[0]), sizeof(conf_name[0]),
	    compare_ci_name);

	if (cname == NULL) {
		r = CPCB_BAD_NAME_ERR;
	} else {
		ci = (enum conf_item)(cname - conf_name);

		if (get_conf_val(ci) != NULL) {
			r = CPCB_DUP_VAL_ERR;
		} else if ((s = strndup(v, vlen)) == NULL) {
			r = CPCB_ALLOC_ERR;
		} else {
			set_conf_val(ci, s);
			r = CPCB_NO_ERR;
		}
	}

	if (a != NULL) {
		*(int *)a = r;
	}

	return r;
}


/*
 * Return non-zero if and only none of the required configuration (file)
 * entries is unset.
 */
static void
check_config(uint16_t *invalid, uint16_t *missing)
{
	enum conf_item	 i;
	uint16_t	 nv;  /* Bitset of items without values. */

	nv = 0;

	for (i = 0; i < sizeof(conf_name) / sizeof(conf_name[0]); i++) {
		if (get_conf_val(i) == NULL) {
			nv |= SNB(i);
		}
	}

	/* At present, we check only that a value is set. */
	*invalid = 0;
	*missing = nv;
}


/*
 * Copies to ts (of size sz) a timestamp (of format TIMESTAMP_FMT) for the
 * current time.  Aborts on failure.
 */
static void
make_current_timestamp(char *ts, size_t sz)
{
	struct	tm	 now;

	xutc_tm_now(&now);

	if (strftime(ts, sz, TIMESTAMP_FMT, &now) == 0) {
		fprintf(stderr, "%s: fatal: strftime failed\n", getprogname());
		exit(EXIT_FAILURE);
	}
}


/*
 * Returns true if and only if (no error occurs and) timestamp ts is in the
 * future.  ts must be the empty string (in which case zero is returned) or
 * a valid timestamp (of format TIMESTAMP_FMT).
 */
static bool
is_future_timestamp(const char *ts)
{
	char	 buf[TIMESTAMP_LEN + 1];
	bool	 future;

	future = false;

	if (ts[0] != '\0') {
		make_current_timestamp(buf, sizeof(buf));
		future = strcmp(ts, buf) > 0;
	}

	return future;
}


/*
 * Builds an appropriate self-referencing URL based on the specified values.
 * The URL is relative unless it requires HTTPS and that is not currently
 * in use.
 */
static char *
url(const struct kreq *r, unsigned int flags, char *s, size_t sz,
    enum action a, const char *sid, const char *pid,
    const char *opts, ...)
{
	const char	*sep;
#ifdef FEATURE_THRES
	const char	*thres;
	char	*ue;
#endif /* FEATURE_THRES */

	/* DMSG(r, "%s()", __func__); */

	sep = (flags & URL_F_ESC) ? ESCD_SEP : SEP;

#ifdef FEATURE_THRES
	thres = r->fieldmap[K_THRES] != NULL ? r->fieldmap[K_THRES]->parsed.s
	    : NULL;
	if (thres != NULL && (thres[0] == '\n' || thres[0] == '\r')) {
		thres = NULL;
	}
#endif /* FEATURE_THRES */

	/* If requested or necessary for the action, make absolute URL. */
	if (((flags & URL_F_ABS) && r->scheme == KSCHEME_HTTPS)
	    || (r->scheme != KSCHEME_HTTPS
	    && (action_table[a].rreq & RR_SCHEME_HTTPS))) {
		/* Make an absolute URL with scheme "https". */
		snprintf(s, sz, SCHEME_HTTPS "://%s", r->host);
		if (r->scheme == KSCHEME_HTTPS && r->port
		    != STD_PORT_N_HTTPS) {
			snprintf(s + strlen(s), sz - strlen(s), ":%u",
			r->port);
		} else if (r->scheme != KSCHEME_HTTPS
		    && strcmp(get_conf_val(CI_HTTPS_PORT), STD_PORT_S_HTTPS)
		    != 0) {
			snprintf(s + strlen(s), sz - strlen(s), ":%s",
			    get_conf_val(CI_HTTPS_PORT));
		}
	} else if (flags & URL_F_ABS) {
		/* Make an absolute URL with scheme "http". */
		snprintf(s, sz, SCHEME_HTTP "://%s", r->host);
		if (r->port != STD_PORT_N_HTTP) {
			snprintf(s + strlen(s), sz - strlen(s), ":%u",
			    r->port);
		}
	} else {
		s[0] = '\0';
	}

	/* Add the script name. */
	snprintf(s + strlen(s), sz - strlen(s), "%s?" KN_ACTION EQ "%s",
	    r->pname, action_table[a].name);

	if (sid != NULL && sid[0] != '\0') {
		snprintf(s + strlen(s), sz - strlen(s), "%s" KN_SID EQ "%s",
		    sep, sid);
	}
	if (pid != NULL && pid[0] != '\0') {
		snprintf(s + strlen(s), sz - strlen(s), "%s" KN_PID EQ "%s",
		    sep, pid);
	}

	if (opts != NULL && opts[0] != '\0') {
		va_list ap;

		va_start(ap, opts);
		vsnprintf(s + strlen(s), sz - strlen(s), opts, ap);
		va_end(ap);
	}

	/*
	 * There is corresponding code in render_common() [case MCR_URL_EXTRA]
	 * and render_common_form() [case MCR_FORM_EXTRA].
	 */

#ifdef FEATURE_THEMES
	/* Ignore invalid theme specification the in request. */
	if (r->fieldmap[K_THEME] != NULL) {
		snprintf(s + strlen(s), sz - strlen(s), "%s" KN_THEME EQ "%s",
		    sep, r->fieldmap[K_THEME]->parsed.s);
	}
#endif /* FEATURE_THEMES */

#ifdef FEATURE_THRES
	if (thres != NULL) {
		ue = kutil_urlencode(thres);
		if (ue == NULL) {
			msg(r, MSG_ERROR, "%s() kutil_urlencode failed for"
			    " thres", __func__);
		} else {
			snprintf(s + strlen(s), sz - strlen(s), "%s" KN_THRES
			    EQ "%s", sep, ue);
			free(ue); /* ue = NULL; */
		}
	}
#endif /* FEATURE_THRES */

#ifdef DEBUG
	if (rt_debug) {
		/* Keep debugging. */
		snprintf(s + strlen(s), sz - strlen(s), "%s" KN_DEBUG EQ "on",
		    sep);
	}
#endif /* DEBUG */

	return s;
}


/*
 * Outputs as HTML the time specified by tp with strftime(3) format fmt.
 */
static void
output_time(struct kreq *r, const struct tm *tp, const char *fmt)
{
	char	 buf[2048];

	if (strftime(buf, sizeof(buf), fmt, tp) == 0) {
		msg(r, MSG_ERROR, "%s() strftime failed", __func__);
		khttp_puts(r, FS_H_UNKNOWN);
	} else {
		khttp_puts(r, buf);
	}
}


/*
 * Outputs as HTML the current time with strftime(3) format fmt.
 */
static void
output_current_time(struct kreq *r, const char *fmt)
{
	struct	tm	 ts;

	xutc_tm_now(&ts);
	output_time(r, &ts, fmt);
}


/*
 * Outputs an unordered list (<ul>) of the descriptions (key_desc[]) of the
 * keys for which bits are set in kbad.
 */
static void
print_bad_keys(struct kreq *r, uint64_t kbad)
{
	uint64_t	 k;

	DMSG(r, "%s() bad = 0x%.16llx", __func__, kbad);

	khttp_puts(r, "<ul>\n");

	k = 0;
	while (kbad && k < sizeof(keys) / sizeof(struct kvalid)) {
		if (kbad & SNB(k)) {
			khttp_puts(r, "<li>");
			khttp_puts(r, key_desc[k]);
			khttp_puts(r, "</li>\n");

			kbad &= ~SNB(k);
		}
		k++;
	}

	khttp_puts(r, "</ul>\n");
}


/*
 * Template rendering function specific to CAPTCHAs.  *a must be a valid
 * struct captcha_details.
 */
static void
render_captcha(struct kreq *r, __unused char c, const char *m, size_t len,
    const void *a)
{
	const struct	captcha_details	*cd;

	DMSG(r, "%s(..., \"%.*s\", ...)", __func__, (int)len, m);

	cd = (const struct captcha_details *)a;

	if (STRNEQ(m, len, MCR_CAPTCHA_TXT)) {
		output_as_html(r, cd->text);
	} else if (STRNEQ(m, len, MCR_CAPTCHA_CHK)) {
		output_as_html(r, cd->check);
	} else {
		msg(r, MSG_ERROR, "%s() unhandled macro \"%.*s\"", __func__,
		    (int)len, m);
	}
}


/*
 * If t (of length len) is not of zero length, and the concatenation of p and t
 * is a safe relative path, the file specified by that concatenation (relative
 * to the templates directory) is processed as a template with macro-delimiting
 * character c, rendering function cb, and reference a; otherwise, an error
 * message is logged (with macro name m).
 */
static void
include_template(struct kreq *r, char c, const char *m, const char *p,
    const char *t, size_t tlen,
    void (*cb)(struct kreq *, char, const char *, size_t, const void *),
    const void *a)
{
	char	 incl[PATH_MAX];

	if (tlen == 0) {
		msg(r, MSG_ERROR, "%s() empty path for macro \"%s\"",
		    __func__, m);
	} else {
		snprintf(incl, sizeof(incl), "%s%.*s", p, (int)tlen, t);
		if (!is_safe_rel_path(t, tlen)) {
			msg(r, MSG_ERROR, "%s() unsafe path \"%.*s\" ignored"
			    " for macro \"%s\"", __func__, (int)tlen, t, m);
		} else {
			use_template_file(r, c, incl, cb, a);
		}
	}
}


/*
 * Expands the first (if b) or second (if not b) part of a conditional macro
 * using the macro-delimiting character after c.
 */
static void
expand_cond_macro(struct kreq *r, bool b, char c, const char *alt, size_t len,
    void (*cb)(struct kreq *, char, const char *, size_t, const void *),
    const void *a)
{
	const char	*sep;
	const char	*deeper;
	char	 as[3];

	as[0] = T_MACRO_ALT_BRA;
	as[1] = c;
	as[2] = T_MACRO_ALT_KET;

	/* (Attempt to) locate the separator between alternatives. */
	sep = memmem(alt, len, as, sizeof(as));

	/* Consider the macro marker character to use for the alternative. */
	if (c == '\0') {
		/* The maximum depth has been reached. */
		msg(r, MSG_ERROR, "%s() maximum macro depth reached",
		    __func__);
	} else if ((deeper = strchr(get_conf_val(CI_TEMPLATE_MACRO_CHARS), c))
	    == NULL) {
		/* Something is very wrong. */
		msg(r, MSG_ERROR, "%s() '%c' not found in \"%s\"", __func__, c,
		    get_conf_val(CI_TEMPLATE_MACRO_CHARS));
	} else {
		/* Use the next delimiting character, even if NUL. */
		deeper++;

		if (b) {
			if (sep == NULL || sep > alt) {
				use_template_buf(r, *deeper, alt, sep != NULL
				    ? (size_t)(sep - alt) : len, cb, a);
			}
		} else {
			if (sep != NULL && sep + sizeof(as) < alt + len) {
				use_template_buf(r, *deeper, sep + sizeof(as),
				    len - (sep - alt) - sizeof(as), cb, a);
			}
		}
	}
}


/*
 * Renders the "latest" articles.
 */
static void
sub_render_latest_list(struct kreq *r, char c)
{
	struct	entry	 latest[ARTICLES_FRONT];
	struct	entry	 e;
	char	 articles_dir[PATH_MAX];

	DMSG(r, "%s()", __func__);

	make_articles_dir_path(articles_dir, sizeof(articles_dir));
	find_articles(r, articles_dir, latest, ARTICLES_FRONT, false, &poll);
	memset(&e, 0, sizeof(e));

	prepare_and_render_al(r, c, &e, latest, ARTICLES_FRONT,
	    TMPL_AL_EARLIER);
}


/*
 * Template rendering function for macros common to most contexts.
 */
static void
render_common(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	char	 scratch[PATH_MAX];
#ifdef FEATURE_THEMES
	const char	*theme;
	size_t	 tlen;
#endif /* FEATURE_THEMES */
#ifdef FEATURE_THRES
	const char	*thres;
	char	*enc;
#endif /* FEATURE_THRES */
	size_t	 p;

	/* DMSG(r, "%s(..., \"%.*s\", ...)", __func__, (int)len, m); */

#ifdef FEATURE_THRES
	thres = r->fieldmap[K_THRES] != NULL ? r->fieldmap[K_THRES]->parsed.s
	    : NULL;
	if (thres != NULL && (thres[0] == '\n' || thres[0] == '\r')) {
		thres = NULL;
	}
#endif /* FEATURE_THRES */

	if (STRNEQ(m, len, MCR_SCRIPT_NAME)) {
		khttp_puts(r, r->pname);
	} else if (STRNEQ(m, len, MCR_SCHEME)) {
		khttp_puts(r, scheme_str(r));
	} else if (STRNEQ(m, len, MCR_URL_EXTRA)) {
		/*
		 * There is corresponding code in url() and
		 * render_common_form() [case MCR_FORM_EXTRA].
		 */
#ifdef FEATURE_THRES
		if (thres != NULL) {
			enc = kutil_urlencode(thres);
			if (enc == NULL) {
				msg(r, MSG_ERROR, "%s() kutil_urlencode failed"
				    " for thres", __func__);
			} else {
				khttp_puts(r, ESCD_SEP KN_THRES EQ);
				output_as_html(r, enc);
				free(enc); /* enc = NULL; */
			}
		}
#endif /* FEATURE_THRES */

#ifdef FEATURE_THEMES
		/* Ignore invalid theme specification in the request. */
		if (r->fieldmap[K_THEME] != NULL) {
			khttp_puts(r, ESCD_SEP KN_THEME EQ);
			khttp_puts(r, r->fieldmap[K_THEME]->parsed.s);
		}
#endif /* FEATURE_THEMES */

#ifdef DEBUG
		if (rt_debug) {
			khttp_puts(r, ESCD_SEP KN_DEBUG EQ "on");
		}
#endif /* DEBUG */

	} else if (STRNEQ(m, len, MCR_ID)) {
		if (KNOWN_USER()) {
#ifdef COMPAT_LEGACY_ID
			output_as_html(r, user.id);
#else /* COMPAT_LEGACY_ID */
			khttp_puts(r, user.id);
#endif /* COMPAT_LEGACY_ID */
		}
	} else if (STRNEQ(m, len, MCR_GROUP)) {
		if (KNOWN_USER()) {
			output_as_html(r, user.group);
		}
	} else if (STRNEQ(m, len, MCR_SERVER_NAME)) {
		khttp_puts(r, r->host);
	} else if (STRNEQ(m, len, MCR_SERVER_PORT)) {
		snprintf(scratch, sizeof(scratch), "%u", r->port);
		khttp_puts(r, scratch);
	} else if (STRNEQ(m, len, MCR_REMOTE_ADDR)) {
		khttp_puts(r, r->remote);
	} else if (STRNEQ(m, len, MCR_VERSION)) {
		khttp_puts(r, FWOBAC_VERSION);
	} else if (STRNEQ(m, len, MCR_LATEST_LIST)) {
		sub_render_latest_list(r, c);
#ifdef FEATURE_THRES
	} else if (STRNEQ(m, len, MCR_THRES)) {
		if (thres != NULL) {
			output_as_html(r, thres);
		}
#endif /* FEATURE_THRES */
	} else if (m[0] == '\0') {
		/* Provide a way to have doubled macro character in output. */
		khttp_putc(r, c);
		khttp_putc(r, c);
	} else if (len >= (p = strlen(MCR_PREFIX_TIME_FMT))
	    && strncmp(m, MCR_PREFIX_TIME_FMT, p) == 0) {
		if (len > p) {
			/* Make a NUL-terminated copy of the format. */
			snprintf(scratch, sizeof(scratch), "%.*s",
			    (int)(len - p), &m[p]);
			output_current_time(r, scratch);
		} else {
			output_current_time(r, DEFAULT_TIME_FMT);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_HTTPS))
	    && strncmp(m, MCR_PREFIX_COND_HTTPS, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_HTTPS "\"", __func__);
		} else {
			expand_cond_macro(r, r->scheme == KSCHEME_HTTPS, c,
			    &m[p], len - p, render_common, a);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_AUTH))
	    && strncmp(m, MCR_PREFIX_COND_AUTH, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_AUTH "\"", __func__);
		} else {
			expand_cond_macro(r, KNOWN_USER(), c, &m[p], len - p,
			    render_common, a);
		}
	} else if (STRNEQ(m, len, MCR_ACTION)) {
		if (r->fieldmap[K_ACTION] != NULL) {
			khttp_puts(r, action_table[
			    r->fieldmap[K_ACTION]->parsed.i].name);
		}
#ifdef FEATURE_THEMES
	} else if (STRNEQ(m, len, MCR_THEME)) {
		get_theme(r, &theme, &tlen);
		snprintf(scratch, sizeof(scratch), "%.*s", (int)tlen, theme);
		khttp_puts(r, scratch);
#endif /* FEATURE_THEMES */
	} else if (len >= (p = strlen(MCR_PREFIX_COND_DEBUG))
	    && strncmp(m, MCR_PREFIX_COND_DEBUG, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_DEBUG "\"", __func__);
		} else {
			expand_cond_macro(r,
#ifdef DEBUG
			    rt_debug,
#else /* DEBUG */
			    false,
#endif /* DEBUG */
			    c, &m[p], len - p, render_common, a);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_DEBUG_B))
	    && strncmp(m, MCR_PREFIX_COND_DEBUG_B, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_DEBUG_B "\"", __func__);
		} else {
			expand_cond_macro(r,
#ifdef DEBUG
			    true,
#else /* DEBUG */
			    false,
#endif /* DEBUG */
			    c, &m[p], len - p, render_common, a);
		}
	} else {
		msg(r, MSG_WARNING, "%s() unhandled macro \"%.*s\"", __func__,
		    (int)len, m);
	}
}


/*
 * Template rendering function for macros common to forms.  Where relevant,
 * a must be of type (const struct field_errs *).
 */
static void
render_common_form(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	struct	captcha_details	 cap_info;

	DMSG(r, "%s(..., \"%.*s\", ...)", __func__, (int)len, m);

	if (STRNEQ(m, len, MCR_FORM_EXTRA)) {
		/*
		 * There is corresponding code in url() and render_common()
		 * [case MCR_URL_EXTRA].
		 *
		 * This case differs from the other two insomuch as it
		 * could be handled by form templates, but we elect not
		 * to follow that approach because of the burden it would
		 * place on the template maintainer, and because the code is
		 * conditionally compiled.
		 */

#ifdef DEBUG
		if (rt_debug) {
			khttp_puts(r, "<input type=\"hidden\""
			    " name=\"" KN_DEBUG "\" value=\"on\">\n");
		}
#endif /* DEBUG */

#ifdef FEATURE_THEMES
		/*
		 * K_THEME is unusual as it is "global" (rather than related to
		 * the purpose of the form) and its presence depends on
		 * compilation options.
		 */
		if (r->fieldmap[K_THEME] != NULL) {
			khttp_puts(r, "<input type=\"hidden\""
			    " name=\"" KN_THEME "\" value=\"");
			khttp_puts(r, r->fieldmap[K_THEME]->parsed.s);
			khttp_puts(r, "\">\n");
		}
#endif /* FEATURE_THEMES */

		/* XXX THRES should be handled here (?) */
	} else if (STRNEQ(m, len, MCR_CAPTCHA)) {
		if (ANON_USER()) {
			/*
			 * User is not logged in, so generate and show
			 * a CAPTCHA.
			 */
			if (generate_captcha(get_conf_val(CI_CAPTCHA_KEY),
			    CAPTCHA_LEN, cap_info.text, sizeof(cap_info.text),
			    cap_info.check, sizeof(cap_info.check))
			    != CAPTCHA_ERR_NONE) {
				fprintf(stderr, "%s: fatal: CAPTCHA generation"
				    " failed\n", getprogname());
				exit(EXIT_FAILURE);
			}
			use_template_file(r, c, TMPL_CAPTCHA, render_captcha,
			    &cap_info);
		}
	} else if (STRNEQ(m, len, MCR_F_ERR_MSG)) {
		if (((const struct field_errs *)a)->invalid) {
			use_template_file(r, c, TMPL_F_ERR_INVALID,
			    render_common_form, a);
		}
		if (((const struct field_errs *)a)->missing) {
			use_template_file(r, c, TMPL_F_ERR_MISSING,
			    render_common_form, a);
		}
	} else if (STRNEQ(m, len, MCR_F_ERR_INVALID_LIST)) {
		if (a == NULL) {
			msg(r, MSG_ERROR, "%s() unable to handle macro"
			    " \"MCR_F_ERR_INVALID_LIST\" because a is NULL",
			    __func__);
		} else {
			print_bad_keys(r,
			    ((const struct field_errs *)a)->invalid);
		}
	} else if (STRNEQ(m, len, MCR_F_ERR_MISSING_LIST)) {
		if (a == NULL) {
			msg(r, MSG_ERROR, "%s() unable to handle macro"
			    " \"MCR_F_ERR_MISSING_LIST\" because a is NULL",
			    __func__);
		} else {
			print_bad_keys(r,
			    ((const struct field_errs *)a)->missing);
		}
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Template rendering function for macros common to many contexts involving a
 * value of type struct entry.  a is interpreted as a pointer to such a struct.
 */
static void
render_common_entry(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	const struct	entry	*e;
	char	 s[1024];
	size_t	 n;

	e = (const struct entry *)a;

	if (STRNEQ(m, len, MCR_SID)) {
		if (e->sid[0] != '\0') {
			khttp_puts(r, e->sid);
		}
	} else if (STRNEQ(m, len, MCR_PID)) {
		if (e->pid[0] != '\0') {
			khttp_puts(r, e->pid);
		}
	} else if (STRNEQ(m, len, MCR_SUBJECT)) {
		if (e->subject[0] != '\0') {
			khttp_puts(r, e->subject);
		}
	} else if (STRNEQ(m, len, MCR_SUBJECT_RAW)) {
		if (e->subject[0] != '\0') {
			output_as_html(r, e->subject);
		}
	} else if (STRNEQ(m, len, MCR_ID_AUTHOR)) {
		if (e->id_author[0] != '\0') {
			output_as_html(r, e->id_author);
		}
	} else if (STRNEQ(m, len, MCR_ADDR_AUTHOR)) {
		if (e->addr_author[0] != '\0') {
			khttp_puts(r, e->addr_author);
		}
	} else if (STRNEQ(m, len, MCR_ID_EDITOR)) {
		if (e->id_editor[0] != '\0') {
			output_as_html(r, e->id_editor);
		}
	} else if (STRNEQ(m, len, MCR_ADDR_EDITOR)) {
		if (e->addr_editor[0] != '\0') {
			khttp_puts(r, e->addr_editor);
		}
	} else if (STRNEQ(m, len, MCR_NAME)) {
		if (e->name[0] != '\0') {
			output_as_html(r, e->name);
		}
	} else if (STRNEQ(m, len, MCR_EMAIL)) {
		if (e->email[0] != '\0') {
			output_as_html(r, e->email);
		}
	} else if (STRNEQ(m, len, MCR_HREF)) {
		if (e->href[0] != '\0') {
			output_as_html(r, e->href);
		}
	} else if (STRNEQ(m, len, MCR_DEPT)) {
		if (e->dept[0] != '\0') {
			output_as_html(r, e->dept);
		}
	} else if (STRNEQ(m, len, MCR_TOPIC)) {
		if (e->topic[0] != '\0') {
			output_as_html(r, e->topic);
		}
	} else if (len >= (n = strlen(MCR_PREFIX_TIME_A_FMT))
	    && strncmp(m, MCR_PREFIX_TIME_A_FMT, n) == 0) {
		if (e->time_authored.tm_zone != NULL) {
			if (len > n) {
				/* Make a NUL-terminated copy of the format. */
				snprintf(s, sizeof(s), "%.*s", (int)(len - n),
				    &m[n]);
				output_time(r, &e->time_authored, s);
			} else {
				output_time(r, &e->time_authored,
				    DEFAULT_TIME_FMT);
			}
#ifdef COMPAT_LEGACY_R
		/*
		 * Old content files may not have a value, so try falling back
		 * to something which ought to be available.
		 */
		} else if (e->time_edited.tm_zone != NULL) {
			DMSG(r, "%s() falling back to time_edited for"
			    " \"MCR_PREFIX_TIME_A_FMT\"", __func__);
			if (len > n) {
				/* Make a NUL-terminated copy of the format. */
				snprintf(s, sizeof(s), "%.*s", (int)(len - n),
				    &m[n]);
				output_time(r, &e->time_edited, s);
			} else {
				output_time(r, &e->time_edited,
				    DEFAULT_TIME_FMT);
			}
#endif /* COMPAT_LEGACY_R */
		} else {
			msg(r, MSG_ERROR, "%s() value unavailable for \""
			    "MCR_PREFIX_TIME_A_FMT\"", __func__);
		}
	} else if (len >= (n = strlen(MCR_PREFIX_TIME_E_FMT))
	    && strncmp(m, MCR_PREFIX_TIME_E_FMT, n) == 0) {
		if (e->time_edited.tm_zone != NULL) {
			if (len > n) {
				/* Make a NUL-terminated copy of the format. */
				snprintf(s, sizeof(s), "%.*s", (int)(len - n),
				    &m[n]);
				output_time(r, &e->time_edited, s);
			} else {
				output_time(r, &e->time_edited,
				    DEFAULT_TIME_FMT);
			}
		} else {
			msg(r, MSG_ERROR, "%s() value unavailable for \""
			    "MCR_PREFIX_TIME_E_FMT\"", __func__);
		}
	} else if (STRNEQ(m, len, MCR_BODY)) {
		if (e->body != NULL && e->body[0] != '\0') {
			if (e->flags & FLAG_PLAIN) {
				output_as_html(r, e->body);
			} else {
				khttp_puts(r, e->body);
			}
		}
	} else if (STRNEQ(m, len, MCR_CONTENT0)) {
		if (e->body != NULL && e->body[0] != '\0') {
			output_as_html(r, e->body);
		}
	} else if (STRNEQ(m, len, MCR_MORE)) {
		if (e->more != NULL && e->more[0] != '\0') {
			if (e->flags & FLAG_PLAIN) {
				output_as_html(r, e->more);
			} else {
				khttp_puts(r, e->more);
			}
		}
	} else if (STRNEQ(m, len, MCR_CONTENT1)) {
		if (e->more != NULL && e->more[0] != '\0') {
			output_as_html(r, e->more);
		}
	} else if (len >= (n = strlen(MCR_PREFIX_COND_NE_NAME))
	    && strncmp(m, MCR_PREFIX_COND_NE_NAME, n) == 0) {
		if (len == n) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_NE_NAME "\"", __func__);
		} else {
			expand_cond_macro(r, e->name[0] != '\0', c, &m[n],
			    len - n, render_common_entry, a);
		}
	} else if (len >= (n = strlen(MCR_PREFIX_COND_NE_HREF))
	    && strncmp(m, MCR_PREFIX_COND_NE_HREF, n) == 0) {
		if (len == n) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_NE_HREF "\"", __func__);
		} else {
			expand_cond_macro(r, e->href[0] != '\0', c, &m[n],
			    len - n, render_common_entry, a);
		}
	} else if (len >= (n = strlen(MCR_PREFIX_COND_HREF_URL))
	    && strncmp(m, MCR_PREFIX_COND_HREF_URL, n) == 0) {
		if (len == n) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_HREF_URL "\"", __func__);
		} else {
			expand_cond_macro(r, e->href[0] != '\0'
			    && might_be_url(e->href), c, &m[n], len - n,
			    render_common_entry, a);
		}
#ifdef FEATURE_MODERATE
	} else if (STRNEQ(m, len, MCR_MODCOUNT)) {
		snprintf(s, sizeof(s), "%d", e->modcount);
		khttp_puts(r, s);
	} else if (STRNEQ(m, len, MCR_MODSUM)) {
		snprintf(s, sizeof(s), "%d", e->modsum);
		khttp_puts(r, s);
#endif /* FEATURE_MODERATE */
	} else {
		render_common(r, c, m , len, a);
	}
}


/*
 * Rendering function which, if m is MCR_EXTRA, treats arg a as a string to
 * output as-is, or invokes render_common() otherwise.
 */
static void
render_extra_simple(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	if (STRNEQ(m, len, MCR_EXTRA)) {
		khttp_puts(r, (const char *)a);
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Rendering function which, if m is MCR_EXTRA, treats arg a as a string to
 * output as HTML, or invokes render_common() otherwise.
 */
static void
render_extra_html(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	if (STRNEQ(m, len, MCR_EXTRA)) {
		output_as_html(r, (const char *)a);
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Rendering function which, if m is MCR_EXTRA, treats arg a as a template to
 * use, or invokes render_common() otherwise.
 */
static void
render_extra_template(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	if (STRNEQ(m, len, MCR_EXTRA)) {
		use_template_file(r, c, (const char *)a, render_common, NULL);
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Callback function for tmpl_file for output of non-macro data.
 */
static int
output_plain(const void *d, size_t len, void *a)
{
	khttp_write((struct kreq *)a, d, len);

	return 0;
}


/*
 * Callback function for tmpl_file for output of macro replacements.
 */
static int
output_macro(const void *s, size_t len, void *a)
{
	struct	macro_cb_info	*i;
	char	 pre[PATH_MAX];
#ifdef FEATURE_THEMES
	const char	*theme;
	size_t	 tlen;
#endif /* FEATURE_THEMES */
	size_t	 p;
	const char	*m;

	i = (struct macro_cb_info *)a;
	m = (const char *)s;

	if (len >= (p = strlen(MCR_PREFIX_INCLUDE))
	    && strncmp(m, MCR_PREFIX_INCLUDE, p) == 0) {
		include_template(i->r, i->c, MCR_PREFIX_INCLUDE, "",
		    &m[p], len - p, i->f, i->a);
	} else if (len >= (p = strlen(MCR_PREFIX_INCLUDE_G))
	    && strncmp(m, MCR_PREFIX_INCLUDE_G, p) == 0) {
		snprintf(pre, sizeof(pre), FMT_TMPL_INCLUDE_G,
		    KNOWN_USER() ? user.group : PSEUDO_GROUP_ANON);
		include_template(i->r, i->c, MCR_PREFIX_INCLUDE_G, pre,
		    &m[p], len - p, i->f, i->a);
#ifdef FEATURE_THEMES
	} else if (len >= (p = strlen(MCR_PREFIX_INCLUDE_T))
	    && strncmp(m, MCR_PREFIX_INCLUDE_T, p) == 0) {
		get_theme(i->r, &theme, &tlen);
		snprintf(pre, sizeof(pre), FMT_TMPL_INCLUDE_T,
		    (int)tlen, theme);
		include_template(i->r, i->c, MCR_PREFIX_INCLUDE_T, pre,
		    &m[p], len - p, i->f, i->a);
#endif /* FEATURE_THEMES */
	} else {
		i->f(i->r, i->c, (const char *)s, len, i->a);
	}

	return 0;
}


/*
 * Outputs template buffer buf (of length len) with macro substitutions.
 * Macro-delimiting character is c, and macro replacement is performed by cb
 * (with reference a).
 */
static int
use_template_buf(struct kreq *r, char c, const char *buf, size_t len,
    void (*cb)(struct kreq *, char, const char *, size_t, const void *),
    const void *a)
{
	struct	macro_cb_info	 cbi;

	/* DMSG(r, "%s() template = \"%s\"", __func__, tn); */

	if (cb == NULL) {
		return tmpl_buf(buf, len, c, output_plain, r, NULL, NULL);
	} else {
		cbi.r = r;
		cbi.f = cb;
		cbi.a = a;
		cbi.c = c;

		return tmpl_buf(buf, len, c, output_plain, r, output_macro,
		    &cbi);
	}
}


/*
 * Outputs template file tn with macro substitutions.  Macro-delimiting
 * character is c, and macro replacement is performed by cb (with reference a).
 * If the first character of tn is not '/', it is treated as relative to the
 * template directory (configuration item CI_TEMPLATE_DIR).
 */
static int
use_template_file(struct kreq *r, char c, const char *tn,
    void (*cb)(struct kreq *, char, const char *, size_t, const void *),
    const void *a)
{
	struct	macro_cb_info	 cbi;
	char	 tp[PATH_MAX];

	/* DMSG(r, "%s() template = \"%s\"", __func__, tn); */

	if (tn[0] == '/') {
		strlcpy(tp, tn, sizeof(tp));
	} else {
		snprintf(tp, sizeof(tp), "%s/%s",
		    get_conf_val(CI_TEMPLATE_DIR), tn);
	}

	if (cb == NULL) {
		return tmpl_file(tp, c, output_plain, r, NULL, NULL);
	} else {
		cbi.r = r;
		cbi.f = cb;
		cbi.a = a;
		cbi.c = c;

		return tmpl_file(tp, c, output_plain, r, output_macro, &cbi);
	}
}


/*
 * Template rendering function for login form.
 */
static void
render_login(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	enum action	 act;

	DMSG(r, "%s() %.*s", __func__, (int)len, m);

	if (STRNEQ(m, len, MCR_INTRO)) {
		act = get_action(r);
		if (act == A_AUTH) {
			/* User is back here due to authentication failure. */
			use_template_file(r, c, TMPL_LOGIN_FAILED,
			    render_common, NULL);
		} else if (act == A_CONFIRM) {
			/* Account registration has been confirmed. */
			use_template_file(r, c, TMPL_LOGIN_REG_CONFRMD,
			    render_common, NULL);
		}
	} else if (STRNEQ(m, len, MCR_ID)) {
		if (r->fieldmap[K_ID] != NULL) {
			output_as_html(r, r->fieldmap[K_ID]->parsed.s);
		}
	} else {
		render_common_form(r, c, m, len, a);
	}
}


/*
 * Template rendering function for RSS output.
 */
static void
render_rss(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	struct	a_rend_info	 details;
	size_t	 i;
	size_t	 n;

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	if (STRNEQ(m, len, MCR_ITEMS)) {
		if (r->fieldmap[K_ITEMS] != NULL) {
			n = r->fieldmap[K_ITEMS]->parsed.i;
			/* Constrain value to acceptable range. */
			if (n < RSS_ITEMS_MIN) {
				n = RSS_ITEMS_MIN;
			} else if (n > RSS_ITEMS_MAX) {
				n = RSS_ITEMS_MAX;
			}
		} else {
			n = RSS_ITEMS_DEFAULT;
		}

		/* Limit to length of array. */
		if (n > sizeof(newest) / sizeof(newest[0])) {
			n = sizeof(newest) / sizeof(newest[0]);
		}

		i = 0;
		while (i < n && newest[i].sid[0] != '\0') {
			details.e = &newest[i];
			details.last_change = 0;
			details.last_read = 0;
			details.cmts_count = count_comments(details.e->sid);
			details.cmts_open = are_comments_open(details.e->sid);
			use_template_file(r, c, TMPL_SUMMARY_ITEM,
			    render_article, &details);

			i++;
		}
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Rendering function for the articles for one day within an article list.
 * For macro MCR_AL_DAY_ARTICLE_LIST, renders the articles for the given
 * day (using template TMPL_AL_ARTICLE_ITEM with render_article()); for other
 * macros, calls render_article with the first article for the day (if there
 * is one) to allow showing the date.
 */
static void
render_al_day_item(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	const struct	al_article_list	*day_list;
	size_t	 n;

	DMSG(r, "%s()", __func__);

	day_list = (const struct al_article_list *)a;

	if (STRNEQ(m, len, MCR_AL_DAY_ARTICLE_LIST)) {
		/* Render articles for the day. */
		DMSG(r, "%s() articles for day = %zu", __func__, day_list->na);
		for (n = 0; n < day_list->na; n++) {
			use_template_file(r, TMDCP, TMPL_AL_ARTICLE_ITEM,
			    render_article, &day_list->dai[n]);
		}
	} else {
		/* Call render_article with first article for the day. */
		if (day_list->na == 0) {
			msg(r, MSG_WARNING, "%s() no articles for day",
			    __func__);
		} else {
			render_article(r, c, m, len, &day_list->dai[0]);
		}
	}
}


/*
 * Rendering function for an article list (such as the "earlier" or "latest"
 * list).  Provides special handling for macros MCR_PREFIX_COND_A_FOUND and
 * MCR_AL_DAY_LIST, and passes other macros to render_common().
 */
static void
render_al(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	struct	al_article_list	 day_i;
	const struct	al_info	*ali;
	size_t	 n;

	DMSG(r, "%s()", __func__);

	if (STRNEQ(m, len, MCR_AL_DAY_LIST)) {
		/* Render info/articles for each day. */
		ali = (const struct al_info *)a;
		if (ali == NULL || ali->len == 0 || ali->ari == NULL
		    || ali->di == NULL || ali->di[0].cnt == 0) {
			msg(r, MSG_NOTE, "%s() zero articles", __func__);
		} else {
			n = 0;
			while (n < ali->len && ali->di[n].cnt > 0) {
				day_i.dai = &ali->ari[ali->di[n].fst];
				day_i.na = ali->di[n].cnt;
				use_template_file(r, TMDCP, TMPL_AL_DAY_ITEM,
				    render_al_day_item, &day_i);
				n++;
			}
		}
	} else if (len >= (n = strlen(MCR_PREFIX_COND_A_FOUND))
	    && strncmp(m, MCR_PREFIX_COND_A_FOUND, n) == 0) {
		ali = (const struct al_info *)a;

		if (len == n) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_A_FOUND "\"", __func__);
		} else {
			expand_cond_macro(r, ali->len > 0, c, &m[n], len - n,
			    render_al, a);
		}
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Writes a summary of articles of aa.  aa must be of size an and the articles
 * must be ordered by sid (ascending or descending).  An sid of the empty
 * string denotes end-of-list.
 *
 * Articles are grouped by (distinct) day.  Conceptually, we render a list
 * of per-day article lists, but the use of templates means that the output
 * need not reflect this.
 */
static void
prepare_and_render_al(struct kreq *r, __unused char c, const struct entry *e,
    struct entry aa[], size_t an, const char *t)
{
	struct	al_info	 al;
	size_t	 days;
	size_t	 i;

	DMSG(r, "%s()", __func__);

	/* Determine number of articles. */
	i = 0;
	while (i < an && aa[i].sid[0] != '\0') {
		i++;
	}

	if (i == 0) {
		DMSG(r, "%s(): zero articles", __func__);
	} else {
		al.len = i;
		al.ari = NULL;
		al.di  = NULL;
		days = 0;

		/*
		 * In addition to the rendering information for each article,
		 * we need to know the number of distinct days having articles,
		 * and, for each of those, the number of articles and index of
		 * the first one.
		 */

		/*
		 * Allocate article and day information structures.
		 *
		 * There can be no more days that articles, so, to save double
		 * processing, allocate sufficient day details for one article
		 * per day.
		 */

		al.ari = (struct a_rend_info *)kreallocarray(NULL, al.len,
		    sizeof(struct a_rend_info));
		al.di = (struct al_day_info *)kreallocarray(NULL, al.len,
		    sizeof(struct al_day_info));

		/* Initialise the per-day article counts. */
		for (i = 0; i < al.len; i++) {
			al.di[i].cnt = 0;
		}

		for (i = 0; i < al.len; i++) {
			/* Set up the required information for the article. */
			aa[i].flags |= e->flags;
			al.ari[i].e = &aa[i];
			al.ari[i].last_change = find_newest_change(aa[i].sid,
			    false);
			if (KNOWN_USER()) {
				if (!get_lastread(aa[i].sid,
				    &al.ari[i].last_read, 0)) {
					al.ari[i].last_read = 0;
				}
			} else {
				al.ari[i].last_read = 0;
			}

			al.ari[i].cmts_count = count_comments(aa[i].sid);
			al.ari[i].cmts_open = are_comments_open(aa[i].sid);

			/* Set up the day information. */
			if (days == 0 || strncmp(aa[i].sid,
			    al.ari[al.di[days - 1].fst].e->sid, 8) != 0) {
				/* This is the first article for a day. */
				al.di[days].fst = i;
				al.di[days].cnt = 1;
				days++;
			} else {
				/* Increment counter for day. */
				al.di[days - 1].cnt++;
			}
		}

		DMSG(r, "%s(): articles = %zu, days = %zu", __func__, al.len,
		    days);

		use_template_file(r, TMDCP, t, render_al, &al);

		free(al.di);  /* al.di = NULL; */
		free(al.ari);  /* al.ari = NULL; */
	}
}


/*
 * Template rendering function used in do_action_front() and
 * do_action_submissions().
 */
static void
render_front(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	struct	a_rend_info	 ai;
	const struct	entry	*e;
	char	 s[8192];
	char	 scratch[1024];
	FILE	*f;
	size_t	 i;
#ifdef FEATURE_VOTE
	int	 val;
#endif /* FEATURE_VOTE */
	bool	 hdr;

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	e = (const struct entry *)a;

	if (STRNEQ(m, len, MCR_ARTICLES)) {
		for (i = 0; (e->flags & FLAG_SUBMISSION || i < ARTICLES_FRONT)
		    && i < sizeof(newest) / sizeof(newest[0]) &&
		    newest[i].sid[0] != '\0'; ++i) {
			newest[i].flags |= e->flags;
			ai.e = &newest[i];
			if (e->flags & FLAG_SUBMISSION) {
				ai.last_change = 0;
				ai.last_read = 0;
				ai.cmts_count = 0;
				ai.cmts_open = false;
			} else {
				ai.last_change = find_newest_change(ai.e->sid,
				    false);
				if (KNOWN_USER()) {
					if (!get_lastread(ai.e->sid,
					    &ai.last_read, 0)) {
						ai.last_read = 0;
					}
				} else {
					ai.last_read = 0;
				}

				ai.cmts_count = count_comments(ai.e->sid);
				ai.cmts_open = are_comments_open(ai.e->sid);
			}
			use_template_file(r, c, (e->flags & FLAG_SUBMISSION)
			    ? TMPL_SUBMISSION_ITEM : TMPL_ARTICLE_ITEM,
			    render_article, &ai);
		}
	} else if (STRNEQ(m, len, MCR_OLDER_LIST)) {
		if (e->flags & FLAG_SUBMISSION) {
			msg(r, MSG_WARNING, "%s() ignoring macro \""
			    MCR_OLDER_LIST "\" for submissions", __func__);
		} else {
			prepare_and_render_al(r, c, e, &newest[ARTICLES_FRONT],
			    ARTICLES_OLDER_LIST, TMPL_AL_EARLIER);
		}
	} else if (STRNEQ(m, len, MCR_POLL)) {/*XXX*/
		hdr = true;

		if (poll.sid[0] == '\0') {
			return;
		}
		if ((f = fopen(poll.fn, "r")) == NULL) {
			return;
		}
#ifdef FEATURE_VOTE
		/* XXX this should use a template */
		snprintf(s, sizeof(s), "<form method=\"post\" action=\"%s\">"
		    "\n", url(r, URL_F_ESC, scratch, sizeof(scratch), A_VOTE,
			NULL, NULL, NULL));
		khttp_puts(r, s);
		khttp_puts(r, "<b>");
		output_as_html(r, poll.subject);
		khttp_puts(r, "</b><br>\n");

#ifdef DEBUG
		if (rt_debug) {
			khttp_puts(r, "<input type=\"hidden\""
			    " name=\"" KN_DEBUG "\" value=\"on\">\n");
		}
#endif /* DEBUG */

		snprintf(s, sizeof(s), "<input type=\"hidden\" name=\"" KN_SID
		    "\" value=\"%s\">\n", poll.sid);
		khttp_puts(r, s);
		val = 0;
		while (fgets(s, sizeof(s), f) != NULL) {
			if (hdr) {
				if (s[0] == '\n') {
					hdr = false;
				}
				continue;
			}
			khttp_puts(r, "<input type=\"radio\" name=\"" KN_VAL
			    "\" value=\"");
			snprintf(scratch, sizeof(scratch), "%d\"> ", val++);
			khttp_puts(r, scratch);
			khttp_puts(r, s);
			khttp_puts(r, "<br>\n");
		}
		khttp_puts(r, "<br><input type=\"submit\" value=\"Vote\">\n");
		snprintf(s, sizeof(s), "<a href=\"%s\">[ %d comments ]</a>\n"
		    "</form>", url(r, URL_F_ESC, scratch, sizeof(scratch),
		    A_ARTICLE, poll.sid, NULL, NULL),
		    count_comments(poll.sid));
		khttp_puts(r, s);
#else /* FEATURE_VOTE */
		khttp_puts(r, "<p><b>");
		output_as_html(r, poll.subject);
		khttp_puts(r, "</b><br>\n");

		while (fgets(s, sizeof(s), f) != NULL) {
			if (hdr) {
				if (s[0] == '\n') {
					hdr = false;
				}
				continue;
			}
			khttp_puts(r, s);
			khttp_puts(r, "<br>\n");
		}
		khttp_puts(r, "</p>\n");
		snprintf(s, sizeof(s), "<a href=\"%s\">[ %u comments ]</a>\n",
		    url(r, URL_F_ESC, scratch, sizeof(scratch), A_ARTICLE,
			poll.sid, NULL, NULL),
		    count_comments(poll.sid));
		khttp_puts(r, s);
#endif /* FEATURE_VOTE */
		fclose(f);
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Template rendering function for submission, article (and its comments),
 * or comment (and its comments).
 */
static void
render_article(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	struct	entry	 comments[256];
#ifdef FEATURE_A_PREV_NEXT_LINKS
	struct	entry	 prev, next;
	struct	entry	 p;
#endif /* FEATURE_A_PREV_NEXT_LINKS */
	struct	tm	 time_sid;
	struct	c_rend_info	 ci;
	const struct	a_rend_info	*ai;
	char	 s[8192];
	char	 fn[PATH_MAX];
	char	 u[1024];
	int	 v[32];
	FILE	*f;
	char	*t;
#ifdef FEATURE_THRES
	const char	*thres;
#endif /* FEATURE_THRES */
	double	 pc;
	time_t	 t_ignored;
	size_t	 i;
	unsigned int	 prev_level;
#ifdef FEATURE_A_PREV_NEXT_LINKS
	unsigned int	 ni;
#endif /* FEATURE_A_PREV_NEXT_LINKS */
	int	 line, votesum;
	bool	 body;

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	ai = (const struct a_rend_info *)a;

#ifdef FEATURE_THRES
	thres = r->fieldmap[K_THRES] != NULL ? r->fieldmap[K_THRES]->parsed.s
	    : NULL;
	if (thres != NULL && (thres[0] == '\n' || thres[0] == '\r')) {
		thres = NULL;
	}
#endif /* FEATURE_THRES */

	if (STRNEQ(m, len, MCR_BODY)) {
		body = false;
		line = 0;
		votesum = 0;

		if (ai->e->body != NULL) {
			render_common_entry(r, c, m, len, ai->e);
			return;
		}
		if ((f = fopen(ai->e->fn, "r")) == NULL) {
			msg(r, MSG_ERROR, "%s() fopen: %s: %s", __func__,
			    ai->e->fn, strerror(errno));
			khttp_puts(r, FS_H_INT_ERR);
			return;
		}
		if (ai->e->flags & FLAG_POLL) {
			read_poll(ai->e->sid, v, sizeof(v) / sizeof(v[0]),
			    &votesum);
			khttp_puts(r, "<table>\n");
		}
		while (fgets(s, sizeof(s), f) != NULL) {
			if (body) {
				if (ai->e->flags & FLAG_POLL) {
					pc = votesum ? ((double)v[line] *
					    100.0 / (double)votesum) : 0.0;

					snprintf(fn, sizeof(fn), "<tr><td>%s"
					    "</td><td>%.1f%% (%d votes)</td>"
					    "</tr>\n", s, pc, v[line]);
					khttp_puts(r, fn);
					snprintf(fn, sizeof(fn), "<tr><td "
					    "colspan=\"2\"><table><tr><td "
					    "width=\"%d%%\"><br/></td><td "
					    "width=\"%d%%\"><br></td></tr>"
					    "</table></td></tr>\n</tr>\n",
					    (int)(pc + 0.5),
					    100 - (int)(pc + 0.5));
					khttp_puts(r, fn);
					line++;
				} else {
					khttp_puts(r, s);
				}
			} else if (s[0] == '\n' && s[1] == '\0') {
				body = true;
			}
		}
		fclose(f);
		if (ai->e->flags & FLAG_POLL) {
			snprintf(fn, sizeof(fn), "</table>\n<p>Total votes:"
			    " %d\n", votesum);
			khttp_puts(r, fn);
		}
	} else if (STRNEQ(m, len, MCR_MORE)) {
		if (ai->e->more != NULL) {
			render_common_entry(r, c, m, len, ai->e);
		} else {
			if (ai->e->flags & FLAG_SUBMISSION) {
				make_submission_more_path(ai->e->sid, fn,
				    sizeof(fn));
			} else {
				make_article_more_path(ai->e->sid, fn,
				    sizeof(fn));
			}
			if ((f = fopen(fn, "r")) != NULL) {
				while (fgets(s, sizeof(s), f) != NULL) {
					khttp_puts(r, s);
				}
				fclose(f);
			}
		}
	} else if (STRNEQ(m, len, MCR_COMMENTS)) {
		if (ai->e->flags & FLAG_SUBMISSION) {
			return;
		}
		if (KNOWN_USER()) {
			/* Update last-read time. */
			(void)get_lastread(ai->e->sid, &t_ignored, 1);
		}
		ci.last_read = ai->last_read;
		ci.cmts_open = ai->cmts_open;
		memset(&comments, 0, sizeof(comments));
#ifdef FEATURE_SHOW_THREAD
		find_comments(r, ai->e->sid, ai->e->pid, comments,
		    sizeof(comments) / sizeof(comments[0]));
#else /* FEATURE_SHOW_THREAD */
		find_comments(r, ai->e->sid, comments,
		    sizeof(comments) / sizeof(comments[0]));
#endif /* FEATURE_SHOW_THREAD */
		prev_level = 0;
		i = 0;
		while (i < sizeof(comments) / sizeof(comments[0])
		    && comments[i].pid[0] != '\0') {
			if (comments[i].level + 1 < prev_level) {
				/* We need to outdent. */
				use_template_file(r, c, TMPL_COMMENT_ITEM_END,
				    render_common, NULL);
				while (prev_level > comments[i].level + 1) {
					use_template_file(r, c,
					    TMPL_COMMENT_LIST_END,
					    render_common, NULL);
					use_template_file(r, c,
					    TMPL_COMMENT_ITEM_END,
					    render_common, NULL);
					prev_level--;
				}
			} else if (comments[i].level + 1 > prev_level) {
				/* We need to indent. */
				while (prev_level < comments[i].level + 1) {
					use_template_file(r, c,
					    TMPL_COMMENT_LIST_BEGIN,
					    render_common, NULL);
					prev_level++;
				}
			} else {
				/* Equal, so just close previous item. */
				use_template_file(r, c, TMPL_COMMENT_ITEM_END,
				    render_common, NULL);
			}
			use_template_file(r, c, TMPL_COMMENT_ITEM_BEGIN,
			    render_common, NULL);
			ci.e = &comments[i];
			use_template_file(r, c, TMPL_COMMENT_ITEM,
			    render_article_comment, &ci);

			i++;
		}

		/* Now close whatever is open. */
		if (prev_level > 0) {
			/* Close immediate list item. */
			use_template_file(r, c, TMPL_COMMENT_ITEM_END,
			    render_common, NULL);
		}
		while (prev_level > 1) {
			/* Close list and enclosing item. */
			use_template_file(r, c, TMPL_COMMENT_LIST_END,
			    render_common, NULL);
			use_template_file(r, c, TMPL_COMMENT_ITEM_END,
			    render_common, NULL);
			prev_level--;
		}
		if (comments[0].pid[0] != '\0') {
			/* There are comments, so close outermost list. */
			use_template_file(r, c, TMPL_COMMENT_LIST_END,
			    render_common, NULL);
		}
	} else if (STRNEQ(m, len, MCR_COUNT)) {
		if (ai->e->flags & FLAG_SUBMISSION) {
			msg(r, MSG_WARNING, "%s() ignoring MCR_COUNT for"
			    " submission sid=\"%s\"", __func__, ai->e->sid);
		} else {
			snprintf(s, sizeof(s), "%u", ai->cmts_count);
			khttp_puts(r, s);
		}
	} else if (STRNEQ(m, len, MCR_STATUS)) {
		if (ai->e->flags & FLAG_SUBMISSION) {
			msg(r, MSG_WARNING, "%s() ignoring MCR_STATUS for"
			    " submission sid=\"%s\"", __func__, ai->e->sid);
		} else {
			if (ai->cmts_count == 0) {
				strlcpy(s, FS_H_COMMENTS_0, sizeof(s));
			} else if (ai->cmts_count == 1) {
				strlcpy(s, FS_H_COMMENTS_1, sizeof(s));
			} else {
				snprintf(s, sizeof(s), FMT_H_COMMENTS_N,
				    ai->cmts_count);
			}

			/* Unnecessary, but avoids compiler warning. */
			if (ai->cmts_count > 0) {
				/* Append time of latest change. */
				make_duration_str(xtime() - ai->last_change,
				    u, sizeof(u));
				snprintf(s + strlen(s), sizeof(s) - strlen(s),
				    FMT_H_COMMENTS_T, u);
			}
			output_as_html(r, s);
		}
#ifdef FEATURE_A_PREV_NEXT_LINKS
	} else if (STRNEQ(m, len, MCR_NAVIGATION)) {
		if (ai->e->flags & FLAG_SUBMISSION) {
			return;
		}

		ni = 0;
		if (ai->e->pid[0] != '\0') {
			find_comment_adjacent(r, ai->e->sid, ai->e->pid, &prev,
			    &next);
		} else {
			find_article_adjacent(r, ai->e->sid, &prev, &next);
		}

		if ((prev.pid[0] != '\0' || (ai->e->pid[0] == '\0'
		    && prev.sid[0] != '\0')) && !(ai->e->flags & FLAG_POLL)) {
			khttp_puts(r, ni++ > 0 ? NAV_ITEM_SEP : "");
			khttp_puts(r, FS_H_PREV SP);
			print_nav_item(r, prev.subject,
			    url(r, URL_F_ESC, u, sizeof(u), A_ARTICLE,
			    prev.sid, prev.pid, NULL));
		}
		if (ai->e->pid[0] == '\0') {
			khttp_puts(r, SP FS_H_SEP SP);
		} else {
			find_parent(r, ai->e->sid, ai->e->pid, &p);
			khttp_puts(r, SP FS_H_UP_PRE SP);
			khttp_puts(r, ni++ > 0 ? NAV_ITEM_SEP : "");
			print_nav_item(r, p.pid[0] != '\0' ? p.subject
			    : ai->e->subject, url(r, URL_F_ESC, u, sizeof(u),
			    A_ARTICLE, p.sid, p.pid, NULL));
			khttp_puts(r, SP FS_H_UP_POST SP);
		}
		if ((next.pid[0] != '\0' || (ai->e->pid[0] == '\0'
		    && next.sid[0] != '\0')) && !(ai->e->flags & FLAG_POLL)) {
			khttp_puts(r, NAV_ITEM_SEP);
			print_nav_item(r, next.subject,
			    url(r, URL_F_ESC, u, sizeof(u), A_ARTICLE,
			    next.sid, next.pid, NULL));
			khttp_puts(r, SP FS_H_NEXT);
		}

#ifdef FEATURE_THRES
		/* XXX see template threshold_form.html */
		if (thres != NULL && thres[0] != '\0'
		    && thres_eval(thres, 0,0 ) < 0) {
			khttp_puts(r, "syntax error");
		}
#endif /* FEATURE_THRES */

#endif /* FEATURE_A_PREV_NEXT_LINKS */
	} else if (len >= (i = strlen(MCR_PREFIX_TIME_S_FMT))
	    && strncmp(m, MCR_PREFIX_TIME_S_FMT, i) == 0) {
		if ((t = strptime(ai->e->sid, TIMESTAMP_FMT,
		    &time_sid)) == NULL || *t != '\0') {
			msg(r, MSG_ERROR, "%s() strptime failed for sid",
			    __func__);
		} else {
			time_sid.tm_isdst = 0;
			time_sid.tm_gmtoff = 0;
			time_sid.tm_zone = utc_name;

			if (len > i) {
				/* Make a NUL-terminated copy of the format. */
				snprintf(s, sizeof(s), "%.*s", (int)(len - i),
				    &m[i]);
				output_time(r, &time_sid, s);
			} else {
				output_time(r, &time_sid, DEFAULT_TIME_FMT);
			}
		}
#ifdef COMPAT_LEGACY_R
	/*
	 * If we are dealing with article files in legacy format, we may not
	 * have a header value for time authored.  If that is the case, use
	 * the sid time rather than falling through to render_common_entry()
	 * (which would use the time edited or give no output).
	 */
	} else if (len >= (i = strlen(MCR_PREFIX_TIME_A_FMT))
	    && strncmp(m, MCR_PREFIX_TIME_A_FMT, i) == 0
	    && ai->e->time_authored.tm_zone == NULL) {
		snprintf(s, sizeof(s), "%s%.*s", MCR_PREFIX_TIME_S_FMT,
		    (int)(len - i), &m[i]);
		render_article(r, c, s, strlen(s), a);
#endif /* COMPAT_LEGACY_R */
	} else if (len >= (i = strlen(MCR_PREFIX_COND_MORE))
	    && strncmp(m, MCR_PREFIX_COND_MORE, i) == 0) {
		if (len == i) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_MORE "\"", __func__);
		} else if (ai->e->pid[0] != '\0') {
			msg(r, MSG_ERROR, "%s() macro \"" MCR_PREFIX_COND_MORE
			    "\" used for comment", __func__);
		} else {
			if (ai->e->more == NULL) {
				if (ai->e->flags & FLAG_SUBMISSION) {
					make_submission_more_path(ai->e->sid,
					    fn, sizeof(fn));
				} else {
					make_article_more_path(ai->e->sid, fn,
					    sizeof(fn));
				}
			}
			expand_cond_macro(r, ai->e->more != NULL
			    || access(fn, F_OK) == 0, c, &m[i], len - i,
			    render_article, a);
		}
	} else if (len >= (i = strlen(MCR_PREFIX_COND_SUBM))
	    && strncmp(m, MCR_PREFIX_COND_SUBM, i) == 0) {
		if (len == i) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_SUBM "\"", __func__);
		} else {
			expand_cond_macro(r, ai->e->flags & FLAG_SUBMISSION, c,
			    &m[i], len - i, render_article, a);
		}
	} else if (len >= (i = strlen(MCR_PREFIX_COND_DELAYED))
	    && strncmp(m, MCR_PREFIX_COND_DELAYED, i) == 0) {
		if (len == i) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_DELAYED "\"", __func__);
		} else {
			expand_cond_macro(r,
			    is_future_timestamp(ai->e->pubdate), c, &m[i],
			    len - i, render_article, a);
		}
	} else if (len >= (i = strlen(MCR_PREFIX_COND_CMTS_OP))
	    && strncmp(m, MCR_PREFIX_COND_CMTS_OP, i) == 0) {
		if (len == i) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_CMTS_OP "\"", __func__);
		} else {
			expand_cond_macro(r, ai->cmts_open, c, &m[i], len - i,
			    render_article, a);
		}
	} else if (len >= (i = strlen(MCR_PREFIX_COND_NEW))
	    && strncmp(m, MCR_PREFIX_COND_NEW, i) == 0) {
		if (len == i) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_NEW "\"", __func__);
		} else if (ANON_USER()) {
			/* If not logged in, neither alternative applies. */
		} else {
			expand_cond_macro(r, ai->last_read == 0, c, &m[i],
			    len - i, render_article, a);
		}
	} else if (len >= (i = strlen(MCR_PREFIX_COND_NEW_C))
	    && strncmp(m, MCR_PREFIX_COND_NEW_C, i) == 0) {
		if (len == i) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_NEW_C "\"", __func__);
		} else if (ANON_USER()) {
			/* If not logged in, neither alternative applies. */
		} else {
			expand_cond_macro(r, ai->cmts_count > 0
			    && ai->last_change > ai->last_read, c, &m[i],
			    len - i, render_article, a);
		}
	} else {
		render_common_entry(r, c, m, len, ai->e);
	}
}


#ifdef FEATURE_A_PREV_NEXT_LINKS
/*
 * Outputs txt as HTML as a link to URL u.
 */
static void
print_nav_item(struct kreq *r, const char *txt, const char *u)
{
	khttp_puts(r, "<a href=\"");
	khttp_puts(r, u);
	khttp_puts(r, "\">");

	output_as_html(r, txt);

	khttp_puts(r, "</a>");
}
#endif /* FEATURE_A_PREV_NEXT_LINKS */


/*
 * Returns true if and only if p is a (presentation format) IPv4 or IPv6
 * Internet address.
 */
static bool
is_inet_addr(const char *p)
{
	struct	in_addr	 a4;
	struct	in6_addr	 a6;
	bool	 r;

	r = false;

	if (p != NULL && p[0] != '\0') {
		if (isdigit((unsigned char)p[0])) {
			r = inet_pton(AF_INET, p, &a4) == 1;
		}

		if (!r && (p[0] == ':' || isxdigit((unsigned char)p[0]))) {
			r = inet_pton(AF_INET6, p, &a6) == 1;
		}
	}

	return r;
}


/*
 * Template rendering function for a comment.
 */
static void
render_article_comment(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	const struct	c_rend_info	*ci;
	struct	tm	 ts;
	char	 s[8192];
	FILE	*f;
	size_t	 p;
	bool	 in_body;

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	ci = (const struct c_rend_info *)a;

	if (STRNEQ(m, len, MCR_CLIENT)) {
		if (ci->e->id_author[0] != '\0') {
			output_as_html(r, ci->e->id_author);
		} else if (ci->e->addr_author[0] != '\0') {
			khttp_puts(r, ci->e->addr_author);
#ifdef COMPAT_LEGACY_R
		} else {
			/* e->host is (registered) username or remote host */
			output_as_html(r, ci->e->host);
#endif /* COMPAT_LEGACY_R */
		}
#ifndef FEATURE_COMMENT_SUBJECT
	} else if (STRNEQ(m, len, MCR_SUBJECT)) {
		msg(r, MSG_WARNING, "%s() ignoring macro \"" MCR_SUBJECT "\""
		    " in non-FEATURE_COMMENT_SUBJECT build", __func__);
#endif /* !FEATURE_COMMENT_SUBJECT */
	} else if (STRNEQ(m, len, MCR_BODY)) {
		if (ci->e->body != NULL) {
			render_common_entry(r, c, m, len, &ci->e);
		} else if ((f = fopen(ci->e->fn, "r")) == NULL) {
			msg(r, MSG_ERROR, "%s() fopen: %s: %s", __func__,
			    ci->e->fn, strerror(errno));
			khttp_puts(r, FS_H_INT_ERR);
		} else {
			in_body = false;
			while (fgets(s, sizeof(s), f) != NULL) {
				if (in_body) {
					khttp_puts(r, s);
				} else if (s[0] == '\n' && s[1] == '\0') {
					in_body = true;
				}
			}
			fclose(f);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_NEW))
	    && strncmp(m, MCR_PREFIX_COND_NEW, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_NEW "\"", __func__);
		} else if (ANON_USER()) {
			/* If not logged in, neither alternative applies. */
		} else {
			memcpy(&ts, &ci->e->time_edited, sizeof(ts));
			/* mktime() returns -1 on error, so this is OK. */
			expand_cond_macro(r, mktime(&ts) > ci->last_read, c,
			    &m[p], len - p, render_article_comment, a);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_CMT_REP))
	    && strncmp(m, MCR_PREFIX_COND_CMT_REP, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_CMT_REP "\"", __func__);
		} else {
			expand_cond_macro(r, ci->e->flags & FLAG_REPORTED, c,
			    &m[p], len - p, render_article_comment, a);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_CMTS_OP))
	    && strncmp(m, MCR_PREFIX_COND_CMTS_OP, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_CMTS_OP "\"", __func__);
		} else {
			expand_cond_macro(r, ci->cmts_open, c, &m[p], len - p,
			    render_article_comment, a);
		}
#ifdef XXX
		if ((SNB(A_BLOCK_ADDR) | SNB(A_LOCK_ACCT)) & ent_act) {
			ipa = is_inet_addr(e->host);
			if (ipa && (SNB(A_BLOCK_ADDR) & ent_act)) {
				khttp_puts(r, " [block address"
				    " - <abbr>NYI</abbr>]");
			} else if (!ipa && (SNB(A_LOCK_ACCT) & ent_act)) {
				khttp_puts(r, " [lock account"
				    " - <abbr>NYI</abbr>]");
			}
		}
#endif /* XXX */

	} else {
		render_common_entry(r, c, m, len, ci->e);
	}
}


/*
 * Template rendering function for user registration document.
 */
static void
render_register(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	bool	 was_err;

	DMSG(r, "%s() %.*s", __func__, (int)len, m);

	was_err = get_action(r) != A_REGISTER;  /* Check whether was error. */

	if (STRNEQ(m, len, MCR_INTRO)) {
		if (was_err) {
			use_template_file(r, c, TMPL_F_ERR, render_register,
			    a);
		}
	} else if (STRNEQ(m, len, MCR_F_ERR_MSG)) {
		/* We override the usual behaviour to handle a special case. */
		if (a == NULL || (((const struct field_errs *)a)->invalid
		    || ((const struct field_errs *)a)->missing)) {
			render_common_form(r, c, m, len, a);
		} else {
			/* Special case : ID in use */
			use_template_file(r, c, TMPL_F_ERR_ID_IN_USE,
			    render_common, a);
		}
	} else if (STRNEQ(m, len, MCR_ID)) {
		if (was_err && r->fieldmap[K_ID] != NULL) {
			output_as_html(r, r->fieldmap[K_ID]->parsed.s);
		}
	} else if (STRNEQ(m, len, MCR_PW)) {
		if (was_err && r->fieldmap[K_PW] != NULL) {
			output_as_html(r, r->fieldmap[K_PW]->parsed.s);
		}
	} else if (STRNEQ(m, len, MCR_PWCONFIRM)) {
		/* Note deliberate use of K_PW rather than K_PWCONFIRM. */
		if (was_err && r->fieldmap[K_PW] != NULL) {
			output_as_html(r, r->fieldmap[K_PW]->parsed.s);
		}
	} else if (STRNEQ(m, len, MCR_NAME)) {
		if (was_err && r->fieldmap[K_NAME] != NULL) {
			output_as_html(r, r->fieldmap[K_NAME]->parsed.s);
		}
	} else if (STRNEQ(m, len, MCR_EMAIL)) {
		if (was_err && r->fieldmap[K_EMAIL] != NULL) {
			output_as_html(r, r->fieldmap[K_EMAIL]->parsed.s);
		}
	} else if (STRNEQ(m, len, MCR_HREF)) {
		if (was_err && r->fieldmap[K_HREF] != NULL) {
			output_as_html(r, r->fieldmap[K_HREF]->parsed.s);
		}
	} else {
		render_common_form(r, c, m, len, a);
	}
}


/*
 * Rendering function for common parts of submit, reply, and edit
 * forms/documents.
 */
static void
render_common_contrib_or_edit(struct kreq *r, char c, const char *m,
    size_t len, const void *a)
{
	const struct	req_ent_info	*rei;
#ifdef XXX
	const char	*s;
#endif /* XXX */
	char	*h;
	size_t	 p;
#ifdef XXX
	enum action	 actn;
#endif /* XXX */

	DMSG(r, "%s() %.*s", __func__, (int)len, m);

	rei = (const struct req_ent_info *)a;

	if (STRNEQ(m, len, MCR_INTRO)) {
		/* Report any errors */
		if (rei->b.invalid || rei->b.missing) {
			use_template_file(r, c, TMPL_F_ERR, render_common_form,
			    &rei->b);
		}
#ifdef XXX
	/* XXX this stuff should not be needed */
	} else if (STRNEQ(m, len, MCR_NAME)) {
		actn = get_action(r);
		s = user.id;  /* Initialise to avoid compiler warnings. */

		/* Not mandatory, but fill initially (if available). */
		if (actn == A_REPLY || actn == A_SUBMIT) {
			if (r->fieldmap[K_NAME] != NULL) {
				/* Was supplied in URL. */
				s = r->fieldmap[K_NAME]->parsed.s;
			} else if (KNOWN_USER()) {
				/* That of logged in user. */
				s = user.id;
			}
		} else /* (actn == A_REPLYPOST || actn == A_SUBMITPOST) */ {
			s = rei->e.name;
		}

		if (s[0] != '\0') {
			output_as_html(r, s);
		}
	} else if (STRNEQ(m, len, MCR_EMAIL)) {
		actn = get_action(r);
		s = user.email;  /* Initialise to avoid compiler warnings. */

		/* Not mandatory, but fill initially (if available). */
		if (actn == A_REPLY || actn == A_SUBMIT) {
			if (r->fieldmap[K_EMAIL] != NULL) {
				/* Was supplied in URL. */
				s = r->fieldmap[K_EMAIL]->parsed.s;
			} else if (KNOWN_USER()) {
				/* That of logged in user. */
				s = user.email;
			}
		} else /* (actn == A_REPLYPOST || actn == A_SUBMITPOST) */ {
			s = rei->e.email;
		}

		if (s[0] != '\0') {
			output_as_html(r, s);
		}
	} else if (STRNEQ(m, len, MCR_HREF)) {
		actn = get_action(r);
		s = user.href;  /* Initialise to avoid compiler warnings. */

		/* Not mandatory, but fill initially (if available). */
		if (actn == A_REPLY || actn == A_SUBMIT) {
			if (r->fieldmap[K_HREF] != NULL) {
				/* Was supplied in URL. */
				s = r->fieldmap[K_HREF]->parsed.s;
			} else if (KNOWN_USER()) {
				/* That of logged in user. */
				s = user.href;
			}
		} else /* (actn == A_REPLYPOST || actn == A_SUBMITPOST) */ {
			s = rei->e.href;
		}

		if (s[0] != '\0') {
			output_as_html(r, s);
		}
	} else if (STRNEQ(m, len, MCR_SUBJECT)) {
		if (get_action(r) == A_SUBMIT) {
			if (r->fieldmap[K_SUBJECT] != NULL) {
				output_as_html(r,
				    r->fieldmap[K_SUBJECT]->parsed.s);
			}
		} else {
			if (rei->e.subject[0] != '\0') {
				output_as_html(r, rei->e.subject);
			}
		}
#endif /* XXX */
	} else if (STRNEQ(m, len, MCR_CONTENT0)) {
		/*
		 * To avoid the client losing a large amount of work, use the
		 * string whether valid or not.  If value isn't a valid C
		 * string, any valid leading substring will suffice.
		 */
		if (r->fieldmap[K_CONTENT0] != NULL) {
			output_as_html(r, r->fieldmap[K_CONTENT0]->parsed.s);
		} else if (r->fieldnmap[K_CONTENT0] != NULL
		    && r->fieldnmap[K_CONTENT0]->valsz > 0) {
			output_as_html(r, r->fieldnmap[K_CONTENT0]->val);
		}
	} else if (STRNEQ(m, len, MCR_CONTENT1)) {
		/*
		 * To avoid the client losing a large amount of work, use the
		 * string whether valid or not.  If value isn't a valid C
		 * string, any valid leading substring will suffice.
		 */
		if (r->fieldmap[K_CONTENT1] != NULL) {
			output_as_html(r, r->fieldmap[K_CONTENT1]->parsed.s);
		} else if (r->fieldnmap[K_CONTENT1] != NULL
		    && r->fieldnmap[K_CONTENT1]->valsz > 0) {
			output_as_html(r, r->fieldnmap[K_CONTENT1]->val);
		}
	} else if (STRNEQ(m, len, MCR_BODY)) {
		if (r->fieldmap[K_CONTENT0] != NULL) {
			if (rei->e.flags & FLAG_PLAIN) {
				h = alloc_text_to_html(
				    r->fieldmap[K_CONTENT0]->parsed.s, true);
				if (h == NULL) {
					msg(r, MSG_CRITICAL, "%s()"
					    " alloc_text_to_html failed",
					    __func__);
				} else {
					khttp_puts(r, h);
					free(h);
				}
			} else {
				khttp_puts(r,
				    r->fieldmap[K_CONTENT0]->parsed.s);
			}
		}
	} else if (STRNEQ(m, len, MCR_MORE)) {
		/* There is not any "more" for new contributions. */
		msg(r, MSG_WARNING, "%s() ignoring unexpected use of macro \""
		    MCR_MORE "\" for new contribution", __func__);
	} else if (len >= (p = strlen(MCR_PREFIX_CHCKD_TYPE))
	    && strncmp(m, MCR_PREFIX_CHCKD_TYPE, p) == 0) {
		if (len > p) {
			if (((rei->e.flags & FLAG_PLAIN)
			    && STRNEQ(&m[p], len - p, KV_TYPE_PLAIN))
			    || (!(rei->e.flags & FLAG_PLAIN)
			    && STRNEQ(&m[p], len - p, KV_TYPE_HTML))) {
				khttp_puts(r, "checked=\"checked\"");/*XXX*/
			}
		} else {
			msg(r, MSG_WARNING, "%s() missing type name with"
			    " macro \"" MCR_PREFIX_CHCKD_TYPE "\"", __func__);
		}
	} else if (len >= (p = strlen(MCR_PREFIX_COND_PRV_BODY))
	    && strncmp(m, MCR_PREFIX_COND_PRV_BODY, p) == 0) {
		if (len == p) {
			msg(r, MSG_ERROR, "%s() no alternatives with macro \""
			    MCR_PREFIX_COND_PRV_BODY "\"", __func__);
		} else {
			/*
			 * Preview body should not be shown if HTML fragment
			 * is invalid unless user has sufficent entitlements.
			 * This avoids displaying invalid HTML (to general
			 * users).
			 */
			expand_cond_macro(r,
			    (rei->e.flags & FLAG_PLAIN)
			    || ((SNB(A_EDIT_A) | SNB(A_EDIT_S)) & ent_act)
			    || !(rei->b.invalid & SNB(K_CONTENT0)), c, &m[p],
			    len - p, render_common_contrib_or_edit, a);
		}
	} else if (STRNEQ(m, len, MCR_FORM_EXTRA)
	    || STRNEQ(m, len, MCR_CAPTCHA)) {
		/*
		 * These are form-related macros rather than entry-related
		 * ones, so deal with them as special cases.
		 */
		render_common_form(r, c, m, len, &rei->b);
	} else {
		render_common_entry(r, c, m, len, &rei->e);
	}
}


/*
 * Rendering function for reply form/document.
 */
static void
render_reply(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
#ifdef FEATURE_COMMENT_SUBJECT
	struct	entry	 en;
#endif /* FEATURE_COMMENT_SUBJECT */
#if defined(FEATURE_COMMENT_SUBJECT) || defined(FEATURE_QUOTE_PARENT_COMMENT)
	const struct	req_ent_info	*rei;
#endif /* defined(F..._C..._SUBJECT) || defined(F..._QUOTE_PARENT_C...) */
#ifdef FEATURE_QUOTE_PARENT_COMMENT
	char	 b[2048];
	char	 x[2048];
	char	 fn[PATH_MAX];
	FILE	*f;
	size_t	 i;
	bool	 in_body;
#endif /* FEATURE_QUOTE_PARENT_COMMENT */

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	if (STRNEQ(m, len, MCR_PREVIEW)) {
		/* Display only if requested. */
		if (r->fieldmap[K_PREVIEW] != NULL) {
			use_template_file(r, c, TMPL_REPLY_PREVIEW,
			    render_common_contrib_or_edit, a);
		}
	} else if (STRNEQ(m, len, MCR_SUBJECT)) {
#ifdef FEATURE_COMMENT_SUBJECT
		rei = (const struct req_ent_info *)a;
		if (rei->e.subject[0] != '\0') {
			output_as_html(r, rei->e.subject);
		} else {
			/*
			 * XXX A cleaner approach would be to set the subject
			 * in do_action_reply()
			 */
			memset(&en, 0, sizeof(en));
			if (rei->e.pid[0] != '\0') {
				if (find_pid_path(rei->e.sid, rei->e.pid,
				    en.fn, sizeof(en.fn)) != 0) {
					msg(r, MSG_ERROR, "%s() find_pid_path"
					    " failed: sid = \"%s\", pid ="
					    " \"%s\"", __func__, rei->e.sid,
					    rei->e.pid);
					return;
				}
				strlcat(en.fn, "/comment", sizeof(en.fn));
			} else {
				make_article_file_path(rei->e.sid, en.fn,
				    sizeof(en.fn));
			}
			if (read_file(r, &en) != 0) {
				msg(r, MSG_ERROR, "%s() read_file failed:"
				    " \"%s\"", __func__, en.fn);
				return;
			}
			if (strncmp(en.subject, "Re: ", 4) != 0 && /*XXX*/
			    strcmp(en.subject, FS_T_NO_SUBJ) != 0) {
				khttp_puts(r, "Re: ");
			}
			khttp_puts(r, en.subject);
		}
#else /* FEATURE_COMMENT_SUBJECT */
		msg(r, MSG_WARNING, "%s() ignoring macro \"" MCR_SUBJECT "\""
		    " in non-FEATURE_COMMENT_SUBJECT build", __func__);
#endif /* FEATURE_COMMENT_SUBJECT */
#ifdef FEATURE_QUOTE_PARENT_COMMENT
	} else if (STRNEQ(m, len, MCR_CONTENT0)) {
		rei = (const struct req_ent_info *)a;
		if (rei->e.body != NULL || !(rei->e.flags & FLAG_PLAIN)) {
			render_common_contrib_or_edit(r, c, m, len, a);
		} else if (rei->e.pid[0] != '\0') {
			/*
			 * XXX A cleaner approach would be to add the quoted
			 * parent comment to the entry struct in
			 * do_action_reply()
			 */

			/*
			 * This is a comment on a comment (rather than on the
			 * article, so quote the parent.
			 */

			if (find_pid_path(rei->e.sid, rei->e.pid, fn,
			    sizeof(fn)) != 0) {
				msg(r, MSG_ERROR, "%s() parent comment (pid ="
				    " %s) not found", __func__, rei->e.pid);
			} else {
				strlcat(fn, "/" COMMENT_FILE, sizeof(fn));
				if ((f = fopen(fn, "r")) == NULL) {
					msg(r, MSG_ERROR, "%s() fopen \"%s\""
					    " failed: %s", __func__, fn,
					    strerror(errno));
				} else {
					in_body = false;
					while (fgets(b, sizeof(b), f) != NULL)
					    {
						while ((i = strlen(b)) > 0
						    && (b[i - 1] == '\n' ||
						    b[i - 1] == '\r')) {
							b[i - 1] = '\0';
						}
						if (in_body) {
							khttp_puts(r, "> ");
							de_html(b, x,
							    sizeof(x));
							khttp_puts(r, x);
							khttp_puts(r, "\n");
						} else if (b[0] == '\0') {
							in_body = true;
						}
					}
					khttp_puts(r, "\n");
					fclose(f);
				}
			}
		}
#endif /* FEATURE_QUOTE_PARENT_COMMENT */
	} else {
		render_common_contrib_or_edit(r, c, m, len, a);
	}
}


/*
 * Rendering function for submit form/document.
 */
static void
render_submit(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	if (STRNEQ(m, len, MCR_PREVIEW)) {
		/* Display only if requested. */
		if (r->fieldmap[K_PREVIEW] != NULL) {
			use_template_file(r, c, TMPL_SUBMIT_PREVIEW,
			    render_common_contrib_or_edit, a);
		}
	} else{
		render_common_contrib_or_edit(r, c, m, len, a);
	}
}


/*
 * Rendering function for edit form/document.
 */
static void
render_edit(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	const struct	req_ent_info	*rei;
	char	 s[8192];
	char	 fn[PATH_MAX];
	FILE	*f;
	size_t	 p;
	enum action	 actn;
	bool	 body;

	DMSG(r, "%s() %.*s", __func__, (int)len, m);

	rei = (const struct req_ent_info *)a;
	actn = get_action(r);

	if (STRNEQ(m, len, MCR_PREVIEW)) {
		use_template_file(r, c, TMPL_EDIT_PREVIEW,
		    render_common_contrib_or_edit, a);
	} else if (STRNEQ(m, len, MCR_AUTHOR_INFO)) {
		if (rei->e.id_author[0] != '\0') {
			output_as_html(r, rei->e.id_author);
		} else {
			khttp_puts(r, FS_H_UNKNOWN);
		}
		khttp_puts(r, FS_H_AT);
		if (rei->e.addr_author[0] != '\0') {
			khttp_puts(r, rei->e.addr_author);
		} else {
			khttp_puts(r, FS_H_UNKNOWN);
		}
	} else if (STRNEQ(m, len, MCR_EDITOR_INFO)) {
		if (rei->e.id_editor[0] != '\0'
		    || rei->e.addr_editor[0] != '\0') {
			if (rei->e.id_editor[0] != '\0') {
				output_as_html(r, rei->e.id_editor);
			} else {
				khttp_puts(r, FS_H_UNKNOWN);
			}
			khttp_puts(r, FS_H_AT);
			if (rei->e.addr_editor[0] != '\0') {
				khttp_puts(r, rei->e.addr_editor);
			} else {
				khttp_puts(r, FS_H_UNKNOWN);
			}
		} else {
#ifdef COMPAT_LEGACY_R
			if (rei->e.host[0] != '\0') {
				output_as_html(r, rei->e.host);
			} else {
				khttp_puts(r, FS_H_UNKNOWN);
			}
#else /* COMPAT_LEGACY_R */
			khttp_puts(r, FS_H_UNKNOWN);
#endif /* COMPAT_LEGACY_R */
		}
	} else if (len >= (p = strlen(MCR_PREFIX_SEL_TOPIC))
	    && strncmp(m, MCR_PREFIX_SEL_TOPIC, p) == 0) {
		if (len > p) {
			if (STRNEQ(&m[p], len - p, rei->e.topic)) {
				khttp_puts(r, "selected");/*XXX*/
			}
		} else {
			msg(r, MSG_WARNING, "%s() missing topic name with"
			    " macro \"" MCR_PREFIX_SEL_TOPIC "\"", __func__);
		}
	} else if (STRNEQ(m, len, MCR_CONTENT0)) {
		if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
			render_common_contrib_or_edit(r, c, m, len, a);
		} else {
			/*
			 * XXX Logically, the data should be loaded into the
			 * entry in do_common_edit() [rather than this
			 * approach].
			 */
			body = false;

			if ((f = fopen(rei->e.fn, "r")) == NULL) {
				msg(r, MSG_ERROR, "%s() fopen: %s: %s",
				    __func__, rei->e.fn, strerror(errno));
				khttp_puts(r, FS_H_INT_ERR);
				return;
			}
			while (fgets(s, sizeof(s), f) != NULL) {
				if (!body) {
					if (s[0] == '\n' && s[1] == '\0') {
						body = true;
					}
				} else {
					output_as_html(r, s);
				}
			}
			fclose(f);
		}
	} else if (STRNEQ(m, len, MCR_CONTENT1)) {
		if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
			render_common_contrib_or_edit(r, c, m, len, a);
		} else {
			/*
			 * XXX Logically, the data should be loaded into the
			 * entry in do_common_edit() [rather than this
			 * approach].
			 */
			if (rei->e.flags & FLAG_SUBMISSION) {
				make_submission_more_path(rei->e.sid, fn,
				    sizeof(fn));
			} else {
				make_article_more_path(rei->e.sid, fn,
				    sizeof(fn));
			}
			if ((f = fopen(fn, "r")) == NULL) {
				return;
			}
			while (fgets(s, sizeof(s), f) != NULL) {
				output_as_html(r, s);
			}
			fclose(f);
		}
	} else if (STRNEQ(m, len, MCR_CHECKED_POLL)) {
		if (rei->e.flags & FLAG_POLL) {
			khttp_puts(r, "checked=\"checked\"");/*XXX*/
		}
	} else if (STRNEQ(m, len, MCR_PUBDATE)) {
		if (rei->e.pubdate[0] != '\0') {
			khttp_puts(r, rei->e.pubdate);
		}
	} else {
		render_common_contrib_or_edit(r, c, m, len, a);
	}
}


/*
 * Template rendering function for the preferences form document (TMPL_PREFS).
 */
static void
render_prefs(struct kreq *r, char c, const char *m, size_t len, const void *a)
{
	const struct	field_errs	*e;

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	/*
	 * We may be here because of a rejected form submission, in which case
	 * we should prefer _valid_ values from the request to those from the
	 * user database.
	 */

	if (STRNEQ(m, len, MCR_INTRO_PREFS)) {
		/* We could provide user with instructions here. */

		if (get_action(r) == A_PREFSPOST) {
			use_template_file(r, c, TMPL_F_ERR, render_common_form,
			    a);
		}
	} else if (STRNEQ(m, len, MCR_INTRO_PASSWORD)) {
		/* We could provide user with instructions here. */

		if (get_action(r) == A_CHANGEPW) {
			/* We are here because of problem in form data. */
			use_template_file(r, c, TMPL_F_ERR, render_common_form,
			    a);
		}
	} else if (STRNEQ(m, len, MCR_NAME)) {
		output_as_html(r, get_action(r) == A_PREFSPOST
		    && r->fieldmap[K_NAME] != NULL
		    ? r->fieldmap[K_NAME]->parsed.s : user.name);
	} else if (STRNEQ(m, len, MCR_EMAIL)) {
		khttp_puts(r, get_action(r) == A_PREFSPOST
		    && r->fieldmap[K_EMAIL] != NULL
		    ? r->fieldmap[K_EMAIL]->parsed.s : user.email);
	} else if (STRNEQ(m, len, MCR_HREF)) {
		e = (const struct field_errs *)a;

		if (get_action(r) == A_PREFS) {
			output_as_html(r, user.href);
		} else if (e->invalid & SNB(K_HREF)) {
			output_as_html(r, user.href);
		} else if (e->missing & SNB(K_HREF)) {
			/* Empty string is OK, so emit no output. */
		} else {
			/* Valid (but possibly empty) value. */
			output_as_html(r, r->fieldmap[K_HREF]->parsed.s);
		}
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Stores into a, which must be of length no less than sz, details of the most
 * recent articles from directory path.  Articles with future timestamps (sids)
 * are excluded unless the current user has permissions to view submissions.
 * Polls are included if and only if polls is true.  If p is non-NULL, the
 * first poll found (if any) is stored in *p.
 *
 * XXX It would be cleaner to:
 * - use a flag for articles vs submissions (rather than path)
 * - use a flag for inclusion of future articles
 */
static void
find_articles(const struct kreq *r, char *path, struct entry *a, size_t sz,
    bool polls, struct entry *p)
{
	char	*path_argv[2];
	char	 ts_now[TIMESTAMP_LEN + 1];
	FTS	*fts;
	FTSENT	*e;
	size_t	 i;

	path_argv[0] = path;
	path_argv[1] = NULL;

	i = 0;
	memset(a, 0, sz * sizeof(*a));
	if (p != NULL) {
		memset(p, 0, sizeof(*p));
	}
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT,
	    compare_name_des_fts);
	if (fts == NULL) {
		msg(r, MSG_ERROR, "%s() fts_open: \"%s\": %s", __func__, path,
		    strerror(errno));
		return;
	} else if ((e = fts_read(fts)) == NULL || !(e->fts_info & FTS_D)) {
		msg(r, MSG_ERROR, "%s() fts_read: \"%s\": %s", __func__, path,
		    strerror(errno));
		return;
	} else if ((e = fts_children(fts, FTS_NAMEONLY)) == NULL) {
		if (errno != 0) {
			msg(r, MSG_ERROR, "%s() fts_children: %s: %s",
			    __func__, path, strerror(errno));
		}
		return;
	}
	/* Skip future articles for non-privileged users. */
	if (!(SNB(A_SUBMISSIONS) & ent_act)) {
		make_current_timestamp(ts_now, sizeof(ts_now));
		for (; e != NULL; e = e->fts_link) {
			if ((e->fts_info & FTS_D)
			    && e->fts_namelen == TIMESTAMP_LEN
			    && is_all_digits(e->fts_name)
			    && strcmp(e->fts_name, ts_now) <= 0) {
				break;
			}
		}
	}
	for (; e != NULL && i < sz; e = e->fts_link) {
		if (!(e->fts_info & FTS_D) || e->fts_namelen != TIMESTAMP_LEN
		    || !is_all_digits(e->fts_name)) {
			continue;
		}
		snprintf(a[i].fn, sizeof(a[i].fn), "%s/%s/" ARTICLE_FILE,
		    path, e->fts_name);/*XXX*/
		if (read_file(r, &a[i]) != 0) {
			/* Reading failed. */
			memset(&a[i], 0, sizeof(a[i]));
			continue;
		}
		strlcpy(a[i].sid, e->fts_name, sizeof(a[i].sid));
		if (a[i].flags & FLAG_POLL) {
			if (p != NULL && p->sid[0] == '\0') {
				memcpy(p, &a[i], sizeof(*p));
			}
			if (!polls) {
				memset(&a[i], 0, sizeof(a[i]));
				continue;
			}
		}
		i++;
	}
	fts_close(fts);
}


#ifdef FEATURE_A_PREV_NEXT_LINKS
/*
 * Assigns to *prev and *next details of the articles (if any) chronologically
 * before and after (respectively) sid (by publication timestamp/sid).
 */
static void
find_article_adjacent(const struct kreq *r, const char *sid,
    struct entry *prev, struct entry *next)
{
	FTS	*fts;
	FTSENT	*e;
	char	*path_argv[2];
	char	 articles_dir[PATH_MAX];
	char	 ts_now[TIMESTAMP_LEN + 1];
	bool	 found;

	make_articles_dir_path(articles_dir, sizeof(articles_dir));
	path_argv[0] = articles_dir;
	path_argv[1] = NULL;

	make_current_timestamp(ts_now, sizeof(ts_now));

	memset(prev, 0, sizeof(*prev));
	memset(next, 0, sizeof(*prev));

	/* Find next article. */
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT,
	    compare_name_asc_fts);
	if (fts == NULL ||
	    (e = fts_read(fts)) == NULL || !(e->fts_info & FTS_D) ||
	    (e = fts_children(fts, FTS_NAMEONLY)) == NULL) {
		return;
	}
	for (found = false; e != NULL; e = e->fts_link) {
		if (!(e->fts_info & FTS_D) || e->fts_namelen != TIMESTAMP_LEN
		    || !is_all_digits(e->fts_name)) {
			continue;
		}
		/* For non-editors, skip future articles. */
		if (!(SNB(A_SUBMISSIONS) & ent_act)
		    && strcmp(e->fts_name, ts_now) > 0) {
			continue;
		}
		if (found) {
			make_article_file_path(e->fts_name, next->fn,
			    sizeof(next->fn));
			strlcpy(next->sid, e->fts_name, sizeof(next->sid));
			if (read_file(r, next) != 0
			    || next->flags & FLAG_POLL) {
				memset(next, 0, sizeof(*next));
			} else {
				break;
			}
		} else if (strcmp(e->fts_name, sid) == 0) {
			found = true;
		}
	}
	fts_close(fts);

	/* Find previous article. */
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT,
	    compare_name_des_fts);
	if (fts == NULL ||
	    (e = fts_read(fts)) == NULL || !(e->fts_info & FTS_D) ||
	    (e = fts_children(fts, FTS_NAMEONLY)) == NULL) {
		return;
	}
	for (found = false; e != NULL; e = e->fts_link) {
		if (!(e->fts_info & FTS_D) || e->fts_namelen != TIMESTAMP_LEN
		    || !is_all_digits(e->fts_name)) {
			continue;
		}
		/* For non-editors, skip future articles. */
		if (!(SNB(A_SUBMISSIONS) & ent_act) && ts_now[0] != '\0'
		    && strcmp(e->fts_name, ts_now) > 0) {
			continue;
		}
		if (found) {
			make_article_file_path(e->fts_name, prev->fn,
			    sizeof(prev->fn));
			strlcpy(prev->sid, e->fts_name, sizeof(prev->sid));
			if (read_file(r, prev) != 0
			    || (prev->flags & FLAG_POLL)) {
				memset(prev, 0, sizeof(*prev));
			} else {
				break;
			}
		} else if (strcmp(e->fts_name, sid) == 0) {
			found = true;
		}
	}
	fts_close(fts);
}
#endif /* FEATURE_A_PREV_NEXT_LINKS */


/*
 * Copies to the elements of a, the details of (up to sz) comments related to
 * the article or comment identified by sid/pid, as a flattened hierarchy
 * ordered by pid.
 */
#ifdef FEATURE_SHOW_THREAD
static void
find_comments(const struct kreq *r, const char *sid, const char *pid,
    struct entry *a, size_t sz)
#else /* FEATURE_SHOW_THREAD */
static void
find_comments(const struct kreq *r, const char *sid, struct entry *a,
    size_t sz)
#endif /* FEATURE_SHOW_THREAD */
{
	char	 path[8192];
	char	*path_argv[2];
	FTS	*fts;
	FTSENT	*e;
	const char	*p;
	size_t	 i, j;

	path_argv[0] = path;
	path_argv[1] = NULL;

	make_article_dir_path(sid, path, sizeof(path));
#ifdef FEATURE_SHOW_THREAD
	if (pid[0] != '\0') {
		fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
		if (fts == NULL) {
			return;
		}
		while ((e = fts_read(fts)) != NULL) {
			if (e->fts_info & FTS_D && strcmp(e->fts_name, pid)
			    == 0) {
				strlcpy(path, e->fts_path, sizeof(path));
				break;
			}
		}
		fts_close(fts);
		if (e == NULL) {
			return;
		}
	}
#endif /* FEATURE_SHOW_THREAD */
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, &compare_fts);
	if (fts == NULL) {
		return;
	}
	i = 0;
	while ((e = fts_read(fts)) != NULL && i < sz) {
		if (e->fts_info & FTS_F && strcmp(e->fts_name, COMMENT_FILE)
		    == 0) {
			p = strrchr(e->fts_path, '/');
			j = 0;

			strlcpy(a[i].sid, sid, sizeof(a[i].sid));
			while (*--p != '/') {
				;
			}
			p++;
			while (*p != '/' && j < sizeof(a[i].pid) - 1) {
				a[i].pid[j++] = *p++;
			}
			a[i].pid[j] = 0;
			strlcpy(a[i].fn, e->fts_path, sizeof(a[i].fn));
			if (read_file(r, &a[i]) == 0) {
#ifdef FEATURE_SHOW_THREAD
				a[i].level = e->fts_level - (pid[0] != '\0'
				    ? 1 : 2);
#else /* FEATURE_SHOW_THREAD */
				a[i].level = e->fts_level - 2;
#endif /* FEATURE_SHOW_THREAD */
				i++;
			}
		}
	}
	fts_close(fts);
}


#ifdef FEATURE_A_PREV_NEXT_LINKS
/*
 * In the absence of error, fills *p with details for the parent comment of
 * comment sid/pid.  If error (including the comment not being found) occurs,
 * the entry is zeroed.
 */
static void
find_parent(const struct kreq *r, const char *sid, const char *pid,
    struct entry *p)
{
	FTS	*fts;
	FTSENT	*e;
	char	 path[8192];
	char	*path_argv[2];
	short	 level;

	path_argv[0] = path;
	path_argv[1] = NULL;

	memset(p, 0, sizeof(*p));
	strlcpy(p->sid, sid, sizeof(p->sid));

	make_article_dir_path(sid, path, sizeof(path));
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts == NULL) {
		return;
	}
	while ((e = fts_read(fts)) != NULL) {
		if (!(e->fts_info & FTS_D)) {
			continue;
		}
		if (strcmp(e->fts_name, pid) == 0) {
			level = e->fts_level;
			break;
		}
	}
	fts_close(fts);
	if (e == NULL) {
		return;
	}
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts == NULL) {
		return;
	}
	while ((e = fts_read(fts)) != NULL) {
		if (!(e->fts_info & FTS_D)) {
			continue;
		}
		if (e->fts_level > 0 && e->fts_level < level) {
			strlcpy(p->pid, e->fts_name, sizeof(p->pid));
			snprintf(p->fn, sizeof(p->fn), "%s/comment",
			    e->fts_path);
		}
		if (strcmp(e->fts_name, pid) == 0) {
			break;
		}
	}
	fts_close(fts);
	if (e == NULL) {
		p->pid[0] = '\0';
	}
	if (p->pid[0] != '\0') {
		if (read_file(r, p) != 0) {
			memset(p, 0, sizeof(*p));
		}
	}
}
#endif /* FEATURE_A_PREV_NEXT_LINKS */


#ifdef FEATURE_A_PREV_NEXT_LINKS
/*
 * Assigns to *prev and *next details of the comments (if any) chronologically
 * before and after (respectively) sid,pid.
 */
static void
find_comment_adjacent(const struct kreq *r, const char *sid, const char *pid,
    struct entry *prev, struct entry *next)
{
	FTS	*fts;
	FTSENT	*e;
	char	 path[8192];
	char	*path_argv[2];
	short	 level;
	bool	 found;

	path_argv[0] = path;
	path_argv[1] = NULL;

	memset(prev, 0, sizeof(*prev));
	memset(next, 0, sizeof(*next));
	strlcpy(prev->sid, sid, sizeof(prev->sid));
	strlcpy(next->sid, sid, sizeof(next->sid));

	make_article_dir_path(sid, path, sizeof(path));
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts == NULL) {
		return;
	}
	while ((e = fts_read(fts)) != NULL) {
		if (!(e->fts_info & FTS_D)) {
			continue;
		}
		if (strcmp(e->fts_name, pid) == 0) {
			level = e->fts_level;
			break;
		}
	}
	fts_close(fts);
	if (e == NULL) {
		return;
	}
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts == NULL) {
		return;
	}
	found = false;
	while ((e = fts_read(fts)) != NULL) {
		if (e->fts_level <= 0 || e->fts_level > level ||
		    !(e->fts_info & FTS_D))
			continue;
		if (strcmp(e->fts_name, pid) == 0) {
			found = true;
			continue;
		}
		if (!found) {
			strlcpy(prev->pid, e->fts_name, sizeof(prev->pid));
			snprintf(prev->fn, sizeof(prev->fn), "%s/comment",
			    e->fts_path);
		} else {
			strlcpy(next->pid, e->fts_name, sizeof(next->pid));
			snprintf(next->fn, sizeof(next->fn), "%s/comment",
			    e->fts_path);
			break;
		}
	}
	fts_close(fts);
	if (prev->pid[0] != '\0') {
		if (read_file(r, prev) != 0) {
			/* Reading failed. */
			memset(prev, 0, sizeof(*prev));
		}
	}
	if (next->pid[0] != '\0') {
		if (read_file(r, next) != 0) {
			/* Reading failed. */
			memset(next, 0, sizeof(*next));
		}
	}
}
#endif /* FEATURE_A_PREV_NEXT_LINKS */


/*
 * Reads the votes for poll sid.  The count counts for each alternative are
 * assigned to a (which must be of size no less than siz).  The total number
 * of votes is assigned to *sum.
 */
static void
read_poll(const char *sid, int *a, int siz, int *sum)
{
	char	 s[8192];
	char	 host[128];
	char	 fn[PATH_MAX];
	FILE	*f;
	int	 i;
	int	 val;

	for (i = 0; i < siz; ++i) {
		a[i] = 0;
	}
	*sum = 0;
	make_article_poll_path(sid, fn, sizeof(fn));
	if ((f = fopen(fn, "r")) == NULL) {
		return;
	}
	while (fgets(s, sizeof(s), f) != NULL) {
		if (sscanf(s, "%127s %d", host, &val) == 2 &&
		    val >= 0 && val < siz) {
			a[val]++;
		}
	}
	for (i = 0; i < siz; ++i) {
		*sum += a[i];
	}
	fclose(f);
}


/*
 * Maintains a file for rate-limiting RSS request responses for IP addresses.
 * Returns true if and only if (no error occurs and) client has exceeded the
 * permitted limit for requests over the last 24 hours.
 */
static bool
rss_rate_limit(const struct kreq *r, const char *client)
{
	struct	stat	 sb;
	char	 s[8192];
	char	 fni[PATH_MAX], fno[PATH_MAX];
	FILE	*fi, *fo;
	time_t	 now, ts;
	unsigned long	 count;
	bool	 exceeded;

	/* Ensure RSS log directory exists. */
	make_rss_log_dir_path(s, sizeof(s));
	if (stat(s, &sb) == 0) {
		if (!S_ISDIR(sb.st_mode)) {
			msg(r, MSG_CRITICAL, "%s() %s is not a directory",
			    __func__, s);
			return false;
		}
	} else if (errno == ENOENT) {
		if (mkdir(s, 0775) == -1) {
			msg(r, MSG_ERROR, "%s() mkdir \"%s\": %s", __func__, s,
			    strerror(errno));
			return false;
		}
	} else {
		msg(r, MSG_CRITICAL, "%s() stat \"%s\": %s", __func__, s,
		    strerror(errno));
		return false;
	}

	make_rss_log_file_path(client, fni, sizeof(fni));
	snprintf(fno, sizeof(fno), "%s.tmp", fni);

	if ((fo = fopen(fno, "w+")) == NULL) {
		msg(r, MSG_ERROR, "%s() fopen: %s: %s", __func__, fno,
		    strerror(errno));
		return false;
	}

	now = xtime();
	count = 0;
	if ((fi = fopen(fni, "r")) != NULL) {
		while (fgets(s, sizeof(s), fi) != NULL) {
			if (sscanf(s, "%lld", &ts) == 1 && ts
			    >= (now - 60 * 60 * 24)) {
				fprintf(fo, "%lld\n", ts);
				count++;
			}
		}
		fclose(fi);
	}
	exceeded = true;
	if (count <= 24 * 4) {/*XXX use define */
		/* Fewer than four request per hour for the last day. */
		fprintf(fo, "%lld\n", now);
		exceeded = false;
	}
	fclose(fo);
	if (rename(fno, fni) != 0) {
		msg(r, MSG_ERROR, "%s() rename: %s: %s", __func__, client,
		    strerror(errno));
	}

	return exceeded;
}


#ifdef FEATURE_VOTE
/*
 * Updates the score file for poll sid by recording vote for alternative val
 * from IP address addr.  returns zero if and only if no error occurs and there
 * is no recorded previous vote from address addr.
 */
static int
vote_poll(const char *addr, const char *sid, int val)
{
	char	 s[8192];
	char	 fn[PATH_MAX];
	char	 host[128];
	FILE	*f;

	make_article_poll_path(sid, fn, sizeof(fn));
	if ((f = fopen(fn, "a+")) == NULL) {
		return 1;
	}
	rewind(f);
	while (fgets(s, sizeof(s), f) != NULL) {
		if (sscanf(s, "%127s", host) != 1) {
			continue;
		}
		if (strcmp(host, addr) == 0) {
			fclose(f);
			return 1;
		}
	}
	fprintf(f, "%s %d\n", addr, val);
	fclose(f);
	return 0;
}
#endif /* FEATURE_VOTE */


/*
 * Metadata lines are of the form:
 * - for a flag:
 *       field_name:  # no space required after colon
 * - for name/value:
 *       field_name: value  # space required after colon
 *
 * Attempts to match the initial substring of s against a metadata field for
 * name (which is of flag format iff is_flag).  If matched, returns pointer
 * to start of value; otherwise, returns NULL.
 */
static const char *
md_match(const char *s, const char *name, bool is_flag)
{
	const char	*p;
	size_t	 len;

	len = strlen(name);

	p = NULL;
	if (strncmp(s, name, len) == 0 && s[len] == ':') {
		if (is_flag) {
			p = &s[len + 1];
		} else if (s[len + 1] == ' ') {
			p = &s[len + 2];
		}
	}

	return p;
}


/*
 * Reads an article (unpublished or published) or comment from the file
 * specified by e->fn, and fills other fields of e accordingly.  Returns
 * zero if and only if no error occurs.
 */
static int
read_file(const struct kreq *r, struct entry *e)
{
	struct	tm	 ti;
	char	 s[8192];
#ifdef FEATURE_MODERATE
	char	 fn[PATH_MAX];
	char	 host[128];
#endif /* FEATURE_MODERATE */
#ifdef COMPAT_LEGACY_R
	char	 cdt[32];
#endif /* COMPAT_LEGACY_R */
	const char	*v;
	char	*p;
	FILE	*f;
#ifdef FEATURE_MODERATE
	int	 val;
#endif /* FEATURE_MODERATE */

	DMSG(r, "%s() file = \"%s\" sid = \"%s\" pid = \"%s\"", __func__,
	    e->fn, e->sid, e->pid);

	e->name[0] = e->email[0] = e->href[0] = e->subject[0] = e->dept[0]
	    = e->topic[0] = '\0';
	e->pubdate[0] = '\0';
#ifdef COMPAT_LEGACY_R
	e->host[0] = '\0';
#endif /* COMPAT_LEGACY_R */
#ifdef FEATURE_MODERATE
	e->modcount = e->modsum = e->modable = 0;
#endif /* FEATURE_MODERATE */
	memset(&e->time_authored, 0, sizeof(e->time_authored));
	memset(&e->time_edited, 0, sizeof(e->time_edited));
#ifdef COMPAT_LEGACY_R
	cdt[0] = '\0';
#endif /* COMPAT_LEGACY_R */

	/* Only editors may see "future" files. */
	if (!((SNB(A_EDIT_A) | SNB(A_EDIT_S)) & ent_act) && e->sid[0] != '\0'
	    && is_future_timestamp(e->sid)) {
		DMSG(r, "%s() rejecting attempt by non-editor to see"
		    "\"future\" file (sid = %s)", __func__, e->sid);
		return 1;
	}
	if ((f = fopen(e->fn, "r")) == NULL) {
		msg(r, MSG_ERROR, "%s() fopen \"%s\" - %s", __func__, e->fn,
		    strerror(errno));
		return 1;
	}
	while (fgets(s, sizeof(s), f) != NULL) {
		if (s[0] == '\n' || s[0] == '\0') {
			break;
		}
		s[strlen(s) - 1] = '\0';

		if ((v = md_match(s, MD_F_NAME, false)) != NULL) {
			strlcpy(e->name, v, sizeof(e->name));
		} else if ((v = md_match(s, MD_F_ADDR_AUTHOR, false))
		    != NULL) {
			strlcpy(e->addr_author, v, sizeof(e->addr_author));
		} else if ((v = md_match(s, MD_F_ADDR_EDITOR, false))
		    != NULL) {
			strlcpy(e->addr_editor, v, sizeof(e->addr_editor));
		} else if ((v = md_match(s, MD_F_ID_AUTHOR, false)) != NULL) {
			strlcpy(e->id_author, v, sizeof(e->id_author));
		} else if ((v = md_match(s, MD_F_ID_EDITOR, false)) != NULL) {
			strlcpy(e->id_editor, v, sizeof(e->id_editor));
#ifdef COMPAT_LEGACY_R
		} else if ((v = md_match(s, MD_F_HOST, false)) != NULL) {
			strlcpy(e->host, v, sizeof(e->host));
#endif /* COMPAT_LEGACY_R */
		} else if ((v = md_match(s, MD_F_EMAIL, false)) != NULL) {
			strlcpy(e->email, v, sizeof(e->email));
		} else if ((v = md_match(s, MD_F_HREF, false)) != NULL) {
			strlcpy(e->href, v, sizeof(e->href));
		} else if ((v = md_match(s, MD_F_SUBJECT, false)) != NULL) {
			strlcpy(e->subject, v, sizeof(e->subject));
		} else if ((v = md_match(s, MD_F_TIME_AUTHORED, false))
		    != NULL) {
			if ((p = strptime(v, TIMESTAMP_FMT,
			    &e->time_authored)) == NULL || *p != '\0') {
				msg(r, MSG_ERROR, "%s() strptime failed for"
				    " time authored", __func__);
				fclose(f);
				return 1;
			} else {
				e->time_authored.tm_isdst = 0;
				e->time_authored.tm_gmtoff = 0;
				e->time_authored.tm_zone = utc_name;
			}
		} else if ((v = md_match(s, MD_F_TIME_EDITED, false))
		    != NULL) {
			if ((p = strptime(v, TIMESTAMP_FMT,
			    &e->time_edited)) == NULL || *p != '\0') {
				msg(r, MSG_ERROR, "%s() strptime failed for"
				    " time edited", __func__);
				fclose(f);
				return 1;
			} else {
				e->time_edited.tm_isdst = 0;
				e->time_edited.tm_gmtoff = 0;
				e->time_edited.tm_zone = utc_name;
			}
#ifdef COMPAT_LEGACY_R
		} else if ((v = md_match(s, MD_F_DATE, false)) != NULL) {
			/* Save the value for later. */
			strlcpy(cdt, v, sizeof(cdt));
#endif /* COMPAT_LEGACY_R */
		} else if ((v = md_match(s, MD_F_DEPT, false)) != NULL) {
			strlcpy(e->dept, v, sizeof(e->dept));
		} else if ((v = md_match(s, MD_F_TOPIC, false)) != NULL) {
			strlcpy(e->topic, v, sizeof(e->topic));
		} else if (md_match(s, MD_F_POLL, true) != NULL) {
			e->flags |= FLAG_POLL;
		} else if ((v = md_match(s, MD_F_PUBDATE, false)) != NULL
		    && strlen(v) == TIMESTAMP_LEN) {
			strlcpy(e->pubdate, v, sizeof(e->pubdate));
			/* Set time of last edit to match pubdate. */
			memset(&ti, 0, sizeof(ti));
			if (strptime(e->pubdate, TIMESTAMP_FMT, &ti) != NULL) {
				ti.tm_isdst = 0;
				ti.tm_gmtoff = 0;
				ti.tm_zone = utc_name;
				memcpy(&e->time_edited, &ti,
				    sizeof(e->time_edited));
			}
		}
	}
	fclose(f);

#ifdef COMPAT_LEGACY_R
	/*
	 * We've finished reading metadata headers.  If no value for time of
	 * last edit has been saved, convert that from "date" (or fail if that
	 * was not found).
	 */
	if (e->time_edited.tm_zone == NULL) {
		if (cdt[0] == '\0') {
			msg(r, MSG_ERROR, "%s() no time of last edit"
			    " available", __func__);
			return 1;
		} else {
			if ((p = strptime(cdt, CTIME_FMT, &e->time_edited))
			    == NULL || *p != '\0') {
				msg(r, MSG_ERROR, "%s() strptime failed for"
				    " \"date\"", __func__);
				return 1;
			} else {
				e->time_edited.tm_isdst = 0;
				e->time_edited.tm_gmtoff = 0;
				e->time_edited.tm_zone = utc_name;
			}
		}
	}
#endif /* COMPAT_LEGACY_R */

	if ((p = strrchr(e->fn, '/')) == NULL
	    || strcmp(p + 1, COMMENT_FILE) != 0) {
		/* Not a comment. */
		return 0;
	}

	/* Check whether comment has been reported to editors. */
	strlcpy(s, e->fn, sizeof(s));
	s[p - e->fn + 1] = '\0';
	strlcat(s, COMMENT_REPORT, sizeof(s));
	if (access(s, F_OK) == 0) {
		e->flags |= FLAG_REPORTED;
	}

#ifdef FEATURE_MODERATE
	e->modable = strcmp(e->host, r->remote) ? 1 : 0;
	if (KNOWN_USER() && strncmp(user.id, e->host) == 0) {
		e->modable = 0;
	}
	snprintf(fn, sizeof(fn), "%s.mod", e->fn);/*XXX*/
	if ((f = fopen(fn, "r")) == NULL) {
		return 0;
	}
	while (fgets(s, sizeof(s), f) != NULL) {
		if (sscanf(s, "%127s %d", host, &val) != 2) {
			continue;
		}
		if (val) {
			e->modcount++;
			e->modsum += val;
		}
		if (strcmp(host, r->remote) == 0 ||
		    (KNOWN_USER() && strncmp(user.id, host) == 0))
		    {
			e->modable = 0;
		}
	}
	fclose(f);
#endif /* FEATURE_MODERATE */

	return 0;
}


/*
 * Returns the number of comments (direct or indirect) on article sid.
 */
static unsigned int
count_comments(const char *sid)
{
	char	 path[8192];
	char	*path_argv[2];
	FTS	*fts;
	FTSENT	*e;
	unsigned int	 count;

	path_argv[0] = path;
	path_argv[1] = NULL;

	make_article_dir_path(sid, path, sizeof(path));
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts == NULL) {
		return 0;
	}
	count = 0;
	while ((e = fts_read(fts)) != NULL) {
		if (e->fts_info & FTS_F && strcmp(e->fts_name, COMMENT_FILE)
		    == 0) {
			count++;
		}
	}
	fts_close(fts);

	return count;
}


/*
 * Looks up last-read time for article sid for the current user.  If a value
 * is found, it is assigned to *t and non-zero is returned; if not found,
 * zero is returned.  If update is non-zero, the stored value is updated/set
 * to the current time.
 */
static int
get_lastread(const char *sid, time_t *t, int update)
{
	char	 s[8192], filename[PATH_MAX];
	char	*u, *d;
	FILE	*f;
	long	 pos;

	make_last_read_path(sid, filename, sizeof(filename));
	if ((f = fopen(filename, update ? "r+" : "r")) != NULL) {
		pos = ftell(f);
		while (fgets(s, sizeof(s), f) != NULL) {
			d = s;
			if ((u = strsep(&d, " \t")) != NULL && d != NULL
			    && strcasecmp(user.id, u) == 0) {
				d = strsep(&d, "\n\r ");
				*t = strtonum(d, 0, LLONG_MAX, NULL);
				if (update) {
					fseek(f, pos + strlen(user.id) + 1,
					    SEEK_SET);
					fprintf(f, "%.10lld", xtime());
				}
				fclose(f);
				return 1;
			}
			pos = ftell(f);
		}
		fclose(f);
	}
	if (update) {
		if ((f = fopen(filename, "a")) != NULL) {
			fprintf(f, "%s %.10lld\n", user.id, xtime());
			fclose(f);
		}
	}
	*t = 0;

	return 0;
}


/*
 * Returns time of most recent modification for article sid, or zero on error.
 *
 * If thorough is true, time is that of any directory or regular file (other
 * than LAST_READ) in the hierarchy for sid; if thorough is false, time is that
 * of article content, or comment or comment moderation.
 *
 * Only in the former case is comment deletion detected.
 */
static time_t
find_newest_change(const char *sid, bool thorough)
{
	struct	stat	 sb;
	char	 path[PATH_MAX];
	char	*path_argv[2];
	FTS	*fts;
	FTSENT	*e;
	time_t	 t;

	path_argv[0] = path;
	path_argv[1] = NULL;

	t = 0;

	if (thorough) {
		make_article_dir_path(sid, path, sizeof(path));
		fts = fts_open(path_argv, FTS_LOGICAL, NULL);
		if (fts != NULL) {
			while ((e = fts_read(fts)) != NULL) {
				if ((e->fts_info & FTS_D)
				    || ((e->fts_info & FTS_F) &&
				    strcmp(e->fts_name, LAST_READ) != 0)) {
					if (e->fts_statp != NULL
					    && e->fts_statp->st_mtime > t) {
						t = e->fts_statp->st_mtime;
					}
				}
			}
		}
	} else {
		make_article_file_path(sid, path, sizeof(path));
		if (stat(path, &sb) == 0 && sb.st_mtime > t) {
			t = sb.st_mtime;
		}
		make_article_more_path(sid, path, sizeof(path));
		if (stat(path, &sb) == 0 && sb.st_mtime > t) {
			t = sb.st_mtime;
		}
		make_article_dir_path(sid, path, sizeof(path));
		fts = fts_open(path_argv, FTS_LOGICAL, NULL);
		if (fts == NULL) {
			return t;
		}
		while ((e = fts_read(fts)) != NULL) {
			if ((e->fts_info & FTS_F)
#ifdef FEATURE_MODERATE
			    && (strcmp(e->fts_name, COMMENT_FILE) == 0
			    || strcmp(e->fts_name, COMMENT_MOD) == 0)) {
#else /* FEATURE_MODERATE */
			    && strcmp(e->fts_name, COMMENT_FILE) == 0) {
#endif /* FEATURE_MODERATE */
				if (e->fts_statp != NULL
				    && e->fts_statp->st_mtime > t) {
					t = e->fts_statp->st_mtime;
				}
			}
		}
	}

	(void)fts_close(fts);

	return t;
}


/*
 * Returns the (numerical) maximum pid used for a comment on article sid,
 * or zero if there are no comments or an error occurs.
 */
static long long
find_max_pid(const char *sid)
{
	char	 path[8192];
	char	*path_argv[2];
	FTS	*fts;
	FTSENT	*e;
	long long	 maxpid, pid;

	path_argv[0] = path;
	path_argv[1] = NULL;

	maxpid = 0;
	make_article_dir_path(sid, path, sizeof(path));
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts != NULL) {
		while ((e = fts_read(fts)) != NULL) {
			if (e->fts_level > 0 && (e->fts_info & FTS_D)
			    && e->fts_namelen <= TIMESTAMP_LEN
			    && isdigit((unsigned char)e->fts_name[0])) {
				/*
				 * If the first character is a digit, then
				 * strtonum() will give an error if any char is
				 * not a digit.  As strtonum() returns zero on
				 * error, we don't have to bother with further
				 * checking for errors.
				 */
				pid = strtonum(e->fts_name, 0, LLONG_MAX,
				    NULL);

				if (pid > maxpid) {
					maxpid = pid;
				}
			}
		}
		fts_close(fts);
	}

	return maxpid;
}


/*
 * Attempts to identify the full path of (the directory of) comment pid on
 * article sid.  On success, the path is copied to path (which must be of
 * size no less than sz) and zero is returned; on failure, path is set to
 * the empty string and non-zero is returned.
 */
static int
find_pid_path(const char *sid, const char *pid, char *path, size_t sz)
{
	char	*path_argv[2];
	FTS	*fts;
	FTSENT	*e;

	path_argv[0] = path;
	path_argv[1] = NULL;

	make_article_dir_path(sid, path, sz);
	fts = fts_open(path_argv, FTS_LOGICAL|FTS_NOSTAT, NULL);
	if (fts == NULL) {
		path[0] = '\0';
		return 1;
	}
	while ((e = fts_read(fts)) != NULL) {
		if ((e->fts_info & FTS_D) && strcmp(e->fts_name, pid) == 0) {
			strlcpy(path, e->fts_path, sz);
			break;
		}
	}
	fts_close(fts);
	if (e == NULL) {
		path[0] = '\0';
		return 1;
	}

	return 0;
}


/*
 * Attempts to create a new submission directory and open an article file
 * therein.  On success, ce->fn is set to the path of the article file,
 * and a (write-only) file descriptor for the file is returned; on failure,
 * -1 is returned and ce->fn is undefined.
 */
static int
create_submission(struct entry *ce)
{
	struct	tm	 ts;
	char	 sid[TIMESTAMP_LEN + 1];
	int	 fd, retry;

	fd = -1;

	retry = 0;
	while (retry < 3) {
		xutc_tm_now(&ts);
		if (strftime(sid, sizeof(sid), TIMESTAMP_FMT, &ts) != 0) {
			make_submission_dir_path(sid, ce->fn, sizeof(ce->fn));
			if (mkdir(ce->fn, 0775) == 0) {
				strlcat(ce->fn, "/" ARTICLE_FILE,
				    sizeof(ce->fn));
				fd = open(ce->fn,
				    O_CREAT|O_TRUNC|O_WRONLY|O_EXLOCK, 0644);
				if (fd >= 0) {
					break;
				}
			}
		}
		retry++;
		sleep(1);
	}

	return fd;
}


/*
 * In the absence of error, creates a new comment directory and file
 * for the comment in the specified entry, and returns a file descriptor
 * for the comment file; if error occurs, returns -1.
 */
static int
create_comment(struct entry *ce)
{
	char	 path[PATH_MAX];
	int	 fd, retry;

	if (ce->pid[0] == '\0') {
		make_article_dir_path(ce->sid, path, sizeof(path));
	} else if (find_pid_path(ce->sid, ce->pid, path, sizeof(path)) != 0) {
		return -1;
	}
	fd = -1;
	retry = 0;
	while (retry < 3) {
		snprintf(ce->fn, sizeof(ce->fn), "%s/%lld", path,
		    find_max_pid(ce->sid) + 1 + retry * 10);
		if (mkdir(ce->fn, 0775) != -1) {
			strlcat(ce->fn, "/" COMMENT_FILE, sizeof(ce->fn));
			fd = open(ce->fn, O_CREAT|O_TRUNC|O_WRONLY|O_EXLOCK,
			    0664);
			if (fd >= 0) {
				break;
			} else {
				(void)rmdir(ce->fn);
			}
		}
		retry++;
	}

	return fd;
}


#ifdef FEATURE_MODERATE
/*
 * Records a vote (down or up) val for comment sid,pid from address addr and
 * the name of the user (if authenticated).  Returns zero if and only if
 * successful.  The existence of a previous vote from the address or user
 * causes failure.
 */
static int
moderate_comment(const char *addr, const char *sid, const char *pid, int val)
{
	char	 s[8192];
	char	 path[PATH_MAX];
	char	 host[128];
	FILE	*f;

	if (find_pid_path(sid, pid, path, sizeof(path)) != 0) {
		return 1;
	}
	strlcat(path, "/" COMMENT_MOD, sizeof(path));
	if ((f = fopen(path, "a+")) == NULL) {
		return 1;
	}
	rewind(f);
	while (fgets(s, sizeof(s), f) != NULL) {
		if (sscanf(s, "%127s", host) != 1) {
			continue;
		}
		/* XXX address or username */
		if (strcmp(host, addr) == 0 || (KNOWN_USER()
		    && strncmp(host, user.id) == 0)) {
			fclose(f);
			return 1;
		}
	}
	fprintf(f, "%s %d\n", addr, val);
	if (KNOWN_USER()) {
		fprintf(f, "%s 0\n", user.id);
	}
	fclose(f);
	return 0;
}
#endif /* FEATURE_MODERATE */


/*
 * Attempts to delete directory hierarchy for submission sid, returning zero
 * on success, or non-zero on error.
 */
static int
delete_submission(const char *sid)
{
	char	 path[PATH_MAX];

	make_submission_dir_path(sid, path, sizeof(path));

	return delete_dir(path);
}


/*
 * Attempts to delete directory hierarchy for article sid, returning zero
 * on success, or non-zero on error.
 */
static int
delete_article(const char *sid)
{
	char	 path[PATH_MAX];

	make_article_dir_path(sid, path, sizeof(path));

	return delete_dir(path);
}


/*
 * In the absence of error, moves replies to comment sid/pid up one level in
 * the directory hierarchy before deleting the directory for sid/pid, and
 * returns zero; if error occurs, the directory for sid/pid is left in an
 * undefined state, and non-zero is returned.
 */
static int
delete_comment(const char *sid, const char *pid)
{
	struct	dirent	 entry;
	struct	dirent	*dp;
	char	 above[PATH_MAX];
	char	 below[PATH_MAX];
	char	 parent[PATH_MAX];
	char	 path[PATH_MAX];
	DIR	*dirp;
	char	*p;
	int	 result;

	result = 0;  /* Assume OK until proven otherwise. */

	if (find_pid_path(sid, pid, path, sizeof(path)) != 0) {
		result = 1;
	} else {
		strlcpy(parent, path, sizeof(parent));
		if ((p = strrchr(parent, '/')) == NULL) {
			result = 1;
		} else {
			*p = '\0';

			/* Now, move all sub-comments of pid up one level. */
			dirp = opendir(path);
			if (dirp == NULL) {
				result = 1;
			} else {
				while (result == 0
				    && readdir_r(dirp, &entry, &dp) == 0
					&& dp != NULL) {
					if (dp->d_type == DT_DIR
					    && dp->d_namlen <= TIMESTAMP_LEN
					    && is_all_digits(dp->d_name)
					    && strtonum(dp->d_name, 1,
					    LLONG_MAX, NULL) != 0) {
						snprintf(below, sizeof(below),
						    "%s/%s", path, dp->d_name);
						snprintf(above, sizeof(above),
						    "%s/%s", parent,
						    dp->d_name);
						if (rename(below, above) != 0)
						    {
							result = 1;
						}
					}
				}
				(void)closedir(dirp);

				/*
				 * In the absence of error, delete the
				 * remaining directory hierarchy.
				 */
				if (result == 0) {
					if (delete_dir(path) != 0) {
						result = 1;
					}
				}
			}
		}
	}

	return result;
}


#ifdef FEATURE_QUOTE_PARENT_COMMENT
/*
 * Copies to d (which must be of size no less than sz) as much as will fit of
 * a plain text version of the HTML string s, leaving d NUL-terminated.
 */
static void
de_html(const char *s, char *d, size_t sz)
{
	size_t	 p;
	int	 opened;

	opened = 0;

	for (p = 0; *s != '\0' && p < sz - 1; ++s) {
		switch (*s) {

		case '<':
			opened++;
		break;

		case '>':
			if (opened > 0) {
				opened--;
			}
		break;

		case '&':
			if (strncmp(s, "&amp;", 5) == 0) {
				if (!opened) {
					d[p++] = '&';
				}
				s += 4;
			} else if (strncmp(s, "&quot;", 6) == 0) {
				if (!opened) {
					d[p++] = '"';
				}
				s += 5;
			} else if (strncmp(s, "&lt;", 4) == 0) {
				if (!opened) {
					d[p++] = '<';
				}
				s += 3;
			} else if (strncmp(s, "&gt;", 4) == 0) {
				if (!opened) {
					d[p++] = '>';
				}
				s += 3;
			} else {
				while (*s != '\0' && *s != ';') {
					s++;
				}
			}
		break;

		default:
			if (opened <= 0) {
				d[p++] = *s;
			}
		break;

		}
	}

	d[p] = '\0';
}
#endif /* FEATURE_QUOTE_PARENT_COMMENT */


/*
 * Generates output consisting of an HTML-escaped version of s.
 */
static void
output_as_html(struct kreq *r, const char *s)
{
	size_t	 i, j;

	j = 0;
	i = 0;
	while (s[i] != '\0') {
		switch (s[i]) {

		case '"' :
			if (i > j) {
				khttp_write(r, &s[j], i - j);
			}
			j = i + 1;
			khttp_puts(r, "&quot;");
		break;


		case '&' :
			if (i > j) {
				khttp_write(r, &s[j], i - j);
			}
			j = i + 1;
			khttp_puts(r, "&amp;");
		break;


		case '<' :
			if (i > j) {
				khttp_write(r, &s[j], i - j);
			}
			j = i + 1;
			khttp_puts(r, "&lt;");
		break;


		case '>' :
			if (i > j) {
				khttp_write(r, &s[j], i - j);
			}
			j = i + 1;
			khttp_puts(r, "&gt;");
		break;

		default :
			/* Keep acumulating plain text. */
		break;

		}

		i++;
	}
	if (i > j) {
		khttp_write(r, &s[j], i - j);
	}
}


#ifdef FEATURE_A_PREV_NEXT_LINKS
/*
 * Ordering function for fts_open() which compares entry names for ascending
 * order.
 */
static int
compare_name_asc_fts(const FTSENT **a, const FTSENT **b)
{
	return strcmp((*a)->fts_name, (*b)->fts_name);
}
#endif /* FEATURE_A_PREV_NEXT_LINKS */


/*
 * Ordering function for fts_open() which compares entry names for descending
 * order.
 */
static int
compare_name_des_fts(const FTSENT **a, const FTSENT **b)
{
	return strcmp((*b)->fts_name, (*a)->fts_name);
}


/* Numerical order PIDs find_comments().  Non-PIDs and errors are first. */
static int
compare_fts(const FTSENT **a, const FTSENT **b)
{
	long long	 na, nb;

	if (!is_all_digits((*a)->fts_name)) {
		na = 0;
	} else {
		/* PIDs start at 1, so zero for error is fine. */
		na = strtonum((*a)->fts_name, 1, LLONG_MAX, NULL);
	}
	if (!is_all_digits((*b)->fts_name)) {
		nb = 0;
	} else {
		/* PIDs start at 1, so zero for error is fine. */
		nb = strtonum((*b)->fts_name, 1, LLONG_MAX, NULL);
	}

	return na == nb ? 0 : (na < nb ? -1 : 1);
}


/*
 * Copies to s (which must be of size no less than sz) a textual representation
 * of time t.
 */
static void
make_duration_str(time_t t, char *s, size_t sz)
{
	long	 days;
	int	 hours;
	int	 minutes;

	days = 0;
	hours = 0;
	minutes = 0;

	if (t >= 24 * 60 * 60) {
		days = t / (24 * 60 * 60);
		t = t % (24 * 60 * 60);
	}
	if (t >= 60 * 60) {
		hours = t / (60 * 60);
		t = t % (60 * 60);
	}
	minutes = t / 60;

	if (days != 0 && hours != 0) {
		snprintf(s, sz, FMT_DS_DHM, days, hours, minutes);
	} else if (days != 0) {
		snprintf(s, sz, FMT_DS_DM, days, minutes);
	} else if (hours != 0) {
		snprintf(s, sz, FMT_DS_HM, hours, minutes);
	} else {
		snprintf(s, sz, FMT_DS_M, minutes);
	}
}


/*
 * Attempts to build entry structure rei->e from the attributes of request r.
 * Any errors are recorded in rei->b.
 */
static void
build_entry_from_request(struct kreq *r, struct req_ent_info *rei)
{
	char	 ts_now[TIMESTAMP_LEN + 1];
	time_t	 t;
	const char	*s;
	char	*h;
	enum action	 actn;

	memset(&rei->e, 0, sizeof(rei->e));
	rei->b.invalid = 0;
	rei->b.missing = 0;

	actn = get_action(r);

	if (actn == A_EDITPOST_A) {
		rei->e.flags |= FLAG_EDIT;
	} else if (actn == A_EDITPOST_S) {
		rei->e.flags |= FLAG_EDIT;
		rei->e.flags |= FLAG_SUBMISSION;
	}

	/*
	 * K_SID: The form always contains a value, but we consider it missing
	 * (rather than invalid) if empty.
	 */
	if (r->fieldmap[K_SID] != NULL) {
		strlcpy(rei->e.sid, r->fieldmap[K_SID]->parsed.s,
		    sizeof(rei->e.sid));
	} else if (r->fieldnmap[K_SID] != NULL
	    && r->fieldnmap[K_SID]->valsz > 0) {
		rei->b.invalid |= SNB(K_SID);
		DMSG(r, "%s() K_SID invalid", __func__);
	} else {
		if (rei->e.flags & FLAG_EDIT) {
			rei->b.missing |= SNB(K_SID);
			DMSG(r, "%s() K_SID missing", __func__);
		}
	}

	/*
	 * K_PID: The form always contains a value, but we consider it to be
	 * OK if empty.
	 */
	if (r->fieldmap[K_PID] != NULL) {
		strlcpy(rei->e.pid, r->fieldmap[K_PID]->parsed.s,
		    sizeof(rei->e.pid));
	} else if (r->fieldnmap[K_PID] != NULL
	    && r->fieldnmap[K_PID]->valsz > 0) {
		rei->b.invalid |= SNB(K_PID);
		DMSG(r, "%s() K_PID invalid", __func__);
	}

	/* K_NAME may be left blank, but must pass basic checks if given. */
	if (r->fieldmap[K_NAME] != NULL) {
		strlcpy(rei->e.name, r->fieldmap[K_NAME]->parsed.s,
		    sizeof(rei->e.name));
	} else if (r->fieldnmap[K_NAME] != NULL
	    && r->fieldnmap[K_NAME]->valsz > 0) {
		rei->b.invalid |= SNB(K_NAME);
		DMSG(r, "%s() K_NAME invalid", __func__);
	}

	/* K_EMAIL is optional but must be valid if specified. */
	if (r->fieldmap[K_EMAIL] != NULL) {
		strlcpy(rei->e.email, r->fieldmap[K_EMAIL]->parsed.s,
		    sizeof(rei->e.email));
	} else if (r->fieldnmap[K_EMAIL] != NULL
	    && r->fieldnmap[K_EMAIL]->valsz > 0) {
		rei->b.invalid |= SNB(K_EMAIL);
		DMSG(r, "%s() K_EMAIL invalid", __func__);
	}

	/* K_HREF is also optional, but must pass basic checks if given. */
	if (r->fieldmap[K_HREF] != NULL) {
		if (user_is_valid_href(r->fieldmap[K_HREF]->parsed.s)) {
			strlcpy(rei->e.href, r->fieldmap[K_HREF]->parsed.s,
			    sizeof(rei->e.href));
		} else {
			kutil_invalidate(r, r->fieldmap[K_HREF]);
			rei->b.invalid |= SNB(K_HREF);
			DMSG(r, "%s() K_HREF invalid", __func__);
		}
	} else if (r->fieldnmap[K_HREF] != NULL
	    && r->fieldnmap[K_HREF]->valsz > 0) {
		rei->b.invalid |= SNB(K_HREF);
		DMSG(r, "%s() K_HREF invalid", __func__);
	}

	/* K_SUBJECT handling depends on the action. */
	if (actn == A_SUBMIT || actn == A_SUBMITPOST
#ifdef FEATURE_COMMENT_SUBJECT
	    || actn == A_REPLY || actn == A_REPLYPOST
#endif /* FEATURE_COMMENT_SUBJECT */
	    ) {
		/*
		 * K_SUBJECT field content is plain text rather than HTML.
		 * K_SUBJECT is mandatory for A_SUBMITPOST.
		 */
		if (r->fieldmap[K_SUBJECT] != NULL) {
			s = r->fieldmap[K_SUBJECT]->parsed.s;

			if ((h = alloc_text_to_html(s, false)) == NULL) {
				msg(r, MSG_CRITICAL, "%s() alloc_text_to_html"
				    " failed", __func__);
				kutil_invalidate(r, r->fieldmap[K_SUBJECT]);
				rei->b.invalid |= SNB(K_SUBJECT);
			} else if (strlen(h) >= sizeof(rei->e.subject)) {
				kutil_invalidate(r, r->fieldmap[K_SUBJECT]);
				rei->b.invalid |= SNB(K_SUBJECT);
				DMSG(r, "%s() K_SUBJECT invalid (too long as"
				    " HTML)", __func__);
				free(h);
			} else {
				strlcpy(rei->e.subject, h,
				    sizeof(rei->e.subject));
				free(h);
			}
		} else if (r->fieldnmap[K_SUBJECT] != NULL) {
#ifdef FEATURE_COMMENT_SUBJECT
			if ((actn == A_SUBMIT || actn == A_SUBMITPOST)
			    || r->fieldnmap[K_SUBJECT]->valsz > 0) {
				rei->b.invalid |= SNB(K_SUBJECT);
				DMSG(r, "%s() K_SUBJECT invalid", __func__);
			}
#elsif /* FEATURE_COMMENT_SUBJECT */
			rei->b.invalid |= SNB(K_SUBJECT);
			DMSG(r, "%s() K_SUBJECT invalid", __func__);
#endif /* FEATURE_COMMENT_SUBJECT */
		} else if (actn == A_SUBMITPOST) {
			rei->b.missing |= SNB(K_SUBJECT);
			DMSG(r, "%s() K_SUBJECT missing", __func__);
		}
	} else if (actn == A_EDIT_S || actn == A_EDITPOST_S
	    || actn == A_EDIT_A || actn == A_EDITPOST_A) {
		/*
		 * K_SUBJECT field content is HTML.
		 * K_SUBJECT is mandatory for A_EDITPOST_A, A_EDITPOST_S.
		 */
		if (r->fieldmap[K_SUBJECT] != NULL) {
			strlcpy(rei->e.subject,
			    r->fieldmap[K_SUBJECT]->parsed.s,
			    sizeof(rei->e.subject));
		} else if (r->fieldnmap[K_SUBJECT] != NULL) {
			rei->b.invalid |= SNB(K_SUBJECT);
			DMSG(r, "%s() K_SUBJECT invalid", __func__);
		} else if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
			rei->b.missing |= SNB(K_SUBJECT);
			DMSG(r, "%s() K_SUBJECT missing", __func__);
		}
	} else {
		msg(r, MSG_CRITICAL, "%s() unexpected action %d", __func__,
		    actn);
	}

	/* K_TYPE may be missing or empty. */
	if (r->fieldmap[K_TYPE] != NULL) {
		if (r->fieldmap[K_TYPE]->parsed.i != 0) {
			rei->e.flags |= FLAG_PLAIN;
		} else {
			rei->e.flags &= ~FLAG_PLAIN;
		}
	} else if (r->fieldnmap[K_TYPE] != NULL) {
		rei->b.invalid |= SNB(K_TYPE);
		DMSG(r, "%s() K_TYPE invalid", __func__);
	} else {
		/* Missing */
		rei->b.missing |= SNB(K_TYPE);
		DMSG(r, "%s() K_TYPE missing", __func__);
	}

	/* K_POLL */
	if (r->fieldmap[K_POLL] != NULL
	    && strcmp(r->fieldmap[K_POLL]->parsed.s, "yes") == 0) {
		/*XXX*/
		rei->e.flags |= FLAG_POLL;
	}

	/* K_PUBDATE */
	if (r->fieldmap[K_PUBDATE] != NULL) {
		/* K_PUBDATE value is valid only if in future. */
		make_current_timestamp(ts_now, sizeof(ts_now));
		if (strcmp(r->fieldmap[K_PUBDATE]->parsed.s, ts_now) <= 0) {
			kutil_invalidate(r, r->fieldmap[K_PUBDATE]);
			rei->b.invalid |= SNB(K_PUBDATE);
			DMSG(r, "%s() K_PUBDATE value not in future",
			    __func__);
		} else {
			strlcpy(rei->e.pubdate,
			    r->fieldmap[K_PUBDATE]->parsed.s,
			    sizeof(rei->e.pubdate));
		}
	} else if (r->fieldnmap[K_PUBDATE] != NULL
	    && r->fieldnmap[K_PUBDATE]->valsz > 0) {
		rei->b.invalid |= SNB(K_PUBDATE);
		DMSG(r, "%s() K_PUBDATE value invalid", __func__);
	}

	/* K_TS_AUTHORED */
	if (r->fieldmap[K_TS_AUTHORED] != NULL) {
		/* The value has been validated already, so just convert. */
		if (strptime(r->fieldmap[K_TS_AUTHORED]->parsed.s,
		    TIMESTAMP_FMT, &rei->e.time_authored) != NULL) {
			/* Failed; just press on without the value. */
			msg(r, MSG_ERROR, "%s() strptime failed for \"%s\"",
			    __func__, r->fieldmap[K_TS_AUTHORED]->parsed.s);
			memset(&rei->e.time_authored, 0,
			    sizeof(rei->e.time_authored));
		}
	} else if (r->fieldnmap[K_TS_AUTHORED] != NULL
	    && r->fieldnmap[K_TS_AUTHORED]->valsz > 0) {
		rei->b.invalid |= SNB(K_TS_AUTHORED);
		DMSG(r, "%s() K_TS_AUTHORED invalid", __func__);
	}

	/* K_ID_AUTHOR - ID of original contributor */
	if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
		if (r->fieldmap[K_ID_AUTHOR] != NULL) {
			strlcpy(rei->e.id_author,
			    r->fieldmap[K_ID_AUTHOR]->parsed.s,
			    sizeof(rei->e.id_author));
		} else if (r->fieldnmap[K_ID_AUTHOR] != NULL) {
			if (r->fieldnmap[K_ID_AUTHOR]->valsz > 0) {
				rei->b.invalid |= SNB(K_ID_AUTHOR);
				DMSG(r, "%s() K_ID_AUTHOR invalid", __func__);
			}
		}
	}

	/* K_ADDR_AUTHOR - IP adrress of original contributor */
	if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
		if (r->fieldmap[K_ADDR_AUTHOR] != NULL) {
			strlcpy(rei->e.addr_author,
			    r->fieldmap[K_ADDR_AUTHOR]->parsed.s,
			    sizeof(rei->e.addr_author));
		} else if (r->fieldnmap[K_ADDR_AUTHOR] != NULL) {
			if (r->fieldnmap[K_ADDR_AUTHOR]->valsz > 0) {
				rei->b.invalid |= SNB(K_ADDR_AUTHOR);
				DMSG(r, "%s() K_ADDR_AUTHOR invalid",
				    __func__);
			}
		}
	}

	/* K_ID_EDITOR - ID of original contributor */
	if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
		if (r->fieldmap[K_ID_EDITOR] != NULL) {
			strlcpy(rei->e.id_editor,
			    r->fieldmap[K_ID_EDITOR]->parsed.s,
			    sizeof(rei->e.id_editor));
		} else if (r->fieldnmap[K_ID_EDITOR] != NULL) {
			if (r->fieldnmap[K_ID_EDITOR]->valsz > 0) {
				rei->b.invalid |= SNB(K_ID_EDITOR);
				DMSG(r, "%s() K_ID_EDITOR invalid", __func__);
			}
		}
	}

	/* K_ADDR_EDITOR - IP adrress of original contributor */
	if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
		if (r->fieldmap[K_ADDR_EDITOR] != NULL) {
			strlcpy(rei->e.addr_editor,
			    r->fieldmap[K_ADDR_EDITOR]->parsed.s,
			    sizeof(rei->e.addr_editor));
		} else if (r->fieldnmap[K_ADDR_EDITOR] != NULL) {
			if (r->fieldnmap[K_ADDR_EDITOR]->valsz > 0) {
				rei->b.invalid |= SNB(K_ADDR_EDITOR);
				DMSG(r, "%s() K_ADDR_EDITOR invalid",
				    __func__);
			}
		}
	}

	/* K_CONTENT0 - the article/comment content. */
	if (r->fieldmap[K_CONTENT0] != NULL) {
		rei->e.body = r->fieldmap[K_CONTENT0]->val;
	} else if (r->fieldnmap[K_CONTENT0] != NULL) {
		rei->b.invalid |= SNB(K_CONTENT0);
		DMSG(r, "%s() K_CONTENT0 invalid", __func__);
	} else {
		rei->b.missing |= SNB(K_CONTENT0);
		DMSG(r, "%s() K_CONTENT0 missing", __func__);
	}

	/*
	 * K_CONTENT1 - optional additional content when editing articles
	 * (unpublished or published).
	 */
	if (actn == A_EDITPOST_A || actn == A_EDITPOST_S) {
		if (r->fieldmap[K_CONTENT1] != NULL
		    && r->fieldmap[K_CONTENT1]->valsz > 0) {
			rei->e.more =
			    kstrdup(r->fieldmap[K_CONTENT1]->val);
		} else if (r->fieldnmap[K_CONTENT1] != NULL
		    && r->fieldnmap[K_CONTENT1]->valsz > 0) {
			rei->b.invalid |= SNB(K_CONTENT1);
			DMSG(r, "%s() K_CONTENT1 invalid", __func__);
		}
	}

	/* Ignore CAPCTHA here. */

#ifdef COMPAT_LEGACY_R
	strlcpy(rei->e.host, r->remote, sizeof(rei->e.host));
#endif /* COMPAT_LEGACY_R */

	if (rei->e.flags & FLAG_EDIT) {
		/* Assuming normal usage, this is from the edit form. */
		if (r->fieldmap[K_TOPIC] != NULL) {
			s = r->fieldmap[K_TOPIC]->parsed.s;
			if (strchr(s, '\r') != NULL || strchr(s, '\n')
			    != NULL) {
				kutil_invalidate(r, r->fieldmap[K_TOPIC]);
				rei->b.invalid |= SNB(K_TOPIC);
				DMSG(r, "%s() K_TOPIC invalid", __func__);
			} else {
				strlcpy(rei->e.topic, s, sizeof(rei->e.topic));
			}
		} else if (r->fieldnmap[K_TOPIC] != NULL) {
			rei->b.invalid |= SNB(K_TOPIC);
			DMSG(r, "%s() K_TOPIC invalid", __func__);
		} else {
			rei->b.missing |= SNB(K_TOPIC);
			DMSG(r, "%s() K_TOPIC missing", __func__);
		}

		if (r->fieldmap[K_DEPT] != NULL) {
			s = r->fieldmap[K_DEPT]->parsed.s;
			if (strchr(s, '\r') != NULL
			    || strchr(s, '\n') != NULL) {
				kutil_invalidate(r, r->fieldmap[K_DEPT]);
				rei->b.invalid |= SNB(K_DEPT);
				DMSG(r, "%s() K_DEPT invalid", __func__);
			} else {
				strlcpy(rei->e.dept, s, sizeof(rei->e.dept));
			}
		} else if (r->fieldnmap[K_DEPT] != NULL) {
			rei->b.invalid |= SNB(K_DEPT);
			DMSG(r, "%s() K_DEPT invalid", __func__);
		} else {
			rei->b.missing |= SNB(K_DEPT);
			DMSG(r, "%s() K_DEPT missing", __func__);
		}
	}

	if (!(rei->b.invalid || rei->b.missing)) {
		/* Time of last edit - use current time */
		t = xtime();
		if (gmtime_r(&t, &rei->e.time_edited) == NULL) {
			msg(r, MSG_CRITICAL, "%s() gmtime failed", __func__);
			memset(&rei->e.time_edited, 0,
			    sizeof(rei->e.time_edited));
		} else {
			rei->e.time_edited.tm_isdst = 0;
			rei->e.time_edited.tm_gmtoff = 0;
			rei->e.time_edited.tm_zone = utc_name;
		}
	}
}


/*
 * Based on any cookie K_AUTH_COOKIE in the HTTP request r, fills out
 * information regarding the user (*u) and the actions which the user is
 * entitled to perform (*ent).
 */
static void
check_auth_cookie(struct kreq *r, struct user *u, uint64_t *ent)
{
	struct	auth_cookie	*cookie;
	char	 vm[4096];
	char	 m[1024];
	time_t	 now;
	bool	 lckd;

	DMSG(r, "%s()", __func__);

	/* Intialise information to be returned. */

	memset(u, 0, sizeof(*u));
	*ent = get_action_entitlements(r, PSEUDO_GROUP_ANON);

	if (r->scheme != KSCHEME_HTTPS) {
		DMSG(r, "not checking for auth cookie because not using"
		    " HTTPS");
		if (r->cookiemap[K_AUTH_COOKIE] != NULL) {
			msg(r, MSG_NOTE, "%s() ignoring cookie sent without"
			    " HTTPS", __func__);
		}
	} else if (r->cookienmap[K_AUTH_COOKIE] != NULL) {
		DMSG(r, "%s() sandboxed auth cookie validation failed",
		    __func__);
	} else if (r->cookiemap[K_AUTH_COOKIE] == NULL) {
		DMSG(r, "%s() sandboxed auth cookie validation not attempted",
		    __func__);
	} else {
		DMSG(r, "%s() sandboxed auth cookie validation succeeded",
		    __func__);

		cookie = (struct auth_cookie*)r->cookiemap[K_AUTH_COOKIE]->val;

		/* Reproduce the cookie message component. */
		snprintf(m, sizeof(m), "%s/%lld/%s", cookie->id, cookie->t,
		    cookie->remote);

		/* Make a logging-safe version of the cookie message. */
		(void)strnvis(vm, m, sizeof(vm), VFLAGS);

		if (!fdk_hmac_validate(get_conf_val(CI_COOKIE_KEY), m,
		    strlen(m), cookie->hmac)) {
			/* HMAC validation failed. */
			msg(r, MSG_WARNING, "%s() HMAC verification failed for"
			    " cookie message \"%s\"", __func__, vm);
		} else {
			DMSG(r, "%s() HMAC verification succeeded for cookie"
			    " message \"%s\"", __func__, vm);

			now = xtime();
			if (cookie->t < now) {
				DMSG(r, "%s() ignoring expired cookie \"%s\"",
				    __func__, vm);
			} else if (strcmp(cookie->remote, r->remote)) {
				DMSG(r, "%s() cookie address mismatch for"
				    " \"%s\"", __func__, vm);
			} else if (user_find(cookie->id, u, &lckd)
			    != USER_ERR_NONE) {
				msg(r, MSG_ERROR, "%s() cookie with valid HMAC"
				    " from unknown user: \"%s\"", __func__,
				    vm);
				memset(u, 0, sizeof(*u));
			} else if (lckd) {
				msg(r, MSG_NOTE, "%s() cookie with valid HMAC"
				    " from locked account: \"%s\"", __func__,
				    vm);
				memset(u, 0, sizeof(*u));
			} else if (strcmp(u->group, GROUP_PART_REG) == 0) {
				msg(r, MSG_NOTE, "%s() ignoring valid cookie"
				    " from incompletely-registered user:"
				    "\"%s\"", __func__, vm);
				memset(u, 0, sizeof(*u));
			} else {
				/* We have a valid cookie, so set rights. */
				*ent = get_action_entitlements(r, u->group);
			}
		}
	}

	if (r->cookiemap[K_AUTH_COOKIE] != NULL) {
		/* Securely wipe sensitive information. */
		explicit_bzero(r->cookiemap[K_AUTH_COOKIE]->val,
		    r->cookiemap[K_AUTH_COOKIE]->valsz);
		/* Just on principle, mark as invalid. */
		kutil_invalidate(r, r->cookiemap[K_AUTH_COOKIE]);
	}
}


/*
 * Return true if and only if entry is directory with name which looks like a
 * timestamp for index_year and is lexicographically no later than index_now.
 */
static int
index_filter(const struct dirent *de)
{
	struct	tm	 ts;

	/* XXX we could check here that a non-empty article file exists. */

	return (de->d_type == DT_DIR && de->d_namlen == TIMESTAMP_LEN
	    && strncmp(de->d_name, index_year, sizeof(index_year) - 1) == 0
	    && strptime(de->d_name, TIMESTAMP_FMT, &ts) != NULL)
	    && strcmp(de->d_name, index_now) <= 0;
}


/*
 * Return true if and only if path refers to a regular file which is
 * of size greater than zero.
 */
static bool
is_nonempty_file(const char *path)
{
	struct	stat	 si;

	return (stat(path, &si) == 0 && S_ISREG(si.st_mode) && si.st_size > 0);
}


/*
 * Copies to subj (of size sz) the subject of article identified by ts
 * (or "?" if the subject cannot be determined).
 */
static void
get_subject(const char *ts, char *subj, size_t sz)
{
	char	 path[PATH_MAX];
	char	 buf[1024];
	FILE	*f;
	size_t	 ll;
	bool	 done;

	strlcpy(subj, "?", sz);  /* until we do better */

	make_article_file_path(ts, path, sizeof(path));
	if (is_nonempty_file(path)) {
		f = fopen(path, "r");
		if (f != NULL) {
			done = false;
			while (!done && fgets(buf, sizeof(buf), f) != NULL) {
				ll = strlen(buf);

				/* Delete any trailing '\n'. */
				if (ll > 0 && buf[ll - 1] == '\n') {
					buf[ll - 1] = '\0';
					ll--;
				}

				if (ll == 0) {
					/* End of article "headers". */
					done = true;
				} else {
					if (strncmp(buf, "subject: ", 9) == 0)
					    {
						if (ll > 9) {
							/* Non-empty */
							strlcpy(subj, &buf[9],
							    sz);
						}
						done = true;
					}
				}
			}
			fclose(f);
		}
	}
}


/*
 * Returns true if and only if no archive index template exists for year yr,
 * such a template exists but is less recent than the time-of-last-change
 * indicator file for yr, or a relevant error occurs.
 */
static bool
need_new_index_template(int yr)
{
	struct	stat	st_ch;
	struct	stat	st_it;
	char	 path_ch[PATH_MAX];
	char	 path_it[PATH_MAX];
	int	 err;
	bool	 nn;

	nn = true;
	snprintf(path_it, sizeof(path_it), FMT_TEMPLATE_ARCHIVE_INDEX,
	    get_conf_val(CI_CACHE_DIR), yr);
	err = stat(path_it, &st_it);
	if (err != 0) {
		/* Not found or other problem, so we need new one. */
		nn = true;
	} else {
		make_last_article_change_path(yr, path_ch, sizeof(path_ch));
		err = stat(path_ch, &st_ch);
		if (err != 0) {
			/*
			 * Not found or other problem.  As we have a template,
			 * it's reasonable to assume we don't need a new one.
			 */
			nn = false;
		} else {
			/* We should compare the modification times. */
			nn = st_ch.st_mtim.tv_sec > st_it.st_mtim.tv_sec
			    || (st_ch.st_mtim.tv_sec == st_it.st_mtim.tv_sec
			    && st_ch.st_mtim.tv_nsec > st_it.st_mtim.tv_nsec);
		}
	}

	return nn;
}


/*
 * Macro rendering function for archive index templates.
 *
 * Strictly, this should be multiple separate functions.  However, as it is
 * needed in specialised circumstances only, we simplify by combining into
 * a single callback.
 */
static int
write_ai_meta_macro(const void *m, size_t len, void *ref)
{
	struct	ai_tmpl_info	*ti;
	char	 s0[2048];
	char	 s1[2048];
	char	 tpath[PATH_MAX];
	const char	*alt;
	const char	*sep;
	const char	*deeper;
	size_t	 n;
	int	 i;
	int	 o;
	char	 as[3];

	ti = (struct ai_tmpl_info *)ref;

	if (STRNEQ((const char *)m, len, MCR_SUBJECT)) {
		get_subject(ti->de[ti->i]->d_name, s0, sizeof(s0));
		write_string(ti->fd, s0);
	} else if (STRNEQ((const char *)m, len, MCR_SID)) {
		write_string(ti->fd, ti->de[ti->i]->d_name);
	} else if (STRNEQ((const char *)m, len, MCR_AI_MONTH_LIST)) {
		snprintf(tpath, sizeof(tpath), "%s/%s",
		    get_conf_val(CI_TEMPLATE_DIR), TMPL_AI_META_MONTH);
		for (i = 0; i < 12; i++) {
			ti->yt.tm_mon = i;
			tmpl_file(tpath, TMDCP, write_data, &ti->fd,
			    write_ai_meta_macro, ti);
		}
	} else if (STRNEQ((const char *)m, len, MCR_AI_MONTH_ARTICLES)) {
		snprintf(tpath, sizeof(tpath), "%s/%s",
		    get_conf_val(CI_TEMPLATE_DIR), TMPL_AI_META_MONTH_ART);
		/* Calculate starting offset. */
		o = 0;
		for (i = 0; i < ti->yt.tm_mon; i++) {
			o += ti->mas[i];
		}
		for (ti->i = o; ti->i < o + ti->mas[ti->yt.tm_mon]; ti->i++)
		    {
			tmpl_file(tpath, TMDCP, write_data, &ti->fd,
			    write_ai_meta_macro, ti);
		}
	} else if (len == 0) {
		/* Write macro-delimiting sequence (using handy buffer). */
		as[0] = TMDCP;
		as[1] = as[0];
		as[2] = '\0';
		write_string(ti->fd, as);
	} else if (len >= (n = strlen(MCR_PREFIX_TIME_I_FMT))
	    && strncmp(m, MCR_PREFIX_TIME_I_FMT, n) == 0) {
		if (len > n) {
			/* Make a NUL-terminated copy of the format. */
			snprintf(s0, sizeof(s0), "%.*s", (int)(len - n),
			    &((const char *)m)[n]);
		} else {
			strlcpy(s0, DEFAULT_TIME_FMT, sizeof(s0));
		}
		if (strftime(s1, sizeof(s1), s0, &ti->yt) == 0) {
			msg(ti->r, MSG_ERROR, "%s() strftime failed",
			    __func__);
		} else {
			write_string(ti->fd, s1);
		}
	} else if (len >= (n = strlen(MCR_PREFIX_COND_A_FOUND))
	    && strncmp(m, MCR_PREFIX_COND_A_FOUND, n) == 0) {
		if (len == n) {
			msg(ti->r, MSG_ERROR, "%s() no alternatives with macro"
			    " \"" MCR_PREFIX_COND_A_FOUND "\"", __func__);
		} else {
			alt = &((const char *)m)[n];

			as[0] = T_MACRO_ALT_BRA;
			as[1] = ti->c;
			as[2] = T_MACRO_ALT_KET;

			/* Attempt to locate separator. */
			sep = memmem(alt, len, as, sizeof(as));

			/* Consider macro marker character to use. */
			if (ti->c == '\0') {
				/* The maximum depth has been reached. */
				msg(ti->r, MSG_ERROR, "%s() maximum macro"
				    " depth reached", __func__);
			} else if ((deeper =
			    strchr(get_conf_val(CI_TEMPLATE_MACRO_CHARS),
			    ti->c)) == NULL) {
				/* Something is very wrong. */
				msg(ti->r, MSG_ERROR, "%s() '%c' not found in"
				    " \"%s\"", __func__, ti->c,
				    get_conf_val(CI_TEMPLATE_MACRO_CHARS));
			} else {
				/* Use next character, even if NUL. */
				deeper++;
				ti->c = *deeper;

				if (ti->mas[ti->yt.tm_mon] > 0) {
					if (sep == NULL || sep > alt) {
						tmpl_buf(alt, sep != NULL
						    ? (size_t)(sep - alt)
						    : len,
						    (unsigned char)ti->c,
						    write_data, &ti->fd,
						    write_ai_meta_macro, ti);
					}
				} else {
					if (sep != NULL && sep + sizeof(as)
					    < alt + len) {
						tmpl_buf(sep + sizeof(as),
						    ((const char *)m + len)
						    - (sep + sizeof(as)),
						    (unsigned char)ti->c,
						    write_data, &ti->fd,
						    write_ai_meta_macro, ti);
					}
				}
			}

			ti->c = as[1];  /* Restore from convenient copy. */
		}
	} else {
		msg(ti->r, MSG_ERROR, "%s() unhandled macro \"%.*s\"",
		    __func__, (int)len, (const char *)m);
	}

	return 0;  /* Assume/pretend all went well. */
}


/*
 * Writes, to fd, a dynamically-generated template for the index (by month) of
 * articles for year yr.
 */
static void
write_ai_template(const struct kreq *r, int fd, int yr)
{
	struct	 ai_tmpl_info	 ai_info;
	char	 articles_dir[PATH_MAX];
	char	 tpath[PATH_MAX];
	int	 i;
	int	 m;
	int	 total;

	DMSG(r, "%s()", __func__);

	make_articles_dir_path(articles_dir, sizeof(articles_dir));

	snprintf(index_year, sizeof(index_year), "%.4d", yr);
	make_current_timestamp(index_now, sizeof(index_now));
	total = scandir(articles_dir, &ai_info.de, index_filter, alphasort);
	if (total == -1) {
		msg(r, MSG_ERROR, "%s() scandir returned -1", __func__);
	} else {
		memset(&ai_info.yt, 0, sizeof(ai_info.yt));
		ai_info.yt.tm_hour = 12;
		ai_info.yt.tm_mday = 1;
		ai_info.yt.tm_year = yr - 1900;
		ai_info.yt.tm_isdst = 0;
		ai_info.yt.tm_gmtoff = 0;
		ai_info.yt.tm_zone = utc_name;

		/* Determine the article count for each month. */
		for (m = 0; m <= 11; m++) {
			ai_info.mas[m] = 0;
		}
		for (i = 0; i < total; i++) {
			ai_info.mas[SID_MONTH(ai_info.de[i]->d_name) - 1]++;
		}

		ai_info.c  = TMDCP;
		ai_info.fd = fd;
		ai_info.r = r;

		snprintf(tpath, sizeof(tpath), "%s/%s",
		    get_conf_val(CI_TEMPLATE_DIR), TMPL_AI_META_YEAR);
		tmpl_file(tpath, TMDCP, write_data, &ai_info.fd,
		    write_ai_meta_macro, &ai_info);

		/* Free scandir() results. */
		for (i = 0; i < total; i++) {
			free(ai_info.de[i]);  /* ai_info.de[i] = NULL; */
		}
		free(ai_info.de);  /* ai_info.de = NULL; */
	}
}


/*
 * Creates a new archive index template file for year yr.
 */
static void
create_new_index_template(const struct kreq *r, int yr)
{
	char	 path_tmp[PATH_MAX];
	char	 path_dst[PATH_MAX];
	int	 fd;

	DMSG(r, "%s()", __func__);

	strlcpy(path_tmp, "/tmp/template.XXXXXXXX", sizeof(path_tmp));

	fd = mkstemp(path_tmp);  /* (Attempt to) create temporary file. */
	if (fd == -1) {
		msg(r, MSG_CRITICAL, "%s() mkstemp() failed", __func__);
	} else {
		/* Save to temporary file an index template for year yr. */
		write_ai_template(r, fd, yr);
		close(fd);

		snprintf(path_dst, sizeof(path_dst),
		    FMT_TEMPLATE_ARCHIVE_INDEX, get_conf_val(CI_CACHE_DIR),
		    yr);
		DMSG(r, "%s() renaming template \"%s\" ==> \"%s\"", __func__,
		    path_tmp, path_dst);

		/* (Attempt to) install saved template in cache. */
		if (rename(path_tmp, path_dst) != 0) {
			msg(r, MSG_CRITICAL, "%s() rename() %s", __func__,
			    strerror(errno));
		}
	}
}


/*
 * Macro rendering function for archive index document.
 */
static void
render_ai_year(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	if (STRNEQ(m, len, MCR_AI_YEAR)) {
		khttp_puts(r, (const char *)a);
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * Rendering function for article (archive) index.
 */
static void
render_archive_index(struct kreq *r, char c, const char *m, size_t len,
    const void *a)
{
	const struct	ai_years	*yrs;
	char	 path[PATH_MAX];
	char	 s[256];
	int	 y;

	/* DMSG(r, "%s() %.*s", __func__, (int)len, m); */

	yrs = (const struct ai_years *)a;

	if (STRNEQ(m, len, MCR_AI_YEAR)) {
		snprintf(s, sizeof(s), "%d", yrs->target);
		khttp_puts(r, s);
	} else if (STRNEQ(m, len, MCR_AI_YEAR_INDEX)) {
		/* Template should be in place. */
		snprintf(path, sizeof(path), FMT_TEMPLATE_ARCHIVE_INDEX,
		    get_conf_val(CI_CACHE_DIR), yrs->target);
		use_template_file(r, c, path, render_common, NULL);
	} else if (STRNEQ(m, len, MCR_AI_EACH_YEAR)) {
		for (y = yrs->first; y <= yrs->last; y++) {
			snprintf(s, sizeof(s), "%d", y);
			use_template_file(r, c, TMPL_AI_YEAR_LIST_ITEM,
			    render_ai_year, s);
		}
	} else {
		render_common(r, c, m, len, a);
	}
}


/*
 * In an If-Modified-Since head is present in request r and can be parsed,
 * returns the time therein, otherwise returns zero.
 */
static time_t
get_if_modified_since(struct kreq *r)
{
	struct	tm	 hv;
	time_t	 ims;
	const char	*s;
	const char	*t;

	ims = 0;

	if ((s = get_header_stringne(r, KREQU_IF_MODIFIED_SINCE)) != NULL) {
		if ((t = strptime(s, RFC_822_FMT, &hv)) == NULL
		    || *t != '\0') {
			msg(r, MSG_NOTE, "%s() strptime failed for"
			    " If-Modified-Since header \"%s\"", __func__, s);
		} else if ((ims = mktime(&hv)) == -1) {
			msg(r, MSG_NOTE, "%s() mktime failed for"
			    " If-Modified-Since header \"%s\"", __func__, s);
			ims = 0;
		}
	}

	return ims;
}


/*
 * Handler for action A_ABOUT.  Generates a response containing the "about"
 * document (based on template TMPL_ABOUT).
 *
 * This could be handled by A_SHOW_DOC, but is retained for backward
 * compatibility, the sake of convention, and in case we wish to include
 * dynamic content in the future.
 */
static void
do_action_about(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_ABOUT, render_common, NULL);
}


/*
 * Returns the current year (UTC).
 */
static int
current_year(void)
{
	struct	tm	 ts;

	xutc_tm_now(&ts);

	return ts.tm_year + 1900;
}


/*
 * Returns the year of the earliest (published) article based on sid/timestamp.
 */
static int
year_of_earliest_article(const struct kreq *r)
{
	struct	dirent	 entry;
	struct	dirent	*dp;
	struct	tm	 ts;
	DIR	*d;
	char	 articles[PATH_MAX];
	int	 y;

	DMSG(r, "%s()", __func__);

	y = current_year();
	make_articles_dir_path(articles, sizeof(articles));

	if ((d = opendir(articles)) == NULL) {
		msg(r, MSG_CRITICAL, "%s() opendir(\"%s\"): %s", __func__,
		    articles, strerror(errno));
	} else {
		/* Initialise to avoid spurious cppcheck error report. */
		ts.tm_year = 0;

		while (readdir_r(d, &entry, &dp) == 0 && dp != NULL) {
			if (dp->d_namlen == TIMESTAMP_LEN
			    && strptime(dp->d_name, TIMESTAMP_FMT, &ts)
			    != NULL) {
				if (ts.tm_year + 1900 < y) {
					y = ts.tm_year + 1900;
				}
			}
		}
		(void)closedir(d);
	}

	return y;
}


/*
 * Handler for action A_ARCHIVE.  Displays the archive for one year:
 * - if a valid year (field K_YEAR) is specified, that number contstrained
 *   between the year of earliest article and the current year, or
 * - the current year by default
 *
 * For efficiency, maintains a cached template.
 */
static void
do_action_archive(struct kreq *r)
{
	struct	ai_years	 yrs;

	DMSG(r, "%s()", __func__);

	yrs.first = year_of_earliest_article(r);
	yrs.last = current_year();

	if (r->fieldmap[K_YEAR] != NULL) {
		/* Use given value constrained to reasonable range. */
		if (r->fieldmap[K_YEAR]->parsed.i < yrs.first) {
			DMSG(r, "%s() year too early", __func__);
			yrs.target = yrs.first;
		} else if (r->fieldmap[K_YEAR]->parsed.i > yrs.last) {
			DMSG(r, "%s() year in future", __func__);
			yrs.target = yrs.last;
		} else {
			yrs.target = r->fieldmap[K_YEAR]->parsed.i;
		}
	} else {
		/* Field missing or invalid, so use current year. */
		yrs.target = yrs.last;
	}

	/* Create/update template if needed. */
	if (need_new_index_template(yrs.target)) {
		create_new_index_template(r, yrs.target);
	}

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_AI_YEAR_INDEX_DOC,
	    render_archive_index, &yrs);
}


/*
 * Handler for action A_ARTICLE.  In the absence of error, displays the
 * requested article.
 */
static void
do_action_article(struct kreq *r)
{
	struct	entry	 e;
	struct	a_rend_info	 ai;
	time_t	 if_modified_since;
	time_t	 last_modified;
	const char	*sid;

	DMSG(r, "%s()", __func__);

	/*
	 * For pledge(2), we need at most PROMISES_RW; the 'W' is for updating
	 * the time of last reading for user [get_lastread() in
	 * render_article()].  This applies if and only if user is logged in,
	 * so if not, narrow the pledge().
	 */
	if (ANON_USER()) {
		xpledge(PROMISES_R);
	}

	if (r->fieldmap[K_SID] != NULL) {
		/*
		 * Well-formed value for K_SID (although there may be
		 * no such article).
		 */
		sid = r->fieldmap[K_SID]->parsed.s;
		DMSG(r, "%s() valid K_SID value \"%s\"", __func__, sid);
	} else if (r->fieldnmap[K_SID] != NULL) {
		/* Invalid value for K_SID. */
		DMSG(r, "%s() invalid K_SID value", __func__);
		resp_not_found(r);
		return;
	} else {
		/*
		 * No value for K_SID in request.  Rather than reporting
		 * an error, just ignore the problem and redirect.
		 */
		DMSG(r, "%s() K_SID value not specified", __func__);
		resp_redirect_to_front(r, NULL);
		return;
	}

	/*
	 * find_newest_change(sid, false) is appropriate for showing the
	 * "4d3hm3m ago" blurb but NOT for handling If-Modified-Since because
	 * it does not detect comment deletion.  (This is particularly
	 * problematic during development/testing!)
	 */

	if_modified_since = get_if_modified_since(r);
	if (if_modified_since != 0 && if_modified_since >=
	    (last_modified = find_newest_change(sid, true))) {
		resp_not_modified(r, last_modified);

		return;
	}

	memset(&e, 0, sizeof(e));
	strlcpy(e.sid, sid, sizeof(e.sid));
	make_article_file_path(sid, e.fn, sizeof(e.fn));

	if (read_file(r, &e) != 0) {
		/* Article does not exist. */
		DMSG(r, "%s() article \"%s\" not found", __func__, sid);
		resp_not_found(r);
		return;
	}

	ai.e = &e;
	ai.last_change = find_newest_change(sid, false);
	if (KNOWN_USER()) {
		if (!get_lastread(sid, &ai.last_read, 1)) {
			ai.last_read = 0;
		}
	} else {
		ai.last_read = 0;
	}

	ai.cmts_count = count_comments(sid);
	ai.cmts_open = are_comments_open(sid);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

#ifdef FEATURE_SHOW_THREAD
	if (r->fieldmap[K_PID] != NULL) {
		strlcpy(e.pid, r->fieldmap[K_PID]->parsed.s, sizeof(e.pid));
		use_template_file(r, TMDCP, TMPL_COMMENTS_DOC, render_article,
		    &ai);
	} else {
		use_template_file(r, TMDCP, TMPL_ARTICLE, render_article, &ai);
	}
#else /* FEATURE_SHOW_THREAD */
	use_template_file(r, TMDCP, TMPL_ARTICLE, render_article, &ai);
#endif /* FEATURE_SHOW_THREAD */
}


/*
 * Handler for action A_AUTH.  In normal use, called as a result of submission
 * of login form.  If credentials are valid, client is redirected (with cookie)
 * to the front page.
 */
static void
do_action_auth(struct kreq *r)
{
	struct	user	 uinfo;
	char	 as[2048];
	char	 m[1024];
	char	 hmb64[FDK_HMAC_B64_SIZE];
	int	 uerr;
	bool	 checked;

	DMSG(r, "%s()", __func__);

	checked = false;

	if (r->fieldmap[K_PW] != NULL) {
		if (r->fieldmap[K_ID] != NULL) {
			/* Check the supplied credentials. */
			uerr = user_check_pw(r->fieldmap[K_ID]->parsed.s,
			    r->fieldmap[K_PW]->parsed.s, &uinfo);

			DMSG(r, "%s() user_check_pw() returned %d", __func__,
			    uerr);

			checked = true;
		}

		/* Securely wipe the password. */
		explicit_bzero(r->fieldmap[K_PW]->val,
		    r->fieldmap[K_PW]->valsz);
		kutil_invalidate(r, r->fieldmap[K_PW]);
	}

	if (!checked) {
		msg(r, MSG_INFO, "%s() invalid ID and/or password", __func__);
		/* Set fake code for appropriate response. */
		msg(r, MSG_NOTE, "%s() faking USER_ERR_NO_SUCH_USER",
		    __func__);
		uerr = USER_ERR_NO_SUCH_USER;
	} else if (uerr == USER_ERR_NONE
	    && strcmp(uinfo.group, GROUP_PART_REG) == 0) {
		/* User has not completed registration. */
		msg(r, MSG_NOTE, "%s() attempted authentication by"
		    " incompletely-registered user"
#ifdef COMPAT_LEGACY_ID
		    , __func__);
#else /* COMPAT_LEGACY_ID */
		    " \"%s\"", __func__, uinfo.id);
#endif /* COMPAT_LEGACY_ID */

		/* Set fake code for appropriate response. */
		msg(r, MSG_NOTE, "%s() faking USER_ERR_NO_SUCH_USER",
		    __func__);
		uerr = USER_ERR_NO_SUCH_USER;
	}

	switch (uerr) {

	case USER_ERR_NONE :
		/* Credentials were correct (and user is registered). */
		msg(r, MSG_INFO, "%s() correct credentials for user"
#ifdef COMPAT_LEGACY_ID
		    , __func__);
#else /* COMPAT_LEGACY_ID */
		    " \"%s\"", __func__, r->fieldmap[K_ID]->parsed.s);
#endif /* COMPAT_LEGACY_ID */

		/* Construct a cookie and send it to client. */
		snprintf(m, sizeof(m), "%s/%lld/%s",
		    r->fieldmap[K_ID]->parsed.s, xtime() + 24 * 60 * 60,
		    r->remote);
		fdk_hmac_generate_b64(get_conf_val(CI_COOKIE_KEY), m,
		    strlen(m), hmb64);

		if (hmb64[0] == '\0') {
			msg(r, MSG_CRITICAL, "%s() fdk_hmac_generate_b64()"
			    " failed for \"%s\"", __func__, m);
			resp_internal_error(r);
		} else {
			DMSG(r, "%s() key = \"%s\", msg = \"%s\","
			    " HMAC=\"%s\"", __func__,
			    get_conf_val(CI_COOKIE_KEY), m, hmb64);
			snprintf(as, sizeof(as), "%s/%s", m , hmb64);
			explicit_bzero(hmb64, sizeof(hmb64));

			resp_redirect_to_front(r, as);

			explicit_bzero(as, sizeof(as));
		}
	break;

	case USER_ERR_NO_SUCH_USER :
	case USER_ERR_PW_INCORRECT :
#ifdef COMPAT_LEGACY_ID
		msg(r, MSG_NOTE, "%s() authentication failed (%d)", __func__,
		    uerr);
#else /* COMPAT_LEGACY_ID */
		if (uerr != USER_ERR_PW_INCORRECT) {
			msg(r, MSG_NOTE, "%s() authentication failed (no such"
			    " user)", __func__);
		} else {
			msg(r, MSG_NOTE, "%s() authentication failed for"
			    " \"%s\" (invalid password)", __func__, uinfo.id);
		}
#endif /* COMPAT_LEGACY_ID */
		/*
		 * Ensure that both K_ID and K_PW fields are invalidated, so
		 * that login form behaviour is consistent between errors.
		 */
		if (r->fieldmap[K_ID] != NULL) {
			kutil_invalidate(r, r->fieldmap[K_ID]);
		}

		/* Present user with login form (again). */
		DMSG(r, "%s() (re-)displaying login form", __func__);
		resp_login(r);
	break;

	default :
		/* Internal error. */
		msg(r, MSG_CRITICAL, "%s() user_check_pw() returned %d",
		    __func__, uerr);
		resp_internal_error(r);
	break;

	}
}


/*
 * Handler for actions A_BLOCK_ADDR, A_UNBLOCK_ADDR.
 *
 * XXX dummy implementation only, at present.
 */
static void
do_common_blockunblock_addr(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	msg(r, MSG_INFO, "%s() NYI", __func__);
	resp_ok(r, TMPL_S_200_NYI, NULL);
}


/*
 * Handler for action A_CHANGEPW.  In the absence of error, updates the
 * user's password.
 */
static void
do_action_changepw(struct kreq *r)
{
	struct	field_errs	 fe;
	int	 err;

	DMSG(r, "%s()", __func__);

	fe.invalid = 0;
	fe.missing = 0;

	/* Check given password. */
	if (r->fieldmap[K_PW] != NULL) {
		if (!user_is_valid_password(r->fieldmap[K_PW]->parsed.s)) {
			DMSG(r, "%s() invalid K_PW", __func__);
			kutil_invalidate(r, r->fieldmap[K_PW]);
			fe.invalid |= SNB(K_PW);
		}
	} else if (r->fieldnmap[K_PW] != NULL) {
		DMSG(r, "%s() invalid K_PW", __func__);
		fe.invalid |= SNB(K_PW);
	} else {
		DMSG(r, "%s() missing K_PW", __func__);
		fe.missing |= SNB(K_PW);
	}

	/* Check given confirmation password. */
	if (r->fieldmap[K_PWCONFIRM] != NULL) {
		if (!user_is_valid_password(
		    r->fieldmap[K_PWCONFIRM]->parsed.s)) {
			DMSG(r, "%s() invalid K_PWCONFIRM", __func__);
			kutil_invalidate(r, r->fieldmap[K_PWCONFIRM]);
			fe.invalid |= SNB(K_PWCONFIRM);
		}
	} else if (r->fieldnmap[K_PWCONFIRM] != NULL) {
		DMSG(r, "%s() invalid K_PWCONFIRM", __func__);
		fe.invalid |= SNB(K_PWCONFIRM);
	} else {
		DMSG(r, "%s() missing K_PWCONFIRM", __func__);
		fe.missing |= SNB(K_PWCONFIRM);
	}

	/* If both passwords are OK, check whether they match. */
	if (!(fe.invalid || fe.missing) && strcmp(r->fieldmap[K_PW]->parsed.s,
	    r->fieldmap[K_PWCONFIRM]->parsed.s) != 0) {
		/* Passwords do not match. */
		DMSG(r, "%s() non-matching K_PW and K_PWCONFIRM", __func__);
		kutil_invalidate(r, r->fieldmap[K_PW]);
		fe.invalid |= SNB(K_PW);
		kutil_invalidate(r, r->fieldmap[K_PWCONFIRM]);
		fe.invalid |= SNB(K_PWCONFIRM);
	}

	if (fe.invalid || fe.missing) {
		/* Go back to the form. */
		DMSG(r, "%s() (re-)displaying prefs form", __func__);
		resp_prefs(r, &fe);
	} else {
		/* Attempt to change the password. */
		err = user_change_pw(user.id, r->fieldmap[K_PW]->parsed.s);

		/* Wipe passwords. */
		explicit_bzero(r->fieldmap[K_PW]->val,
		    r->fieldmap[K_PW]->valsz);
		explicit_bzero(r->fieldmap[K_PWCONFIRM]->val,
		    r->fieldmap[K_PWCONFIRM]->valsz);
		kutil_invalidate(r, r->fieldmap[K_PW]);
		kutil_invalidate(r, r->fieldmap[K_PWCONFIRM]);

		if (err != USER_ERR_NONE) {
			msg(r, MSG_ERROR, "%s() user_change_pw() returned %d",
			    __func__, err);
			resp_internal_error(r);
		} else {
			msg(r, MSG_NOTE, "%s() password changed for \"%s\"",
			    __func__, LSAFE_ID);
			resp_ok(r, TMPL_S_200_PW_CHANGED, NULL);
		}
	}
}


/*
 * Handler for actions A_CLOSE_COMMENTS, A_OPEN_COMMENTS.  In the absence of
 * error, closes or opens comment submission for the specified article.
 */
static void
do_common_openclose_comments(struct kreq *r)
{
	uint64_t	 invalid;
	uint64_t	 missing;

	DMSG(r, "%s()", __func__);

	invalid = 0;
	missing = 0;

	/* Check K_SID. */
	if (r->fieldmap[K_SID] != NULL) {
		if (!article_exists(r->fieldmap[K_SID]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_SID]);
			invalid |= SNB(K_SID);
		}
	} else if (r->fieldnmap[K_SID] != NULL) {
		invalid |= SNB(K_SID);
	} else {
		missing |= SNB(K_SID);
	}

	if (invalid) {
		/* Respond with "not found" rather than just "bad request". */
		resp_not_found(r);
	} else if (missing) {
		/* This should not happen in normal use. */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, invalid, missing);
		resp_bad_request(r);
	} else {
		set_comments_open(r, r->fieldmap[K_SID]->parsed.s,
		    get_action(r) == A_OPEN_COMMENTS);
		resp_redirect_to_article(r, r->fieldmap[K_SID]->parsed.s);
	}
}


/*
 * Handler for action A_CONFIRM, the final step in user account registration.
 * In the absence of error, changes group of specified user from GROUP_PART_REG
 * to GROUP_USER.
 */
static void
do_action_confirm(struct kreq *r)
{
	struct	user	 u;
	char	 s[1024];
	const char	*id, *hmac;
	char	*escdi;
	int	 uerr;
	bool	 lckd;

	DMSG(r, "%s()", __func__);

	id = r->fieldmap[K_ID] != NULL ? r->fieldmap[K_ID]->parsed.s : NULL;
	hmac = r->fieldmap[K_H] != NULL ? r->fieldmap[K_H]->parsed.s : NULL;

	if (id == NULL || hmac == NULL
	    || !fdk_hmac_validate_b64(get_conf_val(CI_REGISTRATION_KEY), id,
	    strlen(id), hmac)) {
		msg(r, MSG_NOTE, "%s() invalid id/hmac (%s/%s)", __func__,
		    id != NULL ? id : "<invalid/missing>",
		    hmac != NULL ? hmac : "<invalid/missing>");
		resp_forbidden(r, TMPL_S_403_CNFMN_INVAL);

		return;
	}

	memset(&u, 0, sizeof(u));
	strlcpy(u.id, id, sizeof(u.id));

	if ((uerr = user_find(id, &u, &lckd)) != USER_ERR_NONE) {
		if (uerr == USER_ERR_NO_SUCH_USER) {
			/*
			 * User not present in password file.  This should not
			 * happen (given that the HMAC has been checked)!
			 */
			msg(r, MSG_WARNING, "%s() user not found", __func__);
			resp_forbidden(r, TMPL_S_403_CNFMN_BAD_ID);
		} else {
			msg(r, MSG_ERROR, "%s() user_find() failed [%d]",
			    __func__, uerr);
			resp_internal_error(r);
		}

		return;
	}

	if (lckd) {
		/* Account is locked. */
		msg(r, MSG_WARNING, "%s() account is locked", __func__);
		resp_forbidden(r, TMPL_S_403_CNFMN_LOCKED);

		return;
	} else if (strcmp(u.group, GROUP_PART_REG) != 0) {
		/* Account already activated. */
		resp_ok(r, TMPL_S_200_ALREADY_ACT, NULL);

		return;
	}

	strlcpy(u.group, GROUP_USER, sizeof(u.group));
	if ((uerr = user_update(&u)) != USER_ERR_NONE) {
		msg(r, MSG_CRITICAL, "%s() user_update() returned %d",
		    __func__, uerr);
		resp_internal_error(r);
	} else {
		msg(r, MSG_NOTE, "%s() id \"%s\" confirmed", __func__, u.id);
		DMSG(r, "%s() redirecting to login form", __func__);
		if ((escdi = kutil_urlencode(u.id)) == NULL) {
			msg(r, MSG_ERROR, "%s() kutil_urlencode() failed",
			    __func__);
			resp_redirect_to_front(r, NULL);
		} else {
			resp_client_redirect(r, url(r, URL_F_ABS, s, sizeof(s),
			    A_LOGIN, NULL, NULL, SEP KN_ID EQ "%s", escdi),
			    NULL);
			free(escdi);  /* id = NULL; */
		}
	}
}


/*
 * Handler for action A_DEAUTH.  Logs out client by setting empty auth cookie.
 */
static void
do_action_deauth(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	msg(r, MSG_INFO, "%s() deauthorising", __func__);
	resp_redirect_to_front(r, "");
}


/*
 * Handler for action A_DELETE_C.  In the absence of error, deletes the
 * specified comment.
 */
static void
do_action_delete_comment(struct kreq *r)
{
	const char	*sid;
	const char	*pid;
	uint64_t	 invalid;
	uint64_t	 missing;

	DMSG(r, "%s()", __func__);

	sid = NULL;
	pid = NULL;
	invalid = 0;
	missing = 0;

	/* Check K_SID. */
	if (r->fieldmap[K_SID] != NULL) {
		sid = r->fieldmap[K_SID]->parsed.s;
	} else if (r->fieldnmap[K_SID] != NULL) {
		invalid |= SNB(K_SID);
	} else {
		missing |= SNB(K_SID);
	}

	/* Check K_PID. */
	if (r->fieldmap[K_PID] != NULL) {
		pid = r->fieldmap[K_PID]->parsed.s;
	} else if (r->fieldnmap[K_PID] != NULL) {
		invalid |= SNB(K_PID);
	} else {
		missing |= SNB(K_PID);
	}

	if (invalid || missing) {
		resp_not_found(r);
	} else if (delete_comment(sid, pid) != 0) {
		msg(r, MSG_ERROR, "%s() delete_comment() failed for sid %s"
		    " pid %s", __func__, sid, pid);
		resp_internal_error(r);
	} else {
		msg(r, MSG_NOTE, "%s() deleted comment sid %s pid %s",
		    __func__, sid, pid);
		resp_redirect_to_article(r, sid);
	}
}


/*
 * Handler for actions A_DELETE_A, A_DELETE_S.  If field K_ASK specifies a
 * valid value for false, deletes the specified article or submission (in the
 * absence of error), otherwise requests confirmation for such a deletion.
 */
static void
do_common_delete_sa(struct kreq *r)
{
	const char	*sid;
	char	 u[1024];
	int	 e;
	bool	 ask, is_subm;

	DMSG(r, "%s()", __func__);

	/*
	 * Check whether this is (unpublished) submission
	 * or (published) article.
	 */
	is_subm = get_action(r) == A_DELETE_S;

	/* Check K_SID. */
	if (r->fieldmap[K_SID] != NULL) {
		sid = r->fieldmap[K_SID]->parsed.s;
	} else if (r->fieldnmap[K_SID] != NULL) {
		sid = NULL;
	} else {
		sid = NULL;
	}

	/* Check K_ASK. */
	if (r->fieldmap[K_ASK] != NULL) {
		ask = r->fieldmap[K_ASK]->parsed.i != 0;
	} else if (r->fieldnmap[K_ASK] != NULL) {
		/* Value was invalid so err on the side of caution. */
		ask = true;
	} else {
		/* Value not specified: consider it false. */
		ask = false;
	}

	if (sid == NULL) {
		msg(r, MSG_WARNING, "%s() invalid or missing K_SID", __func__);
		resp_not_found(r);
	} else if (ask) {
		/* Request confirmation before deleting. */
		emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
		emit_http_end_headers(r);

		use_template_file(r, TMDCP, is_subm ? TMPL_S_200_ASK_DEL_S
		    : TMPL_S_200_ASK_DEL_A, render_extra_simple, sid);
	} else if (is_subm) {
		if (!submission_exists(sid)) {
			msg(r, MSG_WARNING, "%s() submission \"%s\" not found",
			    __func__, sid);
			resp_not_found(r);
		} else if ((e = delete_submission(sid)) != 0) {
			msg(r, MSG_CRITICAL, "%s() failed to delete submission"
			    " \"%s\": %s", __func__, sid, strerror(e));
			resp_internal_error(r);
		} else {
			msg(r, MSG_NOTE, "%s() deleted submission \"%s\"",
			    __func__, sid);
			resp_client_redirect(r, url(r, URL_F_ABS, u, sizeof(u),
			    A_SUBMISSIONS, NULL, NULL, NULL), NULL);
		}
	} else /* if (!is_subm) */ {
		if (!article_exists(sid)) {
			msg(r, MSG_WARNING, "%s() article \"%s\" not found",
			    __func__, sid);
			resp_not_found(r);
		} else if ((e = delete_article(sid)) != 0) {
			msg(r, MSG_CRITICAL, "%s() failed to delete article"
			    " \"%s\": %s", __func__, sid, strerror(e));
			resp_internal_error(r);
		} else {
			/* Avoid stale index fragment. */
			notify_article_change(r, sid);

			msg(r, MSG_NOTE, "%s() deleted article \"%s\"",
			    __func__, sid);
			resp_redirect_to_front(r, NULL);
		}
	}
}


/*
 * Sends an HTTP status 200 response with HTML document containing the
 * "edit" form for unpublished and published articles.
 */
static void
resp_edit(struct kreq *r, const struct req_ent_info *d)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, (d->e.flags & FLAG_SUBMISSION)
	    ? TMPL_EDIT_S : TMPL_EDIT_A, render_edit, d);
}


/*
 * Handler for actions A_EDIT_A and A_EDIT_S.  In the absence of error,
 * generates the form document for editing an unpublished or published
 * article.
 */
static void
do_common_edit(struct kreq *r)
{
	struct	req_ent_info	 rei;

	DMSG(r, "%s()", __func__);

	/* Initialise request entry information. */
	memset(&rei.e, 0, sizeof(rei.e));
	rei.b.invalid = 0;
	rei.b.missing = 0;

	/* Setup flags. */
	rei.e.flags |= FLAG_EDIT;
	if (get_action(r) == A_EDIT_S) {
		rei.e.flags |= FLAG_SUBMISSION;
	}

	/* Try to get the sid from the request. */
	if (r->fieldmap[K_SID] != NULL) {
		strlcpy(rei.e.sid, r->fieldmap[K_SID]->parsed.s,
		    sizeof(rei.e.sid));
	} else if (r->fieldnmap[K_SID] != NULL) {
		rei.b.invalid |= SNB(K_SID);
	} else {
		rei.b.missing |= SNB(K_SID);
	}

	if (rei.b.invalid || rei.b.missing) {
		/* This should not happen in normal use. */
		msg(r, MSG_WARNING, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, rei.b.invalid, rei.b.missing);
		resp_bad_request(r);
	} else {
		if (rei.e.flags & FLAG_SUBMISSION) {
			make_submission_file_path(rei.e.sid, rei.e.fn,
			    sizeof(rei.e.fn));
		} else {
			make_article_file_path(rei.e.sid, rei.e.fn,
			    sizeof(rei.e.fn));
		}

		if (read_file(r, &rei.e) != 0) {
			msg(r, MSG_ERROR, "%s() read_file() failed for %s \""
			    "%s\"", __func__, (rei.e.flags & FLAG_SUBMISSION)
			    ? "submission" : "article", rei.e.fn);

			resp_not_found(r);
		} else {
			resp_edit(r, &rei);
		}
	}
}


/*
 * Writes to f the metadata and (main) content of (article or comment) e.
 * is_new indicates whether the article/comment is new (as opposed to edited),
 * which affects some of the (metadata) headers.
 */
static void
write_file(const struct kreq *r, FILE *f, const struct entry *e, bool is_new)
{
	struct	tm	 ts;
	char	 scratch[64];
	size_t	 len;

	/* Firstly,  deal with fields independed of is_new. */

	fprintf(f, MD_FMT, MD_F_NAME, e->name);
	fprintf(f, MD_FMT, MD_F_EMAIL, e->email);
	fprintf(f, MD_FMT, MD_F_HREF, e->href);
	fprintf(f, MD_FMT, MD_F_SUBJECT, e->subject);

#ifdef COMPAT_LEGACY_W
	/*
	 * The "host" field is ambiguous.  If user is logged in, the user ID
	 * used, otherwise the client address is used.
	 */
	if (KNOWN_USER()) {
		/* User is logged-in, so use ID. */
		fprintf(f, MD_FMT, MD_F_HOST, user.id);
	} else {
		fprintf(f, MD_FMT, MD_F_HOST, r->remote);
	}
#endif /* COMPAT_LEGACY_W */

	/* Details for most recent edit(or) are also independent of is_new. */
	/* Record ID. */
	if (KNOWN_USER()) {
		fprintf(f, MD_FMT, MD_F_ID_EDITOR, user.id);
	}
	/* Record IP address. */
	fprintf(f, MD_FMT, MD_F_ADDR_EDITOR, r->remote);

	/* Record time of last edit, using the current time */
	xutc_tm_now(&ts);
	if (strftime(scratch, sizeof(scratch), TIMESTAMP_FMT, &ts) == 0) {
		msg(r, MSG_CRITICAL, "%s() strftime failed", __func__);
	} else {
		fprintf(f, MD_FMT, MD_F_TIME_EDITED, scratch);
	}

#ifdef COMPAT_LEGACY_W
	if (strftime(scratch, sizeof(scratch), CTIME_FMT, &ts) == 0) {
		msg(r, MSG_CRITICAL, "%s() strftime failed", __func__);
	} else {
		fprintf(f, MD_FMT, MD_F_DATE, scratch);
	}
#endif /* COMPAT_LEGACY_W */

	/* Fields relating to authorship depend on is_new. */

	if (is_new) {
		/* Record ID and address of current client (if available). */
		if (KNOWN_USER()) {
			/* User is logged-in, so record ID. */
			fprintf(f, MD_FMT, MD_F_ID_AUTHOR, user.id);
		}
		fprintf(f, MD_FMT, MD_F_ADDR_AUTHOR, r->remote);
		/* Use current time as time of authoring. */
		if (strftime(scratch, sizeof(scratch), TIMESTAMP_FMT, &ts)
		    == 0) {
			msg(r, MSG_CRITICAL, "%s() strftime failed", __func__);
		} else {
			fprintf(f, MD_FMT, MD_F_TIME_AUTHORED, scratch);
		}

		/* Record default as topic if an article. */
		if (e->sid[0] == '\0') {
			fprintf(f, MD_FMT, MD_F_TOPIC,
			    get_conf_val(CI_DEFAULT_TOPIC));
		}
	} else {
		/* Record ID and address of original contributor. */
		if (e->id_author[0] != '\0') {
			fprintf(f, MD_FMT, MD_F_ID_AUTHOR, e->id_author);
		}
		if (e->addr_author[0] != '\0') {
			fprintf(f, MD_FMT, MD_F_ADDR_AUTHOR, e->addr_author);
		}
		/* Record time authored. */
		if (e->time_authored.tm_zone != NULL) {
			if (strftime(scratch, sizeof(scratch), TIMESTAMP_FMT,
			    &e->time_authored) == 0) {
				msg(r, MSG_CRITICAL, "%s() strftime failed",
				    __func__);
			} else {
				fprintf(f, MD_FMT, MD_F_TIME_AUTHORED,
				    scratch);
			}
		}

		/* Record the topic. */
		fprintf(f, MD_FMT, MD_F_TOPIC, e->topic);
	}

	/* Deal with fields present only in editing case. */

	if (!is_new) {
		if (e->pubdate[0] != '\0') {
			fprintf(f, MD_FMT, MD_F_PUBDATE, e->pubdate);
		}

		fprintf(f, MD_FMT, MD_F_DEPT, e->dept);
		if (e->flags & FLAG_POLL) {
			fprintf(f, MD_FLAG_FMT, MD_F_POLL);
		}
	}

	/*
	 * Finally, deal with the article/comment content data.
	 * A blank line separates metadata from content.
	 */
	fprintf(f, "\n");
	/*
	 * If new content from a user with suitable entitlements, or if editing
	 * existing content, write as-is.  Otherwise, cleanse.
	 */
	if (!is_new || ((SNB(A_EDIT_A) | SNB(A_EDIT_S)) & ent_act)) {
		if (e->body != NULL) {
			fprintf(f, "%s", e->body);
			/* Ensure body ends in '\n'. */
			if ((len = strlen(e->body)) > 0
			    && e->body[len - 1] != '\n') {
				fprintf(f, "\n");
			}
		}
	} else {
		rhf_print_cleansed(f, e->body, ATTR_REL_NO_FOLLOW);
		fprintf(f, "\n");
	}

	fflush(f);
}


/*
 * Handler for actions EDITPOST_A and A_EDITPOST_S.  Handles the submission of
 * forms used for editing (unpublished) submissions and (published) articles.
 */
static void
do_common_editpost(struct kreq *r)
{
	struct	req_ent_info	 ei;
	struct	entry	 o;
	char	 u[1024];
	FILE	*f;

	DMSG(r, "%s()", __func__);

	build_entry_from_request(r, &ei);

	if (((ei.e.flags & FLAG_SUBMISSION)
	    && ((ei.b.invalid | ei.b.missing) & ~EDIT_S_FORM_USER_K_MASK))
	    || (!(ei.e.flags & FLAG_SUBMISSION)
	    && ((ei.b.invalid | ei.b.missing) & ~EDIT_A_FORM_USER_K_MASK))) {
		/* There exist field errors which are not user-correctable. */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, ei.b.invalid, ei.b.missing);
		resp_bad_request(r);
	} else if (r->fieldmap[K_PREVIEW] != NULL
	    || ei.b.invalid || ei.b.missing) {
		/*
		 * The user requested a preview, and/or (all) field errors
		 * are user-correctable.
		 */
		resp_edit(r, &ei);

		return;
	}

	/* Save the (submitted or published) article. */

	if (ei.e.flags & FLAG_SUBMISSION) {
		make_submission_file_path(ei.e.sid, ei.e.fn, sizeof(ei.e.fn));
	} else {
		make_article_file_path(ei.e.sid, ei.e.fn, sizeof(ei.e.fn));
	}

	memset(&o, 0, sizeof(o));
	strlcpy(o.fn, ei.e.fn, sizeof(o.fn));
	if (access(o.fn, R_OK) != 0 || read_file(r, &o) != 0) {
		msg(r, MSG_ERROR, "%s() access/read_file: %s", __func__, o.fn);
		resp_internal_error(r);

		return;
	}
	if (o.name[0] != '\0') {
		strlcpy(ei.e.name, o.name, sizeof(ei.e.name));
	} else {
		strlcpy(ei.e.name, user.id, sizeof(ei.e.name));
	}

	if ((f = fopen(ei.e.fn, "w")) == NULL) {
		msg(r, MSG_ERROR, "%s() fopen: %s: %s", __func__, ei.e.fn,
		    strerror(errno));
		resp_internal_error(r);

		return;
	}

	/* Write out the edited content. */
	write_file(r, f, &ei.e, false);
	fclose(f);

	if (ei.e.flags & FLAG_SUBMISSION) {
		make_submission_more_path(ei.e.sid, ei.e.fn, sizeof(ei.e.fn));
	} else {
		make_article_more_path(ei.e.sid, ei.e.fn, sizeof(ei.e.fn));
	}

	if (access(ei.e.fn, R_OK) == 0 && unlink(ei.e.fn) != 0) {
		xpledge(PROMISES_MIN);
		msg(r, MSG_ERROR, "%s() unlink: %s: %s", __func__, ei.e.fn,
		    strerror(errno));

		return;
	}
	if (ei.e.more != NULL && ei.e.more[0] != '\0') {
		if ((f = fopen(ei.e.fn, "w")) == NULL) {
			xpledge(PROMISES_MIN);
			msg(r, MSG_ERROR, "%s() fopen: %s: %s", __func__,
			    ei.e.fn, strerror(errno));

			return;
		} else {
			fprintf(f, "%s", ei.e.more);
			fflush(f);
			fclose(f);
		}
	}

	if (ei.e.flags & FLAG_SUBMISSION) {
		msg(r, MSG_NOTE, "%s() saved submission sid %s", __func__,
		    ei.e.sid);
		resp_client_redirect(r, url(r, URL_F_ABS, u, sizeof(u),
		    A_SUBMISSIONS, NULL, NULL, NULL), NULL);
	} else {
		notify_article_change(r, ei.e.sid);

		msg(r, MSG_NOTE, "%s() saved published sid %s", __func__,
		    ei.e.sid);
		resp_redirect_to_front(r, NULL);
	}
}


#ifdef UNDEADLY_TRANSITION
/*
 * Deals with action A_ERRATA by redirecting to the (externally-generated)
 * errata (RSS) document.  This is provided for the transition from the
 * Undeadly-specific past, in which the errata RSS feed was generated
 * on-the-fly.
 */
static void
do_action_errata(struct kreq *r)
{
	char	*u;

	DMSG(r, "%s()", __func__);

	DMSG(r, "%s() redirecting to search engine", __func__);
	(void)kasprintf(&u, FMT_URL_ERRATA, scheme_str(r),
	    get_conf_val(CI_DOMAIN));

	resp_client_redirect_perm(r, u, NULL);
	free(u); /* u = NULL; */
}
#endif /* UNDEADLY_TRANSITION */


/*
 * Handler for action A_FRONT.  In the absence of errors, generates the
 * "front" page.
 */
static void
do_action_front(struct kreq *r)
{
	struct	entry	 e;
	char	 articles_dir[PATH_MAX];

	DMSG(r, "%s()", __func__);

	make_articles_dir_path(articles_dir, sizeof(articles_dir));
	find_articles(r, articles_dir, newest,
	    ARTICLES_FRONT + ARTICLES_OLDER_LIST, false, &poll);
	memset(&e, 0, sizeof(e));

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_FRONT, render_front, &e);
}


/*
 * Handler for actions A_LOCK_ACCT and A_UNLOCK_ACCT.  In the absence of error,
 * locks or unlocks the account specified by field K_ID.
 */
static void
do_common_lockunlock_account(struct kreq *r)
{
	const char	*id;
	int	 lerr;

	DMSG(r, "%s()", __func__);

	id = NULL;

	if (r->fieldmap[K_ID] != NULL) {
		id = r->fieldmap[K_ID]->parsed.s;
	} else if (r->fieldnmap[K_ID] != NULL) {
		DMSG(r, "%s() invalid K_ID", __func__);
	} else {
		DMSG(r, "%s() missing K_ID", __func__);
	}

	if (id == NULL) {
		resp_bad_request(r);
	} else if (get_action(r) == A_LOCK_ACCT) {
		lerr = user_account_lock(id);
		if (lerr != USER_ERR_NONE) {
			msg(r, MSG_ERROR, "%s() user_account_lock(\"%s\")"
			    " returned %d", __func__, id, lerr);
			resp_internal_error(r);
		} else {
			resp_ok(r, TMPL_S_200_LOCKED, id);
		}
	} else {
		lerr = user_account_unlock(id);
		if (lerr != USER_ERR_NONE) {
			msg(r, MSG_ERROR, "%s() user_account_unlock(\"%s\")"
			    " returned %d", __func__, id, lerr);
			resp_internal_error(r);
		} else {
			resp_ok(r, TMPL_S_200_UNLOCKED, id);
		}
	}
}


/*
 * Sends an HTTP status 200 response with HTML document containing the
 * login form (based on template TMPL_LOGIN).
 */
static void
resp_login(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_LOGIN, render_login, NULL);
}


/*
 * Handler for action A_LOGIN.  Generates a response containing the login form.
 */
static void
do_action_login(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	if (user.id[0] != '\0') {
		/* User is already authenticated. */
		msg(r, MSG_NOTE, "%s() \"%s\" alreadly authorised", __func__,
		    LSAFE_ID);
		resp_redirect_to_front(r, NULL);
	} else {
		resp_login(r);
	}
}


#ifdef FEATURE_MODERATE
/*
 * Handler for action A_MODERATE.  Down- or up-votes the specified comment.
 */
static void
do_action_moderate(struct kreq *r)
{
	uint64_t	 invalid;
	uint64_t	 missing;

	DMSG(r, "%s()", __func__);

	invalid = 0;
	missing = 0;

	/* Check K_SID. */
	if (r->fieldmap[K_SID] != NULL) {
		if (!article_exists(r->fieldmap[K_SID]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_SID]);
			invalid |= SNB(K_SID);
		}
	} else if (r->fieldnmap[K_SID] != NULL) {
		invalid |= SNB(K_SID);
	} else {
		missing |= SNB(K_SID);
	}

	/* Check K_PID. */
	if (r->fieldmap[K_PID] != NULL) {
		if (r->fieldmap[K_SID] != NULL &&
		    !comment_exists(r->fieldmap[K_SID]->parsed.s,
		    r->fieldmap[K_PID]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_PID]);
			invalid |= SNB(K_PID);
		}
	} else if (r->fieldnmap[K_PID] != NULL) {
		invalid |= SNB(K_PID);
	} else {
		missing |= SNB(K_PID);
	}

	/* Check K_VAL.  We accept only -1 and 1. */
	if (r->fieldmap[K_VAL] != NULL) {
		if (r->fieldmap[K_VAL]->parsed.i != -1
		    && r->fieldmap[K_VAL]->parsed.i != 1) {
			kutil_invalidate(r, r->fieldmap[K_VAL]);
			invalid |= SNB(K_VAL);
		}
	} else if (r->fieldnmap[K_VAL] != NULL) {
		invalid |= SNB(K_VAL);
	} else {
		missing |= SNB(K_VAL);
	}

	if ((SNB(K_SID) | SNB(K_PID)) & invalid) {
		/*
		 * K_SID and/or K_PID invalid.  Respond with "not found"
		 * rather than just "bad request".
		 */
		resp_not_found(r);
	} else if (invalid || missing) {
		/* This should not happen in normal use. */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, invalid, missing);
		resp_bad_request(r);
	} else if (moderate_comment(r->remote, r->fieldmap[K_SID]->parsed.s,
	    r->fieldmap[K_PID]->parsed.s,
	    r->fieldmap[K_VAL]->parsed.i) != 0) {
		msg(r, MSG_ERROR, "%s() moderate_comment() failed", __func__);
		resp_internal_error(r);
	} else {
		DMSG(r, "%s() moderated sid=%s pid=%s val=%d",
		    r->fieldmap[K_SID]->parsed.s,
		    r->fieldmap[K_PID]->parsed.s,
		    r->fieldmap[K_VAL]->parsed.i);
		resp_redirect_to_article(r, r->fieldmap[K_SID]->parsed.s);
	}
}
#endif /* FEATURE_MODERATE */


/*
 * Sends an HTTP status 200 response with HTML document containing the
 * user preferences/password form (based on template TMPL_PREFS).
 */
static void
resp_prefs(struct kreq *r, const struct field_errs *e)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_PREFS, render_prefs, e);
}


/*
 * Handler for action A_PREFS.  Generates the user preferences/password form.
 */
static void
do_action_prefs(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	resp_prefs(r, NULL);
}


/*
 * Handler for action A_PREFSPOST.  Deals with POST request containing
 * (updated) user preferences.  Updates stored preferences and redirects
 * client to front page, or reports error.
 */
static void
do_action_prefspost(struct kreq *r)
{
	struct	field_errs	 fe;
	int	 uerr;

	DMSG(r, "%s()", __func__);

	fe.invalid = 0;
	fe.missing = 0;

	if (r->fieldmap[K_NAME] != NULL) {
		/* Name is valid. */
	} else if (r->fieldnmap[K_NAME] != NULL) {
		fe.invalid |= SNB(K_NAME);
	} else {
		fe.missing |= SNB(K_NAME);
	}

	if (r->fieldmap[K_EMAIL] != NULL) {
		if (!user_is_valid_email(r->fieldmap[K_EMAIL]->parsed.s)) {
			DMSG(r, "%s() invalid email address", __func__);
			kutil_invalidate(r, r->fieldmap[K_EMAIL]);
			fe.invalid |= SNB(K_EMAIL);
		}
	} else if (r->fieldnmap[K_EMAIL] != NULL) {
		fe.invalid |= SNB(K_EMAIL);
	} else {
		fe.missing |= SNB(K_EMAIL);
	}

	if (r->fieldmap[K_HREF] != NULL) {
		if (!user_is_valid_href(r->fieldmap[K_HREF]->parsed.s)) {
			DMSG(r, "%s() invalid href", __func__);
			kutil_invalidate(r, r->fieldmap[K_HREF]);
			fe.invalid |= SNB(K_HREF);
		}
	} else if (r->fieldnmap[K_HREF] != NULL) {
		fe.invalid |= SNB(K_HREF);
	} else {
		/* Do not set bit in fe.missing because value is optional. */
	}

	if (fe.invalid || fe.missing) {
		/* Go back to the form. */
		DMSG(r, "%s() (re-)displaying prefs form", __func__);
		resp_prefs(r, &fe);
	} else {
		strlcpy(user.name, r->fieldmap[K_NAME]->parsed.s,
		    sizeof(user.name));
		strlcpy(user.email, r->fieldmap[K_EMAIL]->parsed.s,
		    sizeof(user.email));

		if (r->fieldmap[K_HREF] != NULL) {
			strlcpy(user.href, r->fieldmap[K_HREF]->parsed.s,
			    sizeof(user.href));
		} else {
			user.href[0] = '\0';
		}

		if ((uerr = user_update(&user)) != USER_ERR_NONE) {
			msg(r, MSG_ERROR, "%s() user_update() id \"%s\""
			    " returned %d", __func__, LSAFE_ID, uerr);
			resp_internal_error(r);
		} else {
			resp_redirect_to_front(r, NULL);
		}
	}
}


/*
 * Handler for action A_PUBLISH.  Publishes a submission - that is, makes a
 * published article from an unpublished one.  This involves assigning a
 * new sid.
 */
static void
do_action_publish(struct kreq *r)
{
	struct entry	 e;
	char	 fns[PATH_MAX], fnp[PATH_MAX];
	char	 ts_now[TIMESTAMP_LEN + 1];
	const char	*sid;
	int	 retry;

	DMSG(r, "%s()", __func__);

	if (r->fieldmap[K_SID] != NULL) {
		sid = r->fieldmap[K_SID]->parsed.s;
	} else if (r->fieldnmap[K_SID] != NULL) {
		/* This should not happen in normal use. */
		msg(r, MSG_WARNING, "%s() invalid sid", __func__);
		resp_bad_request(r);

		return;
	} else {
		/* This should not happen in normal use. */
		msg(r, MSG_WARNING, "%s() missing sid", __func__);
		resp_bad_request(r);

		return;
	}

	make_submission_file_path(sid, e.fn, sizeof(e.fn));
	if (read_file(r, &e) != 0) {
		msg(r, MSG_ERROR, "%s() read_file: \"%s\"", __func__, e.fn);
		resp_not_found(r);

		return;
	}

	make_submission_dir_path(sid, fns, sizeof(fns));
	retry = 0;
	while (retry < 3) {
		make_current_timestamp(ts_now, sizeof(ts_now));

		if (e.pubdate[0] == '\0' || strcmp(e.pubdate, ts_now) < 0) {
			strlcpy(e.pubdate, ts_now, sizeof(e.pubdate));
		}

		/* Create path for candidate (published) article directory. */
		make_article_dir_path(e.pubdate, fnp, sizeof(fnp));

		if (mkdir(fnp, 0775) == 0) {
			if (rename(fns, fnp) != 0) {
				msg(r, MSG_ERROR, "%s() rename: \"%s\" \"%s\""
				    ": %s", __func__, fns, fnp,
				    strerror(errno));
				resp_internal_error(r);
				return;
			}

			/* Open comments for article. */
			set_comments_open(r, e.pubdate, 1);

			/* Report the change. */
			notify_article_change(r, e.pubdate);

			break;
		} else {
			msg(r, MSG_ERROR, "%s() mkdir: \"%s\": %s", __func__,
			    fnp, strerror(errno));
		}
		retry++;
		sleep(1);
	}

	if (retry >= 3) {
		msg(r, MSG_ERROR, "%s() sid allocation failed", __func__);
		resp_internal_error(r);
	} else {
		msg(r, MSG_NOTE, "%s() published \"%s\"", __func__, fnp);
		resp_redirect_to_front(r, NULL);
	}
}


/*
 * Sends an HTTP status 200 response with HTML document containing the
 * "registration" form for creating a user account.
 */
static void
resp_register(struct kreq *r, const struct field_errs *fe)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_REG, render_register, fe);

	xpledge(PROMISES_MIN);
}


/*
 * Handler for action A_REGISTER.  Generates the user registration form.
 */
static void
do_action_register(struct kreq *r)
{
	DMSG(r, "%s()", __func__);

	resp_register(r, NULL);
}


/*
 * Writes len bytes of d to the file descriptor to which a points.
 * Returns zero on success, non-zero on failure.
 */
static int
write_data(const void *d, size_t len, void *a)
{
	return write(*(int *)a, d, len) == -1;
}


/*
 * Writes s (excluding terminating NUL) to fd.  Returns zero on success,
 * non-zero on failure.
 */
static int
write_string(int fd, const char *s)
{
	return write(fd, s, strlen(s)) == -1;
}


/*
 * Macro substitution for registration email.  Returns zero on success,
 * non-zero on error.
 */
static int
write_reg_mail_macro(const void *m, size_t len, void *a)
{
	const struct	reg_mail_info	*mi;
	int	 res;

	mi = (const struct reg_mail_info *)a;

	if (STRNEQ(m, len, MCR_EMAIL)) {
		res = write_string(mi->fd, mi->rcpt);
	} else if (STRNEQ(m, len, MCR_ID)) {
		res = write_string(mi->fd, mi->id);
	} else if (STRNEQ(m, len, MCR_HMAC)) {
		res = write_string(mi->fd, mi->hmac);
	} else if (STRNEQ(m, len, MCR_REG_ADDR)) {
		res = write_string(mi->fd, mi->remote);
	} else if (STRNEQ(m, len, MCR_REG_UA)) {
		res = write_string(mi->fd, mi->ua);
	} else if (STRNEQ(m, len, MCR_REG_REF)) {
		res = write_string(mi->fd, mi->ref);
	} else if (STRNEQ(m, len, MCR_REG_SNOM)) {
		res = write_string(mi->fd, mi->snom);
	} else if (STRNEQ(m, len, MCR_REG_SRVR)) {
		res = write_string(mi->fd, mi->srvr);
	} else {
		/* Macro is unknown. */
		res = 1;  /* failure */
	}

	return res;
}


/*
 * Callback function [for mail_send()] for writing a user registration email
 * message.
 */
static int
write_registration_mail(const int fd, void *ref)
{
	char	 tpath[PATH_MAX];
	int	 n;

	snprintf(tpath, sizeof(tpath), "%s/%s", get_conf_val(CI_TEMPLATE_DIR),
	    TMPL_REG_MAIL);

	((struct reg_mail_info *)ref)->fd = fd;
	n = fd;

	return tmpl_file(tpath, TMDCP, write_data, &n, write_reg_mail_macro,
	    ref);
}


/*
 * Sends the user registration email informing the partially-registered user
 * that confirmation (of account creation) is required.  Returns the result
 * of calling mail_send().
 */
static int
send_conf_req_email(struct kreq *r, const char *id, const char *email,
    const char *hmac)
{
	struct	reg_mail_info	 info;

	DMSG(r, "%s()", __func__);

	info.id = id;
	info.hmac = hmac;
	info.rcpt = email;
	info.snom = r->pname;
	info.srvr = r->host;

	info.remote = r->remote;
	if (r->reqmap[KREQU_USER_AGENT] != NULL
	    && r->reqmap[KREQU_USER_AGENT]->val != NULL
	    && r->reqmap[KREQU_USER_AGENT]->val[0] != '\0') {
		info.ua = r->reqmap[KREQU_USER_AGENT]->val;
	} else {
		info.ua = "unknown";
	}
	if (r->reqmap[KREQU_REFERER] != NULL && r->reqmap[KREQU_REFERER]->val
	    != NULL && r->reqmap[KREQU_REFERER]->val[0] != '\0') {
		info.ref = r->reqmap[KREQU_REFERER]->val;
	} else {
		info.ref = "unknown";
	}

	info.fd = -1;

	return mail_send(write_registration_mail, &info);
}


/*
 * Handler for action A_REGISTERPOST.  In the absence of error, creates the
 * user account (as specified in the registration form) with membership in
 * group GROUP_PART_REG, and sends email containing the URL for confirmation
 * of registration.
 */
static void
do_action_registerpost(struct kreq *r)
{
	struct	user	 u;
	struct	field_errs	 fe;
	char	 b64[FDK_HMAC_B64_SIZE];
	const char	*login, *pw, *name, *email, *href;
	const char	*captcha, *cksum;
	char	*h_enc;
	int	 err;

	DMSG(r, "%s()", __func__);

	fe.invalid = 0;
	fe.missing = 0;

	login = NULL;
	if (r->fieldmap[K_ID] != NULL) {
#ifdef COMPAT_LEGACY_ID
		/* We can, at least, prevent creation of new bizarre IDs. */
		if (user_is_valid_id(r->fieldmap[K_ID]->parsed.s)) {
			login = r->fieldmap[K_ID]->parsed.s;
		} else {
			DMSG(r, "%s() K_ID value invalid (for new account)",
			    __func__);
			kutil_invalidate(r, r->fieldmap[K_ID]);
			fe.invalid |= SNB(K_ID);
		}
#else /* COMPAT_LEGACY_ID */
		login = r->fieldmap[K_ID]->parsed.s;
#endif /* COMPAT_LEGACY_ID */
	} else if (r->fieldnmap[K_ID] != NULL) {
		fe.invalid |= SNB(K_ID);
	} else {
		fe.missing |= SNB(K_ID);
	}

	pw = NULL;
	if (r->fieldmap[K_PW] != NULL) {
		if (user_is_valid_password(r->fieldmap[K_PW]->parsed.s)) {
			pw = r->fieldmap[K_PW]->parsed.s;
		} else {
			kutil_invalidate(r, r->fieldmap[K_PW]);
			fe.invalid |= SNB(K_PW);
		}
	} else if (r->fieldnmap[K_PW] != NULL) {
		fe.invalid |= SNB(K_PW);
	} else {
		fe.missing |= SNB(K_PW);
	}
	if (r->fieldmap[K_PWCONFIRM] != NULL) {
		if (!user_is_valid_password(
		    r->fieldmap[K_PWCONFIRM]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_PWCONFIRM]);
			fe.invalid |= SNB(K_PWCONFIRM);
		}
	} else if (r->fieldnmap[K_PWCONFIRM] != NULL) {
		fe.invalid |= SNB(K_PWCONFIRM);
	} else {
		fe.missing |= SNB(K_PWCONFIRM);
	}
	/* Check for non-matching passwords. */
	if (pw != NULL && r->fieldmap[K_PWCONFIRM] != NULL
	    && strcmp(pw, r->fieldmap[K_PWCONFIRM]->parsed.s) != 0) {
		/* Both are invalid. */
		kutil_invalidate(r, r->fieldmap[K_PW]);
		fe.invalid |= SNB(K_PW);
		pw = NULL;
		kutil_invalidate(r, r->fieldmap[K_PWCONFIRM]);
		fe.invalid |= SNB(K_PWCONFIRM);
	}
	/* We don't need the confirmation password any more. */
	if (r->fieldmap[K_PWCONFIRM] != NULL) {
		explicit_bzero(r->fieldmap[K_PWCONFIRM]->val,
		    r->fieldmap[K_PWCONFIRM]->valsz);
	} else if (r->fieldnmap[K_PWCONFIRM] != NULL) {
		explicit_bzero(r->fieldnmap[K_PWCONFIRM]->val,
		    r->fieldnmap[K_PWCONFIRM]->valsz);
	}

	if (r->fieldmap[K_NAME] != NULL) {
		name = r->fieldmap[K_NAME]->parsed.s;
	} else if (r->fieldnmap[K_NAME] != NULL) {
		name = NULL;
		fe.invalid |= SNB(K_NAME);
	} else {
		name = NULL;
		fe.missing |= SNB(K_NAME);
	}

	email = NULL;
	if (r->fieldmap[K_EMAIL] != NULL) {
		if (user_is_valid_email(r->fieldmap[K_EMAIL]->parsed.s)) {
			email = r->fieldmap[K_EMAIL]->parsed.s;
		} else {
			kutil_invalidate(r, r->fieldmap[K_EMAIL]);
			fe.invalid |= SNB(K_EMAIL);
		}
	} else if (r->fieldnmap[K_EMAIL] != NULL) {
		fe.invalid |= SNB(K_EMAIL);
	} else {
		fe.missing |= SNB(K_EMAIL);
	}

	href = NULL;
	if (r->fieldmap[K_HREF] != NULL) {
		if (user_is_valid_href(r->fieldmap[K_HREF]->parsed.s)) {
			href = r->fieldmap[K_HREF]->parsed.s;
		} else {
			kutil_invalidate(r, r->fieldmap[K_HREF]);
			fe.invalid |= SNB(K_HREF);
		}
	} else if (r->fieldnmap[K_HREF] != NULL) {
		fe.invalid |= SNB(K_HREF);
	} else {
		fe.missing |= SNB(K_HREF);
	}

	/* Registration always requires use of CAPTCHA. */
	captcha = NULL;
	if (r->fieldmap[K_CAPTCHA] != NULL) {
		if (strlen(r->fieldmap[K_CAPTCHA]->parsed.s) == CAPTCHA_LEN) {
			captcha = r->fieldmap[K_CAPTCHA]->parsed.s;
		} else {
			kutil_invalidate(r, r->fieldmap[K_CAPTCHA]);
			fe.invalid |= SNB(K_CAPTCHA);
		}
	} else if (r->fieldnmap[K_CAPTCHA] != NULL) {
		fe.invalid |= SNB(K_CAPTCHA);
	} else {
		fe.missing |= SNB(K_CAPTCHA);
	}
	cksum = NULL;
	if (r->fieldmap[K_CAPTCHA_CHK] != NULL) {
		if (strlen(r->fieldmap[K_CAPTCHA_CHK]->parsed.s) > 0) {
			cksum = r->fieldmap[K_CAPTCHA_CHK]->parsed.s;
		} else {
			kutil_invalidate(r, r->fieldmap[K_CAPTCHA_CHK]);
			fe.invalid |= SNB(K_CAPTCHA_CHK);
		}
	} else if (r->fieldnmap[K_CAPTCHA_CHK] != NULL) {
		fe.invalid |= SNB(K_CAPTCHA_CHK);
	} else {
		fe.missing |= SNB(K_CAPTCHA_CHK);
	}

	/* If no errors, check the CAPTCHA result. */
	if (!(fe.invalid || fe.missing)) {
		err = verify_captcha(get_conf_val(CI_CAPTCHA_KEY), captcha,
		    cksum);
		DMSG(r, "%s() verify_captcha() returned %d", __func__, err);

		switch (err) {

		case CAPTCHA_ERR_NONE :
			DMSG(r, "%s() CAPTCHA verification succeeded",
			    __func__);
		break;

		case CAPTCHA_ERR_VS_NOT_WF :
			kutil_invalidate(r, r->fieldmap[K_CAPTCHA_CHK]);
			fe.invalid |= SNB(K_CAPTCHA_CHK);
			msg(r, MSG_NOTE, "%s() poorly-formed K_CAPTCHA_CHK"
			    " field", __func__);
		break;

		case CAPTCHA_ERR_MISMATCH :
		case CAPTCHA_ERR_TIME :
			/* Client failed the CAPTCHA. */
			kutil_invalidate(r, r->fieldmap[K_CAPTCHA]);
			fe.invalid |= SNB(K_CAPTCHA);

			msg(r, MSG_NOTE, "%s() would-be registrant \"%s\""
			    " failed CAPTCHA (%s)", __func__, login != NULL
			    ? login : "<invalid or missing>",
			    err == CAPTCHA_ERR_TIME ? "time" : "mismatch");
		break;

		default :
			/* Should not happen. */
			msg(r, MSG_ERROR, "%s() unexpected verify_captcha()"
			    " result", __func__);
			resp_internal_error(r);
			return;
		break;

		}
	}

	if (fe.invalid || fe.missing) {/* XXX refactor? */
		if ((fe.invalid | fe.missing) & ~REG_FORM_USER_K_MASK) {
			/* This should not happen in normal use. */
			msg(r, MSG_WARNING, "%s() invalid = 0x%.16llx; missing"
			    " = 0x%.16llx", __func__, fe.invalid, fe.missing);
			resp_bad_request(r);
		} else {
			DMSG(r, "%s() invalid = 0x%.16llx; missing ="
			    " 0x%.16llx", __func__, fe.invalid, fe.missing);

			resp_register(r, &fe);
		}
		return;
	}

	/* Set up user details. */
	memset(&u, 0, sizeof(u));
	strlcpy(u.id, login, sizeof(u.id));
	strlcpy(u.group, GROUP_PART_REG, sizeof(u.group));
	strlcpy(u.name, name, sizeof(u.name));
	strlcpy(u.email, email, sizeof(u.email));
	strlcpy(u.href, href, sizeof(u.href));

	/* Add the (not yet confirmed) user account. */
	err = user_add(&u, pw);

	if (err != USER_ERR_NONE) {
		if (err == USER_ERR_USER_EXISTS) {
			msg(r, MSG_NOTE, "%s() id \"%s\" already used",
			    __func__, login);

			DMSG(r, "%s() (re-)displaying register form",
			    __func__);

			resp_register(r, &fe);
		} else {
			msg(r, MSG_ERROR, "%s() user_add() for \"%s\" returned"
			    " %d", __func__, u.id, err);
			resp_internal_error(r);
		}

		return;
	}

	msg(r, MSG_NOTE, "%s() user added: id=\"%s\", email=\"%s\"", __func__,
	    login, email);

	fdk_hmac_generate_b64(get_conf_val(CI_REGISTRATION_KEY), login,
	    strlen(login), b64);
	if (b64[0] == '\0') {
		msg(r, MSG_CRITICAL, "%s() fdk_hmac_generate_b64() failed",
		    __func__);
		resp_internal_error(r);
		return;
	} else if ((h_enc = kutil_urlencode(b64)) == NULL) {
		explicit_bzero(b64, sizeof(b64));
		msg(r, MSG_CRITICAL, "%s() kutil_urlencode() failed",
		    __func__);
		resp_internal_error(r);
		return;
	}

	DMSG(r, "%s() registration details: id=\"%s\", email=\"%s\", hmac="
	    "\"%s\"", __func__, login, email, h_enc);

	err = send_conf_req_email(r, login, email, h_enc);

	explicit_bzero(h_enc, strlen(h_enc));

	if (err != MAIL_ERR_NONE) {
		msg(r, MSG_ERROR, "%s() mail_send() failed [%d]", __func__,
		    err);
		resp_internal_error(r);
	} else {
		msg(r, MSG_NOTE, "%s() registration confirmation email sent to"
		    " %s", __func__, email);
		resp_ok(r, TMPL_S_200_A_EMAIL_SENT, NULL);
	}
}


/*
 * Handler for actions A_REPORT and A_CLOSE_REPORT.
 */
static void
do_common_openclose_report(struct kreq *r)
{
	uint64_t	 invalid;
	uint64_t	 missing;

	DMSG(r, "%s()", __func__);

	invalid = 0;
	missing = 0;

	/* Check K_SID. */
	if (r->fieldmap[K_SID] != NULL) {
		if (!article_exists(r->fieldmap[K_SID]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_SID]);
			invalid |= SNB(K_SID);
		}
	} else if (r->fieldnmap[K_SID] != NULL) {
		invalid |= SNB(K_SID);
	} else {
		missing |= SNB(K_SID);
	}

	/* Check K_PID. */
	if (r->fieldmap[K_PID] != NULL) {
		if (r->fieldmap[K_SID] != NULL &&
		    !comment_exists(r->fieldmap[K_SID]->parsed.s,
		    r->fieldmap[K_PID]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_PID]);
			invalid |= SNB(K_PID);
		}
	} else if (r->fieldnmap[K_PID] != NULL) {
		invalid |= SNB(K_PID);
	} else {
		missing |= SNB(K_PID);
	}

	if (invalid && !missing) {
		/* Respond with "not found" rather than just "bad request". */
		resp_not_found(r);
	} else if (invalid || missing) {
		/* This should not happen in normal use. */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, invalid, missing);
		resp_bad_request(r);
	} else {
		set_report_comment(r, r->fieldmap[K_SID]->parsed.s,
		    r->fieldmap[K_PID]->parsed.s, get_action(r) == A_REPORT);
		resp_redirect_to_article(r, r->fieldmap[K_SID]->parsed.s);
	}
}


/*
 * Handler for action A_RSS.  Generates RSS feed of articles.
 */
static void
do_action_rss(struct kreq *r)
{
	struct	entry	 e;
	char	 articles_dir[PATH_MAX];
	time_t	 if_modified_since, latest, m;
	size_t	 i, n;
	bool	 handled;

	DMSG(r, "%s()", __func__);

	if (rss_rate_limit(r, r->remote)) {
		/* Client has exceeded permitted request rate. */
		resp_service_unavailable(r);
	} else {
		/* Collect the articles of interest. */
		n = RSS_ITEMS_MAX > (sizeof(newest) / sizeof(newest[0]))
		    ? (sizeof(newest) / sizeof(newest[0])) : RSS_ITEMS_MAX;
		make_articles_dir_path(articles_dir, sizeof(articles_dir));
		find_articles(r, articles_dir, newest, n, false, NULL);

		handled = false;

		/* Deal with If-Modified-Since request header (if any). */
		if_modified_since = get_if_modified_since(r);
		if (if_modified_since != 0) {
			/* Determine latest change of any of the articles. */
			/*
			 * XXX Make this faster by restricting information
			 * available for rendering?
			 */
			latest = 0;
			i = 0;
			while (i < n && newest[i].sid[0] != '\0') {
				if ((m = find_newest_change(newest[i].sid,
				    true)) > latest) {
					latest = m;
				}
				i++;
			}

			if (latest > 0 && latest <= if_modified_since) {
				resp_not_modified(r, latest);
				handled = true;
			}
		}

		if (!handled) {
			emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_XML);
			emit_http_end_headers(r);

			memset(&e, 0, sizeof(e));
			use_template_file(r, TMDCP, RSS_SUMMARY, render_rss,
			    &e);
		}
	}
}


#ifdef FEATURE_SEARCH
/*
 * Deals with action A_SEARCH by redirecting to a search engine.  If the value
 * of K_Q is (defined and) valid, it is used as (the non-site component of) the
 * the query string.
 */
static void
do_action_search(struct kreq *r)
{
	const char	*q;
	char	*q_enc, *u;

	DMSG(r, "%s()", __func__);

	q_enc = NULL;
	q = r->fieldmap[K_Q] != NULL ? r->fieldmap[K_Q]->parsed.s : NULL;

	if (q != NULL) {
		/* We have some query text. */
		q_enc = kutil_urlencode(q);
	}
	if (q != NULL && q_enc == NULL) {
		msg(r, MSG_ERROR, "%s() kutil_urlencode failed", __func__);
		resp_internal_error(r);
	} else {
		DMSG(r, "%s() redirecting to search engine", __func__);
		(void)kasprintf(&u, FMT_SE, scheme_str(r),
		    get_conf_val(CI_DOMAIN), q_enc != NULL ? q_enc : "");
		if (q_enc != NULL) {
			free(q_enc); /* q_enc = NULL */
		}

		resp_client_redirect(r, u, NULL);
		free(u); /* u = NULL; */
	}
}
#endif /* FEATURE_SEARCH */


/*
 * Handler for action A_SHOW_DOC.  Displays the requested document or an
 * appropriate error.
 */
static void
do_action_show_doc(struct kreq *r)
{
	char	 doc[PATH_MAX];
	const char	*td;
	const char	*tpc;

	DMSG(r, "%s()", __func__);

	tpc = NULL;

	if (r->fieldmap[K_TOPIC] != NULL) {
		if (!is_safe_rel_path(r->fieldmap[K_TOPIC]->parsed.s,
		    r->fieldmap[K_TOPIC]->valsz)) {
			kutil_invalidate(r, r->fieldmap[K_TOPIC]);
		} else {
			tpc = r->fieldmap[K_TOPIC]->parsed.s;
		}
	} else if (r->fieldnmap[K_TOPIC] != NULL) {
		DMSG(r, "%s() invalid K_TOPIC", __func__);
	} else {
		DMSG(r, "%s() missing K_TOPIC", __func__);
	}

	if (tpc == NULL) {
		resp_bad_request(r);
	} else {
		td = get_conf_val(CI_TEMPLATE_DIR);
		snprintf(doc, sizeof(doc), "%s/" FMT_TMPL_DOC, td, tpc);

		if (access(doc, F_OK) != 0) {
			resp_not_found(r);
		} else {
			emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
			emit_http_end_headers(r);

			use_template_file(r, TMDCP, &doc[strlen(td) + 1],
			    render_common, NULL);
		}
	}
}


/*
 * Handler for action A_SUBMISSION.  In the absence of error, generates an
 * HTTP response containing the unpublished article identified by the sid
 * in the request.
 */
static void
do_action_submission(struct kreq *r)
{
	struct	entry	 e;
	struct	a_rend_info	 ai;
	const char	*sid;

	DMSG(r, "%s()", __func__);

	if (r->fieldmap[K_SID] == NULL) {
		/* Value is invalid or missing. */
		DMSG(r, "%s() invalid or missing K_SID", __func__);
		resp_not_found(r);
	} else {
		/*
		 * Well-formed value for K_SID (although there may be
		 * no such article).
		 */
		sid = r->fieldmap[K_SID]->parsed.s;
		DMSG(r, "%s() valid K_SID value \"%s\"", __func__, sid);

		memset(&e, 0, sizeof(e));
		strlcpy(e.sid, sid, sizeof(e.sid));
		e.flags |= FLAG_SUBMISSION;
		make_submission_file_path(sid, e.fn, sizeof(e.fn));

		if (read_file(r, &e) != 0) {
			/* Submission does not exist. */
			DMSG(r, "%s() submission \"%s\" not found", __func__,
			    sid);
			resp_not_found(r);
		} else {
			ai.e = &e;

			/* Zero the fields applicable to articles only. */
			ai.last_change = 0;
			ai.last_read = 0;
			ai.cmts_count = 0;
			ai.cmts_open = false;

			emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
			emit_http_end_headers(r);

			use_template_file(r, TMDCP, TMPL_SUBMISSION,
			    render_article, &ai);
		}
	}
}


/*
 * Handler for action A_SUBMISSIONS.  In the absence of error, generates an
 * HTTP response containing the list of (unpublished) submissions.
 * Assumes pre-filtering of non-entitled users.
 */
static void
do_action_submissions(struct kreq *r)
{
	struct	entry	 e;
	char	 path[PATH_MAX];

	DMSG(r, "%s()", __func__);

	make_submissions_dir_path(path, sizeof(path));

	find_articles(r, path, newest, sizeof(newest) / sizeof(newest[0]),
	    true, NULL);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	memset(&e, 0, sizeof(e));
	e.flags = FLAG_SUBMISSION;
	use_template_file(r, TMDCP, TMPL_SUBMISSIONS, render_front, &e);
}


/*
 * Sends an HTTP status 200 response with HTML document containing the
 * "reply" form for comments.
 */
static void
resp_reply(struct kreq *r, const struct req_ent_info *d)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_REPLY, render_reply, d);
}


/*
 * Handler for action A_REPLY.  In the absence of error, generates an HTTP
 * response containing a form document for the submission of a comment (on an
 * article or existing comment).
 */
static void
do_action_reply(struct kreq *r)
{
	struct	req_ent_info	 ei;

	DMSG(r, "%s()", __func__);

	/* Get what we can from the request. */
	build_entry_from_request(r, &ei);

	/* If comments are not open, forbid the request. */
	if (r->fieldmap[K_SID] != NULL
	    && !are_comments_open(r->fieldmap[K_SID]->parsed.s)) {
		msg(r, MSG_WARNING, "%s() blocking attempt to comment on"
		    " article %s because comments are not open", __func__,
		    r->fieldmap[K_SID]->parsed.s);
		resp_forbidden(r, TMPL_S_403_CMTS_CLOSED);
	} else if ((ei.b.invalid & (SNB(K_SID) | SNB(K_PID)))
	    || (ei.b.missing & SNB(K_SID))) {
		/*
		 * The request is bad if K_SID and/or K_PID is invalid, of if
		 * K_SID is missing.  (K_PID is missing for a comment on an
		 * article.)  This should not happen in normal use.
		 */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__,
		    ei.b.invalid & (SNB(K_SID) | SNB(K_PID)),
		    ei.b.missing & SNB(K_SID));
		resp_bad_request(r);
	} else {
		/* Use defaults if K_TYPE is invalid or missing. */
		if ((ei.b.invalid | ei.b.missing) & SNB(K_TYPE)) {
			ei.e.flags |= FLAG_PLAIN;
			ei.b.invalid &= ~SNB(K_TYPE);
			ei.b.missing &= ~SNB(K_TYPE);
		}

		/*
		 * If the user is logged in, use the available information for
		 * invalid/missing fields.
		 */
		if (KNOWN_USER()) {
			if (ei.e.name[0] == '\0') {
				strlcpy(ei.e.name, user.name,
				    sizeof(ei.e.name));
			}
			if (ei.e.email[0] == '\0') {
				strlcpy(ei.e.email, user.email,
				    sizeof(ei.e.email));
			}
			if (ei.e.href[0] == '\0') {
				strlcpy(ei.e.href, user.href,
				    sizeof(ei.e.href));
			}

			/* Clear problems */
			ei.b.invalid &= ~(SNB(K_NAME) | SNB(K_EMAIL)
			    | SNB(K_HREF));
			ei.b.missing &= ~(SNB(K_NAME) | SNB(K_EMAIL)
			    | SNB(K_HREF));
		}

		/* XXX this is where we should quote parent if no subject. */

		/* Clear (other) invalid/missing bits. */
		DMSG(r, "%s() invalid = 0x%.16llx; missing = 0x%.16llx"
		    " - clearing", __func__, ei.b.invalid, ei.b.missing);
		ei.b.invalid = 0;
		ei.b.missing = 0;

		/* Send comment form document as response. */
		resp_reply(r, &ei);
	}
}


/*
 * Sends an HTTP status 200 response with HTML document containing the
 * "submit" form for new articles.
 */
static void
resp_submit(struct kreq *r, const struct req_ent_info *d)
{
	DMSG(r, "%s()", __func__);

	xpledge(PROMISES_R);

	emit_http_start_headers(r, KHTTP_200, KMIME_TEXT_HTML);
	emit_http_end_headers(r);

	use_template_file(r, TMDCP, TMPL_SUBMIT, render_submit, d);
}


/*
 * Handler for action A_SUBMIT.  In the absence of error, generates an HTTP
 * response containing a form document for the submission of a new article.
 */
static void
do_action_submit(struct kreq *r)
{
	struct	req_ent_info	 ei;

	DMSG(r, "%s()", __func__);

	/* Get what we can from the request. */
	build_entry_from_request(r, &ei);

	/* Use defaults if K_TYPE is invalid or missing. */
	if ((ei.b.invalid | ei.b.missing) & SNB(K_TYPE)) {
		ei.e.flags |= FLAG_PLAIN;
		ei.b.invalid &= ~SNB(K_TYPE);
		ei.b.missing &= ~SNB(K_TYPE);
	}

	/*
	 * If the user is logged in, use the available information for
	 * invalid/missing fields.
	 */
	if (KNOWN_USER()) {
		if (ei.e.name[0] == '\0') {
			strlcpy(ei.e.name, user.name,
			    sizeof(ei.e.name));
		}
		if (ei.e.email[0] == '\0') {
			strlcpy(ei.e.email, user.email,
			    sizeof(ei.e.email));
		}
		if (ei.e.href[0] == '\0') {
			strlcpy(ei.e.href, user.href,
			    sizeof(ei.e.href));
		}

		/* Clear problems */
		ei.b.invalid &= ~(SNB(K_NAME) | SNB(K_EMAIL) | SNB(K_HREF));
		ei.b.missing &= ~(SNB(K_NAME) | SNB(K_EMAIL) | SNB(K_HREF));
	}

	/* Clear (other) invalid/missing bits. */
	DMSG(r, "%s() invalid = 0x%.16llx; missing = 0x%.16llx"
	    " - clearing", __func__, ei.b.invalid, ei.b.missing);
	ei.b.invalid = 0;
	ei.b.missing = 0;

	/* Send submit form document as response. */
	resp_submit(r, &ei);
}


/*
 * Handler for actions A_SUBMITPOST and A_REPLYPOST.  Handles the submission
 * of the forms for new articles (A_SUBMITPOST) and comments (A_REPLYPOST).
 */
static void
do_common_contribpost(struct kreq *r)
{
	struct	req_ent_info	 ei;
	char	*p;
	FILE	*f;
	size_t	 pos;
	int	 fd;
	int	 captcha_err;
	int	 status;
	bool	 is_comment;
	bool	 preview;

	DMSG(r, "%s()", __func__);

	p = NULL;  /* Avoid compiler warning. */
	is_comment = get_action(r) == A_REPLYPOST;

	/* If this is for a comment, forbid if comments are not open. */
	if (is_comment && r->fieldmap[K_SID] != NULL
	    && !are_comments_open(r->fieldmap[K_SID]->parsed.s)) {
		msg(r, MSG_WARNING, "%s() blocking attempt to post comment to"
		    " article %s because comments are not open", __func__,
		    r->fieldmap[K_SID]->parsed.s);
		resp_forbidden(r, TMPL_S_403_CMTS_CLOSED);

		return;
	}

	build_entry_from_request(r, &ei);
	preview = r->fieldmap[K_PREVIEW] != NULL;

	/*
	 * If it's HTML data from a user without editing rights,
	 * it must be validated.
	 */
	if (ei.e.body != NULL && !(ei.e.flags & FLAG_PLAIN)
	    && !((SNB(A_EDIT_A) | SNB(A_EDIT_S)) & ent_act)) {
		if (ei.e.body[0] == '\0') {
			status = -1;
			DMSG(r, "%s() K_CONTENT0 empty", __func__);
		} else {
			rhf_validate(ei.e.body, &status, &pos);
			DMSG(r, "%s() rhf_validate - %d (%zu)", __func__,
			    status, pos);
		}
		if (status != RHF_OK) {
			kutil_invalidate(r, r->fieldmap[K_CONTENT0]);
			ei.b.invalid |= SNB(K_CONTENT0);
			DMSG(r, "%s() K_CONTENT0 invalid", __func__);
		}
	}

	/* If article (c.f. comment), clear invalid/missing K_SID/K_PID. */
	if (!is_comment) {
		ei.b.invalid &= ~(SNB(K_SID) | SNB(K_PID));
		ei.b.missing &= ~(SNB(K_SID) | SNB(K_PID));
	}

	/* Deal with CAPTCHA. */
	if (preview) {
		/* Ignore problems. */
		ei.b.invalid &= ~(SNB(K_CAPTCHA_CHK) | SNB(K_CAPTCHA));
		ei.b.missing &= ~(SNB(K_CAPTCHA_CHK) | SNB(K_CAPTCHA));
	} else if (ANON_USER()) {
		/* CAPTCHA verification is required. */

		captcha_err = CAPTCHA_ERR_NONE;

		if (r->fieldmap[K_CAPTCHA_CHK] == NULL
		    || r->fieldmap[K_CAPTCHA_CHK]->valsz == 0) {
			/* Should not happen in normal use. */
			msg(r, MSG_NOTE, "%s() K_CAPTCHA_CHK"
			    " invalid/empty/missing", __func__);
			if (r->fieldmap[K_CAPTCHA_CHK] != NULL) {
				kutil_invalidate(r,
				    r->fieldmap[K_CAPTCHA_CHK]);
			}
			ei.b.invalid |= SNB(K_CAPTCHA_CHK);
			msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
			    " 0x%.16llx", __func__, ei.b.invalid,
			    ei.b.missing);
			resp_bad_request(r);

			return;
		} else if (r->fieldmap[K_CAPTCHA] == NULL
		    || r->fieldmap[K_CAPTCHA]->valsz == 0
		    || (captcha_err
		    = verify_captcha(get_conf_val(CI_CAPTCHA_KEY),
		    r->fieldmap[K_CAPTCHA]->parsed.s,
		    r->fieldmap[K_CAPTCHA_CHK]->parsed.s))
		    == CAPTCHA_ERR_MISMATCH
		    || captcha_err == CAPTCHA_ERR_TIME) {
			/*
			 * CAPTCHA verification failed.  Rather than just
			 * reporting an error, just mark as invalid so we
			 * can re-display the submission form to allow for
			 * retry.
			 */

			msg(r, MSG_INFO, "%s() verify_captcha() failed",
			    __func__);

			/* Mark as invalid. */
			if (r->fieldmap[K_CAPTCHA] != NULL) {
				kutil_invalidate(r, r->fieldmap[K_CAPTCHA]);
			}
			ei.b.invalid |= SNB(K_CAPTCHA);
		} else if (captcha_err != CAPTCHA_ERR_NONE) {
			/* Other error occurred in CAPTCHA verification. */
			msg(r, MSG_ERROR, "%s() verify_captcha() failed: %d",
			    __func__, captcha_err);

			resp_internal_error(r);

			return;
		}
	}

	if ((is_comment
	    && ((ei.b.invalid | ei.b.missing) & ~REPLY_FORM_USER_K_MASK))
	    || (!is_comment
	    && ((ei.b.invalid | ei.b.missing) & ~SUBMIT_FORM_USER_K_MASK))) {
		/*
		 * Problem(s) were detected in field(s).  One or more of these
		 * fields is outside the set of user-editable fields in the
		 * form, so the request is bad (and possibly mailicious).
		 * Therefore, report it as such.
		 */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, ei.b.invalid, ei.b.missing);

		resp_bad_request(r);

		return;
	} else if (preview || ei.b.invalid || ei.b.missing) {
		/*
		 * A preview may include times of authoring and editing, so
		 * set values of the relevant fields.
		 */
		xutc_tm_now(&ei.e.time_authored);
		memcpy(&ei.e.time_edited, &ei.e.time_authored,
		    sizeof(ei.e.time_edited));

		/*
		 * Preview has been requested and/or their are user-correctible
		 * errrors.  Therefore, redisplay the form document.
		 */
		if (is_comment) {
			resp_reply(r, &ei);
		} else /* if (!is_comment) */ {
			resp_submit(r, &ei);
		}

		return;
	}

	/* If name is missing, use the default value. */
	if (ei.e.name[0] == '\0') {
		strlcpy(ei.e.name, FS_H_ANON_NAME, sizeof(ei.e.name));
	}

	if (ei.e.flags & FLAG_PLAIN) {
		/* HTML-escape body content. */
		p = ei.e.body;
		if ((ei.e.body = alloc_text_to_html(ei.e.body, true)) == NULL)
		    {
			msg(r, MSG_CRITICAL, "%s() alloc_text_to_html()"
			    " failed", __func__);
			resp_internal_error(r);

			return;
		}
	}

	if (is_comment) {
#ifndef FEATURE_COMMENT_SUBJECT
		/* Build does not support user-editable comment subjects. */
#ifdef COMPAT_LEGACY_W
		/*
		 * Data files are to remain compatible with legacy systems, so
		 * set comment subject to that of article.
		 */
		get_subject(ei.e.sid, ei.e.subject, sizeof(ei.e.subject));
#endif /* COMPAT_LEGACY_W */
#endif /* !FEATURE_COMMENT_SUBJECT */
		fd = create_comment(&ei.e);
	} else {
		fd = create_submission(&ei.e);
	}
	if (fd == -1) {
		msg(r, MSG_CRITICAL, "%s() %s() failed", __func__, is_comment
		    ? "create_comment" : "create_submission");
		resp_internal_error(r);
		if (ei.e.flags & FLAG_PLAIN) {
			free(ei.e.body);
			ei.e.body = p;
		}

		return;
	}

	if ((f = fdopen(fd, "w")) == NULL) {
		msg(r, MSG_CRITICAL, "%s() fdopen(): %s", __func__,
		    strerror(errno));
		close(fd);
		resp_internal_error(r);
		if (ei.e.flags & FLAG_PLAIN) {
			free(ei.e.body);
			ei.e.body = p;
		}

		return;
	}

	/* Write the article/comment document content. */
	write_file(r, f, &ei.e, true);
	fclose(f);
	close(fd);

	if (ei.e.flags & FLAG_PLAIN) {
		free(ei.e.body); /* ei.e.body = NULL; */
		ei.e.body = p;
	}

	if (is_comment) {
		msg(r, MSG_NOTE, "%s() submitted comment sid %s pid %s",
		    __func__, ei.e.sid, ei.e.pid[0] != '\0' ? ei.e.pid : "0");
		resp_redirect_to_article(r, ei.e.sid);
	} else {
		msg(r, MSG_NOTE, "%s() submitted article subject \"%s\"",
		    __func__, ei.e.subject);
		resp_redirect_to_front(r, NULL);
	}
}


#ifdef FEATURE_VOTE
/*
 * Handler for action A_VOTE.  Handles voting in polls.
 */
static void
do_action_vote(struct kreq *r)
{
	uint64_t	 invalid;
	uint64_t	 missing;

	DMSG(r, "%s()", __func__);

	invalid = 0;
	missing = 0;

	/* Check K_SID. */
	if (r->fieldmap[K_SID] != NULL) {
		if (!article_exists(r->fieldmap[K_SID]->parsed.s)) {
			kutil_invalidate(r, r->fieldmap[K_SID]);
			invalid |= SNB(K_SID);
		}
	} else if (r->fieldnmap[K_SID] != NULL) {
		invalid |= SNB(K_SID);
	} else {
		missing |= SNB(K_SID);
	}

	/*
	 * Check K_VAL.  We accept only 0 to 255, but the validator function
	 * has already rejected values above the upper bound.
	 */
	if (r->fieldmap[K_VAL] != NULL) {
		if (r->fieldmap[K_VAL]->parsed.i < 0) {
			kutil_invalidate(r, r->fieldmap[K_VAL]);
			invalid |= SNB(K_VAL);
		}
	} else if (r->fieldnmap[K_VAL] != NULL) {
		invalid |= SNB(K_VAL);
	} else {
		missing |= SNB(K_VAL);
	}

	if (invalid || missing) {
		/* This should not happen in normal use. */
		msg(r, MSG_NOTE, "%s() invalid = 0x%.16llx; missing ="
		    " 0x%.16llx", __func__, invalid, missing);
		resp_bad_request(r);
	} else if (vote_poll(r->remote, r->fieldmap[K_SID]->parsed.s,
	    r->fieldmap[K_VAL]->parsed.i) != 0) {
		msg(r, MSG_ERROR, "%s() vote_poll(%s, %s, %lld) failed",
		    __func__, r->remote, r->fieldmap[K_SID]->parsed.s,
		    r->fieldmap[K_VAL]->parsed.i);
		resp_internal_error(r);
	} else {
		DMSG(r, "%s() voted sid=\"%s\" val=\"%lld\"", __func__,
		    r->fieldmap[K_SID]->parsed.s,
		    r->fieldmap[K_VAL]->parsed.i);
		resp_redirect_to_article(r, r->fieldmap[K_SID]->parsed.s);
	}
}
#endif /* FEATURE_VOTE */


/*
 * Return true if and only if environment variable GATEWAY_INTERFACE
 * suggests a CGI environment.
 */
static bool
invoked_as_cgi(void)
{
	const char	*scratch;

	scratch = getenv("GATEWAY_INTERFACE");
	return scratch != NULL && strncmp(scratch, "CGI/", 4) == 0;
}


#ifdef DEBUG
/*
 * If r contains a valid value for field k (assumed to have been validated
 * with valid_boolean(), the value is returned; otherwise, if invalid, vi
 * is returned, or vu if no field k was specified in the request.
 */
static bool
get_boolean_key_val(const struct kreq *r, enum key k, bool vi, bool vu)
{
	bool	 val;

	if (r->fieldmap[k] != NULL) {
		val = r->fieldmap[k]->parsed.i != 0;
	} else if (r->fieldnmap[k] != NULL) {
		val = vi;
	} else {
		val = vu;
	}

	return val;
}
#endif /* DEBUG */


#ifdef COMPAT_BRAINDEAD_FB
/*
 * A certain well-known social networking site percent-encodes URLs, apparently
 * in an attempt at protecction from cross-site scripting (XSS) attacks.
 * Unfortunately, it does so in violation of the RFC insomuch as it
 * percent-encodes reserved characters (such as '&', '%', and '=') in their
 * reserved contexts, thus changing the meaning of the URL.
 *
 * Specifically, we are concerned with requests of the form
 *	fwobac?action=article;sid=YYYYmmddHHMMSS...
 * (or the equivalent with '&' rather than ';'), which get mangled into
 *	fwobac?action=article%3bsid%3dYYYYmmddHHMMSS...
 * (or "%26" for '&') so that kcgi gives us, for the action value
 *	article;sid=YYYYmmddHHMMSS...
 *
 * We deal with (only) this case by identifying the intention and redirecting
 * to the appropriate article.
 */


/*
 * Returns true if and only if the first len characters of s are digits.
 */
static bool
is_all_digits_n(const char *s, size_t len)
{
	size_t	 i;

	i = 0;
	while (i < len && isdigit((unsigned char)s[i])) {
		i++;
	}

	return i >= len;
}


/*
 * Returns true if and only if the start of kp->val is of the form
 *	"article" ['&' | ';'] "sid=" YYYYmmddHHMMSS ['&' | ';' | '\0']
 */
static bool
is_mangled_article_req(const struct kpair *kp)
{
	/*
	 * article;sid=YYYYmmddHHMMSS
	 * 01234567890123456789012345
	 * 12345678901234567890123456
	 */
	const char	*v;
	size_t	 a;
	size_t	 e;
	size_t	 z;

	v = kp->val;
	a = strlen(AN_ARTICLE);
	e = strlen(KN_SID EQ);
	z = a + 1 + e + TIMESTAMP_LEN;

	return kp->valsz >= z && memchr(v, '\0', kp->valsz) == NULL
	    && strncmp(v, AN_ARTICLE, a) == 0 && (v[a] == '&' || v[a] == ';')
	    && strncmp(&v[a + 1], KN_SID EQ, e) == 0
	    && is_all_digits_n(&v[a + 1 + e], TIMESTAMP_LEN)
	    && (v[z] == '&' || v[z] == ';' || v[z] == '\0');
}


/*
 * r must be a request with mangled value for key K_ACTION.
 * Sends a response with a redirection to the intended article.
 */
static void
do_action_mangled_article_req(struct kreq *r)
{
	char	 sid[TIMESTAMP_LEN + 1];

	/* DMSG(r, "%s()", __func__); */

	strlcpy(sid, &r->fieldnmap[K_ACTION]->val[strlen(AN_ARTICLE SEP KN_SID
	    EQ)], sizeof(sid));

	DMSG(r, "%s() redirecting to article %s", __func__, sid);

	resp_redirect_to_article(r, sid);
}
#endif /* COMPAT_BRAINDEAD_FB */


/*
 * Handles action actn according to its entry in action_table: calls pledge()
 * with the relevant promises, then responds with an appropriate error if
 * - HTTPS is required but was not used in the request, or
 * - method POST is required but was not used in the request, or
 * - the user has insufficient rights
 * or dispatches to the hander in the absence of the above errors.
 */
static void
do_action(struct kreq *r, enum action actn)
{
	DMSG(r, "%s() action = \"%s\"", __func__, action_table[actn].name);

	/* pledge() appropriately. */
	xpledge(action_table[actn].promises);

	if ((action_table[actn].rreq & RR_SCHEME_HTTPS)
	    && r->scheme != KSCHEME_HTTPS) {
		/*
		 * We could redirect to the equivalent URL with scheme HTTPS.
		 * However, given that
		 * - this should not occur in normal use, and
		 * - there may be sensitive information involved,
		 * we elect to report the error.
		 */
		msg(r, MSG_INFO, "%s() action \"%s\" attempted without HTTPS",
		    __func__, action_table[actn].name);
		resp_forbidden(r, TMPL_S_403_HTTPS_REQD);
	} else if ((action_table[actn].rreq & RR_METHOD_POST)
	    && r->method != KMETHOD_POST) {
		msg(r, MSG_INFO, "%s() action \"%s\" attempted without POST",
		    __func__, action_table[actn].name);
		resp_forbidden(r, TMPL_S_403_POST_REQD);
	} else if (!(SNB(actn) & ent_act)) {
		/* User has insufficient privileges. */
		msg(r, MSG_NOTE, "%s() action \"%s\" denied because of"
		    " insufficient rights", __func__, action_table[actn].name);
		resp_insufficient_rights(r);
	} else {
		DMSG(r, "%s() dispatching", __func__);
		(*action_table[actn].handler)(r);
	}
}


/*
 * Prints to f a line containing a hexadecimal representation of r, plus
 * string s.
 */
static void
print_group_rights_info_item(FILE *f, const char *s, uint64_t r)
{
	fprintf(f, "# 0x%.*llx\t%s\n", (int)(sizeof(action_table) /
	    sizeof(action_table[0]) + BITS_PER_XDIG - 1) / 4, r, s);
}


/*
 * Prints to f detailed information for possible values for group rights.
 */
static void
print_group_rights_info(FILE *f)
{
	size_t	 a;

	a = 0;

	fprintf(f, "#\n");
	/* Show bit for each action. */
	for (a = 0; a < sizeof(action_table) / sizeof(action_table[0]); a++) {
		print_group_rights_info_item(f, action_table[a].name, SNB(a));
	}

	fprintf(f, "#\n");
	/* Show useful cases. */
	print_group_rights_info_item(f, "read-only", ACTIONS_RO);
	print_group_rights_info_item(f, "registering", ACTIONS_REGISTER);
	print_group_rights_info_item(f, "login", ACTIONS_LOGIN);
	print_group_rights_info_item(f, "submitting", ACTIONS_SUBMIT);
	print_group_rights_info_item(f, "commenting", ACTIONS_COMMENT);
	print_group_rights_info_item(f, "logged-in", ACTIONS_LOGGED_IN);
	print_group_rights_info_item(f, "read-only+submitting", ACTIONS_RO
	    | ACTIONS_SUBMIT);
	print_group_rights_info_item(f, "read-only+submitting+commenting",
	    ACTIONS_RO | ACTIONS_SUBMIT | ACTIONS_COMMENT);
	fprintf(f, "#\n");
	print_group_rights_info_item(f, "default if not logged-in (\""
	    PSEUDO_GROUP_ANON "\")",
	    DEFAULT_RIGHTS_ANON);
	print_group_rights_info_item(f, "default for logged in (\"" GROUP_USER
	    "\")", DEFAULT_RIGHTS_USER);
}


/*
 * Prints to f a sample configuration file, with an entry for each item plus
 * detailed information on group rights.
 */
static void
print_sample_config(FILE *f)
{
	enum conf_item	 i;

	fprintf(f, "# %s\n", CONF_FILE);

	for (i = 0; i < sizeof(conf_name) / sizeof(conf_name[0]); i++) {
		if (i == CI_GROUP_RIGHTS) {
			print_group_rights_info(f);
		}
		fprintf(f, "%s\t?\n", conf_name[i]);
	}
}


/*
 * Prints to the standard error a usage string (for non-CGI invocation)
 */
static void
non_cgi_usage(void)
{
	fprintf(stderr, "# (non-CGI) usage: %s [-c]\n", getprogname());
}


/*
 * main()-equivalent for use when program is invoked outside CGI environment.
 */
static int
non_cgi_main(int argc, char *argv[])
{
	int	 arg;
	int	 status;
	bool	 flag;

	status = EXIT_SUCCESS;
	flag = false;

	arg = 1;

	while (arg < argc && strcmp(argv[arg], "-c") == 0) {
		flag = true;
		arg++;
	}

	if (arg < argc) {
		flag = false;
		status = EXIT_FAILURE;
	}

	if (flag) {
		print_sample_config(stdout);
	} else {
		non_cgi_usage();
	}

	return status;
}


/* Finally, the main program. */
int
main(int argc, char *argv[])
{
	struct	kreq	 r;
	int	 i, j;
	uint16_t	 invalid, missing;

	/* Do security setup, pledging with the most promises we will need. */
	malloc_options = "S";
	xpledge(PROMISES_MAX);

	/* Set stderr to line-buffering to maximise chance of clean logs. */
	if (setvbuf(stderr, NULL, _IOLBF, 0) != 0) {
		fprintf(stderr, "%s: error: setvbuf failed\n", getprogname());
	}

#ifdef DEBUG
	if (sizeof(keys) / sizeof(keys[0]) != sizeof(key_desc)
	    / sizeof(key_desc[0])) {
		fprintf(stderr, "%s: fatal: keys <==> key_desc mismatch\n",
		    getprogname());
		return EXIT_FAILURE;
	}
#endif /* DEBUG */

	/* If invoked without CGI environment, run differently. */
	if (!invoked_as_cgi()) {
		return non_cgi_main(argc, argv);
	}

	umask(007);
	memset(&user, 0, sizeof(user));

	/* Abort if can't load configuration. */
	i = conf_parse_file(CONF_FILE, store_config_entry, &j);
	if (i < 0) {
		fprintf(stderr, "%s: fatal: failed to read/parse config file"
		    " \"%s\" [%d]\n", getprogname(), CONF_FILE, i);
		return EXIT_FAILURE;
	} else if (i > 0) {
		fprintf(stderr, "%s: fatal: error on line %d of config file"
		    "[callback: %d]\n", getprogname(), i, j);
		return EXIT_FAILURE;
	}

	/* Abort if configuration is incomplete. */
	check_config(&invalid, &missing);
	if (invalid || missing) {
		fprintf(stderr, "%s: fatal: invalid [0x%.4x] / missing"
		    " [0x%.4x] configuration file entries\n", getprogname(),
		    invalid, missing);
		return EXIT_FAILURE;
	}

	/* Initialise user database. */
	user_init(get_conf_val(CI_USER_DB_FILE));

	if (chdir("/tmp") != 0) {
		fprintf(stderr, "%s: fatal: failed to chdir to /tmp: %s\n",
		    getprogname(), strerror(errno));
		return EXIT_FAILURE;
	}

	/* Analyze the CGI environment and perform basic checks. */
	if (khttp_parse(&r, keys, sizeof(keys) / sizeof(struct kvalid),
	    NULL, 0, 0) != KCGI_OK) {
		xpledge(PROMISES_MIN);
		fprintf(stderr, "%s: fatal: khttp_parse() failed\n",
		    getprogname());
		return EXIT_FAILURE;
	} else if (r.remote == NULL || r.remote[0] == '\0') {
		xpledge(PROMISES_MIN);
		khttp_free(&r);
		fprintf(stderr, "%s: fatal: REMOTE_ADDR unknown\n",
		    getprogname());
		return EXIT_FAILURE;
	} else if (r.method != KMETHOD_GET && r.method != KMETHOD_HEAD
	    && r.method != KMETHOD_POST) {
		resp_method_not_allowed(&r);
	} else if (r.fullpath != NULL && r.fullpath[0] != '\0') {
		DMSG(&r, "PATH_INFO non-null");
		resp_not_found(&r);
	} else {
#ifdef DEBUG
		rt_debug = get_boolean_key_val(&r, K_DEBUG, 1, 0);
#endif /* DEBUG */

		/* Check authentication cookie. */
		check_auth_cookie(&r, &user, &ent_act);

#ifdef COMPAT_LEGACY_ID
		if (user.id[0] == '\0') {
			vis_id[0] = '\0';
		} else {
			/* Create logging-safe version of ID. */
			(void)strnvis(vis_id, user.id, sizeof(vis_id), VFLAGS);
		}
#endif /* COMPAT_LEGACY_ID */

		if (r.fieldmap[K_ACTION] != NULL) {
			/* Action specification was valid. */
			do_action(&r,
			    (enum action)r.fieldmap[K_ACTION]->parsed.i);
		} else if (r.fieldnmap[K_ACTION] != NULL) {
			/* Action specification was invalid. */
#ifdef COMPAT_BRAINDEAD_FB
			if (is_mangled_article_req(r.fieldnmap[K_ACTION])) {
				xpledge(PROMISES_R);
				DMSG(&r, "%s() mangled article req", __func__);
				do_action_mangled_article_req(&r);
			} else {
				DMSG(&r, "%s() invalid action", __func__);
				resp_not_found(&r);
			}
#else /* COMPAT_BRAINDEAD_FB */
			DMSG(&r, "%s() invalid action", __func__);
			resp_not_found(&r);
#endif /* COMPAT_BRAINDEAD_FB */
		} else {
			/* No action was specified. */
			DMSG(&r, "%s() no action specified; redirecting...",
			    __func__);
			resp_redirect_to_front(&r, NULL);
		}
	}

	/*
	 * Many things, such as the configuration values, could be freed here.
	 * However, as we're about to exit, there's no value in so doing.
	 */

	khttp_free(&r);

	return EXIT_SUCCESS;
}


